/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2022 David Frantz

FORCE is free software_: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http_://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for spectral adjustment
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
D. Scheffler, D. Frantz, and K. Segl (2020). Spectral harmonization and 
red edge prediction of Landsat-8 to Sentinel-2 using land cover optimized 
multivariate regressors. Remote Sensing of Environment 241, 111723. 
https://doi.org/10.1016/j.rse.2020.111723
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "spec-adjust-hl.h"


int spectral_predict(ard_t ard, small *mask_, int nc, int sid);

#define _SPECHOMO_N_SEN_  7 // number of sensors that we can homogenize
#define _SPECHOMO_N_CLS_ 51 // number of clusters, incl. global
#define _SPECHOMO_N_SRC_  6 // number of source bands
#define _SPECHOMO_N_DST_ 10 // number of destination bands
#define _SPECHOMO_N_COF_  7 // number of coefficients
#define _SPECHOMO_N_SIM_ 10 // max. number of close clusters to be used for kNN
#define _SPECHOMO_GOOD_SAM_ 0.0698132 // good clusters are closer than this angle (4°)
#define _SPECHOMO_MEDI_SAM_ 0.2094395 // ok'ish clusters are closer than this angle (12°)
#define _SPECHOMO_POOR_SAM_ 0.2617995 // clusters further away from this angle (15°) should not be used
                                      // also used to compute a weight for each cluster
#define _SPECHOMO_MIN_WEIGHT_ 1.0 - _SPECHOMO_GOOD_SAM_/_SPECHOMO_POOR_SAM_
#define _SPECHOMO_MED_WEIGHT_ 1.0 - _SPECHOMO_MEDI_SAM_/_SPECHOMO_POOR_SAM_

const char _SPECHOMO_SENSOR_[_SPECHOMO_N_SEN_][NPOW_04] = {
  "LND04", "LND05", "LND07", "LND08", "LND09", "MOD01", "MOD02" };

const char _SPECHOMO_SRC_DOMAIN_[_SPECHOMO_N_SRC_][NPOW_10] = {
  "BLUE", "GREEN", "RED", "NIR", "SWIR1", "SWIR2"};

const char _SPECHOMO_DST_DOMAIN_[_SPECHOMO_N_DST_][NPOW_10] = {
  "BLUE", "GREEN", "RED", "REDEDGE1", "REDEDGE2",
  "REDEDGE3", "BROADNIR", "NIR", "SWIR1", "SWIR2"};

const short _SPECHOMO_CENTER_[_SPECHOMO_N_SEN_][_SPECHOMO_N_SRC_][_SPECHOMO_N_CLS_] = {
  {
    { 1000, 1260, 480, 6810, 1072, 693, 308, 1381, 3436, 442, 554, 444, 7408, 3845, 640, 215, 677, 348, 556, 7855, 1567, 1454, 423, 2147, 288, 418, 2077, 914, 1254, 723, 373, 405, 939, 549, 655, 1555, 469, 1258, 489, 5768, 735, 1322, 841, 1809, 2173, 7194, 568, 665, 765, 605, 749 },
    { 1161, 1684, 694, 6520, 1528, 992, 485, 2163, 2122, 676, 799, 677, 7183, 3777, 929, 424, 978, 501, 763, 7590, 1730, 1594, 734, 4608, 435, 654, 3538, 1244, 1493, 1033, 580, 575, 1061, 722, 819, 1592, 674, 1892, 710, 5608, 1072, 2072, 1098, 2046, 2277, 7043, 817, 942, 1182, 856, 1036 },
    { 1226, 1931, 712, 6343, 1932, 1192, 372, 2637, 2803, 614, 992, 614, 7017, 3692, 1162, 296, 1228, 408, 883, 7415, 1714, 1556, 654, 5081, 341, 595, 3578, 1295, 1611, 1265, 412, 463, 1037, 929, 924, 1109, 668, 2290, 785, 5497, 1297, 2496, 1202, 2157, 2293, 6914, 859, 1162, 1534, 1191, 1099 },
    { 1503, 2385, 2935, 6589, 2563, 1849, 4201, 3320, 5508, 2873, 1546, 3430, 7089, 4326, 1587, 5134, 2443, 3171, 2386, 7633, 1733, 3105, 4199, 5161, 3521, 3560, 4485, 3065, 1855, 1863, 4410, 3275, 2325, 1338, 1552, 1055, 2271, 2913, 2699, 5989, 2246, 3024, 2102, 2310, 3019, 7226, 2143, 1683, 2042, 1625, 2393 },
    { 1611, 2843, 2107, 5327, 4116, 2811, 1571, 4833, 463, 1566, 2124, 1907, 5883, 3735, 2017, 1958, 2663, 1480, 2706, 6353, 1569, 2255, 4109, 4997, 1468, 2181, 3904, 2100, 2190, 2756, 1772, 1465, 1716, 3126, 2676, 892, 1631, 4495, 2244, 5094, 2438, 4133, 2301, 1961, 2590, 6079, 2154, 2589, 2429, 3249, 2255 },
    { 1423, 2470, 1192, 4304, 3474, 2209, 622, 4001, 375, 826, 1657, 985, 4837, 3303, 1628, 800, 1771, 661, 1716, 5295, 1249, 1504, 3336, 3803, 616, 1172, 2767, 1367, 1959, 2165, 724, 656, 1286, 3211, 2079, 639, 953, 3830, 1337, 4399, 1769, 3601, 1803, 1553, 2198, 5042, 1426, 2082, 1894, 3167, 1633 }
  },
  {
    { 746, 1383, 306, 6452, 466, 552, 414, 782, 3424, 3808, 842, 529, 1974, 1256, 1051, 661, 583, 6002, 728, 493, 6230, 2107, 377, 383, 551, 1292, 1315, 743, 1486, 1177, 418, 7903, 484, 419, 683, 401, 1745, 633, 696, 580, 1715, 791, 1010, 7433, 1095, 856, 584, 317, 217, 1361, 759 },
    { 1059, 2173, 448, 6200, 674, 764, 650, 1156, 2125, 3741, 971, 759, 2404, 1889, 1305, 934, 796, 5817, 977, 718, 6064, 4464, 582, 540, 732, 1980, 1493, 1055, 1511, 1587, 620, 7618, 695, 731, 1097, 604, 1941, 909, 826, 839, 1985, 1181, 1171, 7226, 1559, 1241, 837, 493, 431, 1509, 1050 },
    { 1278, 2671, 364, 6016, 671, 888, 588, 1482, 2811, 3658, 943, 946, 2332, 2295, 1428, 1167, 946, 5710, 1097, 789, 5913, 4887, 406, 432, 958, 2383, 1553, 1290, 1006, 1836, 535, 7442, 718, 646, 1163, 535, 1832, 1139, 905, 875, 2129, 1502, 1237, 7075, 1970, 1498, 1194, 367, 299, 1481, 1113 },
    { 1900, 3357, 3034, 6236, 2322, 2465, 3582, 2879, 5516, 4242, 2272, 1490, 3210, 2914, 1823, 2249, 1671, 6193, 2061, 2755, 5920, 5095, 4466, 3241, 1364, 2887, 1682, 1920, 927, 2264, 2974, 7661, 2915, 4251, 3301, 3422, 2091, 1661, 1299, 2167, 2571, 1964, 1506, 7373, 2596, 2311, 1624, 4238, 5018, 3124, 2423 },
    { 2988, 4864, 1391, 5003, 1651, 2732, 2169, 2993, 462, 3657, 1635, 2037, 2612, 4499, 2160, 2582, 2465, 5259, 2145, 2265, 4911, 4881, 1755, 1460, 3166, 3927, 1710, 2846, 767, 2738, 1472, 6334, 2134, 4170, 1802, 1933, 1753, 2465, 2647, 2112, 2258, 2360, 1614, 6189, 4169, 2600, 3191, 1633, 1926, 2278, 2286 },
    { 2386, 4001, 613, 3983, 956, 1708, 1160, 1912, 375, 3219, 1179, 1582, 1874, 3829, 1878, 1765, 1836, 4551, 1613, 1348, 3998, 3653, 708, 648, 3250, 3435, 1469, 2209, 555, 2386, 708, 5241, 1210, 3389, 1124, 968, 1260, 1976, 2236, 1400, 1860, 1872, 1427, 5189, 3524, 2024, 3111, 656, 790, 1547, 1655 }
  },
  {
    { 636, 622, 6971, 1244, 340, 207, 3894, 1394, 543, 592, 379, 3640, 1083, 6136, 6684, 1293, 631, 1045, 1624, 394, 1223, 761, 404, 288, 1933, 457, 679, 739, 1342, 559, 841, 1524, 8038, 1983, 739, 974, 5906, 742, 656, 393, 425, 481, 281, 1031, 1386, 501, 583, 359, 547, 1264, 729 },
    { 903, 919, 6795, 1840, 508, 449, 3826, 1543, 761, 821, 548, 2160, 1517, 5996, 6413, 1774, 884, 1630, 1835, 631, 1870, 912, 728, 459, 2350, 669, 834, 1069, 2055, 743, 1229, 1869, 7753, 4275, 1038, 1156, 5737, 1041, 1069, 575, 641, 711, 428, 1202, 1516, 717, 807, 584, 692, 1506, 1016 },
    { 1169, 893, 6663, 2319, 596, 276, 3732, 1483, 984, 1007, 417, 2820, 2001, 5853, 6219, 2136, 1130, 2331, 1860, 529, 2365, 882, 617, 329, 2285, 650, 941, 1396, 2613, 870, 1563, 2016, 7559, 4864, 1301, 1264, 5621, 1320, 1137, 461, 583, 768, 319, 1273, 903, 823, 1248, 371, 941, 1645, 1083 },
    { 1600, 2535, 6983, 2926, 1501, 4980, 4364, 3152, 1522, 2336, 3179, 5550, 2626, 5925, 6493, 2618, 1700, 3170, 1983, 3596, 2853, 2110, 4230, 3666, 3012, 2417, 1450, 2677, 3272, 2103, 2489, 2461, 7769, 5105, 1912, 1586, 6107, 1926, 3320, 2961, 3160, 2859, 2913, 1558, 810, 2412, 1693, 4502, 1341, 1865, 2377 },
    { 2001, 2343, 5853, 4504, 1349, 1953, 3788, 2287, 2101, 2742, 1454, 445, 4234, 4961, 5210, 3104, 2528, 4186, 1754, 2088, 3891, 1574, 4158, 1503, 2481, 1687, 2650, 2883, 4793, 2844, 2720, 2194, 6377, 4953, 2983, 1782, 5187, 2841, 1774, 1390, 2057, 2287, 1327, 1666, 671, 2277, 3300, 1770, 3128, 2241, 2254 },
    { 1610, 1461, 4802, 3877, 912, 779, 3298, 1505, 1644, 1788, 627, 371, 3604, 3943, 4123, 2697, 2005, 2599, 1368, 1049, 3400, 1133, 3405, 604, 1782, 936, 2148, 1870, 4037, 1854, 2010, 1728, 5271, 3645, 2400, 1573, 4415, 2226, 1091, 634, 1061, 1327, 563, 1457, 481, 1385, 3199, 673, 3231, 2023, 1620 }
  },
  {
    { 601, 580, 8068, 891, 330, 4996, 1707, 727, 453, 1049, 511, 276, 1240, 3403, 700, 402, 7258, 415, 1451, 1602, 806, 1539, 683, 499, 285, 658, 2043, 6349, 554, 729, 479, 1028, 7660, 604, 202, 1313, 550, 3766, 1317, 1261, 356, 1312, 402, 448, 3608, 859, 381, 1588, 1357, 2005, 783 },
    { 861, 859, 7780, 1263, 516, 4880, 1927, 993, 698, 1150, 770, 460, 1912, 2121, 996, 708, 6947, 645, 1646, 1830, 1140, 1721, 985, 696, 506, 833, 4453, 6151, 764, 1071, 722, 1192, 7461, 840, 474, 1867, 696, 2184, 1999, 1517, 621, 1847, 760, 772, 3559, 1138, 575, 1936, 2038, 3012, 1093 },
    { 1417, 865, 7579, 1765, 358, 4755, 1914, 1191, 642, 1079, 744, 298, 2386, 2773, 1224, 570, 6725, 541, 1544, 1954, 1406, 1117, 1263, 903, 326, 949, 4860, 5985, 854, 1298, 757, 1255, 7298, 1083, 265, 2263, 944, 2853, 2730, 1650, 367, 2206, 582, 630, 3459, 1234, 409, 2102, 2485, 2932, 1114 },
    { 1971, 2221, 7750, 2414, 3223, 5370, 2068, 1659, 2281, 2505, 3101, 3657, 3114, 4903, 1969, 3757, 6857, 2724, 3233, 2223, 2105, 1060, 2704, 1444, 4253, 1675, 5162, 6403, 2525, 2364, 2761, 1554, 7429, 1619, 5303, 2922, 1390, 6093, 3606, 1940, 4320, 2876, 4336, 3532, 4340, 2258, 3216, 2227, 3136, 4063, 2557 },
    { 3649, 2158, 6298, 3749, 1486, 4570, 1872, 2009, 1571, 1794, 2247, 1479, 4521, 444, 2929, 2307, 5515, 1479, 2369, 2042, 3059, 946, 2896, 1951, 1541, 2728, 5030, 5244, 2799, 2463, 2243, 1672, 6152, 2448, 2007, 3369, 3108, 371, 5064, 2299, 1814, 4335, 4177, 1930, 3728, 2486, 1438, 2268, 4293, 3516, 2372 },
    { 3507, 1453, 5464, 3149, 671, 4056, 1389, 1642, 909, 1376, 1276, 632, 3809, 408, 2272, 1243, 4682, 779, 1475, 1608, 2428, 640, 1878, 1530, 638, 2113, 3648, 4439, 1774, 1796, 1336, 1477, 5222, 1974, 828, 2901, 3245, 345, 3073, 2065, 741, 3784, 3438, 1027, 3365, 1939, 639, 1919, 3713, 2378, 1718 }
  },
  {
    { 601, 580, 8068, 891, 330, 4996, 1707, 727, 453, 1049, 511, 276, 1240, 3403, 700, 402, 7258, 415, 1451, 1602, 806, 1539, 683, 499, 285, 658, 2043, 6349, 554, 729, 479, 1028, 7660, 604, 202, 1313, 550, 3766, 1317, 1261, 356, 1312, 402, 448, 3608, 859, 381, 1588, 1357, 2005, 783 },
    { 861, 859, 7780, 1263, 516, 4880, 1927, 993, 698, 1150, 770, 460, 1912, 2121, 996, 708, 6947, 645, 1646, 1830, 1140, 1721, 985, 696, 506, 833, 4453, 6151, 764, 1071, 722, 1192, 7461, 840, 474, 1867, 696, 2184, 1999, 1517, 621, 1847, 760, 772, 3559, 1138, 575, 1936, 2038, 3012, 1093 },
    { 1417, 865, 7579, 1765, 358, 4755, 1914, 1191, 642, 1079, 744, 298, 2386, 2773, 1224, 570, 6725, 541, 1544, 1954, 1406, 1117, 1263, 903, 326, 949, 4860, 5985, 854, 1298, 757, 1255, 7298, 1083, 265, 2263, 944, 2853, 2730, 1650, 367, 2206, 582, 630, 3459, 1234, 409, 2102, 2485, 2932, 1114 },
    { 1971, 2221, 7750, 2414, 3223, 5370, 2068, 1659, 2281, 2505, 3101, 3657, 3114, 4903, 1969, 3757, 6857, 2724, 3233, 2223, 2105, 1060, 2704, 1444, 4253, 1675, 5162, 6403, 2525, 2364, 2761, 1554, 7429, 1619, 5303, 2922, 1390, 6093, 3606, 1940, 4320, 2876, 4336, 3532, 4340, 2258, 3216, 2227, 3136, 4063, 2557 },
    { 3649, 2158, 6298, 3749, 1486, 4570, 1872, 2009, 1571, 1794, 2247, 1479, 4521, 444, 2929, 2307, 5515, 1479, 2369, 2042, 3059, 946, 2896, 1951, 1541, 2728, 5030, 5244, 2799, 2463, 2243, 1672, 6152, 2448, 2007, 3369, 3108, 371, 5064, 2299, 1814, 4335, 4177, 1930, 3728, 2486, 1438, 2268, 4293, 3516, 2372 },
    { 3507, 1453, 5464, 3149, 671, 4056, 1389, 1642, 909, 1376, 1276, 632, 3809, 408, 2272, 1243, 4682, 779, 1475, 1608, 2428, 640, 1878, 1530, 638, 2113, 3648, 4439, 1774, 1796, 1336, 1477, 5222, 1974, 828, 2901, 3245, 345, 3073, 2065, 741, 3784, 3438, 1027, 3365, 1939, 639, 1919, 3713, 2378, 1718 }
  },
  {
    { 654, 437, 7170, 437, 1347, 363, 4057, 1000, 751, 970, 532, 8149, 4243, 1632, 1498, 365, 1006, 4053, 349, 1191, 780, 455, 194, 1217, 972, 1848, 808, 406, 668, 748, 1005, 351, 5821, 7967, 738, 1157, 391, 1753, 1893, 997, 482, 798, 3254, 484, 1349, 502, 626, 1322, 237, 999, 800 },
    { 943, 734, 6887, 615, 1967, 669, 3952, 1405, 1165, 1101, 751, 7869, 2169, 2171, 1850, 572, 1335, 4006, 599, 1933, 1205, 817, 556, 1512, 1163, 3485, 1258, 747, 1044, 1077, 1170, 577, 5590, 7562, 1045, 1774, 803, 2063, 4512, 1484, 740, 1101, 3209, 818, 2274, 702, 977, 1645, 412, 1688, 1150 },
    { 1213, 637, 6758, 830, 2412, 381, 3866, 1702, 1224, 1034, 838, 7742, 2672, 1678, 1685, 486, 1273, 3986, 373, 2472, 1561, 678, 288, 1645, 1233, 3291, 1533, 428, 1302, 1378, 1238, 425, 5471, 7385, 1758, 2233, 606, 2175, 4883, 1985, 800, 1333, 3133, 706, 2927, 864, 938, 922, 258, 2321, 1180 },
    { 2779, 2534, 7102, 1330, 3076, 4671, 4541, 2424, 2696, 2383, 2085, 7915, 5489, 1826, 3035, 2520, 3039, 4068, 4043, 3241, 2414, 3637, 5488, 1968, 1606, 4352, 2591, 4736, 3134, 2093, 1480, 2767, 5998, 7556, 2485, 2969, 4242, 2320, 5197, 2821, 2591, 2014, 4068, 3064, 3648, 1450, 2508, 791, 2689, 3401, 2630 },
    { 3331, 1936, 6045, 1858, 3573, 1931, 4012, 3830, 2440, 1807, 2862, 6714, 428, 1711, 2449, 1556, 2408, 3381, 1829, 4815, 3026, 2455, 2178, 2316, 1770, 3909, 2872, 2183, 2991, 3519, 1621, 1436, 5146, 6248, 4295, 4587, 3844, 2233, 5110, 4089, 2801, 3305, 3579, 2116, 5103, 1769, 2301, 692, 1294, 4666, 2543 },
    { 2112, 1012, 4747, 1522, 3191, 629, 3415, 3315, 1695, 1346, 1828, 5073, 336, 1227, 1550, 786, 1498, 2672, 628, 4331, 2315, 1251, 735, 2133, 1619, 2721, 2206, 723, 1721, 2975, 1510, 576, 4172, 4826, 4156, 4142, 3120, 1957, 3861, 3456, 1597, 2779, 3113, 1083, 4772, 1422, 1492, 511, 490, 2728, 1833 }
  },
  {
    { 517, 4032, 1170, 551, 458, 5729, 442, 629, 3116, 487, 926, 4244, 833, 1890, 1253, 943, 6999, 1685, 275, 300, 857, 192, 1840, 542, 755, 7944, 1336, 340, 711, 343, 419, 4093, 1106, 362, 501, 453, 711, 1297, 658, 1034, 1355, 8219, 1162, 741, 397, 665, 331, 1492, 1375, 383, 717 },
    { 739, 4045, 1672, 882, 855, 5497, 785, 893, 3081, 748, 1122, 2170, 1273, 4503, 2007, 1385, 6704, 2002, 462, 510, 1000, 553, 3414, 763, 1152, 7543, 1664, 623, 1007, 667, 593, 3983, 1488, 584, 833, 747, 1094, 1451, 853, 1222, 2027, 7938, 1825, 1085, 750, 972, 513, 1824, 2284, 794, 1050 },
    { 978, 4026, 2038, 851, 667, 5376, 647, 1171, 3010, 820, 1214, 2673, 1317, 4866, 2657, 1751, 6580, 2123, 459, 318, 941, 286, 3211, 849, 1507, 7365, 931, 374, 1710, 376, 785, 3895, 1668, 436, 811, 687, 1219, 1480, 950, 1290, 2560, 7807, 2315, 1368, 422, 1216, 358, 1670, 2912, 590, 1093 },
    { 1338, 4139, 2722, 2440, 3979, 5917, 3258, 1745, 3928, 2621, 1602, 5490, 3152, 5187, 3517, 2577, 6945, 2271, 1681, 3285, 2200, 5498, 4301, 2058, 3256, 7537, 803, 4271, 2423, 4711, 1312, 4611, 2283, 2752, 3090, 2758, 2487, 1563, 1543, 1571, 3282, 7985, 3077, 2095, 4826, 2449, 2924, 3153, 3625, 4300, 2452 },
    { -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 },
    { 1473, 2630, 3896, 1293, 1199, 4103, 1076, 2241, 3009, 1647, 1695, 335, 1541, 3840, 4336, 3152, 4647, 1964, 719, 554, 1252, 731, 2666, 1798, 2013, 4810, 521, 641, 4093, 633, 1420, 3477, 2356, 589, 1548, 1098, 1896, 1539, 2235, 1596, 3344, 5109, 4149, 2862, 721, 1983, 582, 1560, 4739, 3250, 1712 },
  }
};

const double _SPECHOMO_COEFS_[_SPECHOMO_N_SEN_][_SPECHOMO_N_DST_][_SPECHOMO_N_COF_][_SPECHOMO_N_CLS_] = {
  {
    {
      { 0.954152762889862, 0.94907933473587, 0.922274708747864, 0.97670990228653, 1.00875294208527, 0.98089987039566, 0.953009605407715, 0.993607223033905, 0.564308047294617, 0.955916047096252, 0.986152172088623, 0.945890545845032, 0.94731730222702, 0.896374106407166, 0.985254168510437, 0.936255276203156, 0.92149692773819, 0.944583117961884, 0.959907233715057, 0.96592766046524, 0.883292019367218, 0.899154901504517, 0.957743883132935, 0.980517685413361, 0.944049179553986, 0.934145569801331, 0.986884355545044, 0.969458639621735, 0.960440099239349, 0.991686463356018, 0.961585998535156, 0.94071364402771, 0.938427269458771, 0.970795691013336, 1.01875901222229, 0.92786055803299, 0.946386873722076, 0.974800050258636, 0.926619291305542, 0.975853562355042, 0.948089420795441, 0.994568169116974, 0.932363510131836, 0.87620621919632, 0.886674404144287, 0.963523328304291, 0.959991335868835, 0.986239194869995, 0.962845623493195, 0.970291554927826, 0.960189998149872 },
      { 0.0402863025665283, 0.0624573566019535, 0.0618130266666412, -0.0526312813162804, 0.0164788663387299, 0.0445019453763962, 0.0545056723058224, 0.0506417788565159, 1.07493352890015, 0.0596030130982399, 0.0296508967876434, 0.0608920268714428, 0.029423676431179, 0.148258477449417, 0.0340978279709816, 0.0577920340001583, 0.0906097292900085, 0.0677787214517593, 0.100165255367756, -0.0292339622974396, 0.202041670680046, 0.163701087236404, 0.0545140616595745, 0.0208322443068027, 0.0638619363307953, 0.060560654848814, 0.0815840512514114, 0.0466293506324291, 0.0332869216799736, 0.035683773458004, 0.0496813468635082, 0.066174291074276, 0.0529126897454262, 0.0564990490674973, 0.0640547201037407, 0.159275829792023, 0.0568414330482483, 0.0422058813273907, 0.0674057304859161, -0.0306065306067467, 0.0759643614292145, 0.0579155758023262, 0.0793044045567513, 0.139805063605309, 0.157222256064415, 0.0105971405282617, 0.0479044243693352, 0.0264190677553415, -0.00393397314473987, 0.0567609705030918, 0.0551056452095509 },
      { 0.0160278659313917, -0.0075277709402144, 0.00613748328760266, 0.0812630504369736, 0.00108857359737158, -0.0181697495281696, 0.000876090372912586, -0.0583856925368309, -0.279684513807297, -0.0146267814561725, -0.0109732812270522, -0.0091131990775466, 0.0246037766337395, -0.0395693890750408, -0.00950089748948812, 0.0147597873583436, -0.0246783252805471, -0.017017675563693, -0.0595677495002747, 0.0664066821336746, -0.0909163877367973, -0.0136682000011206, -0.0072373365983367, -0.0100923608988523, -0.0147726181894541, 0.00165849144104868, -0.0685375556349754, -0.00451065972447395, 0.0245530996471643, -0.0046236515045166, -0.00121944630518556, -0.0144777204841375, 0.0136804180219769, -0.0214083082973957, -0.0781974866986275, -0.0735452026128769, 0.00176932115573436, -0.0158434901386499, -0.00270389695651829, 0.0468460395932198, -0.00776124093681574, -0.0322873555123806, -0.0123854102566838, -0.0331278704106808, -0.0378896966576576, 0.0235703662037849, -0.00639875745400786, -0.007984584197402, 0.0461783520877361, -0.0227770358324051, 0.0108882123604417 },
      { -0.015979902818799, 0.00214660842902958, 0.00336026470176876, -0.00278370804153383, -0.00492391781881452, 0.00176418165210634, 0.000967742351349443, 0.0352106504142284, -0.0324527360498905, 0.000809228513389826, 0.00192214327398688, -0.000691640714649111, 0.00389387481845915, 0.00388546544127166, 0.0184827093034983, 0.00128364970441908, -1.03768588815001e-05, 0.00109023263212293, -0.00156510551460087, -0.00343954912386835, 0.0289674513041973, -0.0187593605369329, -0.000470673723611981, 0.00918452255427837, -0.000502132810652256, 0.000916715711355209, 0.00708888145163655, -0.00130110536701977, -0.00424903118982911, -0.00101853674277663, 0.00353734893724322, 0.000425838137743995, -0.003512152004987, 0.00501851458102465, -0.00505800126120448, 0.00602959003299475, 0.00152842199895531, 0.0161731727421284, 0.00255726277828217, 0.00222252099774778, 0.00267260288819671, 0.0129376696422696, -0.00305292848497629, 0.0124804629012942, 0.00944083370268345, -0.00394497299566865, 0.00200528395362198, 0.00380332535132766, -0.00643083686009049, 0.00600552232936025, 0.00149607646744698 },
      { 0.0241706352680922, -0.00854111276566982, -0.0084362318739295, 0.00155422929674387, -0.011219228617847, 0.00598123762756586, -0.00842682085931301, -0.00132897915318608, 0.138037741184235, -0.00115944561548531, 0.00582201918587089, 0.00736778741702437, -0.0084771066904068, -0.0069909431040287, -0.00397836277261376, -0.00681614829227328, 0.00983233284205198, 0.00231115613132715, 0.000162587079103105, 0.00366881350055337, -0.042820431292057, 0.0192503072321415, 0.00197088019922376, -0.00254912883974612, 0.000171167848748155, 0.00187973841093481, 0.0109723703935742, 0.00372592732310295, 0.00796071346849203, 0.0109810652211308, -0.0104529550299048, 0.00308098108507693, 0.0103304460644722, 0.00187005230691284, 0.0119223045185208, -0.00651711830869317, -0.0066034747287631, -0.00566236814484, -0.00419989181682467, 0.0117298811674118, -0.000664377119392157, -0.00632491707801819, 0.00421496154740453, 0.0118063194677234, 0.00403941189870238, 0.0179560203105211, 0.00417023291811347, 0.0034647670108825, 0.00788921117782593, 0.00876558478921652, -0.000413373607443646 },
      { -0.0159037541598082, 0.0101315528154373, 0.00999878533184528, -0.00584046030417085, 0.0121565330773592, -0.00524913100525737, 0.0163487214595079, -0.00129183777607977, -0.130325853824615, 0.00306715653277934, -0.00338942231610417, -0.0052008181810379, 0.00360536854714155, -0.00284339557401836, -0.00714190257713199, 0.00715671200305223, -0.00643592467531562, -0.00922902394086123, 0.00822501536458731, -0.00328209297731519, 0.0208654887974262, -0.0364398844540119, -0.00167583383154124, 0.00197151978500187, 0.00445178803056479, 0.000589685281738639, -0.00679202610626817, -0.00364657933823764, -0.0168666411191225, -0.0151200070977211, 0.00663328170776367, -0.00645291665568948, -0.0105381142348051, -0.00391995534300804, -0.00304826814681292, 0.00669667730107903, 0.00699702091515064, 0.000170746381627396, 0.00621790578588843, -0.00602288683876395, -0.00907915271818638, -0.00167780462652445, -0.00130787317175418, -0.0134806176647544, -0.0238715503364801, -0.0118432156741619, -0.00318538444116712, -0.00246179802343249, -0.00701367715373635, -0.0101786497980356, -0.011543289758265 },
      { -2.73233819007874, -7.39250993728638, -2.589768409729, 5.60628604888916, 0.497821062803268, -5.88011074066162, -4.44139289855957, -16.1223068237305, -37.7767906188965, -2.72757196426392, -6.67591953277588, -6.00887823104858, -2.9945867061615, -1.53486883640289, -10.0940027236938, -0.908351898193359, -5.95108079910278, -0.770017862319946, 1.91191828250885, 2.49559688568115, 1.00378465652466, -4.68840169906616, -2.2626678943634, 71.8133850097656, -0.751812219619751, -7.97007465362549, 40.1365928649902, -3.31138491630554, 0.153638362884521, -8.36728763580322, -5.11956262588501, -0.776113271713257, -1.52161431312561, -2.51331305503845, 3.10317707061768, -27.1852073669434, -2.91665148735046, -4.38505411148071, -4.17811155319214, -2.4609215259552, -5.9576244354248, -6.97857284545898, -3.21034598350525, 1.18858790397644, -2.25379180908203, -5.34168672561646, -6.4182333946228, -6.49510049819946, -8.71828079223633, -4.39861154556274, -6.6911358833313 }
    },
    {
      { 0.0699474513530731, 0.000447254104074091, -0.0407416261732578, 0.109037727117538, 0.138975769281387, 0.0626815110445023, 0.0719314143061638, 0.0819588676095009, 0.00657040812075138, 0.010160319507122, 0.0818821415305138, 0.00450140377506614, 0.101358443498611, -0.0595822855830193, 0.0712922737002373, -0.0693788453936577, -0.0668748319149017, -0.0788240730762482, -0.108458809554577, 0.0802537873387337, -0.0804312229156494, -0.00327873765490949, -0.0767797008156776, -0.203405871987343, -0.0206879805773497, 0.0599006414413452, -0.135099396109581, 0.0880259945988655, 0.0509278066456318, 0.0800512358546257, 0.0337705835700035, -0.0837195888161659, 0.0167210325598717, 0.138540223240852, 0.0325776040554047, 0.0329815670847893, 0.0560623109340668, 0.0448647513985634, 0.00958298426121473, 0.119366012513638, -0.0101833716034889, 0.0746732726693153, -0.0273828133940697, -0.0532403290271759, -0.082334391772747, 0.161178395152092, 0.0701030418276787, 0.0659445822238922, 0.0228011813014746, 0.141220450401306, 0.077989436686039 },
      { 1.02600061893463, 1.15212750434875, 1.28150522708893, 0.92971533536911, 1.0959540605545, 1.14169919490814, 1.40715777873993, 1.19450855255127, 0.96418970823288, 1.15135836601257, 1.10282611846924, 1.28396201133728, 0.762278079986572, 1.17661356925964, 1.07256472110748, 1.34094715118408, 1.25308787822723, 1.302414894104, 1.33265054225922, 0.734473645687103, 1.33053588867188, 1.20554494857788, 1.31397187709808, 1.58230090141296, 1.41263699531555, 1.28640818595886, 1.4090461730957, 1.04937934875488, 1.02492046356201, 1.13035571575165, 1.36308360099792, 1.26588797569275, 1.04980278015137, 1.07850456237793, 1.11695182323456, 1.40287029743195, 1.08402872085571, 1.18093872070312, 1.24753546714783, 0.879064083099365, 1.20704460144043, 1.14255368709564, 1.17804074287415, 1.22863602638245, 1.26846659183502, 0.722107887268066, 1.05407285690308, 1.11057531833649, 1.03218936920166, 1.05523979663849, 1.05947875976562 },
      { -0.0934537127614021, -0.157595589756966, -0.267203390598297, -0.110333815217018, -0.170836001634598, -0.214841410517693, -0.382690697908401, -0.307782083749771, 0.0262081567198038, -0.19554503262043, -0.196970388293266, -0.263910949230194, 0.174293160438538, -0.118782959878445, -0.151674211025238, -0.309660404920578, -0.233033999800682, -0.280432939529419, -0.237795770168304, 0.176424607634544, -0.307987928390503, -0.159538105130196, -0.246647283434868, -0.551170527935028, -0.387298136949539, -0.309214949607849, -0.287589460611343, -0.135453581809998, -0.0712859258055687, -0.195779204368591, -0.372524291276932, -0.241470769047737, -0.0882823318243027, -0.221928298473358, -0.15166275203228, -0.425678610801697, -0.185803636908531, -0.214803844690323, -0.265599489212036, -0.0137641532346606, -0.189784228801727, -0.231313347816467, -0.171270608901978, -0.231429919600487, -0.184780791401863, 0.119156308472157, -0.151121571660042, -0.195159584283829, -0.077765665948391, -0.20574451982975, -0.0889297127723694 },
      { -0.00919922441244125, 0.00875081215053797, 0.0226037353277206, -0.00502727786079049, -0.0210420712828636, 0.0236859023571014, 0.0016931010177359, 0.0711172595620155, -0.00126600381918252, 0.0297678802162409, 0.022148197516799, 0.0056879660114646, 0.0174207352101803, 0.0202411822974682, 0.0606870390474796, 0.00294115347787738, 0.0109446831047535, 0.00746378721669316, 0.013196817599237, 0.0186908207833767, 0.0575248338282108, -0.0181050729006529, 0.00430421158671379, 0.143644288182259, -0.0177230648696423, 0.00569313671439886, -0.000692640198394656, 0.0228458307683468, 0.0134924985468388, 0.0178524982184172, 0.00114502443466336, 0.0145590025931597, -0.0083286976441741, 0.0456270687282085, 0.022174010053277, 0.0377510376274586, 0.0193433538079262, 0.0318591147661209, 0.0114089613780379, -0.00642452714964747, 0.0238712877035141, 0.0486545152962208, 0.00932602491229773, 0.0447979643940926, 0.0225604102015495, -0.0193474944680929, 0.0131425773724914, 0.0201018117368221, 0.00337696634232998, 0.0522585436701775, 0.0215576626360416 },
      { 0.039199061691761, -0.0154542988166213, -0.030188787728548, 0.134016811847687, -0.0300341714173555, 0.0142403105273843, -0.0130670927464962, -0.00294290552847087, -0.0293575637042522, -0.0104878898710012, 0.0219223760068417, 0.0217441394925117, -0.105524398386478, -0.0262026078999043, -0.0241316705942154, 0.0081799328327179, 0.0211371816694736, 0.0587459839880466, -0.00835048407316208, -0.0205529052764177, 0.0269462987780571, 0.0937554836273193, 0.00343960546888411, -0.0278804134577513, 0.077034093439579, 0.00222762580960989, 0.0112988380715251, -0.000348725327057764, 0.0093856593593955, 0.0206110682338476, 0.0219380185008049, 0.0216317791491747, 0.0684610903263092, -0.000293359335046262, -0.00947089865803719, -0.0145706748589873, -0.00753479357808828, -0.00724488869309425, -0.020259577780962, 0.0117324544116855, -0.0143582774326205, -0.0168827250599861, -0.00569304265081882, 0.0368052423000336, 0.0113414572551847, 0.0302562937140465, 0.022397618740797, 0.00865807291120291, 0.0118362959474325, 0.0147813484072685, -0.0102261854335666 },
      { -0.0263668783009052, 0.0192821528762579, 0.0463058613240719, -0.0528238378465176, 0.0393768921494484, -0.0125127378851175, 0.00827372074127197, 0.00327711459249258, 0.045925110578537, -0.0107876351103187, -0.0167295336723328, -0.0202714037150145, 0.0408728793263435, 0.00021118980657775, -0.00211111106909811, -0.00784134771674871, -0.0076655219309032, -0.104210443794727, 0.00594167876988649, 0.0176056921482086, -0.0300096720457077, -0.131663575768471, -0.00500046461820602, 0.0210866536945105, -0.0731827095150948, 0.0145559692755342, -0.0125069431960583, -0.00643422920256853, -0.0236440394073725, -0.0258950684219599, -0.0182703491300344, -0.0481427237391472, -0.0477761030197144, -0.00770697509869933, 0.00380701757967472, 0.00649250065907836, 0.0400750003755093, -0.00548169482499361, 0.0475634709000587, 0.00490669393911958, 0.000151223051943816, 0.0136280534788966, 0.0104220472276211, -0.0259107425808907, -0.048474594950676, -0.0206999089568853, -0.0069719422608614, 0.000278894061921164, -0.00336774089373648, -0.0247768983244896, -0.0206688735634089 },
      { -2.66025114059448, -8.59470748901367, -19.4411449432373, -44.3928833007812, 11.128360748291, -15.8574409484863, -16.4375686645508, -32.7555885314941, -41.3234748840332, -9.21890354156494, -19.8372917175293, -27.1643180847168, 21.6735363006592, 9.87426376342773, -23.164571762085, 6.77121591567993, -14.8347463607788, -3.12947797775269, -5.18421220779419, -33.2651023864746, -9.20868015289307, 13.0761156082153, 0.231784209609032, -69.2340774536133, -1.22886693477631, -22.9733638763428, -7.34850692749023, -6.27172660827637, 2.72335982322693, -17.7033233642578, -22.1642284393311, -2.78287863731384, 5.53763723373413, -9.26263618469238, -5.32645320892334, -86.2315826416016, -10.9668302536011, -6.30227184295654, -12.2211246490479, 13.9573698043823, -16.3404197692871, -18.1849327087402, -2.2060341835022, -10.8741273880005, 3.72300267219543, 18.4332485198975, -15.0583877563477, -16.23073387146, -15.7103319168091, -13.5120191574097, -12.0921716690063 }
    },
    {
      { 0.040469940751791, 0.0387555658817291, 0.288461983203888, 0.113313846290112, 0.0833131968975067, 0.0640221834182739, 0.163211911916733, 0.0720557123422623, 0.0292535405606031, 0.169972598552704, 0.0737764611840248, 0.222514912486076, 0.10999421775341, 0.0115104680880904, 0.046854741871357, 0.160685956478119, 0.0316922925412655, 0.310383081436157, -0.0300918836146593, 0.112132951617241, -0.0166824236512184, 0.0370767302811146, 0.222216352820396, -0.0272125639021397, 0.117313914000988, 0.201745718717575, 0.0112154362723231, 0.0891489312052727, 0.0403071753680706, 0.0716152340173721, 0.18885526061058, 0.131980180740356, 0.0454945713281631, 0.0545134693384171, 0.0327259674668312, 0.0150084812194109, 0.1011121571064, 0.0460719875991344, 0.168971434235573, 0.131222441792488, 0.0670196264982224, 0.0508528463542461, -0.00201112288050354, -0.0083066551014781, -0.00645954813808203, 0.132538795471191, 0.0739712789654732, 0.0634488165378571, 0.0217163618654013, 0.0517871975898743, 0.045031551271677 },
      { -0.137764289975166, -0.127850651741028, -0.305727005004883, -0.421122014522552, -0.143710076808929, -0.149665668606758, -0.360726922750473, -0.105835251510143, -0.358216911554337, -0.104669965803623, -0.173064842820168, -0.27180027961731, -0.367715388536453, -0.12026309967041, -0.165055766701698, -0.205433964729309, -0.165109202265739, -0.10246392339468, -0.161565586924553, -0.5124232172966, -0.0245728846639395, -0.103880658745766, -0.343890577554703, 0.00876394938677549, -0.278671830892563, -0.354057312011719, -0.0488330163061619, -0.165204897522926, -0.146420389413834, -0.154474318027496, -0.233542129397392, -0.0649359673261642, -0.132928535342216, -0.112343892455101, -0.151051297783852, 0.00141677004285157, -0.0598927065730095, -0.102362982928753, -0.225946128368378, -0.353079050779343, -0.145630672574043, -0.12007150799036, -0.0845940858125687, -0.0784816741943359, -0.0621231533586979, -0.392188876867294, -0.0877657979726791, -0.161827504634857, -0.170338973402977, -0.101886436343193, -0.0911044031381607 },
      { 1.1310920715332, 1.11737930774689, 1.08329963684082, 1.32161641120911, 1.12297248840332, 1.12529063224792, 1.1565181016922, 1.04480111598969, 1.24305653572083, 1.03466773033142, 1.13256251811981, 1.13808298110962, 1.29638528823853, 1.1252030134201, 1.14471364021301, 1.02733659744263, 1.11490952968597, 0.978488862514496, 1.22022545337677, 1.43680083751678, 1.05069017410278, 1.10047781467438, 1.17061161994934, 1.00367426872253, 1.0257294178009, 1.14707505702972, 1.05522859096527, 1.11318778991699, 1.13405632972717, 1.13620126247406, 1.0125048160553, 0.923142910003662, 1.12580215930939, 1.07250261306763, 1.15243363380432, 0.997494518756866, 1.04342365264893, 1.10364949703217, 1.10862469673157, 1.22806763648987, 1.14223098754883, 1.08199965953827, 1.09812116622925, 1.07719767093658, 1.09235727787018, 1.27955281734467, 1.08392822742462, 1.13426077365875, 1.15516173839569, 1.06858682632446, 1.10814547538757 },
      { -0.0400353372097015, -0.0244103576987982, 0.00332659133709967, -0.042353194206953, -0.0416930057108402, -0.0305107515305281, 0.0146112106740475, 0.0103708729147911, -0.0081028463318944, -0.0149597693234682, -0.0289867632091045, 0.00555036682635546, -0.0888553261756897, -0.00563774118199944, -0.0141885075718164, -0.00455702096223831, -0.0119544919580221, -0.0135175734758377, -0.0224501509219408, -0.037770826369524, -0.016480827704072, -0.0250257011502981, -0.00364523916505277, 0.00164525571744889, 0.0560933984816074, 0.0121817346662283, -0.0194904990494251, -0.0221214648336172, -0.0175713002681732, -0.0317789874970913, 0.0198153518140316, -0.0114539703354239, -0.0210293903946877, -0.00614922540262341, -0.027780145406723, 0.00134070112835616, -0.0280914660543203, -0.0198276937007904, -0.00385403889231384, -0.0466321371495724, -0.0236891191452742, -0.0101629327982664, -0.0253562144935131, 0.00300700473599136, -0.00505124405026436, -0.0484422594308853, -0.0102523695677519, -0.028455451130867, -0.0181681085377932, 0.00122092023957521, -0.0155528392642736 },
      { 0.0252580586820841, 0.000247458257945254, -0.0102121224626899, 0.0694852992892265, -0.00554443942382932, 0.0138236107304692, 0.016402518376708, 0.000122758792713284, 0.00383221311494708, -0.0167077239602804, 0.0166637618094683, -0.000163506134413183, 0.0767252668738365, -0.0101805860176682, 0.0141893718391657, 0.0102162100374699, 0.01545216049999, -0.00996067468076944, -0.00207420811057091, 0.00372071587480605, 0.0118037238717079, 0.0357754267752171, 0.00957358255982399, -0.0136741660535336, -0.122992098331451, -0.0313824564218521, 0.00705202482640743, 0.0192153751850128, 0.00531324418261647, 0.0186022315174341, -0.0242973100394011, 0.0278176888823509, 0.0156710632145405, -0.000917295750696212, 0.00215012300759554, -0.00384780368767679, 0.0221830457448959, -0.000879704835824668, -0.0394972488284111, 0.0630018189549446, -0.0110945785418153, 0.00268808729015291, 0.0126792136579752, 0.00889095105230808, -0.0125453863292933, 0.067804828286171, 0.00826144311577082, 0.00836851913481951, -2.04882489924785e-05, 0.00890242587774992, 0.0127335339784622 },
      { -0.0150007922202349, 0.00336940935812891, 0.0174443908035755, -0.0460475571453571, 0.0111525738611817, -0.0115814190357924, -0.0695721507072449, 0.00207184627652168, -0.0259414575994015, 0.0191294569522142, -0.0120444651693106, -0.0315029434859753, -0.0197945702821016, -0.000865682901348919, -0.0198718924075365, 3.77043834305368e-05, -0.00496374862268567, 0.00700479978695512, -0.00974588748067617, -0.00451290234923363, -0.00720639480277896, -0.0435784906148911, -0.00435002101585269, 0.0151362651959062, 0.0918929427862167, 0.0620807334780693, -0.00450501497834921, -0.0217261463403702, -0.0126881310716271, -0.0222269091755152, -0.0251524578779936, -0.0323911346495152, -0.0222371127456427, -0.000282402004813775, -0.00479646911844611, -0.00115087535232306, -0.0300335139036179, -0.00555354775860906, 0.0613311976194382, -0.0226856078952551, 0.00425006030127406, 0.00397385098040104, -0.00179782649502158, -0.00204903422854841, -0.00707030249759555, -0.0407911874353886, -0.0324728637933731, -0.00569305801764131, 0.00551695330068469, -0.0137038975954056, -0.0342186354100704 },
      { -3.10243916511536, -5.11938047409058, -40.7057685852051, 29.2929592132568, -0.909285664558411, -6.94108533859253, -24.0233497619629, -16.2417221069336, 45.929988861084, -20.8784122467041, -8.0362377166748, -43.9895401000977, 6.61102533340454, -1.41011047363281, -14.0986881256104, -0.933007776737213, -13.1302795410156, -52.3622550964355, -8.09228992462158, 45.2353897094727, -1.79328870773315, -3.51952075958252, -30.1224002838135, 3.39114379882812, -48.6153221130371, -44.2673492431641, 18.9238243103027, 4.76648473739624, 0.79725307226181, -8.35221004486084, -28.3904590606689, -18.8614349365234, -7.44122648239136, -3.08257341384888, 0.36680656671524, -35.1159858703613, -17.334508895874, -3.8359546661377, -26.6256790161133, 15.6922960281372, -14.2551050186157, -7.16331481933594, -10.5868253707886, -6.09113216400146, -2.83337926864624, 4.81139183044434, -17.8199367523193, -6.49242973327637, -8.96585178375244, -6.01778554916382, -17.2531337738037 }
    },
    {
      { -0.00553128495812416, -0.139303922653198, -1.68574643135071, 0.566900193691254, -0.116556473076344, -0.0811835154891014, -1.0080953836441, -0.0635349079966545, -0.0879743248224258, -1.146036028862, -0.126263678073883, -1.61442410945892, 0.121778503060341, 0.00670274347066879, -0.219291195273399, -2.21228384971619, -0.0544343627989292, -2.5820837020874, -0.0529536977410316, 0.8121138215065, 0.0447324961423874, -0.147760257124901, -1.84566915035248, 0.541895508766174, -0.454405277967453, -1.60925960540771, 0.0138172646984458, -0.453642547130585, -0.0346159525215626, -0.0792337283492088, -1.48104178905487, -1.14703631401062, -0.185846775770187, -0.0261894576251507, 0.102349974215031, 0.0292922966182232, -0.44827276468277, -0.0164795443415642, -0.949978530406952, 0.223819941282272, -0.217409282922745, -0.0308145601302385, 0.122992664575577, 0.113008327782154, 0.033081416040659, 0.394925147294998, -0.287714093923569, -0.100326739251614, -0.192676618695259, 0.0459587350487709, -0.0505596324801445 },
      { -0.165374264121056, -0.0049798060208559, 1.57566964626312, -0.861450612545013, -0.0643801018595695, -0.0524429939687252, 3.34365010261536, -0.0909823253750801, -1.32086515426636, -0.119671531021595, 0.0437125414609909, 1.41203558444977, -0.894037663936615, -0.318811774253845, 0.169804438948631, 1.90381693840027, 0.184184223413467, 0.0187128894031048, 0.342711240053177, -0.75957065820694, -0.264653205871582, -0.0325290448963642, 2.27098155021667, -1.15008091926575, 2.32665920257568, 2.19989061355591, -0.381990611553192, 0.12356048822403, -0.145504549145699, -0.0259338896721601, 1.84719038009644, -0.224010288715363, -0.0813314318656921, -0.13990581035614, -0.147921785712242, 0.0414335913956165, -0.509933412075043, -0.109094217419624, 0.931085467338562, -0.658015191555023, 0.0445144139230251, -0.0601564683020115, -0.278336733579636, -0.301836639642715, -0.337482631206512, -1.15410053730011, -0.370427787303925, 0.00637349300086498, 0.0867267027497292, -0.266534030437469, -0.35538786649704 },
      { 0.868682861328125, 0.871690511703491, 0.649854362010956, 0.865329265594482, 0.900425910949707, 0.811168849468231, -0.553359806537628, 0.922485530376434, 2.43909382820129, 1.31501746177673, 0.778000295162201, 0.449514716863632, 1.26977622509003, 1.14796376228333, 0.701841711997986, 1.07137560844421, 0.811192035675049, 1.81820547580719, 0.481497883796692, 0.46021431684494, 1.02313494682312, 0.920293867588043, 0.148623898625374, 1.58791434764862, 0.562854290008545, 0.489669173955917, 1.1537172794342, 0.93611866235733, 0.922175526618958, 0.81231814622879, 0.86717677116394, 2.27158141136169, 0.844866096973419, 0.963413655757904, 0.791417300701141, 0.753373444080353, 1.24800443649292, 0.930762887001038, 0.706094861030579, 1.3262951374054, 0.891758441925049, 0.868354678153992, 0.95833545923233, 0.986450612545013, 1.08384132385254, 1.44026660919189, 1.11088061332703, 0.779980480670929, 0.868394315242767, 1.03094053268433, 0.961361229419708 },
      { 0.349308878183365, 0.288151532411575, 0.0386879928410053, 0.472043037414551, 0.25284543633461, 0.30712017416954, -0.110076569020748, 0.217532604932785, 0.0181990880519152, 0.236286878585815, 0.290206283330917, 0.0139515679329634, 1.29794061183929, 0.167025297880173, 0.297888219356537, 0.0605036653578281, 0.183996304869652, 0.185616225004196, 0.256981134414673, 0.373119235038757, 0.18105685710907, 0.223298415541649, 0.0787836834788322, 0.194546654820442, -0.549937427043915, -0.0259225890040398, 0.172759279608727, 0.246344611048698, 0.255279451608658, 0.280466765165329, -0.18836572766304, 0.172488778829575, 0.219122737646103, 0.221135124564171, 0.306413769721985, 0.207416638731956, 0.308342456817627, 0.188420087099075, 0.0790878981351852, 0.342385292053223, 0.228743985295296, 0.306065589189529, 0.275438964366913, 0.213881984353065, 0.166924148797989, 0.504041135311127, 0.161368131637573, 0.293301403522491, 0.229574382305145, 0.17453795671463, 0.21151277422905 },
      { -0.129443868994713, -0.0386823937296867, 0.163161262869835, -0.245914205908775, 0.0332471840083599, -0.0504982732236385, -0.441797256469727, -0.00739067187532783, 0.151984229683876, 0.104414395987988, -0.0354255028069019, 0.0227006133645773, -1.18872547149658, -0.0231745149940252, -0.109898410737514, 0.0198993515223265, 0.042721513658762, 0.0464177578687668, -0.0269937235862017, 0.23004350066185, -0.0327628329396248, -0.191556319594383, -0.0693306997418404, 0.0443525724112988, 1.17377912998199, 0.270584225654602, -0.0134098762646317, -0.121471710503101, -0.0392433889210224, -0.0250276569277048, 0.211156368255615, -0.299359977245331, -0.0537519566714764, -0.00825373735278845, -0.0480511896312237, -0.0320694372057915, -0.105012305080891, 0.0138670047745109, 0.290673732757568, -0.372004300355911, 0.0298689119517803, -0.0684060528874397, -0.0634859800338745, -0.0527247413992882, 0.0574536696076393, -0.463943123817444, 0.00772662600502372, -0.0212551429867744, -0.0149662867188454, 0.000740896910429001, -0.0822508931159973 },
      { 0.0823613181710243, 0.000259477645158768, -0.190429016947746, 0.305153250694275, -0.048624362796545, 0.030512060970068, 1.10537946224213, 0.0007047462859191, -0.0828567519783974, -0.130732923746109, 0.011384641751647, 0.301245272159576, 0.308840423822403, 0.020662946626544, 0.117476865649223, -0.148274153470993, -0.103327915072441, 0.0289461258798838, 0.0372315645217896, -0.0575992241501808, 0.0750737264752388, 0.230054974555969, 0.0193192027509212, -0.0940116569399834, -0.70242965221405, -0.525150954723358, 0.0203072410076857, 0.179093539714813, 0.038873765617609, 0.00747042754665017, 0.526103973388672, 0.407033890485764, 0.167228505015373, -0.0045000440441072, 0.0436265096068382, 0.0260712318122387, 0.178724616765976, -0.0143827171996236, -0.366760164499283, 0.150547832250595, -0.054975189268589, 0.00533397821709514, -0.00211678096093237, 0.0722689107060432, -0.00326859275810421, 0.302022784948349, 0.17764687538147, 0.00336126796901226, -0.0108596505597234, 0.0045588118955493, 0.216886550188065 },
      { -3.05968737602234, -2.25347328186035, 233.096282958984, -418.593017578125, -23.3297271728516, 12.5525960922241, 174.515686035156, -1.75808835029602, -232.482543945312, 117.319709777832, 16.444637298584, 286.5576171875, -30.623929977417, 28.7046680450439, 12.1859188079834, 16.4340648651123, 32.4420471191406, 429.601043701172, 4.25269079208374, -525.406555175781, -13.2288084030151, 39.2583389282227, 216.493927001953, 316.987976074219, 440.161193847656, 248.697448730469, 88.0431823730469, -28.8080539703369, -10.3106546401978, 10.9161672592163, 215.14958190918, 140.37272644043, 53.5787887573242, 1.75629043579102, -35.9123802185059, -157.02668762207, 45.2008285522461, 2.01128578186035, 105.341094970703, -107.931037902832, 29.1597213745117, -0.544696152210236, 50.8069458007812, -6.06479215621948, 3.67219924926758, -36.1334457397461, 44.8700332641602, 9.18795871734619, 0.157948955893517, 3.87554907798767, 63.1565933227539 }
    },
    {
      { 0.0161681193858385, -0.250279724597931, -0.205133438110352, 0.382497936487198, -0.229141220450401, -0.362457901239395, 4.06053638458252, -0.359842121601105, -0.258111208677292, -0.0427736714482307, -0.250228583812714, -0.0317737497389317, 0.0796362236142159, -0.0316031239926815, 0.0704733058810234, -0.943193852901459, -0.732865989208221, -1.05169057846069, -0.482859820127487, 0.503074526786804, -0.000412702385801822, -0.166230916976929, -1.80417501926422, 0.384296029806137, 2.37920546531677, 0.342118591070175, 0.567112803459167, 0.252959460020065, -0.0332681648433208, -0.338492065668106, 2.02617454528809, -0.772763013839722, -0.193129509687424, -0.12982065975666, 0.149711802601814, 0.10216910392046, 0.200425267219543, -0.116808049380779, -0.237302079796791, 0.146625354886055, -0.0380252338945866, -0.0983867421746254, -0.175618559122086, -0.0225081108510494, -0.146006494760513, 0.118763647973537, -0.105044409632683, -0.336812824010849, -0.233760744333267, -0.0263069998472929, 0.294877409934998 },
      { -0.119869224727154, 0.197769790887833, 2.33429193496704, -0.383078753948212, 0.150885909795761, 0.425002455711365, 6.72474098205566, 0.0836605057120323, -2.11864137649536, 0.766856133937836, 0.293604671955109, 1.968674659729, -0.377116322517395, -0.179770663380623, 0.26740974187851, 2.81037306785583, 1.56358349323273, 1.84884297847748, 1.87983906269073, -0.33789724111557, -0.0200245305895805, 0.327900528907776, 2.81485962867737, -0.0856194794178009, 5.86699342727661, 1.87832319736481, 0.0394173078238964, -0.116143621504307, -0.0694746375083923, 0.442208111286163, 3.49644351005554, 1.8777961730957, -0.178224459290504, -0.0360902920365334, 0.128239214420319, 0.0278527606278658, 0.403735637664795, -0.0013964407844469, 2.03586673736572, -0.358596712350845, 0.646477103233337, 0.162021592259407, 0.407549798488617, 0.0477870441973209, 0.143314063549042, -0.566987454891205, 0.160664692521095, 0.401280283927917, 0.651790738105774, -0.296696543693542, -0.217389285564423 },
      { 0.340162634849548, 0.294072687625885, -1.71718311309814, 0.167739704251289, 0.291032195091248, 0.0930305272340775, -6.32505178451538, 0.750772356987, 3.56839275360107, -0.718027889728546, 0.16298346221447, -1.44117140769958, 0.554814219474792, 0.5232013463974, 0.0976102575659752, -1.14058482646942, -0.582803606987, -1.25962579250336, -0.840850532054901, -0.0482519716024399, 0.275903880596161, 0.186673909425735, -0.940399408340454, 0.238826215267181, -5.82441997528076, -1.23604619503021, 0.248854175209999, 0.107321061193943, 0.414938658475876, 0.105731979012489, -4.20173740386963, -0.877187669277191, 0.438098788261414, 0.66277277469635, 0.0463922545313835, 0.124295324087143, -0.618213593959808, 0.333257734775543, -1.36595928668976, 0.542043089866638, -0.115270428359509, 0.481356590986252, 0.0881588086485863, 0.269633054733276, 0.27802562713623, 0.688425719738007, -0.0424390770494938, 0.125500574707985, 0.135854795575142, 0.731340527534485, 0.315940976142883 },
      { 0.890429377555847, 0.736318826675415, 0.723933160305023, 0.822213113307953, 0.74349057674408, 0.795440375804901, 0.440702110528946, 0.436837583780289, 0.0659849867224693, 0.869718074798584, 0.777094006538391, 0.645372927188873, 1.11677598953247, 0.714770615100861, 0.683590292930603, 0.711179614067078, 0.63674134016037, 0.717732012271881, 0.751800000667572, 0.749379277229309, 0.727649092674255, 0.695654928684235, 0.645080089569092, 0.753261387348175, -0.062768891453743, 0.705695390701294, 0.679357588291168, 0.836033821105957, 0.758715748786926, 0.753522932529449, 0.473004281520844, 0.766016125679016, 0.682539999485016, 0.551460325717926, 0.776088953018188, 0.788426280021667, 0.802503049373627, 0.665820181369781, 0.6115842461586, 0.788555204868317, 0.743743240833282, 0.70685487985611, 0.821438670158386, 0.700616419315338, 0.721147894859314, 0.849258482456207, 0.767199516296387, 0.774275064468384, 0.485264301300049, 0.46953147649765, 0.828325867652893 },
      { -0.157899379730225, -0.109521649777889, -0.115489915013313, -0.0620217323303223, -0.129254102706909, -0.15728884935379, -1.33176267147064, -0.0278412159532309, 0.551805794239044, -0.193133279681206, -0.152236953377724, 0.164169475436211, -0.524663269519806, -0.0378193221986294, -0.156877189874649, -0.14231051504612, 0.0591072887182236, 0.176837280392647, -0.129818797111511, 0.193723887205124, 0.00841110665351152, 0.246963858604431, -0.0465250387787819, 0.0139789190143347, 1.35319852828979, 0.00469214376062155, 0.0392914935946465, -0.191470637917519, -0.116857118904591, -0.163666248321533, 0.271609246730804, -0.1059939712286, 0.273285567760468, -0.00387182203121483, -0.109906069934368, 0.0193106159567833, -0.0669335275888443, -0.0390234552323818, -0.17907926440239, -0.198006838560104, -0.0557934567332268, -0.137890607118607, -0.159534618258476, -0.0299353078007698, 0.0659542307257652, -0.239447623491287, -0.0641137138009071, -0.129079923033714, 0.0612845718860626, -0.0068004815839231, -0.210597589612007 },
      { 0.0299199298024178, 0.0904150754213333, 0.277119994163513, 0.0887362360954285, 0.0941406488418579, 0.112288117408752, 2.31876659393311, -0.0101910373196006, -0.47918239235878, 0.11746271699667, 0.114246919751167, 0.0653630122542381, 0.0696436688303947, -0.0151492832228541, 0.11081700026989, -0.0261731296777725, -0.0266963988542557, -0.198694556951523, 0.0848002433776855, -0.0563038066029549, 0.0544327050447464, -0.38879781961441, 0.0301288738846779, -0.0561755709350109, -0.745763719081879, -0.00682714721187949, -0.0434162728488445, 0.227789849042892, 0.0561420470476151, 0.129883363842964, 0.733944952487946, -0.111350640654564, -0.145129546523094, -0.0288553144782782, 0.103444024920464, -0.0121540036052465, 0.266908973455429, 0.0654904618859291, 0.515440583229065, 0.0541556216776371, -0.0891918689012527, -0.0418920591473579, 0.0174427591264248, 0.0816870257258415, -0.100394830107689, 0.133018851280212, 0.195932194590569, 0.0801674053072929, -0.0803452283143997, 0.0349768064916134, 0.142452329397202 },
      { 2.39821171760559, -5.60296249389648, -201.202789306641, -188.184097290039, 30.758602142334, 34.3148345947266, -495.596801757812, 75.7850723266602, -319.617584228516, -35.7960739135742, 18.2689208984375, -285.525115966797, 45.1209182739258, 52.4062919616699, 64.499885559082, -106.361602783203, 17.2031383514404, 20.0847854614258, -139.416793823242, -201.933822631836, -31.7785472869873, 57.396656036377, -93.2794189453125, -48.6981544494629, 33.5141296386719, -321.032287597656, -914.391967773438, -48.4687385559082, -9.42468452453613, 40.4858245849609, -654.098754882812, -62.2000045776367, 49.9651870727539, 8.88720417022705, -73.1533966064453, -134.800765991211, -83.7133102416992, -12.11106300354, -91.1226806640625, -1.01859307289124, 33.6607131958008, 4.62635040283203, 40.6452865600586, -22.4499530792236, 31.9429054260254, 62.5809707641602, -38.7355880737305, 31.4324340820312, 37.3468856811523, 26.7039432525635, -26.5970401763916 }
    },
    {
      { 0.0940380245447159, -0.155274450778961, 0.910254180431366, -0.110663220286369, 0.0917174220085144, -0.313234150409698, 0.285828709602356, -0.369637161493301, 0.246203884482384, 0.858076095581055, -0.0217047035694122, 0.796505570411682, -0.0196909196674824, -0.0948025062680244, 0.373706966638565, 0.826735854148865, -0.762357831001282, 0.192252859473228, -0.470827847719193, -0.318263739347458, -0.0657362788915634, 0.0918976739048958, 0.260937333106995, 0.595941185951233, -0.0662732645869255, 1.70847988128662, 1.1877920627594, 0.923062443733215, -0.0197794865816832, -0.230561718344688, 0.524168610572815, -0.233887642621994, 0.186789691448212, -0.140293821692467, 0.104455903172493, -0.0425097234547138, 0.79210102558136, -0.0915030539035797, 0.651416003704071, 0.0248630009591579, 0.172228544950485, -0.0591536536812782, -0.109170340001583, -0.150464579463005, -0.172486335039139, -0.176500141620636, 0.403910547494888, -0.220834344625473, 0.0793420970439911, 0.0140479756519198, 0.556615829467773 },
      { -0.237422034144402, 0.0619138590991497, 1.0478230714798, 0.160362243652344, -0.0333339273929596, 0.285081177949905, 0.386853784322739, 0.116431124508381, -0.537435173988342, -0.122888557612896, -0.104164958000183, 0.792667508125305, 0.314986675977707, 0.121985889971256, -0.21969772875309, 0.415398716926575, 1.33441281318665, 1.2756187915802, 1.69384467601776, 0.216142311692238, 0.252891629934311, -0.0873556211590767, 0.122607544064522, 0.43433204293251, 0.453097313642502, 0.188286408782005, 0.161381006240845, -0.966206908226013, -0.103953190147877, 0.311821311712265, 0.74379301071167, 1.28799748420715, -0.875075101852417, -0.0380914211273193, 0.193151816725731, 0.045953031629324, -0.226326122879982, -0.0106682693585753, 1.03330826759338, -0.118166871368885, 0.392687618732452, 0.075424998998642, 0.135363027453423, 0.25426983833313, 0.348567277193069, 0.360982924699783, -0.488761812448502, 0.155080616474152, 0.147283509373665, -0.277647227048874, -0.728200435638428 },
      { 0.242340981960297, 0.192004710435867, -1.65044319629669, 0.13167043030262, 0.156933844089508, 0.0581337139010429, -0.271691799163818, 0.506769776344299, 0.186800017952919, -0.504527449607849, 0.218290194869041, -1.11061859130859, 0.0730021521449089, 0.0817079395055771, 0.183414444327354, -1.08842062950134, -0.644574642181396, -1.37159383296967, -0.870310127735138, 0.378162533044815, -0.186737686395645, 0.289468586444855, -0.220512971282005, -0.33314374089241, -0.899492979049683, -1.51966464519501, -0.158466950058937, 0.274248719215393, 0.265128821134567, 0.090760201215744, -0.798325538635254, -1.30771923065186, 0.772207021713257, 0.379907488822937, -0.184813722968102, 0.0337747447192669, -0.693812966346741, 0.0199038721621037, -1.41818189620972, 0.144072294235229, -0.124309696257114, 0.311434864997864, 0.1289282143116, -0.0969003960490227, -0.10176283121109, 0.000476657907711342, 0.0533359833061695, 0.132009744644165, 0.192403689026833, 0.389896333217621, 0.635526418685913 },
      { 0.983322143554688, 0.90293675661087, 1.12091374397278, 0.786462247371674, 0.822509229183197, 0.912636578083038, 0.973436176776886, 0.639531075954437, 1.03536915779114, 1.1553772687912, 0.93310809135437, 0.978540658950806, 0.255868434906006, 0.956687152385712, 0.907157123088837, 0.982941925525665, 0.837695837020874, 0.91922515630722, 0.942123889923096, 0.867701590061188, 1.02023017406464, 0.852792143821716, 1.00892066955566, 0.628452360630035, 0.954265415668488, 1.04362082481384, 0.803371012210846, 1.07536423206329, 0.917588531970978, 0.897967159748077, 1.05595207214355, 1.03769636154175, 0.848687291145325, 0.787788271903992, 0.923842072486877, 0.99564516544342, 1.01421320438385, 0.964505970478058, 0.939641237258911, 0.917019367218018, 0.959939301013947, 0.888199806213379, 0.981063842773438, 0.968885481357574, 0.954240500926971, 0.774311900138855, 1.03856217861176, 0.912013828754425, 0.673574805259705, 0.773808777332306, 1.07374215126038 },
      { -0.0157581549137831, -0.0747932568192482, -0.485135465860367, 0.124279849231243, -0.279313981533051, -0.0951956063508987, -0.279065191745758, -0.00353243388235569, -0.6619992852211, -0.339156419038773, -0.135060697793961, 0.0034368927590549, 0.528624892234802, -0.0814860016107559, -0.137193113565445, 0.0448271259665489, 0.108930058777332, 0.475713193416595, -0.118257418274879, -0.218202024698257, 0.0285600647330284, 0.584676086902618, 0.000817324558738619, -0.0465482547879219, 0.363795757293701, -0.244358494877815, -0.104888588190079, -0.155454099178314, -0.0878571420907974, -0.130200520157814, -0.443428665399551, 0.28576996922493, 0.441624879837036, 0.00380876753479242, -0.0712528377771378, 0.049143273383379, -0.118237130343914, -0.0784069001674652, -0.460045009851456, 0.0297308098524809, -0.141459301114082, -0.0897954627871513, -0.131675153970718, 0.03843729570508, 0.0652619749307632, 0.174796745181084, -0.0330635458230972, -0.121544897556305, 0.0775858089327812, -0.0372995808720589, -0.214367255568504 },
      { -0.0704966932535172, 0.0507217943668365, 0.77365916967392, -0.143207728862762, 0.285063683986664, 0.0594194047152996, 0.390700787305832, -0.0200903099030256, 0.872855246067047, 0.140228375792503, 0.107627637684345, 0.0683779343962669, -0.112036369740963, 0.00653909193351865, 0.038364116102457, -0.125942975282669, -0.0537770204246044, -0.849869787693024, 0.0816779509186745, 0.0339446477591991, -0.0447166226804256, -0.848416686058044, 0.021696213632822, 0.0672619715332985, -0.568758428096771, 0.515092074871063, 0.048262819647789, 0.0960457548499107, 0.0367703139781952, 0.0573982186615467, 0.0817861780524254, -0.88388067483902, -0.418574810028076, -0.0171315707266331, 0.0804392248392105, -0.0503453649580479, 0.397868990898132, 0.118429280817509, 0.87145060300827, -0.0150178782641888, -0.0575533509254456, -0.0587468519806862, 0.00133544602431357, 0.00383651675656438, -0.128083154559135, -0.158024519681931, 0.0175238754600286, 0.0873478353023529, -0.0999593436717987, 0.0559082366526127, -0.0576285719871521 },
      { 8.84366798400879, -8.25189208984375, -412.724243164062, 174.284057617188, 127.386413574219, 36.889575958252, -26.6274719238281, 72.1143951416016, -163.063293457031, -103.930236816406, 1.82276487350464, -324.645843505859, 26.7787494659424, 16.6884441375732, 42.2947425842285, -60.716854095459, 17.6177444458008, -136.440841674805, -155.964874267578, 210.628677368164, -7.79973793029785, 51.8016471862793, -311.287658691406, 198.601409912109, -62.6863555908203, -277.323455810547, -1211.77917480469, -60.9547462463379, -3.41460275650024, 38.7265663146973, 33.268180847168, -78.0103073120117, 31.1625556945801, -0.600024878978729, -38.9484710693359, 17.4651756286621, -94.0782165527344, -55.7817230224609, -117.697967529297, 72.8115997314453, 24.3663692474365, -6.87899446487427, 29.8008995056152, -24.9542999267578, 23.9574813842773, 33.7465171813965, -32.562126159668, 24.8515129089355, 41.0835952758789, 15.8306274414062, -50.2744369506836 }
    },
    {
      { 0.00510082161054015, -0.019398083910346, 0.0910779535770416, 0.10356155782938, 0.00957555416971445, -0.0388740971684456, 0.174608170986176, -0.0486070849001408, 0.0339780040085316, 0.0720664858818054, -0.00647362368181348, 0.102654546499252, 0.0289799533784389, -0.00907632242888212, 0.0262155048549175, 0.0510975457727909, -0.11090425401926, 0.0531332232058048, -0.0272975005209446, 0.154114678502083, -0.0176976956427097, -0.00357164349406958, 0.0859438702464104, -0.0348165556788445, 0.00354540953412652, 0.261029452085495, 0.00694166170433164, 0.0950879007577896, -0.0156811475753784, -0.0317555442452431, 0.201316133141518, -0.0339990258216858, 0.0277194064110518, -0.0342423468828201, 0.0457428358495235, -0.0109185576438904, 0.0800815895199776, -0.0164563376456499, 0.0581613257527351, 0.027538750320673, -0.00416263891384006, -0.0148117681965232, -0.00517774512991309, -0.0214408822357655, -0.0234287828207016, 0.0641157776117325, 0.0307995546609163, -0.0305737406015396, -0.00218477915041149, -0.00342160719446838, 0.046344917267561 },
      { -0.0170798879116774, 0.0172185637056828, 0.108038544654846, -0.0918062627315521, -0.000472176616312936, 0.0275141056627035, -0.0851972624659538, 0.0178098976612091, -0.0742952451109886, -0.038055956363678, -0.0143746705725789, 0.0847072824835777, -0.106042921543121, 0.0194381605833769, -0.0219969917088747, 0.000593878154177219, 0.166307136416435, 0.166166737675667, 0.174983710050583, -0.0779295042157173, 0.045347698032856, 0.0269092041999102, -0.0478766448795795, 0.16691417992115, 0.0243387557566166, 0.0266701430082321, 0.0493722930550575, -0.119902715086937, 0.0140476301312447, 0.0367786586284637, -0.0631003081798553, 0.127420261502266, -0.105568677186966, 0.0310298781841993, -0.0266471840441227, 0.0153720248490572, -0.038368321955204, 0.00727273244410753, 0.111500211060047, -0.0138448141515255, 0.0460251979529858, 0.00661042658612132, 0.013586089015007, 0.039498221129179, 0.0492792353034019, -0.139046251773834, -0.0602344833314419, 0.0177835505455732, 0.0213706027716398, -0.00631943624466658, -0.0836684256792068 },
      { 0.022303169593215, 0.0108680548146367, -0.165218487381935, -0.0501765049993992, 0.0130246635526419, 0.0131605332717299, 0.0371190123260021, 0.0518179014325142, -0.00884731207042933, -0.0202663019299507, 0.0302069950848818, -0.100880958139896, 0.0258404109627008, -0.000592585362028331, 0.0232344213873148, -0.0495388731360435, -0.0677842944860458, -0.221371158957481, -0.113555185496807, -0.133362904191017, -0.0271680448204279, 0.00729317544028163, -0.00874746218323708, -0.110023878514767, -0.139211788773537, -0.208241671323776, -0.0397332757711411, 0.0412138476967812, 0.0129156894981861, 0.0159797351807356, -0.0303213577717543, -0.16480840742588, 0.0921326354146004, 0.0175170265138149, -0.0133832031860948, -0.00373127078637481, -0.0620057508349419, -0.0067075346596539, -0.154293462634087, 0.00886220764368773, -0.00335665233433247, 0.0333538427948952, 0.0166715942323208, -0.0147559391334653, -0.0203314758837223, 0.0555296614766121, 0.0137828746810555, 0.0167061146348715, 0.0186575725674629, 0.021219601854682, 0.080859512090683 },
      { 1.00133538246155, 0.997118771076202, 1.02102589607239, 1.04003059864044, 0.981808543205261, 0.993771255016327, 1.00616633892059, 0.964646100997925, 1.02644503116608, 1.03073501586914, 0.996399939060211, 1.00803625583649, 1.20467495918274, 0.994578182697296, 1.00214946269989, 1.0031042098999, 0.981828570365906, 1.01197278499603, 0.99931401014328, 1.02522230148315, 1.01187884807587, 0.987103402614594, 1.00741326808929, 0.99057400226593, 1.01760613918304, 1.00671136379242, 0.988472640514374, 1.02298355102539, 0.995181262493134, 0.991233050823212, 1.02864921092987, 1.01830220222473, 0.987393736839294, 0.985324203968048, 0.99893593788147, 1.00474655628204, 1.00754082202911, 1.00144445896149, 0.998920977115631, 1.01349151134491, 1.00176131725311, 0.992317199707031, 0.998884379863739, 0.997383117675781, 0.998211622238159, 1.05580496788025, 1.00681722164154, 0.994643867015839, 0.991340637207031, 0.988683581352234, 1.01100015640259 },
      { -0.00632668705657125, -0.00928095448762178, -0.0613124035298824, -0.0262231379747391, -0.0370536893606186, -0.0164255350828171, -0.0138769224286079, -0.00184764643199742, -0.132851228117943, -0.048168957233429, -0.0209084525704384, -0.00328237353824079, -0.218877568840981, -0.00543573265895247, -0.0277913063764572, -0.00704421522095799, 0.00930212996900082, 0.0813515782356262, -0.0172222312539816, 0.0541469305753708, -0.0117046972736716, 0.0627807751297951, -0.00273845135234296, -0.0393826104700565, 0.0437425374984741, -0.0397947058081627, 0.0113540664315224, -0.0240906309336424, -0.0145866610109806, -0.0186921898275614, -0.0555799268186092, 0.0350794158875942, 0.0437885858118534, 6.85315244481899e-05, -0.014833222143352, 0.00438674725592136, -0.0199800189584494, -0.00947562698274851, -0.0549931451678276, -0.0535474009811878, -0.0237762555480003, -0.00874549802392721, -0.0214712768793106, -0.000616198813077062, 0.00353624881245196, -0.0929550006985664, -0.00546344928443432, -0.0173800028860569, -0.0121687119826674, -0.013160414993763, -0.0301751717925072 },
      { -0.00538887968286872, 0.00120599393267184, 0.0846595764160156, 0.0406696014106274, 0.0383578650653362, 0.0099062267690897, -0.00775491865351796, -0.00144199701026082, 0.184115186333656, 0.00587630597874522, 0.0147594250738621, -0.011564252898097, 0.0458035916090012, 0.000168079030117951, 0.0113650802522898, 0.00584411295130849, -0.00424718391150236, -0.205954879522324, 0.0172464679926634, -0.0116498358547688, -0.00252171326428652, -0.0937406197190285, 0.00748223764821887, 0.0293276682496071, -0.10953126847744, 0.0910355225205421, -0.00669289659708738, 0.00493043614551425, 0.0079340199008584, 0.00614689709618688, -0.0554420165717602, -0.115795634686947, -0.0473997816443443, -0.00230836099945009, 0.0176373496651649, -0.0064050080254674, 0.0459219701588154, 0.0137169016525149, 0.0973250791430473, 0.0196689311414957, -0.00196355814114213, -0.00693519320338964, 0.0041591078042984, -0.000295539881335571, -0.0107224937528372, 0.060772567987442, 0.00486319418996572, 0.0106607927009463, -0.0102565977722406, 0.0113745210692286, 0.000631088798400015 },
      { -0.350468158721924, -0.857789754867554, -48.5056610107422, -75.6870346069336, 15.8804759979248, 6.41317701339722, -17.4247703552246, 9.29103374481201, -23.0124168395996, -12.9158563613892, 2.57969856262207, -54.3651962280273, -8.53565692901611, 1.13850975036621, 5.92332983016968, -10.6163206100464, 5.09269142150879, -33.4382362365723, -19.0844593048096, -92.3793182373047, 0.51538872718811, 3.66469836235046, -37.6792373657227, 9.24941062927246, -26.3118019104004, -51.2625045776367, 12.0392265319824, -14.3991594314575, -0.827172875404358, 7.44322395324707, -7.52160692214966, -16.3671932220459, 2.0118362903595, -4.14539861679077, -4.27471351623535, 0.728651285171509, -8.19477939605713, -7.51940250396729, -10.1553592681885, -24.1722946166992, 5.04181575775146, 0.489306837320328, 3.86782288551331, -0.0212693139910698, 1.24905025959015, -8.3881139755249, -1.0444028377533, 5.41647386550903, 8.6667537689209, -3.19726419448853, -4.39256572723389 }
    },
    {
      { -0.00950705539435148, 0.0868098437786102, 0.0530449524521828, -0.488147139549255, 0.201862186193466, 0.171770617365837, 0.533887684345245, 0.183852776885033, -0.141209527850151, 0.032551497220993, 0.148091524839401, 0.24163119494915, -0.1545140594244, 0.0456248410046101, 0.0862152799963951, -0.3900226354599, 0.226572930812836, 0.49365770816803, 0.0333681143820286, -0.755000293254852, -0.0211207680404186, -0.00327880890108645, -0.171771585941315, 0.125476509332657, 0.490047305822372, 0.0812859386205673, -0.113293491303921, -0.0532966032624245, 0.0346830785274506, 0.1740762591362, 0.419270664453506, 0.354237765073776, -0.00897122826427221, 0.10389556735754, 0.00235827942378819, -0.00198724144138396, -0.0267417952418327, 0.0680187344551086, 0.206755220890045, -0.137766763567924, 0.0504489466547966, 0.0984084233641624, -0.0394468493759632, -0.0118622174486518, 0.011715330183506, -0.296106517314911, 0.128220424056053, 0.16776867210865, 0.167465731501579, 0.0631433054804802, -0.0699658989906311 },
      { 0.0603752546012402, -0.043301597237587, -0.700429737567902, 0.419764310121536, -0.140406996011734, -0.123403958976269, -0.0198124200105667, -0.0566281378269196, 0.42400273680687, -0.0904316380620003, -0.126433685421944, -0.424773365259171, 0.568605601787567, -0.113553315401077, -0.16963666677475, -0.208793148398399, -0.518931567668915, -0.239283725619316, -0.516496956348419, 0.371169090270996, 0.0147130284458399, -0.0657220035791397, -0.0304195340722799, -0.0938680768013, -0.130243688821793, -0.486400991678238, -0.0432982221245766, 0.110618732869625, 0.0250381100922823, -0.16635924577713, -0.193634122610092, -0.267491430044174, 0.0989000499248505, -0.0104925604537129, -0.11234125494957, -0.015063252300024, -0.172306060791016, -0.0359630696475506, -0.711498975753784, 0.0834360122680664, -0.173744857311249, -0.049872238188982, -0.0458486378192902, -0.0190835446119308, -0.0278212632983923, 0.75189870595932, -0.119382597506046, -0.205325365066528, -0.346940100193024, 0.0130820805206895, 0.099492184817791 },
      { -0.0606191866099834, -0.102105244994164, 0.540403127670288, 0.237813144922256, -0.0620442740619183, -0.0358893275260925, -0.192965254187584, -0.251883059740067, -0.176382288336754, 0.00214807875454426, -0.0542005114257336, 0.242399618029594, -0.1837028414011, 0.00758833438158035, 0.00626431498676538, 0.513077735900879, 0.180022194981575, 0.148834273219109, 0.347070574760437, 0.662387549877167, -0.0493021309375763, 0.00257593812420964, 0.108482018113136, 0.0446630157530308, -0.0600393861532211, 0.369877278804779, 0.0129738682880998, -0.0566130131483078, -0.110463723540306, -0.0410330295562744, -0.0653562769293785, -0.199795320630074, -0.0530559085309505, -0.177916690707207, 0.0190423857420683, -0.0242509990930557, 0.248732283711433, -0.117232874035835, 0.465720921754837, -0.0555528104305267, 0.0826591625809669, -0.167024478316307, -0.00252443086355925, -0.0325788706541061, -0.0362043604254723, -0.366042912006378, 0.123212687671185, -0.038853470236063, 0.0566553771495819, -0.159261643886566, -0.0120918313041329 },
      { 0.983336567878723, 1.06800925731659, 1.03103470802307, 0.807507932186127, 1.05042707920074, 1.05875205993652, 0.993501722812653, 1.18270349502563, 0.974943041801453, 1.00808441638947, 1.05744957923889, 1.00292229652405, 0.130644828081131, 1.05296647548676, 1.11497235298157, 1.01945352554321, 1.08923768997192, 1.00737309455872, 1.04513955116272, 0.9117711186409, 1.06279408931732, 1.03535187244415, 0.984618842601776, 1.01853108406067, 1.02829003334045, 1.00568163394928, 1.07922053337097, 1.00176858901978, 1.05322408676147, 1.06139969825745, 0.999641120433807, 1.03234994411469, 1.03732085227966, 1.11239886283875, 1.03635692596436, 1.04807376861572, 1.03143489360809, 1.12641775608063, 1.07060813903809, 0.950357794761658, 1.05534470081329, 1.08339893817902, 1.02776801586151, 1.07195913791656, 1.0425751209259, 0.732130467891693, 1.02425825595856, 1.09301519393921, 1.1483199596405, 1.14448535442352, 0.991702795028687 },
      { 0.0783978402614594, 0.0098544005304575, 0.0947898998856544, 0.153502196073532, -0.0264365635812283, 0.0341417081654072, -0.0766714289784431, -0.00451573077589273, 0.464753091335297, 0.153505340218544, 0.0438694879412651, 0.158650800585747, 0.881302952766418, -0.00930855702608824, -0.019042830914259, -0.0223749708384275, 0.0384120531380177, 0.0182450264692307, 0.0393328964710236, -0.314088046550751, -0.00978832598775625, -0.00899368710815907, 0.00734616816043854, 0.0335699208080769, -0.169192954897881, 0.139728978276253, 0.00488761067390442, 0.0841134637594223, 0.0298761539161205, 0.0510832108557224, 0.0193876475095749, 0.0121314292773604, -0.0340855605900288, 0.00314754131250083, 0.0167334545403719, -0.0228149406611919, 0.0527650564908981, -0.0158610045909882, 0.105707205832005, 0.230818241834641, -0.00174518069252372, 0.0417027696967125, 0.065873458981514, -0.0070161446928978, 0.00571767007932067, 0.438799470663071, 0.0579682551324368, 0.0291631165891886, -0.0133023671805859, 0.0181403085589409, 0.0933655574917793 },
      { -0.0466831885278225, 0.00212599639780819, -0.114036157727242, -0.205944746732712, 0.0420809052884579, -0.0332562848925591, 0.196180999279022, 0.0140683697536588, -0.715979814529419, -0.177940145134926, -0.0269843339920044, -0.227557957172394, -0.153853848576546, 0.0246253777295351, 0.00928522273898125, 0.0635134503245354, -0.0325375460088253, -0.0501516610383987, -0.0446213185787201, 0.0727103725075722, -0.00881695002317429, 0.053079005330801, -0.00921474862843752, -0.0101248398423195, 0.275578022003174, -0.159112155437469, -0.0185841750353575, -0.103036142885685, -0.0226853601634502, -0.0345629565417767, -0.00237498665228486, 0.0993273183703423, -0.00375300506129861, -0.00278889480978251, -0.00201227655634284, 0.0203500762581825, -0.13083977997303, 0.00491535756736994, -0.198560193181038, -0.0789093226194382, 0.00804470293223858, 0.00256184348836541, -0.0337765030562878, -0.015845226123929, 0.00921567622572184, -0.276023626327515, -0.12708754837513, -0.0153080774471164, -0.0036260059569031, -0.0418349467217922, -0.100869804620743 },
      { -3.94895839691162, -5.02114200592041, 42.4523315429688, 365.512664794922, 18.0113506317139, -29.5420627593994, 33.5854072570801, -45.0702590942383, 138.816665649414, 10.5636301040649, -26.1322708129883, 47.1542472839355, 26.6865825653076, -12.344123840332, -31.4489593505859, 19.9921188354492, -27.1134300231934, -33.6073188781738, 46.526538848877, 436.183197021484, 5.33108854293823, -5.39072608947754, 165.850967407227, -213.001556396484, 11.3769769668579, 35.988956451416, 59.3405151367188, 13.1267690658569, -3.69882440567017, -30.861780166626, 63.5525550842285, 8.41785430908203, -15.095871925354, -12.3287420272827, 36.3482856750488, 12.3346586227417, 8.17768573760986, 1.04278671741486, 2.34563446044922, 110.095664978027, -24.7207851409912, -10.8286151885986, -7.83109331130981, 5.78513288497925, -6.10740613937378, 23.220251083374, -8.46368503570557, -27.628755569458, -20.939640045166, -1.15320467948914, 0.027873357757926 }
    },
    {
      { 0.221861317753792, 0.050030030310154, 1.11370694637299, -0.547532081604004, 0.439813315868378, -0.0783578008413315, 1.79663026332855, -0.0990196913480759, 0.0065006329677999, 0.558640778064728, 0.0700766518712044, 1.03331303596497, -0.27096763253212, 0.111236028373241, 0.60030734539032, -0.0573627427220345, 0.401475936174393, 0.823980927467346, -0.14777284860611, -0.865634262561798, 0.205499812960625, 0.0267388690263033, 0.0404303595423698, 0.0251687746495008, 1.21472787857056, 1.22735357284546, -0.0570568703114986, 0.498003572225571, 0.121672071516514, 0.00277544208802283, 1.42277729511261, 0.487678050994873, 0.129406049847603, 0.192270368337631, -0.154949739575386, -0.146102398633957, 0.463585257530212, 0.0407678633928299, 0.963185250759125, -0.271611660718918, 0.509841501712799, 0.151998654007912, 0.0385866463184357, 0.166568547487259, 0.0107884109020233, -0.559976875782013, 0.556264102458954, 0.0214869789779186, 0.609348118305206, 0.128049835562706, 0.404436498880386 },
      { -0.310258895158768, -0.155201211571693, -0.81706702709198, 0.477411508560181, -0.16836316883564, 0.356111854314804, 0.292022854089737, 0.0920354798436165, 0.0291130691766739, -0.253904789686203, -0.0563761033117771, -0.542941272258759, 0.708619773387909, -0.67172384262085, -0.665766894817352, -0.0607476346194744, -0.422311723232269, 0.0522295944392681, -0.289701610803604, 0.415897130966187, -0.198059096932411, -0.149918109178543, -0.0743569433689117, -0.283408492803574, 0.481899678707123, -0.749520540237427, -0.0806778892874718, -0.46447029709816, -0.218977332115173, 0.03739845007658, -0.0482203885912895, -0.0541620180010796, -0.467423647642136, -0.271979153156281, -0.214188873767853, 0.202419623732567, -0.410824298858643, -0.00097986776381731, -0.765297174453735, -0.028870927169919, -0.284257918596268, -0.118927344679832, -0.339654296636581, -0.353323459625244, -0.115713641047478, 0.718235313892365, -0.743315160274506, 0.0421664640307426, -0.667471706867218, -0.36384379863739, -0.512909829616547 },
      { 0.159098938107491, 0.13306736946106, -0.0951651260256767, 0.339397698640823, -0.227196484804153, -0.325871199369431, -0.975699245929718, -0.0948900952935219, -0.0240403041243553, -0.28028792142868, -0.106228455901146, -0.140091136097908, 0.0847468674182892, 0.56781017780304, 0.291601240634918, 0.0847831591963768, -0.0375591218471527, -0.526760756969452, 0.516093373298645, 0.845652759075165, 0.0729868337512016, 0.205473259091377, 0.0412877053022385, 0.169535174965858, -1.26670408248901, -0.0742191597819328, 0.0739742442965508, 0.0643303096294403, 0.147838205099106, -0.283624857664108, -1.03379535675049, -0.673608839511871, 0.460707575082779, 0.154220968484879, 0.509683132171631, -0.11378575861454, 0.00420711142942309, -0.500846207141876, 0.0747585818171501, 0.264628916978836, -0.148468226194382, 0.047432504594326, 0.257833808660507, 0.104367025196552, 0.179224714636803, 0.0419994108378887, 0.347225606441498, -0.270554393529892, 0.285605490207672, 0.0872643887996674, 0.447955638170242 },
      { 0.00419024983420968, -0.0270173158496618, 0.0534496530890465, -0.37352454662323, 0.0876316055655479, 0.0921244323253632, -0.0393413715064526, 0.0946665108203888, -0.00468133157119155, 0.0573714785277843, 0.0614579990506172, -0.0321571156382561, -1.405717253685, -0.00457081478089094, 0.0393449366092682, -0.0059762392193079, 0.0760276094079018, 0.00812695920467377, 0.0413086749613285, -0.298071324825287, -0.0798703953623772, -0.0410192720592022, -0.00224711024202406, 0.0152052845805883, -0.0415362231433392, -0.00279914680868387, 0.0826649740338326, 0.0153557397425175, 0.0023120625410229, 0.178249448537827, -0.0232239160686731, 0.0198618452996016, -0.0475935712456703, -0.0139226065948606, 0.095946729183197, 0.0428337603807449, 0.0239939596503973, 0.43439257144928, -0.00293003837577999, -0.200845569372177, 0.0897590070962906, 0.0447832234203815, 0.0374263003468513, 0.122132189571857, 0.00201108446344733, -0.435376435518265, 0.115520797669888, 0.141790732741356, -0.0957457199692726, 0.0435486063361168, 0.0138940755277872 },
      { 1.1452442407608, 1.03041958808899, 1.06583690643311, 1.36387753486633, 0.634058356285095, 1.129714012146, 0.885165631771088, 1.02690672874451, 0.859994947910309, 1.0154904127121, 1.14291846752167, 1.32587277889252, 2.31143689155579, 1.11806499958038, 0.984505116939545, 1.09063971042633, 1.05987858772278, 1.22677874565125, 1.00378096103668, 0.830446779727936, 1.32301259040833, 1.31190323829651, 0.983767151832581, 1.10822057723999, 1.06350290775299, 1.13800287246704, 1.06446671485901, 1.07587039470673, 1.06681370735168, 0.97302120923996, 1.16415703296661, 1.18717896938324, 1.2259110212326, 1.08930611610413, 1.01004958152771, 1.14927065372467, 1.07616806030273, 0.904691100120544, 1.03891479969025, 1.41775918006897, 1.04761350154877, 1.03245723247528, 1.09328365325928, 1.2729754447937, 1.12766098976135, 1.57518982887268, 1.03165757656097, 1.02515745162964, 1.09429407119751, 1.0325767993927, 1.07200968265533 },
      { -0.201021909713745, -0.0275562349706888, 0.0302394349128008, -0.373576402664185, 0.368838608264923, -0.161576569080353, 0.237987071275711, -0.0639740899205208, 0.136970981955528, -0.0590067245066166, -0.114786021411419, -0.321048021316528, -0.344407856464386, -0.131497874855995, -0.0815758630633354, -0.125980377197266, -0.0341877602040768, -0.452532649040222, -0.0896797627210617, -0.00841785781085491, -0.321443140506744, -0.406954675912857, 0.0103467777371407, -0.139278993010521, 0.0275087617337704, -0.0641886740922928, -0.170837730169296, -0.0702849701046944, -0.106606878340244, 0.0581444539129734, -0.0683911144733429, -0.175489395856857, -0.276966840028763, -0.108349680900574, -0.145702615380287, -0.160460233688354, -0.065345898270607, 0.0830662623047829, 0.0356502123177052, -0.222520634531975, -0.0763993188738823, -0.0887786969542503, -0.140202581882477, -0.322971284389496, -0.226255297660828, -0.3802310526371, -0.201501667499542, 0.0151592753827572, -0.0835840478539467, -0.0383925661444664, -0.248467996716499 },
      { 2.05563545227051, -2.14503288269043, -165.41926574707, 496.519195556641, 232.201202392578, -47.2645568847656, -117.68627166748, -8.15823554992676, -30.1346645355225, -10.3293991088867, -80.0838165283203, -139.808853149414, 162.973999023438, 40.2096290588379, -6.54451274871826, 11.9967031478882, -30.0108699798584, -97.7083129882812, -59.436897277832, 655.141906738281, -21.7755699157715, 35.5775451660156, 22.9706573486328, 482.810516357422, -53.523567199707, -127.324119567871, 173.362976074219, -9.89118099212646, -1.19075202941895, -41.3295478820801, -118.343811035156, -24.1988677978516, 9.97845649719238, -5.77293920516968, -46.7215805053711, 8.52595806121826, -25.6056480407715, -78.356330871582, -46.1143951416016, 289.795349121094, -23.0664615631104, -30.4575309753418, 7.69859600067139, -36.2696418762207, 31.1194801330566, 208.742050170898, -7.83441257476807, -57.7108497619629, 1.70241713523865, 47.4776496887207, -21.4597015380859 }
    },
    {
      { 0.000411205517593771, 0.0150657445192337, -0.00183106306940317, -0.0579997971653938, 0.144774869084358, 0.0893037989735603, 0.0394500009715557, 0.103948093950748, -0.000824340793769807, 0.0239041950553656, 0.125216394662857, 0.0301004312932491, 0.0401479080319405, 0.0067406902089715, 0.0773936286568642, 0.0141013795509934, 0.0287620611488819, 0.0162779781967402, -0.0264480356127024, -0.0495126880705357, 0.00611872179433703, 0.0171157382428646, -0.0219662841409445, -0.00148031651042402, 0.0325810611248016, 0.0544765554368496, 0.0111556211486459, 0.0130739007145166, 0.0103982575237751, 0.130831256508827, 0.0277126729488373, 0.0105901900678873, 0.0123592196032405, 0.0519930943846703, 0.345017224550247, 0.000212219500099309, 0.0151666905730963, 0.105198994278908, 0.0245052166283131, -0.015958035364747, 0.108099669218063, 0.125013291835785, 0.00849458668380976, -0.00438682874664664, 0.0278622843325138, -0.00022259347315412, 0.0277908816933632, 0.0979552492499352, 0.0196224488317966, -0.0281927343457937, 0.0228400770574808 },
      { -0.0110614104196429, 0.016013951972127, 0.00946022663265467, 0.0226991772651672, -0.151553750038147, -0.0930474847555161, 0.0149192921817303, -0.0694102719426155, 0.00721397763118148, -0.0188205819576979, -0.162594646215439, 0.00184637971688062, -0.0543241910636425, 0.0580345019698143, -0.0438353270292282, 0.00426964927464724, -0.0489114560186863, 0.00251699308864772, 0.0740536972880363, -0.0266980845481157, 0.00591400498524308, -0.0522784478962421, -0.000916666991543025, 0.00516571011394262, 0.0224911049008369, 0.00352011690847576, 0.0150369219481945, 0.00450126314535737, -0.0222459938377142, -0.120654202997684, 0.0136037617921829, -0.00239598006010056, 0.0048606269992888, -0.053099013864994, -0.188270881772041, -0.0183682292699814, -0.0288268327713013, -0.106586366891861, 0.00867654196918011, 0.0175997000187635, -0.058646697551012, 0.0103108081966639, 0.00959073752164841, 0.00515455938875675, 0.000427266000770032, 0.0641971379518509, -0.0150935389101505, -0.158023461699486, -0.0530866086483002, 0.0743974894285202, 0.0359646864235401 },
      { -0.00174014153890312, -0.049538791179657, -0.0152155356481671, 0.051873192191124, -0.017393946647644, 0.0190396625548601, -0.0382882654666901, -0.139165222644806, -0.00193258712533861, -0.00906145665794611, 0.0292557924985886, -0.0344243682920933, -0.00217997981235385, -0.0598825477063656, -0.0263459570705891, -0.00959661323577166, -0.0269857123494148, -0.0290921963751316, -0.0515235140919685, 0.0864311456680298, -0.00550142815336585, 0.0255772285163403, 0.0083431713283062, -0.00393116706982255, -0.0461140424013138, -0.0318412557244301, -0.0200855322182178, -0.00952419638633728, -2.69976826530183e-05, 0.0324021726846695, -0.0379807129502296, -0.0111827244982123, 0.000377066142391413, 0.0063736904412508, -0.191057324409485, 0.00433756457641721, 0.0123501317575574, 0.013548425398767, -0.0233067125082016, -0.00876790285110474, 0.0836851075291634, -0.111267082393169, -0.027853537350893, 0.000439687661128119, -0.0306660160422325, -0.0650986209511757, -0.0194820668548346, 0.0576765388250351, 0.0577855743467808, -0.0592126250267029, -0.03140988945961 },
      { 0.00456675281748176, 0.0030564155895263, 9.796058293432e-05, -0.0139408577233553, 0.0537226125597954, 0.0107512678951025, -0.00205812626518309, 0.168859213590622, -0.000622451712843031, 0.00175507133826613, 0.0156878270208836, -0.00244523328728974, -0.048253022134304, 0.0118389343842864, 0.0277897324413061, 0.00035998469684273, -0.00727319763973355, 0.000950997404288501, -0.00590623868629336, -0.00881386827677488, 0.00677115051075816, 0.00639427592977881, 0.00177650153636932, 0.0112945772707462, -0.0020554456859827, -0.00363761791959405, 0.00807742681354284, -0.00469267275184393, 0.0124782361090183, -0.00677189091220498, -0.00122613122221082, 0.00175348331686109, 0.000213737424928695, 0.0140419630333781, -0.00749211711809039, 0.00678482092916965, -0.000743853044696152, 0.056304294615984, -0.00361413089558482, -0.00655271438881755, -0.0185380838811398, 0.0833787024021149, 0.00306680798530579, -0.00345599325373769, 0.0241426695138216, -0.0183662232011557, -0.0110578872263432, 0.0146645130589604, -0.0222262963652611, -0.0233712531626225, -0.00251971953548491 },
      { 0.0258708689361811, 0.0108482595533133, 0.0259961355477571, 0.0567264184355736, -0.0392973273992538, -0.0140775926411152, 0.0281695816665888, -0.0419520437717438, 0.0585192814469337, 0.0314738154411316, -0.0150635140016675, 0.0432154536247253, 0.162199974060059, 0.00307949702255428, -0.004052490927279, 0.00806586910039186, 0.0947198271751404, 0.0217084176838398, 0.0307825077325106, 0.0335025303065777, -0.0149826752021909, 0.00253565609455109, 0.0245533082634211, -0.0107962861657143, 0.0223024357110262, 0.038891214877367, -0.0140893897041678, 0.0266680102795362, 0.0158009193837643, 0.0137695400044322, 0.0206477958709002, 0.0184548255056143, 0.00805816240608692, -0.00770738488063216, 0.0623695738613605, 0.000798179826233536, 0.0311748255044222, -0.0308206025511026, 0.0375862456858158, 0.0789561867713928, -0.0015935564879328, -0.0518503747880459, 0.0447853170335293, 0.00221070903353393, -0.0313733592629433, 0.0917357876896858, 0.047016765922308, -0.0147382579743862, 0.0544381216168404, 0.0264702402055264, 0.023009842261672 },
      { 0.977982342243195, 1.0082950592041, 0.983614802360535, 0.948478102684021, 1.02453124523163, 1.00415277481079, 0.977069497108459, 0.995018422603607, 0.935893177986145, 0.969285011291504, 1.01932597160339, 0.962957203388214, 0.921719551086426, 0.994922161102295, 0.994189262390137, 1.01388466358185, 0.915624856948853, 0.993684887886047, 0.98009330034256, 0.975045204162598, 1.00454473495483, 1.00786769390106, 0.965641617774963, 1.01037323474884, 0.994675517082214, 0.961684107780457, 1.0225430727005, 0.977000832557678, 0.984218537807465, 0.979234278202057, 0.99768453836441, 0.991822898387909, 0.989306688308716, 1.00362539291382, 0.982913613319397, 1.00396585464478, 0.977231204509735, 0.98077380657196, 0.965429902076721, 0.948545396327972, 0.961801290512085, 0.998260021209717, 0.95713073015213, 1.002037525177, 1.0198563337326, 0.942762851715088, 0.974739909172058, 1.008460521698, 0.940037369728088, 0.99379026889801, 0.974464952945709 },
      { -0.128893360495567, -4.63058376312256, -2.8960075378418, 27.4140014648438, 34.2815780639648, 13.7071294784546, -1.97282242774963, 0.624226272106171, 0.811776578426361, -2.17361998558044, 7.43235015869141, -5.21949481964111, -20.6166896820068, -19.4262638092041, -5.52937173843384, -4.18611431121826, -3.15819215774536, -3.68882155418396, 1.99779856204987, 48.1151580810547, -1.12137174606323, 0.128604054450989, 11.4070796966553, -25.7445278167725, -2.78303647041321, -4.01878929138184, -32.417552947998, 2.30366706848145, -3.76400256156921, 5.25242185592651, -2.63923740386963, -2.81404137611389, -2.46140360832214, -1.80915248394012, 8.10323715209961, 15.5194797515869, -4.3799295425415, 56.1949806213379, -1.29428327083588, -10.8115930557251, -2.52588105201721, 25.32692527771, 0.195849061012268, -0.387489587068558, -5.90843486785889, -26.7600917816162, -7.37163591384888, 10.0107412338257, -10.3282470703125, -1.58174610137939, -5.16199636459351 }
    }
  },
  {
    {
      { 0.983287036418915, 0.988351345062256, 0.948216795921326, 0.978127181529999, 0.946706533432007, 0.96110999584198, 0.93360847234726, 0.930785477161407, 0.592195451259613, 0.906360745429993, 0.935167670249939, 0.968820452690125, 0.911403834819794, 0.974629402160645, 0.937114834785461, 0.928373038768768, 1.00735974311829, 0.984192252159119, 0.93999981880188, 0.932036399841309, 0.926354765892029, 0.985671520233154, 0.96104109287262, 0.943711698055267, 0.97110641002655, 0.986142158508301, 0.921929597854614, 0.990068197250366, 0.95394504070282, 0.944244861602783, 0.950223505496979, 0.966168522834778, 0.921666264533997, 0.956909656524658, 0.964275300502777, 0.937280893325806, 0.876898109912872, 0.98014771938324, 0.927029073238373, 0.953143000602722, 0.901076853275299, 0.938539028167725, 0.952286243438721, 0.974982261657715, 1.00347173213959, 0.959471642971039, 0.970371186733246, 0.948654115200043, 0.936703026294708, 0.945698380470276, 0.959066808223724 },
      { 0.0386371575295925, 0.0527846999466419, 0.0698574855923653, -0.0460120812058449, 0.0577471852302551, 0.0747496038675308, 0.0628287121653557, 0.0913006514310837, 0.995431244373322, 0.123374715447426, 0.0566198565065861, 0.0515774004161358, 0.139021828770638, 0.0420828424394131, 0.0729949772357941, 0.0977419838309288, 0.0381900332868099, -0.0388571843504906, 0.0780066698789597, 0.0711043402552605, 0.07848209887743, 0.022917028516531, 0.0525006130337715, 0.0666437968611717, 0.0537315905094147, 0.0658135041594505, 0.0680458396673203, 0.0407532043755054, 0.120831616222858, 0.0613677799701691, 0.0606674142181873, -0.0451116561889648, 0.069567084312439, 0.055192906409502, 0.0585149936378002, 0.0690347626805305, 0.169683530926704, 0.033262774348259, 0.122732564806938, 0.0558418929576874, 0.12539529800415, 0.0321716368198395, 0.0390757285058498, -0.0312076490372419, 0.0213897787034512, 0.0580204129219055, 0.0531954877078533, 0.0628255531191826, 0.0582210421562195, 0.0633448138833046, 0.0507863685488701 },
      { -0.0100540621206164, -0.0544171072542667, -0.0155473537743092, 0.0707068219780922, 0.000962431367952377, -0.0431010387837887, -0.00260927295312285, -0.0182439032942057, -0.253696829080582, -0.023233400657773, 0.00290032266639173, -0.019032122567296, -0.0474823378026485, -0.0110452128574252, -0.00246264692395926, -0.0246285330504179, -0.0358192548155785, 0.0466791801154613, -0.0155316991731524, -0.00320933340117335, 0.00159574148710817, -0.0112753063440323, 2.98228223982733e-06, -0.015488937497139, -0.0202626250684261, -0.0272349584847689, 0.0156561192125082, -0.00318913511000574, -0.0660129338502884, -0.004999412689358, -0.0097748190164566, 0.0780529230833054, 0.00354016595520079, -0.00772347627207637, -0.0133065525442362, -0.0132041992619634, -0.0703221932053566, -0.00447809463366866, -0.0462346337735653, -0.00274116592481732, -0.0230376180261374, 0.0165960043668747, 0.0198739431798458, 0.0564221777021885, 0.00961966253817081, -0.00457171536982059, -0.0202689971774817, -0.00271396455354989, 0.0164727158844471, 0.0246601272374392, 0.0106303272768855 },
      { 0.00234316405840218, 0.0352309010922909, -0.000569193973205984, -0.000270195334451273, 0.000972046342212707, -0.00160481373313814, 0.00202150992117822, -0.00106818962376565, -0.0318962335586548, 0.00608478300273418, 0.00178698496893048, 0.00363212684169412, 0.00821902696043253, 0.0119093703106046, -0.0075200404971838, -0.00558307068422437, 0.000517013308126479, 0.000167570615303703, 0.00317099620588124, 0.00160155980847776, -0.00548611069098115, 0.00405108742415905, 0.00140255538281053, 0.000728854851331562, 0.00553205795586109, 0.00896541494876146, -0.00141029513906687, -0.00128527102060616, 0.00779814692214131, 0.00346000562421978, -0.00122481468133628, -0.00248180376365781, 0.00237454916350543, -0.000236188832786866, -0.00153319200035185, 0.000347894994774833, 0.0331722572445869, 0.00423253141343594, -0.00646578287705779, 0.000342906801961362, 0.0081419562920928, 0.0104095665737987, -0.0142155010253191, -0.00203089276328683, -0.0111971907317638, -0.00278210872784257, 0.00698238331824541, 0.00023285998031497, 0.00148950610309839, -0.00731527945026755, 0.00235572783276439 },
      { 0.0019328179769218, -0.000803686038125306, 0.000737409223802388, -0.000581837783101946, -0.00608734972774982, -0.000534094695467502, 0.00237003271467984, 0.00483894813805819, 0.137294963002205, -0.0135153736919165, 0.0101648326963186, 0.00451838597655296, 0.0332232117652893, -0.00527188554406166, 0.00982507970184088, 0.0145221101120114, 0.0127175943925977, 0.0141310663893819, -0.00848497543483973, -0.00584101583808661, 0.00596300512552261, 0.00155654037371278, -0.0072124027647078, 0.00227688695304096, 0.00162490969523787, -0.00571131613105536, -0.00205987296067178, 0.0173437707126141, -0.00542455725371838, -0.00419742846861482, 0.000695259252097458, -0.00126023264601827, -0.00890270620584488, 0.00145700492430478, 0.00481971679255366, -0.00176390050910413, 0.0151126021519303, 0.00485205603763461, 0.00938812084496021, -0.000202564740902744, 0.00247656577266753, 0.0125191193073988, 0.0212054289877415, 0.0116752907633781, -0.0127294193953276, 0.00185761204920709, 0.00689127668738365, -0.00772746466100216, -0.00687590893357992, 0.0151449693366885, 8.2551319792401e-05 },
      { -0.00294668204151094, -0.00480650551617146, -0.00193184800446033, -0.00313884089700878, 0.00716760288923979, 0.0128610301762819, -0.00247146817855537, -0.00737121747806668, -0.129403755068779, -7.807544170646e-05, -0.0116064883768559, -0.00425213854759932, -0.056927964091301, 0.000597888138145208, -0.0084538497030735, -0.0122535629197955, -0.00947166979312897, -0.00590754020959139, 0.00787028949707747, 0.00714399432763457, -0.00803928729146719, -0.00308319157920778, 0.0101299537345767, -0.00646506808698177, -0.00359372585080564, -0.00408202549442649, -0.00247038109228015, -0.0254174266010523, 0.00401863409206271, 0.00452287076041102, 0.00464471243321896, 0.0061552906408906, 0.0114362332969904, -0.00135446281637996, -0.00461494270712137, 0.00920529756695032, -0.0404030904173851, -0.00752977980300784, -0.00623971410095692, 0.00180426333099604, -0.01737966760993, -0.0187957864254713, -0.0158124957233667, -0.0109002003446221, 0.0118117975071073, -0.00120617274660617, -0.00880198273807764, 0.015431753359735, 0.0061341468244791, -0.0346501730382442, -0.0102326711639762 },
      { -6.09722375869751, -11.8036241531372, -0.357008844614029, 2.85317921638489, -2.75676798820496, -0.162332445383072, -7.7587571144104, -3.3046817779541, -33.9985160827637, -0.716310560703278, -2.00609993934631, -4.86195516586304, 3.5936062335968, -5.65955066680908, -2.48300504684448, -6.12209463119507, -5.3881459236145, -3.82200336456299, -4.54961156845093, -3.86150765419006, -1.67552971839905, 81.5744934082031, -5.37480449676514, -0.974778890609741, -2.48948001861572, -6.09067344665527, -1.76984620094299, -8.90085983276367, -18.6810779571533, -6.24583005905151, -1.33017981052399, 9.13687133789062, -3.46326947212219, -1.39351511001587, -2.23754096031189, -5.37913417816162, 1.43183875083923, -6.87870645523071, 2.39633226394653, -6.30754709243774, -1.66945052146912, -10.4543466567993, -2.47996926307678, -0.649707555770874, 5.72441673278809, -7.35027551651001, -3.92691588401794, -2.75697708129883, -1.33434116840363, -2.57027077674866, -6.3579363822937 }
    },
    {
      { 0.0510551258921623, 0.056752935051918, -0.0594167560338974, 0.0292445924133062, 0.0389241687953472, -0.0820483341813087, 0.0502016805112362, -0.0371532067656517, 0.00684659369289875, -0.0457551293075085, 0.00187314196955413, 0.0646771639585495, -0.0602192468941212, 0.0481233447790146, -0.000935002055484802, -0.0572278052568436, 0.0212533120065928, 0.126436173915863, -0.0160241015255451, 0.0179409421980381, 0.00603616330772638, -0.125634640455246, 0.0155194150283933, -0.0757771208882332, 0.132518947124481, 0.0521738976240158, 0.0358721502125263, 0.0799153745174408, 0.0585096031427383, -0.0127393240109086, 0.0120540652424097, 0.0788538455963135, -0.0327148213982582, -0.0791476666927338, -0.0145272016525269, -0.0128061948344111, -0.047180287539959, 0.0644414275884628, 0.0545604228973389, 0.0431781895458698, -0.0772017315030098, -0.061082698404789, 0.0575082041323185, 0.15364134311676, 0.126943618059158, 0.00301137822680175, 0.139845952391624, 0.0816347002983093, -0.0717279762029648, 0.0308254696428776, 0.0631355196237564 },
      { 1.12964510917664, 1.20021891593933, 1.36129939556122, 0.943076133728027, 1.09672105312347, 1.30904173851013, 1.29049432277679, 1.22567975521088, 0.957179009914398, 1.14037919044495, 1.06319570541382, 1.13274145126343, 1.26762664318085, 1.16769969463348, 1.12767601013184, 1.26708912849426, 1.15828716754913, 0.888000428676605, 1.16716623306274, 1.2329888343811, 0.995983600616455, 1.52040553092957, 1.37151694297791, 1.32888376712799, 1.07107377052307, 1.16086745262146, 1.05169749259949, 1.10838663578033, 1.34884250164032, 1.15585696697235, 1.15901374816895, 0.80710381269455, 1.27710258960724, 1.31259548664093, 1.17616665363312, 1.36143374443054, 1.31771385669708, 1.12894880771637, 1.09242415428162, 1.08984339237213, 1.2486572265625, 1.14316189289093, 1.02792119979858, 0.80914044380188, 1.09901332855225, 1.15685153007507, 1.04875195026398, 1.39284861087799, 1.33163774013519, 1.05500018596649, 1.05667185783386 },
      { -0.19384928047657, -0.301225066184998, -0.310765385627747, 0.0157820545136929, -0.184897735714912, -0.215074002742767, -0.305564790964127, -0.19167771935463, 0.0296941008418798, -0.0926708355545998, -0.104296311736107, -0.196860581636429, -0.206448122859001, -0.191668942570686, -0.127211064100266, -0.235941916704178, -0.191839873790741, -0.0235313400626183, -0.170158416032791, -0.241705387830734, 0.0174253638833761, -0.480861872434616, -0.364486575126648, -0.248724222183228, -0.210592404007912, -0.21790324151516, -0.0909982547163963, -0.178526103496552, -0.408833712339401, -0.155034363269806, -0.19044017791748, 0.0324201099574566, -0.2493746727705, -0.242576152086258, -0.171075001358986, -0.337648242712021, -0.284302741289139, -0.184469521045685, -0.146962776780128, -0.12781423330307, -0.179179355502129, -0.135865896940231, -0.0835751220583916, 0.0583597384393215, -0.143238589167595, -0.15046514570713, -0.191471323370934, -0.361202776432037, -0.3072290122509, -0.0646308362483978, -0.0848637595772743 },
      { 0.0213357023894787, 0.0755629912018776, -0.00497126346454024, 0.0139068085700274, 0.0123904598876834, 0.0138708055019379, 0.00684495549649, 0.00830471143126488, -0.00133695476688445, 0.0266895610839128, 0.00697435950860381, 0.0194641072303057, 0.00728894537314773, 0.0218020416796207, -0.00151088018901646, -0.00195015734061599, 0.0153195690363646, -0.0279544498771429, 0.021173931658268, 0.0175768863409758, 0.00570754194632173, 0.0896467417478561, 0.00311580346897244, 0.0145747838541865, 0.0418579690158367, 0.0411915443837643, 0.0168787613511086, 0.024770038202405, 0.0432740561664104, 0.0155864413827658, 0.011798040010035, 0.0411935932934284, 0.00932832527905703, 0.00361702940426767, 0.0124890059232712, -0.00212158798240125, 0.0566491149365902, 0.0211337395012379, 0.0192679669708014, 0.0102809257805347, 0.019136618822813, 0.0358069501817226, -0.00932178366929293, -0.0251898635178804, -0.0371692143380642, 0.0110347205772996, 0.0509267635643482, -0.0104624265804887, 0.000132075438159518, -0.00593016808852553, 0.0222670957446098 },
      { 0.0153773883357644, 0.00208863499574363, 0.0476241447031498, -0.0198502540588379, 0.000955516763497144, -0.0141705609858036, -0.000192753650480881, 0.00724528403952718, -0.0288823395967484, -0.0439525879919529, 0.0667287185788155, 0.0204231198877096, 0.0510803088545799, -0.00686631631106138, 0.00715363444760442, 0.0391859002411366, 0.010057064704597, 0.0450370386242867, -0.0213304217904806, -0.032145019620657, -0.0437382645905018, -0.00451271655037999, 0.00384030770510435, 0.0163054000586271, 0.00112598482519388, -0.0181914996355772, -0.00299647566862404, 0.0226209759712219, -0.0181143842637539, -0.0110435504466295, 0.00665656151250005, 0.0477116107940674, -0.0283554960042238, 0.00237468047998846, 0.0139605291187763, 0.00357323349453509, 0.012495924718678, 0.0163918621838093, -0.00678161904215813, 0.00384277198463678, 0.00989590398967266, 0.0145942261442542, 0.0393455922603607, 0.0199822913855314, -0.0330829955637455, -0.00294103473424911, 0.0122640272602439, 0.0167917646467686, 0.0145404618233442, 0.0800594538450241, -0.00679304450750351 },
      { -0.0129633229225874, -0.0054265228100121, -0.0618803426623344, 0.0158008728176355, 0.0410321429371834, 0.00714977458119392, 0.0106214415282011, -0.012885476462543, 0.0452078990638256, 0.00813135970383883, -0.0588703192770481, -0.0215674005448818, -0.0806934610009193, -0.00408468162640929, -0.00320457527413964, -0.0244580712169409, -0.00508479960262775, -0.0115234721451998, 0.0220926851034164, 0.0465080067515373, 0.00884236674755812, 0.0163027234375477, 0.00405325321480632, -0.0918618589639664, -0.00722710974514484, 0.0105595020577312, -0.00654374109581113, -0.0321864411234856, 0.0132410796359181, 0.0113888755440712, 0.00586369633674622, 0.0074511943385005, 0.0662880763411522, -0.00356910121627152, -0.0167902782559395, 0.037388052791357, -0.0697459951043129, -0.0225673615932465, 0.00375346560031176, 0.000432137050665915, -0.0325214490294456, -0.0238288082182407, -0.0279225930571556, -0.0214532297104597, 0.0399210900068283, -0.00366559368558228, -0.0239052008837461, -0.0102556049823761, -0.00322351651266217, -0.109901584684849, -0.0202541723847389 },
      { -15.2734155654907, -23.8553905487061, 1.47772943973541, 12.3162775039673, -10.1436033248901, -8.22198104858398, -18.7714595794678, -3.76807379722595, -39.3296928405762, 8.92800903320312, 2.90585088729858, -16.4483585357666, 18.2039966583252, -10.2285900115967, -4.0641884803772, -15.3859891891479, -14.87668800354, 8.4199800491333, -9.00525188446045, -12.3702621459961, 20.3973140716553, -109.037185668945, -17.88014793396, -1.4915269613266, -9.37479114532471, -13.6005926132202, -2.78793597221375, -18.1185970306396, -73.9272232055664, -10.280421257019, -5.47097301483154, -55.3440208435059, -21.1992530822754, 1.74567353725433, -16.7227573394775, -23.1593837738037, -12.9171571731567, -16.9750518798828, -5.19464302062988, -13.0355796813965, 3.70898532867432, -22.8610782623291, -1.90694344043732, -14.3530893325806, 23.2852802276611, -13.620213508606, -12.4035367965698, -5.45647239685059, 6.77162790298462, 12.7226457595825, -11.7580537796021 }
    },
    {
      { 0.0619147941470146, 0.0592579580843449, 0.347881525754929, 0.118559345602989, 0.10354745388031, -0.0149501077830791, 0.213498711585999, 0.044752512127161, 0.0290508009493351, 0.0228255745023489, 0.0292501095682383, 0.061795212328434, -0.00444433279335499, 0.046229425817728, 0.016421752050519, 0.047577016055584, 0.0509143620729446, 0.149385467171669, 0.0210590939968824, 0.165391653776169, 0.0642963126301765, 0.000628140638582408, 0.22559829056263, 0.196923524141312, 0.0479609295725822, 0.0413981974124908, 0.016993435099721, 0.0729858800768852, 0.0235794764012098, 0.0314286798238754, 0.133924946188927, 0.102122515439987, 0.293699860572815, 0.235781028866768, 0.215550556778908, 0.298591613769531, -0.010097929276526, 0.063863143324852, 0.0354262329638004, 0.0470242351293564, -0.00662417383864522, -0.00429744180291891, 0.0359496772289276, 0.117883369326591, 0.0778062045574188, 0.0454576797783375, 0.0513285025954247, 0.166290491819382, 0.169657424092293, 0.033360056579113, 0.0429966226220131 },
      { -0.146944850683212, -0.0953746363520622, -0.118688620626926, -0.364417612552643, -0.04856888204813, -0.118283115327358, -0.381085783243179, -0.148767977952957, -0.32607689499855, -0.136311113834381, -0.105083666741848, -0.146110609173775, -0.0467024967074394, -0.0974263027310371, -0.0894545912742615, -0.166483372449875, -0.145894557237625, -0.343410432338715, -0.0712912678718567, -0.207752898335457, -0.211775347590446, -0.0343956537544727, -0.204888984560966, -0.0534277260303497, -0.095604307949543, -0.0976205393671989, -0.105987973511219, -0.146712526679039, -0.0120859369635582, -0.108299732208252, -0.0923739820718765, -0.466137170791626, -0.288823962211609, -0.347957402467728, -0.259915709495544, -0.374372899532318, -0.0315222628414631, -0.145085871219635, -0.141464844346046, -0.0609882362186909, -0.058632668107748, -0.125379249453545, -0.125023871660233, -0.437712281942368, -0.125702187418938, -0.113719798624516, -0.0944493114948273, -0.350801825523376, -0.184234380722046, -0.113535962998867, -0.0823394805192947 },
      { 1.12576711177826, 1.04295814037323, 0.858949840068817, 1.26315796375275, 1.03442299365997, 1.2044233083725, 1.13644516468048, 1.12462472915649, 1.21390104293823, 1.12929773330688, 1.10774827003479, 1.12455523014069, 1.0678790807724, 1.1003965139389, 1.10226547718048, 1.11210191249847, 1.13828706741333, 1.21061456203461, 1.09280788898468, 1.09698283672333, 1.16821944713593, 1.02715301513672, 1.00166594982147, 0.901241660118103, 1.0632358789444, 1.07426786422729, 1.11775350570679, 1.13094592094421, 0.999221801757812, 1.10397601127625, 0.992362320423126, 1.34561789035797, 1.08324611186981, 1.16545617580414, 1.09814274311066, 1.10874772071838, 1.05642330646515, 1.12944674491882, 1.13623893260956, 1.06023037433624, 1.0813136100769, 1.12947702407837, 1.12734758853912, 1.36305618286133, 1.11590051651001, 1.11225891113281, 1.06560111045837, 1.15912270545959, 0.980032622814178, 1.10823655128479, 1.09794640541077 },
      { -0.029748871922493, 0.00940264761447906, 0.0096942400559783, -0.026517853140831, -0.0198566131293774, -0.0121218487620354, 0.024689020588994, -0.0117733804509044, -0.00754177244380116, -0.00532381236553192, -0.0105567490682006, -0.0271031782031059, -0.0110697504132986, -0.024003678932786, -0.0302239265292883, -0.0173133332282305, -0.0353637561202049, -0.0642430484294891, -0.0138767883181572, -0.00515610771253705, -0.0691904053092003, -0.00855307281017303, -0.00389506830833852, -0.00218666484579444, -0.00946482922881842, -0.0129618579521775, -0.0216359794139862, -0.0333049334585667, 0.000661837810184807, -0.0217417012900114, -0.0252721849828959, -0.0339207202196121, -0.0103581771254539, -0.0028856738936156, -0.0140336146578193, -7.23228140486754e-06, 0.00280006579123437, -0.0310762971639633, -0.0307991430163383, -0.0150050120428205, -0.00722239445894957, -0.00687822559848428, -0.0381309986114502, -0.0489328727126122, -0.0441383011639118, -0.0352489538490772, 0.000125498947454616, 0.0316350758075714, -0.00173492380417883, -0.0192144978791475, -0.0168890859931707 },
      { 0.0119047220796347, 0.000262536865193397, -0.0418638624250889, 0.0428895689547062, 0.00748850684612989, -0.017924714833498, -0.0311793349683285, -0.00866369623690844, 0.00386841781437397, -0.0110287107527256, 0.00965298712253571, 0.0121942227706313, 0.00843377877026796, 6.61085914543946e-06, 0.0103124678134918, 0.0185918491333723, 0.0143628409132361, 0.0782417804002762, -0.0204332284629345, -0.0414020754396915, 0.067195400595665, -0.00435940688475966, 0.00825454667210579, -0.00435650674626231, 0.00138484127819538, 0.00348017015494406, -0.00413704803213477, 0.024803452193737, -0.000971536908764392, 0.000660822202917188, 0.0100414920598269, 0.0510528832674026, -0.019985331222415, 0.010086240246892, 0.0231709834188223, -0.0253116879612207, 0.00304152420721948, 0.0108401626348495, 0.00186203583143651, 0.00575678935274482, -0.00507984822615981, 0.00319605437107384, 0.0192845407873392, 0.0293340906500816, -0.00803687050938606, 0.0144760143011808, 0.00762644177302718, -0.0454611591994762, -0.00124676991254091, 0.0251694042235613, 0.0123707912862301 },
      { -0.0117647284641862, -0.00010337710409658, 0.0362162925302982, -0.0419408939778805, -0.0281635038554668, -0.0133298458531499, 0.043714452534914, 0.00560736563056707, -0.026389105245471, 2.30581827054266e-05, -0.0293626133352518, -0.0144018856808543, -0.0187693778425455, -0.00416604895144701, -0.0072498950175941, -0.00760291377082467, -0.0132531682029366, -0.0306399520486593, 0.0129662901163101, 0.0669629722833633, -0.011419827118516, 0.0051611247472465, 0.0115507999435067, -0.0233770459890366, -0.00161489238962531, 3.09723782265792e-06, -0.00191503192763776, -0.0308275427669287, -0.00234867772087455, 0.000294240599032491, 0.0440569333732128, 0.00939642079174519, 0.0581767894327641, -0.00487412698566914, -0.00830749701708555, 0.0893628671765327, -0.0303431041538715, -0.0129508292302489, -3.01119944197126e-05, -0.0116862719878554, -0.00547207240015268, -0.00949038472026587, -0.0152298947796226, -0.0286912824958563, 0.0122671667486429, -0.00656070047989488, -0.0135437324643135, -0.00613385206088424, 0.0134526491165161, -0.0324436724185944, -0.0301362536847591 },
      { -6.17744636535645, -8.87973117828369, -55.7704238891602, 14.429762840271, -17.4222354888916, -14.2383794784546, -49.6703720092773, -8.32763195037842, 48.0624694824219, -0.846945762634277, -8.80495262145996, -6.44264268875122, 1.70555412769318, -6.83427762985229, -2.51352739334106, -12.9487009048462, -6.87581443786621, 14.6344957351685, -13.9795141220093, -30.2831039428711, 2.59277606010437, 64.852294921875, -33.0685348510742, -29.6594066619873, -3.1508903503418, -5.68893718719482, -2.08640217781067, -8.93008804321289, -28.3936462402344, -5.60926914215088, -15.6370992660522, 46.3296775817871, -44.5491714477539, -34.6050605773926, -13.8191585540771, -48.2231903076172, -4.38442707061768, -6.23852348327637, -0.843660771846771, -19.0347366333008, -2.77197694778442, -8.55053520202637, -2.88093233108521, 32.0086479187012, -0.0276255197823048, -12.4505805969238, -5.65181159973145, -44.6645011901855, -0.412947654724121, -7.41749811172485, -17.4690380096436 }
    },
    {
      { -0.087231419980526, -0.0510270074009895, -2.68505120277405, 0.213758707046509, -0.473720252513885, -0.170975089073181, -1.70240449905396, -0.139337480068207, -0.0937265902757645, -0.0351980775594711, -0.111478917300701, -0.125438883900642, 0.0724941045045853, -0.0253355037420988, -0.044166412204504, -0.176404610276222, -0.203271225094795, 0.195481210947037, 0.093556821346283, -0.889673888683319, -0.165069371461868, 0.549368441104889, -1.8223592042923, -1.64047718048096, 0.00258039007894695, 0.00223607174120843, 0.0851513221859932, -0.0799306109547615, 0.0659193322062492, -0.114717371761799, -0.866431832313538, 0.718864738941193, -1.64218807220459, -1.91815447807312, -1.3057473897934, -2.19670462608337, 0.0326413065195084, -0.0973203182220459, 0.127070039510727, -0.146922141313553, 0.0619528256356716, -0.16642090678215, 0.00578717282041907, 0.575945258140564, -0.122481979429722, -0.176447376608849, 0.0595178529620171, -0.815060138702393, -2.1842930316925, -0.123010635375977, -0.0732573717832565 },
      { -0.014417395927012, -0.0924655795097351, 0.466091215610504, -0.952587306499481, -0.449349194765091, 0.33114966750145, 2.56884980201721, 0.137347534298897, -1.23300242424011, -0.20043408870697, -0.1523297727108, 0.0177376307547092, -0.332843124866486, -0.098392091691494, -0.137530982494354, 0.240059167146683, 0.00685739051550627, -0.613684892654419, -0.316562503576279, 0.872328460216522, 0.134672984480858, -0.699219346046448, 1.72699081897736, 0.00558557128533721, -0.185180455446243, -0.106385856866837, -0.265643030405045, -0.0133449779823422, -0.00656031584367156, -0.0357416458427906, 0.00512751890346408, -0.783246099948883, 1.48729300498962, 2.31944155693054, 0.945838153362274, 2.3426661491394, -0.294053286314011, 0.0107468916103244, -0.159799620509148, -0.420193284749985, -0.318037241697311, 0.00877553224563599, -0.168220430612564, -0.781821012496948, -0.0640065670013428, -0.0250962935388088, -0.282104820013046, 3.19888710975647, 1.76471924781799, -0.0108063146471977, -0.314219325780869 },
      { 0.788975715637207, 0.916194319725037, 2.34764313697815, 1.48239028453827, 1.21390748023987, 0.505585491657257, 0.474203109741211, 0.788293421268463, 2.37158203125, 1.08657503128052, 0.883233428001404, 0.780711352825165, 1.13942968845367, 0.923117935657501, 0.940468609333038, 0.791251063346863, 0.731629431247711, 1.27717137336731, 0.926597416400909, 0.776480078697205, 0.804356753826141, 1.3609414100647, 0.802874147891998, 2.20234489440918, 0.969199061393738, 0.890544414520264, 0.932842314243317, 0.801286220550537, 0.74124276638031, 0.883136570453644, 1.54412889480591, 0.91667252779007, 0.645959615707397, 0.135974690318108, 0.921729803085327, 0.533070206642151, 1.11714088916779, 0.764522790908813, 0.813032746315002, 1.20022141933441, 1.06184136867523, 0.856067001819611, 0.843527793884277, 0.788051426410675, 0.88556981086731, 0.89512687921524, 1.02330923080444, -0.475822865962982, 1.3150542974472, 0.883670568466187, 0.947574615478516 },
      { 0.281833648681641, 0.213537827134132, -0.0445431135594845, 0.310761719942093, 0.235801294445992, 0.180672645568848, -0.130123406648636, 0.187046900391579, 0.0185883902013302, 0.182632818818092, 0.169125765562057, 0.283082157373428, 0.144104212522507, 0.188659563660622, 0.241232335567474, 0.19046151638031, 0.332221865653992, 0.365678250789642, 0.203203082084656, 0.0549554526805878, 0.613956689834595, 0.0666119530797005, 0.0492311269044876, 0.0793220922350883, 0.237676620483398, 0.30321341753006, 0.278802961111069, 0.268285036087036, 0.220930904150009, 0.289870321750641, 0.265622645616531, 0.49025622010231, 0.110609404742718, 0.0665175169706345, 0.185604721307755, 0.0503177233040333, 0.15893717110157, 0.285533368587494, 0.290180653333664, 0.167157188057899, 0.187915056943893, 0.273043811321259, 0.331677943468094, 0.437023401260376, 0.263929814100266, 0.303892582654953, 0.170752584934235, -0.331594675779343, 0.0262804012745619, 0.188340753316879, 0.22150619328022 },
      { -0.0414932183921337, -0.00295511703006923, 0.33334094285965, -0.401404172182083, -0.00784795358777046, 0.0542910173535347, 0.297276824712753, 0.0935035943984985, 0.152153059840202, -0.0668185204267502, -0.0225919410586357, -0.0308910179883242, -0.0411453507840633, 0.0169454105198383, -0.0528184846043587, 0.0570847019553185, -0.0342701971530914, -0.341470837593079, 0.0429585911333561, 0.319837659597397, -0.50327342748642, 0.059960525482893, -0.116241924464703, -0.0587193220853806, -0.00343814352527261, -0.070813812315464, -0.0350326225161552, -0.04289660602808, -0.0536406673491001, -0.0380253158509731, -0.137405067682266, -0.195299491286278, 0.206364095211029, -0.0704615190625191, -0.0837070196866989, 0.212898477911949, -0.0552709028124809, -0.0316255055367947, -0.0257610883563757, 0.0224040448665619, 0.020370077341795, -0.0485463365912437, -0.10290353000164, -0.162200957536697, 0.045588243752718, -0.0693427100777626, 0.000890169001650065, 0.295793056488037, 0.146797761321068, -0.124387629330158, -0.0827454850077629 },
      { 0.0370217524468899, 0.000719600939191878, -0.14761920273304, 0.453346163034439, 0.177064269781113, 0.0210512299090624, -0.455869674682617, -0.107338652014732, -0.0834284722805023, 0.0333460234105587, 0.185031160712242, 0.0229737069457769, 0.048067893832922, -0.0184975247830153, 0.0476587414741516, -0.10511452704668, 0.0448828190565109, 0.130899846553802, 0.00462990254163742, -0.410524785518646, 0.0497010610997677, -0.0756026357412338, 0.101025559008121, 0.322685658931732, -0.0122847054153681, 0.0104651125147939, 0.0148217584937811, 0.0397262275218964, 0.0613065026700497, -0.0028761385474354, -0.214499920606613, -0.18981721997261, -0.418971866369247, 0.0255968552082777, 0.00117569684516639, -0.618810057640076, 0.0594279356300831, 0.0277415309101343, 0.0112853096798062, 0.0622634030878544, -0.0133872712031007, 0.0374626778066158, 0.0887473523616791, 0.191933184862137, -0.0607309117913246, 0.00975306052714586, 0.0101292347535491, 0.381986439228058, -0.264447867870331, 0.178235203027725, 0.196390226483345 },
      { 11.2851781845093, -10.3529319763184, 455.733795166016, -81.2650680541992, 47.2251815795898, 49.5678405761719, 284.1640625, 26.6254405975342, -230.283630371094, 22.5131130218506, 59.3451957702637, 15.8196096420288, -3.42253684997559, 9.15186309814453, -0.5370734333992, 33.9855461120605, 4.22313404083252, -132.79362487793, 50.5198364257812, 119.487823486328, -7.5855507850647, -128.411682128906, 269.635772705078, 237.02001953125, 0.68475466966629, -6.32992506027222, -4.78604698181152, 14.2655096054077, -141.728256225586, -2.32157492637634, 100.80110168457, -502.016265869141, 248.388748168945, 254.295684814453, 5.483971118927, 314.750366210938, 0.333280771970749, 10.5899868011475, -27.2408599853516, 50.307487487793, 4.13094615936279, -7.82826995849609, -4.28639650344849, -267.225036621094, -28.4663028717041, 23.9231986999512, 3.75244569778442, 379.720764160156, 13.4252195358276, 62.6947555541992, 62.653003692627 }
    },
    {
      { -0.359315425157547, -0.295525729656219, -0.379832088947296, 0.0707892328500748, 0.0942229554057121, -0.531907737255096, 0.224987879395485, -0.307746022939682, -0.273381680250168, -0.110005728900433, -0.226843729615211, -0.231751039624214, -0.145877942442894, -0.12789885699749, -0.0822040364146233, -0.884437024593353, -0.739747941493988, 0.0848317369818687, -0.00750014837831259, -0.0751122012734413, -0.133162334561348, 0.687800467014313, 1.21465158462524, -1.02147674560547, -0.080488845705986, -0.0400293879210949, 0.228858381509781, -0.318265050649643, 0.0665402263402939, -0.189025714993477, -0.137052580714226, 0.447604209184647, -0.0397122651338577, -1.98530077934265, -0.352800220251083, -0.210175573825836, -0.0150579148903489, -0.348596692085266, 0.239277616143227, -0.0498638637363911, -0.151734501123428, -0.288896143436432, 0.0314518548548222, 0.314812332391739, -0.204761326313019, -0.0214460119605064, -0.014268365688622, 4.85664319992065, -1.26706480979919, -0.03885168582201, 0.257485777139664 },
      { 0.427636742591858, 0.0584728568792343, 2.53647398948669, -0.416761338710785, 0.47774538397789, 1.9297993183136, 2.38109850883484, 1.16017401218414, -1.95909154415131, -0.0385663323104382, -0.0277056060731411, 0.248461484909058, -0.0492601133882999, -0.0127662299200892, 0.042986735701561, 1.64446175098419, 0.787544250488281, -0.328420966863632, 0.429250776767731, 1.95392072200775, 0.100287280976772, -0.592777132987976, 3.88079595565796, 2.64681506156921, -0.12405351549387, 0.101641431450844, -0.442360818386078, 0.420089900493622, 0.0708803907036781, 0.158882990479469, 1.2085942029953, -0.338381707668304, 2.27029299736023, 2.87895107269287, 0.83162260055542, 2.7696008682251, -0.0638356134295464, 0.395175486803055, -0.0292635764926672, 0.212009102106094, 0.112418740987778, 0.566643476486206, -0.108329102396965, -0.291786968708038, 0.119196720421314, 0.388621836900711, -0.32044443488121, 6.64915180206299, 2.63950872421265, -0.102507300674915, -0.181360349059105 },
      { 0.112888880074024, 0.729687929153442, -2.32016921043396, 0.625647842884064, -0.58423787355423, -0.768571197986603, -1.35767936706543, -0.366210013628006, 3.45477390289307, 0.47095113992691, 0.286878198385239, 0.151619523763657, 0.24874746799469, 0.316706627607346, 0.352354615926743, -0.596592366695404, -0.185945317149162, 0.563747525215149, -0.0252752006053925, -1.22460341453552, 0.434957027435303, 0.584976315498352, -4.36452770233154, -1.34810495376587, 0.672528922557831, 0.493742793798447, 0.49483847618103, 0.103797890245914, 0.0942826792597771, 0.324303448200226, -0.789344727993011, 0.190076097846031, -1.58865284919739, -0.929858982563019, -0.220173522830009, -2.15875816345215, 0.335203796625137, 0.137325063347816, 0.19674414396286, 0.0649181455373764, 0.301059037446976, 0.0928981304168701, 0.279121339321136, 0.139265060424805, 0.284710794687271, -0.014871884137392, 0.715228378772736, -5.383141040802, -0.860652267932892, 0.480599939823151, 0.29843932390213 },
      { 0.76286244392395, 0.420408219099045, 0.612496972084045, 0.765584707260132, 0.698302984237671, 0.747839510440826, 0.626423358917236, 0.61623203754425, 0.0663857311010361, 0.739062488079071, 0.719400227069855, 0.762560307979584, 0.870564520359039, 0.697497546672821, 0.744226276874542, 0.680848479270935, 0.806004822254181, 0.811338305473328, 0.775555491447449, 0.675313174724579, 0.825433075428009, 0.881097435951233, 0.657646179199219, 0.735791206359863, 0.596782326698303, 0.742316603660583, 0.791043817996979, 0.780843377113342, 0.803749144077301, 0.755427718162537, 0.773861706256866, 0.815994918346405, 0.656950294971466, 0.634370684623718, 0.760001122951508, 0.608172178268433, 0.726778984069824, 0.798568844795227, 0.747826337814331, 0.774649024009705, 0.742873251438141, 0.572401344776154, 0.86208164691925, 0.798158526420593, 0.757748007774353, 0.772139251232147, 0.433130204677582, -0.121587328612804, 0.644812047481537, 0.714839518070221, 0.840042412281036 },
      { -0.142629191279411, -0.0171374697238207, -0.0187618304044008, -0.242834031581879, 0.0530460104346275, -0.143432855606079, 0.110792681574821, -0.00900687277317047, 0.550278306007385, -0.0901762917637825, 0.282413572072983, -0.14539198577404, 0.104185976088047, -0.0472478047013283, -0.152520880103111, 0.131489828228951, -0.0347042493522167, -0.226972505450249, -0.184362158179283, -0.274272263050079, -0.265433132648468, -0.155099853873253, -0.0889830589294434, -0.120507806539536, -0.0011782554211095, -0.165171191096306, -0.0827963575720787, -0.176058724522591, 0.0101729920133948, -0.109631232917309, -0.139387026429176, -0.0415109321475029, -0.16683903336525, -0.0535917356610298, -0.0992943346500397, 0.0778963714838028, 0.0382452644407749, -0.132628202438354, -0.0620844922959805, -0.148609757423401, 0.022778432816267, 0.11710412800312, -0.134411856532097, -0.00817022006958723, -0.123910039663315, -0.169791355729103, 0.0114649292081594, 0.196243390440941, 0.070707343518734, 0.249311864376068, -0.190062865614891 },
      { 0.107297956943512, 0.00652918824926019, 0.49091038107872, 0.2168979793787, 0.292383313179016, 0.072515569627285, -0.185844287276268, 1.62180185725447e-05, -0.478305011987686, 0.0020997419487685, -0.18958654999733, 0.142220005393028, -0.129965022206306, 0.0661934688687325, 0.103799916803837, -0.16603285074234, 0.0623656399548054, 0.0704113319516182, 0.104705303907394, 0.47760197520256, -0.0383604876697063, 0.161715805530548, 0.49968808889389, -0.1874780356884, -0.0404561869800091, -0.0332807898521423, 0.0380598902702332, 0.127773746848106, -0.00698200147598982, 0.0451731234788895, 0.120188869535923, -0.12540690600872, 0.444716453552246, 0.0438319966197014, 0.111972101032734, 0.319707602262497, 0.00104615872260183, 0.065520204603672, 0.023769648745656, 0.183141142129898, -0.0612353645265102, -0.128513321280479, 0.0710890740156174, 0.0474401637911797, 0.0866285711526871, 0.111122339963913, 0.0458300411701202, 0.93346893787384, -0.0547021068632603, -0.419911324977875, 0.104338109493256 },
      { 38.2779693603516, 32.2711639404297, 57.1903610229492, 33.3863372802734, -83.2500686645508, -142.435516357422, -270.569854736328, 51.7593727111816, -317.216888427734, 47.5921745300293, 34.0214920043945, 18.2464847564697, 32.0160217285156, 5.25028419494629, -10.3551731109619, 19.483757019043, 9.32757472991943, 7.84110355377197, 21.1914024353027, -127.388236999512, 65.7425537109375, -780.778503417969, -516.211181640625, -2.25958871841431, 6.8916163444519, -4.35319185256958, -1.13970053195953, 42.6079177856445, -127.700492858887, -5.58507251739502, -65.1830978393555, -185.21565246582, -215.405654907227, -40.048526763916, -104.319564819336, -277.4638671875, -13.1534271240234, 30.3404903411865, -51.5415687561035, -23.1818904876709, 22.2751903533936, 29.3357639312744, -1.82715690135956, -110.270240783691, 34.9609527587891, 22.8718013763428, 24.2137641906738, -212.952377319336, -105.483505249023, 79.9363403320312, -27.0586566925049 }
    },
    {
      { -0.279424279928207, -0.320058733224869, 0.478883802890778, -0.0867227539420128, 0.650396704673767, -0.235363230109215, 1.72307300567627, -0.214010745286942, 0.242857933044434, -0.10227133333683, 0.0815608203411102, -0.00983322877436876, -0.241675227880478, -0.0901723355054855, -0.0348688624799252, -0.856381475925446, -0.353423088788986, 0.00909099075943232, 0.0756249874830246, 0.759428381919861, -0.039785198867321, 0.432937353849411, 0.742905378341675, 0.226032212376595, -0.113308504223824, -0.0427005141973495, 0.1360003054142, -0.180825755000114, -0.0777398571372032, -0.115808948874474, 0.878027141094208, -0.297422587871552, 1.03402519226074, 0.22003173828125, 0.294624507427216, 0.997744143009186, -0.067638985812664, -0.246690303087234, -0.0549597442150116, 0.262215942144394, -0.249298959970474, -0.169993415474892, 0.084097184240818, -0.177748307585716, 0.133224308490753, 0.14069539308548, 0.018527377396822, 1.07255446910858, 0.605241358280182, 0.358365386724472, 0.509986162185669 },
      { 0.244467437267303, 0.0967988073825836, 1.3526211977005, 0.291384130716324, -0.173778831958771, 1.45243728160858, 0.23308938741684, 0.977148950099945, -0.510085105895996, 0.107450008392334, -0.675711929798126, -0.104575157165527, 0.22927924990654, -0.0474454984068871, -0.105830974876881, 1.35977518558502, 0.326459139585495, -0.11020965129137, 0.139565423130989, 0.985607504844666, -0.0508160553872585, 0.282920211553574, 0.334141790866852, 2.28089308738708, -0.0631627067923546, 0.0777553245425224, -0.343416035175323, 0.22349239885807, 0.0882653445005417, 0.061799418181181, 0.121793061494827, 0.173028111457825, 1.04640030860901, 0.10378223657608, -0.166382119059563, 1.43995821475983, 0.223822444677353, 0.157302409410477, 0.214731112122536, -0.241388708353043, 0.34231299161911, 0.373417168855667, -0.221152439713478, 0.259417444467545, -0.0771543011069298, 0.0724026411771774, -0.281340032815933, -0.082850381731987, 0.531591415405273, -0.928834855556488, -0.700022757053375 },
      { 0.0955381616950035, 0.489286243915558, -1.94262433052063, -0.069867804646492, -0.681036651134491, -0.730656445026398, -1.67419159412384, -0.456212073564529, 0.1661436855793, 0.10938948392868, 0.505178689956665, 0.17719854414463, -0.260578721761703, 0.033459760248661, 0.25213885307312, -0.64186304807663, -0.128829285502434, 0.168222740292549, 0.085168868303299, -1.32583320140839, 0.33075413107872, -0.275567770004272, -0.768006443977356, -1.41265749931335, 0.359100490808487, 0.303989112377167, 0.320378631353378, 0.0899966135621071, 0.0175384543836117, 0.199171140789986, -0.841795086860657, 0.114184558391571, -1.55141043663025, -0.20043471455574, -0.0443051122128963, -2.08717918395996, -0.179107517004013, 0.164542317390442, -0.0455638356506824, 0.143305122852325, -0.0768769457936287, 0.0085928700864315, 0.204873755574226, 0.141874670982361, 0.22539971768856, -0.00740654300898314, 0.366897851228714, -0.328837037086487, -1.04701745510101, 0.820539653301239, 0.577401697635651 },
      { 0.908065319061279, 0.631645917892456, 0.889667451381683, 0.878205299377441, 0.917138993740082, 1.0160391330719, 1.08153343200684, 0.840350568294525, 1.03607928752899, 0.980195879936218, 0.960539162158966, 0.962868213653564, 1.17685663700104, 0.982918739318848, 0.926218330860138, 0.857084155082703, 0.956393539905548, 0.897260129451752, 0.97358113527298, 1.08026146888733, 0.709452152252197, 0.801180422306061, 0.981702029705048, 1.05164313316345, 0.823253095149994, 0.90565288066864, 0.921325087547302, 0.963957369327545, 1.00615727901459, 0.916099011898041, 0.985561549663544, 0.806657612323761, 0.973812997341156, 1.00874328613281, 1.02925992012024, 0.843013405799866, 1.02914929389954, 0.957474052906036, 0.913462460041046, 1.02194631099701, 0.9766885638237, 0.773854851722717, 0.975464761257172, 0.801696717739105, 0.774978578090668, 0.953231036663055, 0.756866812705994, 0.97310334444046, 0.96623158454895, 0.912703573703766, 1.09574663639069 },
      { -0.0967399850487709, 0.00089404807658866, 0.371409505605698, 0.108884453773499, 0.0146237360313535, -0.221955180168152, -0.229067146778107, -0.0400878600776196, -0.650607883930206, -0.130367457866669, 0.475210279226303, -0.155945643782616, 0.192326307296753, -0.0877617597579956, -0.110324047505856, 0.239910736680031, -0.0294366665184498, 0.0188190191984177, -0.223705649375916, -0.657685458660126, 0.0939679071307182, -0.08640968054533, -0.238080903887749, -0.0411239229142666, 0.00152352277655154, -0.107716858386993, -0.0195780508220196, -0.144167050719261, 0.0430046133697033, -0.0760572627186775, -0.014452108182013, 0.0468859672546387, -0.509180963039398, 0.000880120031069964, -0.132019579410553, -0.152730405330658, 0.113046258687973, -0.105800874531269, -0.033878818154335, -0.165894106030464, 0.0279039293527603, 0.12566339969635, -0.000639165635220706, 0.0607317537069321, -0.297033607959747, -0.135748893022537, -0.0247869621962309, -0.441241711378098, 0.0612447783350945, 0.531050264835358, -0.177200615406036 },
      { 0.0546629130840302, -0.00491552893072367, -0.409059911966324, -0.181060314178467, 0.434308409690857, 0.133614376187325, 0.401383459568024, 0.0122296502813697, 0.854058027267456, 0.0254132431000471, -0.521945774555206, 0.124587990343571, -0.248788878321648, 0.121935725212097, 0.074099637567997, -0.246414035558701, 0.0442716218531132, 0.000830891309306026, 0.0891685113310814, 0.883574604988098, -0.0448109842836857, 0.0845138356089592, 0.228986576199532, -1.26019787788391, -0.0215021427720785, -0.0554821565747261, -0.00950764957815409, 0.0483586303889751, -0.043917927891016, 0.0156154632568359, 0.196114435791969, 0.200059413909912, 1.09796059131622, 0.0275930296629667, 0.139879316091537, 0.919335782527924, -0.141045942902565, 0.0195470042526722, 0.0186780337244272, 0.0805561169981956, -0.0558796711266041, -0.14930684864521, -0.0510415099561214, -0.120996333658695, 0.297631353139877, 0.0684070587158203, 0.0617108829319477, 0.844439923763275, -0.0900289639830589, -0.817893028259277, -0.100653029978275 },
      { 35.7974700927734, 36.9868202209473, -79.8143844604492, 73.2079925537109, -93.5440521240234, -207.587326049805, -253.373886108398, 46.9208335876465, -159.787734985352, 17.4956340789795, 7.84659099578857, 7.08867740631104, 35.2005157470703, -40.1775360107422, -11.1457862854004, 12.5721626281738, 7.16256332397461, 92.6903991699219, 7.19202280044556, -158.968566894531, 27.5773048400879, 204.134353637695, 48.5999298095703, -63.8855781555176, -1.88934028148651, -6.17316436767578, 12.0032148361206, 34.2759094238281, 12.6429071426392, -9.24583053588867, -120.016723632812, 242.612594604492, -440.561950683594, -310.588043212891, -101.059646606445, -339.028350830078, -11.3699254989624, 23.4448299407959, -24.1519889831543, -8.72505950927734, 16.3227958679199, 22.6991024017334, 7.75694942474365, 94.5273666381836, 171.893493652344, 14.2831392288208, 12.6920032501221, -43.7143898010254, -51.1542854309082, 58.518424987793, -46.1992416381836 }
    },
    {
      { -0.0292745549231768, -0.0367226228117943, 0.0997787714004517, 0.0330979451537132, 0.0560160055756569, 0.0027194763533771, 0.2360949665308, -0.0537793971598148, 0.0283372104167938, -0.00829871837049723, 0.0127556966617703, 0.000249329634243622, -0.0250502470880747, -0.0125944800674915, -0.0127925695851445, -0.0963270589709282, -0.026209706440568, 0.0160519909113646, 0.000140266682137735, 0.0594391226768494, -0.0165711753070354, 0.011382813565433, 0.21255911886692, 0.0213920120149851, -0.0296721123158932, -0.0127066690474749, -0.00183063349686563, -0.0220090914517641, -0.00989232398569584, -0.0145861515775323, 0.0773809999227524, 0.120662070810795, 0.0995092317461967, 0.0766851603984833, 0.029385432600975, 0.125811159610748, -0.0129771344363689, -0.0274654496461153, 0.00953839160501957, 0.00486082304269075, -0.0297023244202137, -0.013207346200943, 0.00201658415608108, 0.0835864618420601, 0.0184074155986309, -0.0133551256731153, -0.00261115515604615, 0.208897396922112, 0.024580517783761, 0.0382683128118515, 0.0364735648036003 },
      { 0.0214916076511145, 0.0128140803426504, 0.191561043262482, -0.101583480834961, -0.030002411454916, 0.115123458206654, -0.00241096806712449, 0.119890473783016, -0.0606901273131371, 0.0132182138040662, -0.0741982981562614, -0.0189086366444826, 0.0341984815895557, 0.00277790729887784, -0.00343500822782516, 0.138033270835876, 0.0205345544964075, -0.00885592866688967, 0.00797950942069292, 0.0864337161183357, 0.0438034273684025, 0.0622270852327347, -0.14520999789238, 0.18732838332653, 0.0295127425342798, 0.00920033175498247, -0.0110229710116982, 0.0228358767926693, 0.0134611930698156, 0.0167723204940557, -0.0261796116828918, -0.0634421408176422, 0.0827988758683205, -0.0452278070151806, -0.0392932295799255, 0.118342839181423, 0.0346433594822884, 0.0143624870106578, -0.0133147714659572, -0.0257375463843346, 0.0457718670368195, 0.0206070858985186, -0.0110129537060857, -0.0711979866027832, -0.00634629838168621, 0.0194559451192617, -0.00359320221468806, -0.118844956159592, 0.0126272523775697, -0.0967828631401062, -0.0711927339434624 },
      { 0.0128841688856483, 0.0458826646208763, -0.260127484798431, 0.0612491071224213, -0.0522470138967037, -0.0826235637068748, -0.196282833814621, -0.049136471003294, -0.0129307778552175, 0.00408542808145285, 0.0554542317986488, 0.021082878112793, -0.0354160442948341, -0.00670502334833145, 0.0209928043186665, -0.0565287508070469, -0.0140784317627549, 0.0083401333540678, 0.0116005279123783, -0.123262904584408, -0.0247814226895571, -0.0482366569340229, 0.0365923084318638, -0.188435614109039, 0.00980356056243181, 0.0242052935063839, 0.0234347376972437, 0.0132398921996355, -0.00331964623183012, 0.00773571105673909, -0.0528377033770084, -0.0395011864602566, -0.13039056956768, -0.00735006388276815, 0.0110773965716362, -0.219728231430054, -0.0303485468029976, 0.0184199120849371, 0.00950544141232967, 0.0252988208085299, -0.0180793572217226, 0.00764035247266293, 0.0152935329824686, -0.0478248670697212, 0.0172091424465179, 0.0106512028723955, 0.0138475177809596, 0.0231847353279591, -0.0363133996725082, 0.0814165621995926, 0.0644750073552132 },
      { 0.995398342609406, 0.968571901321411, 0.993666768074036, 1.01378297805786, 0.995991885662079, 1.00886332988739, 1.02184557914734, 0.985557079315186, 1.02491223812103, 0.998553037643433, 1.00117838382721, 1.00180184841156, 1.01570880413055, 1.00565707683563, 1.0000376701355, 0.989867985248566, 1.00118041038513, 1.01766765117645, 0.999761700630188, 1.0182820558548, 1.06087076663971, 0.98081374168396, 1.02333998680115, 1.01955127716064, 0.991089522838593, 0.995884299278259, 0.989291250705719, 0.998204469680786, 1.00667810440063, 0.996172726154327, 1.00362539291382, 1.03796279430389, 1.00322675704956, 1.0068427324295, 1.01027321815491, 1.00379967689514, 1.01029813289642, 1.00025153160095, 0.999053716659546, 1.00668632984161, 1.0009937286377, 0.999038696289062, 1.00096476078033, 1.03811490535736, 0.978159189224243, 0.997221052646637, 0.994969427585602, 1.01598727703094, 1.00287055969238, 0.996133685112, 1.01265048980713 },
      { -0.0156621225178242, -0.0023622487206012, 0.0726891383528709, -0.0533811002969742, -0.000469459424493834, -0.0252353325486183, -0.0493677742779255, -0.00396672356873751, -0.115922085940838, -0.0112322410568595, 0.0434284284710884, -0.0203363746404648, 0.0244085323065519, -0.0109541323035955, -0.0144267454743385, 0.0199031643569469, -0.00553324352949858, -0.0474162623286247, -0.0186117812991142, -0.075230598449707, -0.0775867104530334, -0.0249816477298737, -0.0295916553586721, 0.033088106662035, -0.00172476039733738, -0.008270681835711, 0.00130145892035216, -0.0147177884355187, 0.00129471370019019, -0.00560960173606873, 0.00202817562967539, -0.0245695896446705, -0.0541057884693146, -0.00308426050469279, -0.0171079076826572, -0.0290309060364962, 0.00698774075135589, -0.0142954392358661, -0.00726040871813893, -0.0178406443446875, -0.00135834352113307, -0.0120360385626554, -0.00210544839501381, -0.0278381630778313, -0.036000594496727, -0.0177486427128315, -0.0132752712816, -0.0578001402318478, -0.0101776421070099, 0.04911994561553, -0.0204356145113707 },
      { 0.00900865532457829, -3.1697578378953e-05, -0.127906039357185, 0.061849907040596, 0.0411841385066509, 0.0220846086740494, 0.0756931900978088, 0.00258052418939769, 0.160302296280861, 0.0031239481177181, -0.0544453300535679, 0.0153323924168944, -0.0286769140511751, 0.0137482695281506, 0.00751437060534954, -0.0251899287104607, 0.00605477951467037, 0.0162228774279356, 0.00589833594858646, 0.0924375429749489, 0.00208121980540454, 0.013216694816947, -0.0451216660439968, -0.177643775939941, -0.000830948003567755, -0.00606237119063735, -0.00270187784917653, 0.0020979386754334, -0.00389396236278117, -0.00117351685184985, 0.026566905900836, -0.0411643497645855, 0.106963038444519, 0.00760003039613366, 0.0167797412723303, 0.0883143320679665, -0.0129727981984615, 0.00280793011188507, 0.00485399086028337, 0.00639398070052266, -0.00061756611103192, -0.00200193910859525, -0.00604358687996864, 0.0332041233778, 0.037396464496851, 0.00687912665307522, 0.0106128845363855, 0.0582645051181316, 0.0106070712208748, -0.079390324652195, -0.00718152476474643 },
      { 5.47911977767944, 5.84513282775879, -24.0159778594971, -13.9052391052246, -6.99573612213135, -22.7732448577881, -41.1898651123047, 7.98552274703979, -19.3163375854492, 0.349865645170212, -0.628250896930695, 2.5334906578064, 3.39280605316162, -6.46635675430298, -0.154108583927155, 3.17148470878601, 2.19827675819397, -23.9421119689941, 3.0149872303009, -13.5316762924194, -1.62664723396301, 119.777740478516, -18.4887638092041, -18.4734344482422, -3.36900806427002, 0.49560409784317, 1.31165289878845, 5.50910043716431, -0.132541969418526, -1.06566667556763, -16.2110080718994, -80.3678207397461, -45.7244606018066, -31.7297191619873, -11.5371932983398, -50.2978057861328, 0.944720149040222, 5.03856420516968, -2.46574687957764, 1.54749512672424, 0.625477850437164, 4.84230279922485, -0.232273116707802, -43.7258377075195, 18.6634788513184, 3.14952397346497, -3.3243043422699, -18.2752094268799, -8.02901744842529, 4.19663763046265, -3.11904621124268 }
    },
    {
      { 0.182154670357704, 0.170854046940804, 0.450964003801346, -0.170050948858261, -0.00301882205531001, 0.0678932294249535, 0.102919869124889, 0.185422226786613, -0.147305086255074, 0.0581569112837315, -0.0144491754472256, 0.109253570437431, 0.0256917048245668, 0.0867529585957527, -0.0351327657699585, 0.283022224903107, 0.290510803461075, -0.108995147049427, -0.00739349192008376, 0.184264332056046, 0.164773508906364, 0.136070430278778, 0.464345455169678, 0.441657155752182, 0.107790566980839, 0.0678426697850227, -0.0878575965762138, 0.189553275704384, 0.0239598210901022, 0.0582150407135487, 0.074261412024498, -0.743152320384979, 0.00810250733047724, -0.168973937630653, -0.011880487203598, 0.332383364439011, -0.0104818623512983, 0.177591726183891, -0.0150941470637918, 0.0463389605283737, 0.0261570122092962, 0.128360867500305, -0.0279898140579462, -0.492167145013809, 0.208139643073082, 0.0444884821772575, 0.0442324429750443, 0.484689325094223, -0.314633905887604, -0.0639173686504364, -0.0812805518507957 },
      { -0.160478785634041, -0.0453233569860458, -0.314887762069702, 0.541398108005524, -0.18855856359005, -0.565167009830475, -0.547567069530487, -0.423762232065201, 0.437634974718094, -0.123023398220539, 0.0870423391461372, -0.0503678061068058, 0.00477056205272675, -0.0497027896344662, 0.0601796135306358, -0.526368141174316, -0.269567012786865, 0.0742360129952431, -0.106476314365864, -0.709418475627899, -0.358079314231873, -0.119653321802616, -0.261730164289474, -0.353231012821198, -0.0152066703885794, -0.0135902743786573, 0.16033561527729, -0.171756833791733, -0.0490630976855755, -0.0203408524394035, -0.248637348413467, 0.402239084243774, -0.697896659374237, -0.0147195477038622, 0.0693437606096268, -0.733469367027283, 0.0187152847647667, -0.162420704960823, -0.0703400820493698, -0.0491949617862701, -0.0415627136826515, -0.230618581175804, 0.0722560808062553, 0.432417869567871, -0.135274276137352, -0.112837612628937, 0.0470824129879475, -0.052324004471302, -0.258222132921219, 0.0802111625671387, 0.110504344105721 },
      { -0.0408718138933182, -0.25725844502449, 0.0681157335639, -0.347095847129822, 0.276870965957642, 0.375284641981125, 0.423361569643021, 0.14975181221962, -0.176812872290611, -0.00128540443256497, -0.0273784305900335, -0.0831412300467491, -0.022647263482213, -0.10212466865778, -0.10825888812542, 0.162279814481735, 0.0637381449341774, -0.0600745379924774, 0.048958744853735, 0.455512523651123, 0.1530641913414, 0.0694004893302917, -0.0496721416711807, -0.134567067027092, -0.179973796010017, -0.176728770136833, -0.146644279360771, -0.025817884132266, -0.0107706626877189, -0.111095957458019, 0.0671709179878235, 0.249452635645866, 0.53317129611969, 0.104688420891762, -0.0328748226165771, 0.522525608539581, -0.0698914080858231, -0.0476466529071331, -0.0193528290838003, 0.0847823321819305, -0.0364726148545742, 0.0130338231101632, -0.0627997443079948, 0.26862245798111, -0.0558006912469864, 0.0235065370798111, -0.167051762342453, -0.0619212314486504, 0.493546605110168, -0.0909502953290939, -0.0280974972993135 },
      { 1.08649146556854, 1.19120585918427, 1.0364750623703, 0.941726088523865, 1.04834425449371, 1.04711484909058, 1.02763676643372, 1.09664750099182, 0.971347272396088, 1.04756999015808, 1.03477823734283, 1.04729199409485, 1.00295960903168, 1.11066961288452, 1.04195094108582, 1.06165504455566, 1.0348105430603, 0.920164048671722, 1.02799928188324, 1.03456807136536, 0.700000584125519, 1.03398382663727, 1.01295673847198, 1.03279900550842, 1.10181164741516, 1.06747233867645, 1.06026780605316, 1.05239832401276, 1.04483079910278, 1.04664933681488, 1.00566148757935, 0.817648112773895, 1.03976237773895, 0.983278453350067, 0.981858968734741, 1.01834642887115, 1.07185363769531, 1.06007075309753, 1.05689871311188, 1.03523087501526, 1.04318034648895, 1.14307510852814, 0.9962158203125, 0.790818154811859, 1.04224789142609, 1.03467488288879, 1.14964973926544, 0.982232809066772, 1.02971112728119, 1.03569638729095, 1.00368201732635 },
      { 0.022833114489913, -0.000368642125977203, 0.0222809016704559, 0.307630717754364, 0.0319230370223522, 0.0377336256206036, 0.131587162613869, 0.0394208692014217, 0.486985802650452, 0.0121470736339688, -0.0306262001395226, 0.0500956699252129, -0.00346607225947082, -0.00958161149173975, 0.0552126280963421, 0.0239542778581381, 0.0239702127873898, 0.239313468337059, 0.0414266437292099, 0.160822406411171, 0.40775191783905, 0.0236285384744406, 0.045343954116106, -0.00305753038264811, 0.00734934443607926, 0.0456732138991356, 0.00951007287949324, 0.0760479420423508, -0.0230114087462425, 0.0268840342760086, 0.133349552750587, 0.0806569755077362, 0.1102364808321, -0.00203303107991815, 0.0697420462965965, 0.104915924370289, -0.0253230500966311, 0.0511549301445484, 0.00758506497368217, 0.0729243978857994, 0.00998544506728649, -0.0722460672259331, 0.0635944530367851, 0.126332461833954, -0.0290875528007746, 0.072749637067318, 0.0202506277710199, 0.000585241534281522, -0.0389183573424816, -0.0188215002417564, 0.0715478733181953 },
      { -0.0241615995764732, 0.000538820924703032, -0.076097160577774, -0.359804093837738, -0.154991537332535, -0.0539560280740261, -0.204103380441666, -0.0557419583201408, -0.74410355091095, 0.0135560669004917, -0.00848265178501606, -0.0336992479860783, 0.0119506428018212, 0.00361147802323103, -0.0265750158578157, 0.0245495922863483, -0.0254516843706369, -0.0722543001174927, -0.00628551933914423, -0.215449720621109, 0.000738201371859759, -0.00228969100862741, -0.100333116948605, 0.126784637570381, -0.00179146998561919, 0.00698349718004465, -0.0021849162876606, -0.0677516832947731, 0.0188860390335321, 0.00855750124901533, -0.0783768445253372, 0.248670190572739, -0.147487878799438, -0.00169946835376322, -0.0679837092757225, -0.16156193614006, 0.00761158810928464, -0.0325024351477623, 0.00300761568360031, -0.150018587708473, 0.00268542324192822, 0.0565983019769192, -0.0385242663323879, -0.168394953012466, 0.0419108308851719, -0.0583710223436356, -0.0484576784074306, 0.0694146305322647, 0.0583669692277908, 0.075401596724987, -0.0790762975811958 },
      { -29.7635974884033, -27.6819972991943, -30.4682674407959, 72.2514266967773, 8.55252933502197, 58.4545059204102, 29.9883289337158, -36.2264518737793, 143.014190673828, -10.4086256027222, -15.4833002090454, -21.3925628662109, 0.60703432559967, -5.2867283821106, 0.446573525667191, -32.2929916381836, -24.2168865203857, 130.193542480469, -21.6923408508301, 21.1016979217529, 0.838193774223328, -305.285736083984, 48.531177520752, -2.77269744873047, -10.0108652114868, -3.36372089385986, -2.80609488487244, -34.4354438781738, 14.7108879089355, -3.55802965164185, 20.9756984710693, 459.082183837891, 50.9306831359863, 176.392379760742, 51.7804336547852, 33.874568939209, 3.60655426979065, -28.2673721313477, 32.1371917724609, -8.39994812011719, -2.95379042625427, -22.283203125, -3.53432893753052, 246.113952636719, 27.1976890563965, -19.805103302002, -1.69403624534607, 22.745813369751, 19.0480537414551, -17.3391590118408, 1.8287935256958 }
    },
    {
      { -0.060032095760107, -0.0723902359604836, 0.980297088623047, -0.33272859454155, 0.452203631401062, 0.104894384741783, 1.23387336730957, 0.606536626815796, 0.00543143972754478, 0.100904688239098, 0.0577281825244427, 0.0603423193097115, -0.0136654525995255, 0.0652079731225967, 0.0962393134832382, 0.217427432537079, 0.532236695289612, -0.281982570886612, 0.0948613435029984, 0.987426042556763, 0.211028560996056, 0.118693314492702, 1.29263949394226, 0.630468249320984, 0.174864694476128, 0.0505526885390282, 0.257684260606766, 0.0485608465969563, -0.142522856593132, 0.0611283183097839, 0.639277756214142, -0.744594693183899, 1.10150647163391, 0.0330489166080952, 0.296909660100937, 1.20549952983856, 0.169448554515839, -0.0179710946977139, 0.289432018995285, 0.393897205591202, 0.00295346742495894, 0.318351149559021, 0.223333686590195, -0.674814283847809, 0.494165420532227, 0.480868607759476, 0.128903195261955, 1.80114150047302, -0.058619562536478, 0.131639167666435, 0.367417544126511 },
      { 0.241093024611473, 0.0730777308344841, 0.270693004131317, 0.602368175983429, -0.421668916940689, -0.399641364812851, -0.734812557697296, -0.526352763175964, 0.0281714480370283, -0.622962176799774, -0.384961485862732, 0.0702337771654129, -0.119311265647411, -0.0587764084339142, -0.217064216732979, -0.168295904994011, -0.0851085335016251, -0.0208129920065403, -0.361480295658112, -0.741021931171417, -0.741032421588898, -0.387755751609802, -0.0275656152516603, 0.0758076608181, -0.238616749644279, -0.0156109584495425, -0.357977598905563, -0.141010910272598, 0.190633550286293, -0.125595480203629, -0.342519670724869, 0.402746647596359, -0.751001954078674, -0.0661207735538483, 0.0286167860031128, -0.5933678150177, -0.232445731759071, 0.103914715349674, -0.401351422071457, -0.522162020206451, -0.138721510767937, -0.31011289358139, -0.306736379861832, 0.482749223709106, -0.22541506588459, -0.495300501585007, -0.371485680341721, 0.342381596565247, -0.0517269968986511, -0.581978142261505, -0.48354434967041 },
      { -0.331109315156937, -0.0775493681430817, -0.911279737949371, -0.126038059592247, 0.0184554755687714, 0.55950665473938, -0.0975724756717682, 0.0311203207820654, -0.0232234727591276, 0.522105634212494, 0.338021010160446, -0.143047347664833, 0.157919317483902, -0.479424446821213, 0.154388502240181, -0.103505097329617, 0.207513570785522, 0.329342544078827, 0.27864670753479, 0.0462512746453285, 0.687462449073792, 0.226724117994308, -1.00590026378632, -0.705068051815033, 0.135221689939499, 0.000224448347580619, 0.109359167516232, -0.25537109375, -0.08443733304739, 0.115302927792072, -0.317127108573914, 0.286138862371445, -0.0843654349446297, 0.0411316864192486, -0.173743382096291, -0.273584127426147, 0.117142885923386, -0.168593913316727, 0.329354882240295, 0.289903223514557, 0.158190995454788, 0.110074020922184, 0.16689296066761, 0.49482586979866, -0.0836101844906807, 0.088436670601368, 0.0733437836170197, -0.887619972229004, 0.0654139965772629, 0.501335501670837, 0.413475841283798 },
      { 0.10322467237711, 0.0911174342036247, -0.0161396581679583, -0.20157378911972, 0.0248985104262829, 0.0782119631767273, 0.0442247055470943, 0.0243588201701641, -0.0042216656729579, -0.0190169401466846, 0.0170643497258425, 0.0695843026041985, 0.00960851367563009, 0.443639606237411, -0.0305083449929953, -0.0549966171383858, -0.0273222960531712, -0.228343665599823, 0.0326669588685036, 0.0180054660886526, -0.652540147304535, -0.000294804340228438, 0.00223196716979146, 0.0274402871727943, -0.0128725972026587, 0.0225702300667763, 0.0671055838465691, 0.271662354469299, 0.0385084189474583, -0.0175224430859089, -0.00448055705055594, -0.392910093069077, 0.000322179694194347, -0.00129860627930611, -0.0223370715975761, -0.014025779440999, 0.0169943757355213, 0.133833348751068, 0.0475244522094727, 0.0655424445867538, 0.0238803271204233, -0.0417560040950775, 0.0213948711752892, -0.36173677444458, -0.00636529270559549, 0.0301219895482063, 0.0514265336096287, -0.0971249267458916, -0.00537463743239641, -0.0106464149430394, 0.0463254153728485 },
      { 1.10962080955505, 1.02467238903046, 1.15663552284241, 1.39650511741638, 1.07374608516693, 0.951367855072021, 1.09608268737793, 1.06922113895416, 0.859874904155731, 1.15860402584076, 1.25985991954803, 1.1125510931015, 1.21405863761902, 0.889857351779938, 1.0913165807724, 1.16559064388275, 0.726673901081085, 1.33567428588867, 1.05067586898804, 1.01464772224426, 1.66857945919037, 1.10087954998016, 1.11194252967834, 1.20945024490356, 1.08669352531433, 1.03413808345795, 1.10900640487671, 0.85364830493927, 1.144495844841, 1.02470326423645, 1.11009109020233, 1.29906249046326, 1.04423439502716, 0.978720128536224, 1.04011380672455, 1.07247161865234, 1.22954785823822, 1.08341586589813, 1.01100873947144, 1.00919103622437, 1.16748380661011, 1.1261203289032, 1.11325669288635, 1.25902271270752, 0.58654797077179, 1.08034133911133, 1.0303692817688, 1.05328583717346, 1.08495628833771, 1.26743400096893, 1.05679047107697 },
      { -0.0958186611533165, -0.0663906112313271, -0.205666810274124, -0.456627994775772, -0.0604416206479073, -0.103133738040924, -0.136871963739395, -0.0428019501268864, 0.136861622333527, -0.152155086398125, -0.348216563463211, -0.123018220067024, -0.298176854848862, 0.103576689958572, -0.0961994305253029, -0.0408046916127205, 0.17395156621933, -0.177571311593056, -0.0815490335226059, 0.029614195227623, -0.11002154648304, -0.104786798357964, -0.0749518349766731, -0.400557637214661, -0.105556935071945, -0.0698807612061501, -0.17005480825901, 0.181914642453194, -0.170643225312233, -0.0403642877936363, 0.0718474835157394, 0.196314588189125, 0.160210371017456, 0.0146610429510474, -0.0398688577115536, 0.137331232428551, -0.300509482622147, -0.12262012809515, -0.0862025544047356, -0.135953426361084, -0.244474157691002, -0.142535760998726, -0.193512365221977, -0.26717609167099, 0.396755546331406, -0.0887438058853149, -0.0328717678785324, 0.0587770380079746, -0.120191991329193, -0.352880746126175, -0.256624102592468 },
      { -46.2529067993164, -16.7036991119385, -83.5624542236328, 222.787719726562, -24.4429244995117, -83.1296081542969, -117.983184814453, -37.9990272521973, -28.8171463012695, 42.3110160827637, -4.25657606124878, -73.6627731323242, 41.2360877990723, -84.2538986206055, 0.200789719820023, -29.9806804656982, -8.85134983062744, 335.779693603516, -34.1170883178711, -46.9922943115234, 137.651138305664, 448.705047607422, -108.420692443848, -42.8721351623535, -5.46905136108398, -1.48017406463623, 1.72753441333771, -32.9045028686523, 2.97038149833679, -7.45214033126831, -34.0442886352539, 631.111022949219, -167.982467651367, 22.7478561401367, 3.36004114151001, -137.018707275391, -25.1732578277588, -62.9384346008301, -81.6401901245117, 7.56623649597168, 18.1354637145996, -6.7006311416626, 0.375565499067307, 398.220397949219, 326.451965332031, -17.3988857269287, 45.2356376647949, -84.5597076416016, 16.053394317627, 27.5030670166016, -18.3667507171631 }
    },
    {
      { 0.0826874747872353, 0.085453100502491, 0.0200429782271385, 0.0250256489962339, 0.0152111817151308, -0.0122637515887618, 0.0521720200777054, 0.0385669209063053, -6.69744840706699e-05, 0.0186115074902773, 0.00906509533524513, 0.106071971356869, 0.00966802984476089, 0.0970587581396103, 0.00774931954219937, 0.00390145252458751, 0.0797271430492401, -0.00373311340808868, 0.0743691623210907, 0.0292200203984976, 0.113100655376911, 0.0181614495813847, 0.021161375567317, 0.0149180926382542, 0.054700244218111, 0.136399999260902, -0.0113783916458488, 0.143145799636841, -0.0055000283755362, 0.0174186155200005, 0.0245339628309011, -0.0430981367826462, -0.0012188694672659, -0.0112682413309813, 0.020243214443326, 0.0305543728172779, -0.00415204837918282, 0.106560878455639, 0.0804156810045242, 0.0212031882256269, 0.0147221572697163, -0.0217096898704767, 0.0032438759226352, -0.0577258951961994, 0.135524794459343, 0.0479795858263969, -0.0337147526443005, 0.0372242704033852, 0.0162294693291187, -0.00172652862966061, 0.0149784358218312 },
      { -0.120260246098042, -0.0542730316519737, 0.0100648542866111, 0.0271420478820801, -0.0275781434029341, 0.0480448789894581, 0.000759892340283841, -0.0426573008298874, 0.00536793190985918, 0.037595059722662, 0.00222005089744925, -0.144671425223351, 0.0114762606099248, -0.113243453204632, 0.0243929550051689, -0.0692385211586952, -0.16370677947998, 0.0155040128156543, 0.0231587905436754, 0.00131577125284821, -0.163594394922256, -0.0111898900941014, 0.0162858739495277, 0.01736612431705, -0.0598934218287468, 0.00339136854745448, -0.000714033551048487, -0.0937924236059189, -0.00932190474122763, 0.030576391145587, -0.0208196751773357, -0.00117742666043341, 0.0126067688688636, -0.00358342216350138, 0.00758025143295527, 0.0197269599884748, 0.0151241710409522, -0.157412737607956, -0.090442918241024, 0.00470194779336452, -0.00253657693974674, -0.00120553479064256, -0.00952445063740015, -0.00251304567791522, -0.146278858184814, 0.00771512230858207, 0.0607463084161282, 0.00949200056493282, 0.0042002908885479, -0.0259736943989992, 0.047001663595438 },
      { 0.0344887115061283, -0.148696303367615, -0.0418014712631702, -0.0640286356210709, 0.00939745083451271, -0.0374522246420383, -0.0284400619566441, -0.000187721030670218, -0.00137824902776629, -0.061271920800209, -4.31341250077821e-05, 0.0222694221884012, -0.0294129773974419, 0.0185016132891178, -0.0325909517705441, -0.0176245123147964, -0.0772406756877899, -0.0225558206439018, -0.0374633818864822, -0.0191184543073177, -0.0039001980330795, 0.00327777862548828, -0.0358657538890839, -0.0266344770789146, 0.0099307456985116, -0.0833905637264252, 0.00374992820434272, 0.0372366942465305, 0.00307186157442629, -0.0498092919588089, -0.00272272480651736, 0.0522801242768764, -0.0144588593393564, 0.00481863413006067, -0.0124395936727524, -0.0404292456805706, -0.0241558942943811, 0.0397332198917866, 0.000686598010361195, -0.0171792730689049, -0.0295614935457706, 0.00731256837025285, 0.00122916256077588, 0.073440320789814, 0.0151143968105316, -0.0345495976507664, -0.0592187605798244, -0.0358198061585426, -0.0144139165058732, 0.0134340571239591, -0.034791361540556 },
      { 0.0273458771407604, 0.189653500914574, 0.00166569766588509, -0.0112441377714276, 0.0017713678535074, -0.00453749764710665, -0.00435368483886123, 0.00961215421557426, -0.000787565659265965, 0.00152308505494148, 0.00162232120055705, 0.0173249635845423, 0.0150936925783753, 0.0596241615712643, 0.00358729856088758, 0.015130260027945, 0.0485927015542984, -0.0085306279361248, -0.00562960747629404, -0.00105958781205118, -0.016476783901453, 0.00766840530559421, -0.00134899863041937, 0.00223098089918494, 0.0147834690287709, 0.0741738229990005, -0.000757989124394953, -0.0195858869701624, 0.00625760201364756, 0.000690365443006158, 0.00167517201043665, -0.0177775472402573, 0.000824458547867835, 0.000685011793393642, -0.00465842336416245, -0.00152063986752182, 0.00619004853069782, 0.00997924152761698, -0.0217229276895523, -0.00616328464820981, 0.0226253513246775, 0.00910294614732265, 0.00159807538148016, -0.00459103984758258, 0.0282337665557861, 0.0109519567340612, -0.020777178928256, -0.00144999753683805, -0.000270176795311272, 0.00301533774472773, -0.00142098788637668 },
      { -0.0181259289383888, -0.0366354584693909, 0.0150471357628703, 0.118683010339737, 0.0295140333473682, 0.0282058082520962, 0.0394628457725048, 0.0525061823427677, 0.0610293000936508, 0.03574538230896, 0.0129329338669777, -0.00928078219294548, -0.0212045311927795, -0.0335889235138893, 0.0305250771343708, 0.0864024385809898, 0.102837674319744, 0.0875031799077988, -0.00158862862735987, 0.0322900861501694, 0.124239765107632, -0.00904334336519241, 0.0201820284128189, 0.0121889039874077, -0.00511301541700959, -0.0661162808537483, 0.0166219286620617, 0.072508729994297, 0.00259026465937495, 0.0101581960916519, 0.0251603052020073, 0.0581736341118813, 0.0233358647674322, 0.0271851029247046, 0.0403490886092186, 0.0231983531266451, -0.00429958431050181, 0.00337207992561162, 0.0566322393715382, 0.0341721810400486, -0.00910864304751158, 0.0509537197649479, 0.0185884367674589, 0.0384444929659367, -0.0462068878114223, 0.0380484275519848, 0.0185942687094212, 0.0269525721669197, 0.00774603756144643, 0.0100125148892403, 0.0220976211130619 },
      { 1.00155401229858, 0.973653078079224, 1.00537812709808, 0.911865413188934, 0.974975287914276, 0.9828822016716, 0.962372779846191, 0.928254008293152, 0.93386447429657, 0.98289692401886, 0.986051023006439, 1.01107096672058, 1.02863359451294, 0.986875772476196, 0.96941614151001, 0.915370523929596, 0.915048837661743, 0.94526219367981, 0.998855829238892, 0.967894315719604, 0.976793110370636, 1.01179730892181, 0.998545408248901, 0.994567096233368, 1.00079739093781, 1.00457525253296, 0.989859759807587, 0.896552562713623, 1.00097393989563, 1.00079119205475, 0.979312479496002, 0.961764812469482, 0.983874380588531, 0.963346123695374, 0.954257428646088, 0.988637626171112, 1.01605677604675, 0.997285127639771, 0.965822696685791, 0.976233184337616, 1.01111996173859, 0.938170254230499, 0.983903408050537, 0.963024497032166, 1.0302540063858, 0.947452068328857, 1.00454998016357, 0.979411423206329, 1.01966917514801, 1.00799870491028, 0.973453342914581 },
      { 14.5081071853638, 10.7896022796631, -3.67876720428467, -23.0986957550049, -4.03657865524292, 0.927780628204346, -2.68074679374695, -0.757431328296661, 1.0313378572464, -19.9974765777588, -3.12985467910767, 9.63033962249756, -9.90398788452148, 47.7249031066895, -3.56882476806641, -7.07202482223511, -10.2411117553711, -12.376856803894, -4.93355560302734, 0.209504500031471, -25.3799381256104, -23.4481353759766, -2.02967381477356, -2.66966795921326, -1.57384252548218, 19.7285442352295, -0.219545394182205, 2.1788694858551, 12.6980438232422, -4.10666513442993, -2.46752667427063, 38.1283264160156, -2.40337181091309, 13.6982574462891, -0.526745498180389, -3.37223720550537, 0.928637027740479, 9.29304313659668, 5.20367956161499, -6.40767908096313, -6.65659284591675, -13.092188835144, -0.369137853384018, 24.8691158294678, 48.1142158508301, -6.96300792694092, 0.793479681015015, -1.67205369472504, -4.08502578735352, -1.91614723205566, -5.21430969238281 }
    }
  },
  {
    {
      { 0.909986734390259, 0.88829642534256, 0.972735524177551, 0.940046966075897, 0.888769268989563, 0.852122724056244, 0.809826731681824, 0.830339431762695, 0.926758229732513, 0.951664209365845, 0.881052315235138, 0.30770006775856, 0.992385566234589, 0.849790811538696, 0.970993220806122, 0.887139141559601, 0.948846995830536, 0.948158264160156, 0.715900957584381, 0.884775459766388, 0.966454207897186, 0.865698397159576, 0.910567998886108, 0.907074809074402, 0.836675941944122, 0.888258159160614, 0.900739789009094, 0.821689605712891, 0.975225985050201, 0.86685848236084, 0.852671205997467, 0.811135351657867, 0.953025221824646, 0.944706380367279, 0.98438686132431, 0.860226452350616, 0.975686192512512, 0.965167641639709, 0.932741343975067, 0.897134780883789, 0.884471237659454, 0.814658999443054, 0.896056950092316, 0.865474998950958, 0.89078164100647, 0.860943257808685, 0.902029693126678, 0.926250100135803, 0.904026567935944, 0.874789834022522, 0.897536218166351 },
      { 0.127884015440941, 0.13760794699192, -0.0761730968952179, 0.0912440046668053, 0.128766357898712, 0.0987097844481468, 0.251092612743378, 0.259496122598648, 0.0964989587664604, 0.130087330937386, 0.109753988683224, 1.62448465824127, 0.0508621297776699, 0.195242077112198, -0.107583619654179, 0.165723502635956, 0.0726814344525337, 0.0585126876831055, 0.369277685880661, 0.0999567061662674, 0.17136313021183, 0.137916460633278, 0.0982496961951256, 0.109436169266701, 0.244920402765274, 0.128272145986557, 0.333286941051483, 0.196458324790001, 0.0991469770669937, 0.27680042386055, 0.205909356474876, 0.218383014202118, -0.0942210257053375, 0.0601209923624992, 0.0382428653538227, 0.17932365834713, -0.0671706348657608, 0.0967019125819206, 0.115870580077171, 0.119685992598534, 0.123830214142799, 0.132463037967682, 0.138246402144432, 0.158846989274025, 0.222396731376648, 0.141699209809303, 0.143910989165306, 0.0868949815630913, 0.139139994978905, 0.137659341096878, 0.133770793676376 },
      { -0.0231277998536825, -0.0087196072563529, 0.0991958454251289, -0.0231002829968929, -0.0402423702180386, 0.0607120916247368, -0.0515766777098179, -0.020019669085741, -0.0213351938873529, -0.112113520503044, -0.0101762684062123, -0.441506445407867, -0.0943393036723137, -0.0464088469743729, 0.140822067856789, -0.0315118432044983, -0.0108766118064523, 0.0548031739890575, -0.0841093510389328, 0.00041230209171772, -0.0693862736225128, 0.00507718045264482, 0.000592219934333116, -0.0163146238774061, -0.0475721247494221, -0.00178582116495818, -0.193448171019554, -0.0208343677222729, -0.0909577608108521, -0.110995836555958, -0.00915331672877073, -0.0296266451478004, 0.148448809981346, -0.0224497988820076, -0.0200873706489801, 0.0107386577874422, 0.0732476860284805, 0.00410828739404678, -0.0206218268722296, -0.00187881744932383, 0.00461477134376764, 0.0374521277844906, -0.0415741428732872, 0.006884157191962, -0.103354915976524, -0.0295232739299536, -0.0376400426030159, 0.0154123026877642, -0.035565335303545, 0.0276836603879929, 0.0151044577360153 },
      { 0.00442893197759986, -0.00408736243844032, 0.0032926588319242, 0.018132921308279, 0.00143009540624917, 0.00114781514275819, 1.33516123241861e-05, -0.0237245261669159, 0.00415771594271064, 0.00758716324344277, -5.54681573703419e-05, -0.0235171355307102, 0.0461229681968689, 0.000440403557149693, -0.00227412185631692, -0.00259366957470775, -0.00166904623620212, -0.0283923838287592, 0.0151315601542592, -0.000288811512291431, 0.040260910987854, -0.00157904089428484, -0.000985233229584992, 0.00094756594626233, -0.000725352205336094, -0.00118996517267078, -0.0344070494174957, -0.00715769175440073, 0.0430401712656021, -0.00673854630440474, -0.0229268651455641, 0.0192892365157604, -0.0129829561337829, 0.00804672855883837, 0.0293550696223974, -0.0376402698457241, 0.00900210347026587, -0.0273827631026506, -0.00610958738252521, -0.00180037750396878, -5.8014236856252e-05, 0.00547869643196464, -0.0023304910864681, -0.0334641002118587, 0.0178350359201431, 0.00440517300739884, 0.00953665934503078, 0.0012224706588313, 0.00240845535881817, -0.00404849788174033, 0.000151530213770457 },
      { 0.0112122800201178, 0.00322861433960497, 0.0131780775263906, -0.00205112202093005, 0.00667265895754099, -0.00583912106230855, 0.00108241126872599, 0.0273285731673241, 0.0094763208180666, 0.0181949622929096, 0.0123619316145778, 0.0587645135819912, -0.00784866232424974, 0.0103303417563438, 0.00205404497683048, -0.0221493784338236, 0.0098227895796299, 0.0312968902289867, -0.0368994846940041, 0.00974850542843342, -0.0250585954636335, 0.0151614034548402, 0.00315689435228705, -0.00277821905910969, 0.0423278100788593, -0.00932277273386717, 0.022687204182148, 0.0151523016393185, 0.00253381603397429, 0.00408021360635757, 0.0204070657491684, 0.0077986903488636, 0.00538867013528943, 0.00429253885522485, -0.0328896753489971, 0.0150164542719722, 0.0205978788435459, 0.0420781187713146, 0.0104278204962611, -0.00266666477546096, -0.00193644757382572, -0.0123371006920934, 0.00397288799285889, 0.029932726174593, -0.0150477951392531, -0.00614297110587358, 0.0108134588226676, -0.00150733743794262, 0.00318275368772447, 0.004979501478374, 0.0100565450266004 },
      { -0.0103628570213914, -0.00105465087108314, -0.0139639703556895, -0.00716992653906345, 0.00805994961410761, 0.00545204430818558, -0.0111360661685467, -0.0569521822035313, -0.0113578401505947, -0.00265598041005433, -0.0197473429143429, -0.0754277110099792, 0.0292944498360157, -0.0107870725914836, -0.00580876274034381, 0.0254277419298887, -0.0076552308164537, -0.0208764560520649, 0.0127868792042136, -0.0012217661133036, -0.0207048691809177, -0.0246600043028593, -0.00231211865320802, 0.00100235687568784, -0.0872332826256752, 0.0118198925629258, -0.0082571180537343, -0.020790321752429, 0.00135766924358904, -0.00466205785050988, -0.0278294216841459, -0.0336056612432003, -3.72849244740792e-05, -0.00758876139298081, 0.0321111902594566, -0.0197444576770067, -0.0109692122787237, -0.0487978309392929, -0.0083316545933485, -0.000421148695750162, 0.00626283092424273, 0.00426645204424858, -0.000450647465186194, -0.0226539932191372, 0.0121136261150241, 0.0209738798439503, -0.0141499126330018, -0.00225331238470972, -0.00482121715322137, -0.0322020649909973, -0.0332429669797421 },
      { -20.8814945220947, -10.8041524887085, -4.36561965942383, -2.46061110496521, -8.92346286773682, -3.6081120967865, -4.60914754867554, -6.58960199356079, -5.10161542892456, -2.80921792984009, -1.11759734153748, -13.5409317016602, -34.9594421386719, -5.39812135696411, 7.71538639068604, -13.7412624359131, -7.49441385269165, -53.7948036193848, 3.53413796424866, -12.489691734314, -6.21089792251587, -3.52929782867432, -6.38706731796265, -3.11113786697388, 3.00130701065063, -4.40748357772827, 13.6597232818604, -2.53010129928589, -28.3580493927002, 9.76344871520996, -14.7192878723145, -4.88943958282471, 23.6982040405273, 163.283935546875, -13.25341796875, 1.04537081718445, -13.3938016891479, -13.0782251358032, -4.46338081359863, -1.37751054763794, -14.8056478500366, -1.33041930198669, -1.49866318702698, -4.16184234619141, -24.1458988189697, -11.8394384384155, -8.5393238067627, -10.4525260925293, -3.32402920722961, 0.824457764625549, -13.1113185882568 }
    },
    {
      { -0.0810845568776131, -0.0440090633928776, -0.0719956234097481, -0.0686426386237144, -0.051402285695076, -0.107912257313728, -0.0624396875500679, -0.0556441321969032, -0.0736291483044624, -0.0866539105772972, -0.111641690135002, -0.00184922071639448, -0.0404689908027649, -0.0626309290528297, -0.107359938323498, -0.0865461453795433, -0.0632608011364937, -0.00717686954885721, -0.0828422904014587, 0.00606286944821477, -0.067387655377388, -0.0562064796686172, -0.119092129170895, -0.0420615002512932, -0.0678616911172867, -0.0443511269986629, -0.0124462116509676, -0.0918935686349869, -0.0638692826032639, -0.0841375961899757, -0.0996459275484085, -0.0789631977677345, -0.0385149978101254, -0.33964592218399, -0.0490097887814045, -0.0602841377258301, -0.0580639243125916, -0.0590649656951427, -0.0851679742336273, -0.0786112993955612, -0.0448799878358841, -0.0471386574208736, -0.085393950343132, -0.0561911016702652, -0.0290480684489012, -0.0614676252007484, -0.00888883881270885, -0.037316557019949, -0.0116605395451188, -0.051057692617178, -0.0402898378670216 },
      { 1.15258586406708, 1.11264991760254, 1.18055808544159, 1.16904079914093, 1.16080892086029, 1.23378431797028, 1.09425866603851, 1.14988172054291, 1.16866147518158, 1.18101394176483, 1.17802655696869, 0.886472046375275, 1.14668095111847, 1.09736979007721, 1.27299284934998, 1.17561542987823, 1.16194450855255, 1.08875930309296, 1.16492021083832, 1.19572103023529, 1.17093896865845, 1.11200654506683, 1.2297089099884, 1.26189064979553, 1.15041124820709, 1.13979971408844, 1.059774518013, 1.18603587150574, 1.18560934066772, 1.16734254360199, 1.19456970691681, 1.14822864532471, 1.15510225296021, 1.61824417114258, 1.14138185977936, 1.12192380428314, 1.17426478862762, 1.14277875423431, 1.15768492221832, 1.17230069637299, 1.19676530361176, 1.16519951820374, 1.26768410205841, 1.11428809165955, 1.21773135662079, 1.16480314731598, 1.114142537117, 1.25668394565582, 1.11327564716339, 1.10466027259827, 1.11348223686218 },
      { -0.0865653902292252, -0.0808853507041931, -0.102892331779003, -0.0930627286434174, -0.0877095237374306, -0.169886469841003, -0.0390103086829185, -0.0888388678431511, -0.108191788196564, -0.108839020133018, -0.0906121581792831, 0.0754571408033371, -0.144820123910904, -0.039168655872345, -0.204167425632477, -0.0926380380988121, -0.107722751796246, -0.0690569430589676, -0.110025823116302, -0.175324976444244, -0.114300861954689, -0.0630272030830383, -0.115742452442646, -0.192683145403862, -0.0810208767652512, -0.0999739319086075, -0.0465207770466805, -0.0936895757913589, -0.136452049016953, -0.0929890125989914, -0.0937336608767509, -0.0786958783864975, -0.141747161746025, -0.351340085268021, -0.109626814723015, -0.0613153167068958, -0.114051975309849, -0.103889092803001, -0.0802464112639427, -0.0801517441868782, -0.148015111684799, -0.114938862621784, -0.19607649743557, -0.0581856817007065, -0.188860833644867, -0.122254833579063, -0.102812014520168, -0.186712116003036, -0.109599061310291, -0.0539956204593182, -0.0509677492082119 },
      { 0.0137058449909091, 0.00911984965205193, -0.0327862948179245, 0.0029183030128479, 0.0135576408356428, 8.3193983300589e-05, 0.0114141926169395, -0.00570585113018751, 0.00766956806182861, 0.0131301805377007, 0.0111426049843431, 0.000462237338069826, 0.0203131157904863, 0.0253874026238918, -0.00338923837989569, 0.00353163806721568, 0.00590927340090275, 0.0022255361545831, 0.0232582557946444, 0.00303413509391248, 0.0184583514928818, -0.00242375256493688, 0.00177359511144459, 0.00296942610293627, -2.66375045612222e-05, 0.0147773614153266, 0.0105820121243596, 0.00112443091347814, 0.017576040700078, 0.00847318582236767, 0.00141033192630857, 0.00593901704996824, -0.0512297414243221, 0.0664468258619308, 0.0179159007966518, 0.00603367760777473, -0.0294454973191023, 0.0120694981887937, 0.00793504435569048, -0.00011630332301138, 0.00079655135050416, 0.00583537947386503, -0.00483292108401656, -0.00873785745352507, 0.0191191509366035, 0.00860839895904064, 0.0283268038183451, -0.00369360344484448, 0.0225828252732754, 0.00987759791314602, 0.0105836279690266 },
      { -0.000643684063106775, 0.00980732869356871, 0.0359858833253384, 0.00386101100593805, -0.00201181857846677, 0.0138431629166007, -0.00890185497701168, 0.0459335260093212, 0.0148157412186265, -0.00612290529534221, 0.0262353401631117, -0.0570296160876751, 0.00401458516716957, -0.0616019628942013, 0.0657985955476761, -0.00565954344347119, 0.0124329617246985, 0.0148942954838276, 0.0126059306785464, 0.00424659717828035, -0.00563183706253767, 0.036647841334343, 0.00110587268136442, 0.0116683831438422, 0.0200354065746069, -0.0117163583636284, -0.00497148325666785, -0.000685476989019662, 0.00562360743060708, -0.00177460687700659, 0.00563734071329236, 0.00617401162162423, 0.115065313875675, -0.00918026082217693, -0.00807851646095514, -0.00048333709128201, 0.0299311187118292, 0.00735087413340807, 0.00424814829602838, 0.0158703234046698, 0.00484656356275082, -0.00450170272961259, 0.0313305705785751, 0.0187366679310799, -0.00934677384793758, -0.00614767195656896, 0.00658926367759705, 0.031358826905489, -0.0020330548286438, 0.00444065872579813, -0.00081428000703454 },
      { 0.000944148690905422, -0.00783602334558964, -0.0108273653313518, -0.00674008624628186, -0.0185832586139441, -0.0138333244249225, 0.00095591158606112, -0.0572875253856182, -0.0111031774431467, 0.00814769975841045, -0.0724097192287445, 0.069847360253334, 0.0156820081174374, 0.0417402535676956, -0.0196858514100313, 0.00666736718267202, -0.00720419874414802, -0.0056737931445241, -0.00873641856014729, 0.00872706342488527, 0.00468095205724239, -0.0320884808897972, -0.0021809118334204, -0.045804601162672, -0.0276078041642904, 0.00787470117211342, 0.00578525010496378, 0.00312750949524343, -0.00122665055096149, 0.0020201918669045, -0.00773854646831751, -0.00648342724889517, -0.0234645791351795, 0.0211161989718676, 0.0152299245819449, -0.00468321377411485, -0.00534579390659928, -0.000306801754049957, -0.00664381915703416, -0.0175311788916588, 0.0187684111297131, 0.0145585769787431, -0.0252528134733438, -0.0112911220639944, 0.00680571189150214, 0.0218787305057049, -0.0141544612124562, -0.0501209087669849, -0.00139477697666734, -0.0126131782308221, -0.0156528484076262 },
      { -10.1391725540161, -7.01867151260376, 6.79128789901733, 0.503415286540985, -4.69984531402588, 3.75619912147522, 3.89908456802368, 8.73125076293945, -8.60106754302979, -6.27003479003906, -4.45712995529175, -23.6590957641602, -19.0229816436768, 4.09554815292358, -30.8267726898193, -2.80069518089294, -8.38675594329834, -27.6225643157959, -5.51200342178345, -15.8936367034912, -6.04267024993896, 1.28341662883759, 0.55491840839386, -5.16452217102051, 5.73705005645752, -4.15013408660889, -6.71190452575684, -2.35890126228333, -6.69595575332642, -2.13908243179321, -5.01045894622803, 2.10703611373901, -66.9505233764648, -362.801971435547, -11.0775728225708, -0.370742589235306, 11.0164852142334, -9.37808418273926, -9.94343280792236, -2.74258375167847, -16.3703899383545, -6.45443058013916, -6.28885316848755, 0.668324649333954, -36.400520324707, -6.8479962348938, -8.66606712341309, -9.25359058380127, -3.44517397880554, -0.0179662760347128, -6.44424295425415 }
    },
    {
      { 0.00971721578389406, 0.0356892980635166, 0.145426869392395, 0.0307090263813734, 0.0665511712431908, 0.0848902761936188, 0.0242891293019056, 0.0322963148355484, 0.0398223176598549, 0.0263887364417315, 0.0782791301608086, 0.0136648947373033, 0.0524307377636433, 0.037503682076931, 0.124854378402233, 0.0321389995515347, 0.0492636114358902, 0.0648551881313324, -0.0170815754681826, 0.177584171295166, 0.0350350998342037, 0.0300328060984612, 0.143729612231255, 0.124615520238876, -0.00124683405738324, 0.0703384578227997, 0.0397928357124329, 0.019272442907095, 0.0564541071653366, 0.0177760496735573, 0.0405866019427776, -0.005878577940166, 0.16734516620636, -0.0118200452998281, 0.062510758638382, 0.0211327914148569, 0.140787079930305, 0.0555251836776733, 0.16758169233799, 0.0722522065043449, 0.139239400625229, 0.170605942606926, 0.241599604487419, 0.0207685846835375, 0.0221340693533421, 0.0920381844043732, 0.0312838964164257, 0.140620693564415, 0.0308512821793556, 0.0286838039755821, 0.0280679035931826 },
      { -0.0715665891766548, -0.0340723432600498, -0.326363742351532, -0.0540465377271175, -0.125235974788666, -0.101064585149288, -0.111567199230194, -0.058492012321949, -0.0914570242166519, -0.134251892566681, -0.111116245388985, -0.36144757270813, -0.0791374295949936, -0.109499678015709, -0.34183931350708, -0.0760475322604179, -0.0967160612344742, -0.1298488676548, -0.0305969770997763, -0.265760749578476, -0.0603088364005089, -0.0612322688102722, -0.20341169834137, -0.190200880169868, -0.0353048779070377, -0.0408352650702, -0.107529282569885, -0.0929244682192802, -0.0504466891288757, -0.127536296844482, -0.0810135081410408, -0.036471463739872, -0.417408913373947, 0.00917650014162064, -0.115084379911423, -0.0717082098126411, -0.27404648065567, -0.0938313752412796, -0.186573714017868, -0.0622641779482365, -0.185138866305351, -0.143681779503822, -0.134671449661255, -0.0629229918122292, -0.00583786331117153, -0.110957428812981, -0.0478731915354729, -0.14366751909256, -0.0567641369998455, -0.083948090672493, -0.0502845719456673 },
      { 1.074347615242, 1.03179287910461, 1.18603694438934, 1.05553126335144, 1.05766177177429, 0.990672469139099, 1.09448027610779, 1.04689252376556, 1.07270479202271, 1.10473823547363, 1.0265519618988, 1.26491093635559, 1.03379344940186, 1.08167791366577, 1.21397030353546, 1.06593084335327, 1.07817041873932, 1.08791720867157, 1.04038393497467, 1.08019006252289, 1.03612768650055, 1.0642272233963, 1.08948791027069, 1.03183460235596, 1.04502773284912, 1.02661097049713, 1.08805787563324, 1.07995021343231, 1.00165343284607, 1.12443602085114, 1.07712030410767, 1.05229640007019, 1.25919854640961, 0.997985541820526, 1.07473480701447, 1.08068871498108, 1.1408805847168, 1.08296823501587, 1.06324791908264, 1.00217425823212, 1.05830049514771, 1.03771328926086, 0.969160199165344, 1.06747150421143, 0.985611081123352, 1.05031788349152, 1.03336405754089, 1.00613033771515, 1.03493201732635, 1.07521259784698, 1.05839312076569 },
      { -0.0172770563513041, -0.0126558840274811, -0.0401442833244801, -0.0141018657013774, -0.00883239042013884, 0.00194923114031553, -0.00337430997751653, -0.0137944780290127, -0.0190935879945755, -0.0024035528767854, -0.00715082697570324, -0.00712165236473083, -0.00682845525443554, -0.0144909406080842, -0.0246455390006304, -0.0126225464046001, -0.0228569395840168, -0.00898429378867149, -5.75564336031675e-05, 0.00202544638887048, -0.00109947950113565, -0.0119756031781435, 0.00117503176443279, 0.0243356972932816, -0.00599151710048318, -0.0135196875780821, -0.0189898498356342, -0.00518423318862915, 0.00810857489705086, -0.0203025434166193, -0.0186550281941891, -0.00112681568134576, -0.0641783028841019, -0.00344644789583981, -0.00788910128176212, -0.0242411270737648, -0.0421085245907307, -0.0293695162981749, -0.0058847819454968, -0.00798126962035894, 0.00374070880934596, -0.00750004313886166, 0.0103548839688301, -0.0283852778375149, 0.0073680505156517, 0.00855435337871313, 0.00484683969989419, 0.00154159974772483, -0.00444772280752659, -0.0069593614898622, -0.00673376070335507 },
      { 0.0131398048251867, 0.0113478526473045, 0.0595717765390873, 0.00442731147632003, 0.00296759442426264, -0.00365732167847455, 0.000486643344629556, 0.0218230988830328, 0.0120881013572216, -0.00611320836469531, 0.050293043255806, -0.0155487330630422, 0.00581111991778016, -0.00639277137815952, 0.0559706352651119, -0.00391571829095483, 0.0101753128692508, 0.00728741614148021, 0.00698760151863098, 0.00618366478011012, -0.00108982948586345, 0.00634228996932507, 0.0045225122012198, -0.0514620579779148, 0.00426009437069297, 0.0104342475533485, 0.000491835991851985, 0.00186872470658273, 0.00368455448187888, 0.00164208991918713, 0.00537484744563699, -0.00411156797781587, 0.0845112875103951, -0.00399149814620614, -0.00932389870285988, 0.00797257851809263, 0.0556733943521976, 0.0206795874983072, 0.0201240181922913, -0.0029289647936821, 0.00218945159576833, -0.015762934461236, -0.0207673739641905, 0.0138285718858242, -0.00415180809795856, -0.028260936960578, 0.00508078001439571, 0.00966520607471466, -0.000577792117837816, 0.000612811592873186, 0.0109427692368627 },
      { -0.00939090270549059, -0.014965302310884, -0.0231666248291731, -0.00928153935819864, 0.0156245697289705, 0.00836761947721243, -0.00445286650210619, -0.0283390413969755, -0.0116374334320426, 0.00691490015015006, -0.0686482191085815, -0.00366870709694922, 0.00543847260996699, 0.0152440033853054, -0.0296270195394754, 0.00278603821061552, -0.00936572346836329, -0.00823730137199163, -0.00131309207063168, 0.0406034700572491, 0.000681744830217212, -0.0174037776887417, -0.00163378170691431, 0.0528709329664707, -0.00826012901961803, -0.0185274910181761, 0.0016010218532756, -0.0088697662577033, 0.0018599204486236, 0.000504621828440577, -0.0066194199025631, -0.00707841478288174, -0.0243867859244347, 0.00314708426594734, 0.00990418251603842, -0.00998773798346519, -0.0211423542350531, -0.0219209119677544, -0.0148868588730693, 0.0243045892566442, 0.0119774350896478, 0.0420797728002071, 0.00480835931375623, -0.00783807411789894, 0.00110223039519042, 0.0276109967380762, -0.0108891734853387, 0.00321584846824408, -7.89682089816779e-05, -0.00999011006206274, -0.0263541713356972 },
      { -9.88065910339355, -11.7655410766602, 2.18730425834656, -0.524335741996765, -13.7614107131958, -2.63003587722778, -0.606450140476227, -2.94918394088745, -3.37108612060547, -12.5284299850464, -18.4802150726318, 33.1566390991211, -20.6254806518555, 0.315664410591125, 11.0535974502563, -4.50144672393799, -3.81646347045898, -13.9868268966675, -3.19555258750916, -42.1312675476074, -3.60748481750488, -5.83391618728638, -24.1120376586914, -34.2055854797363, -0.231652200222015, -15.5721254348755, -1.52521574497223, -6.40058898925781, -17.9907360076904, -3.14109516143799, -10.3585329055786, -2.14020943641663, 12.5903759002686, 16.1803512573242, -6.58552646636963, -1.22614562511444, 5.96978521347046, -6.55273151397705, -14.5833425521851, -11.0671606063843, -34.1318206787109, -32.2909469604492, -42.4512329101562, -2.03752756118774, -25.6370048522949, -19.7994117736816, -5.79101753234863, -28.2818813323975, -2.13031721115112, -0.354975908994675, -13.4482736587524 }
    },
    {
      { -0.17201828956604, -0.242596641182899, -0.0367136262357235, -0.00859749969094992, -0.305148094892502, -2.00567746162415, 0.0441417209804058, -0.147287383675575, -0.0807929560542107, -0.313991844654083, -0.879083454608917, -0.057895589619875, -0.0999472886323929, -0.1743124127388, 0.0893621891736984, -0.15805871784687, -0.0951845049858093, -0.235360264778137, 0.143389239907265, -2.15487861633301, -0.0240190159529448, -0.163027316331863, -1.83726096153259, -0.863673388957977, 0.0629901438951492, -0.460606783628464, 0.156454935669899, -0.0456017591059208, -0.0692345798015594, 0.16338013112545, -0.222579210996628, 0.116494379937649, 0.094865083694458, 0.659827649593353, -0.146507233381271, -0.00970995146781206, -0.0131719363853335, -0.0924364551901817, -1.52031302452087, -0.705470383167267, -1.45955336093903, -1.3439679145813, -2.77955460548401, 0.0614184625446796, 0.0510572157800198, -0.780847311019897, 0.054824810475111, -1.65898537635803, -0.0151154650375247, -0.0261041019111872, -0.0106385760009289 },
      { 0.100932903587818, -0.472540736198425, -0.5416339635849, -0.108569525182247, 0.253991693258286, 1.69794452190399, -0.346169054508209, -0.0634932219982147, 0.00373582169413567, 0.520926535129547, 0.790802180767059, -0.808695554733276, -0.0576269440352917, 0.116182334721088, -0.530975520610809, 0.0642077699303627, -0.00185582262929529, 0.232280090451241, -0.338160902261734, 2.71722674369812, -0.04780038818717, -0.207984358072281, 2.14024448394775, 2.8047354221344, -0.284525603055954, -0.309953898191452, -0.0975440219044685, 0.0791133493185043, -0.0995745360851288, -0.0113764954730868, 0.000375703821191564, -0.307562232017517, -0.454313576221466, -0.812944769859314, 0.0987701490521431, -0.105455882847309, -0.507016539573669, 0.0207945741713047, 1.05868244171143, 0.186655521392822, 1.51793742179871, 1.07528722286224, 1.40676498413086, -0.262557059526443, 0.0242560300976038, 0.388057798147202, -0.261122971773148, 2.13133478164673, -0.123580403625965, -0.133111640810966, -0.322849690914154 },
      { 0.760172128677368, 1.18145000934601, 1.37525188922882, 0.880119740962982, 0.815900266170502, 1.07630801200867, 1.12746512889862, 0.938929617404938, 0.756967127323151, 0.686846375465393, 1.04240810871124, 2.03174734115601, 1.03656148910522, 0.860845744609833, 1.26077520847321, 0.795775055885315, 0.747379064559937, 0.711099624633789, 1.00882112979889, 0.528817713260651, 0.856561601161957, 0.848316073417664, 0.271057158708572, -0.0956939831376076, 1.06795060634613, 1.13729393482208, 0.737111389636993, 0.775281310081482, 0.944588243961334, 0.637422502040863, 0.831334948539734, 0.991201996803284, 1.12577986717224, 1.32779932022095, 0.766917169094086, 0.791532218456268, 1.40383064746857, 0.777348637580872, 0.896315336227417, 1.41379141807556, 0.705601632595062, 0.853223562240601, 1.18534886837006, 0.861629068851471, 0.710702002048492, 0.950176894664764, 0.976531863212585, 0.61145007610321, 0.922459006309509, 0.873547732830048, 0.90251612663269 },
      { 0.307431608438492, 0.271897375583649, 0.532496333122253, 0.2156952470541, 0.253803461790085, 0.0047832471318543, 0.189594462513924, 0.211352348327637, 0.307820528745651, 0.172395750880241, 0.192215800285339, 0.0217424426227808, 0.170718118548393, 0.287685602903366, 0.508907973766327, 0.298553913831711, 0.312627375125885, 0.24498850107193, 0.207259505987167, 0.0397093631327152, 0.292226582765579, 0.224346086382866, 0.0518148951232433, -0.317129224538803, 0.170889556407928, 0.31078577041626, 0.288940817117691, 0.178510785102844, 0.212972581386566, 0.301175951957703, 0.278822034597397, 0.185931697487831, 0.801082074642181, 0.0869717299938202, 0.238178431987762, 0.327092200517654, 0.445511937141418, 0.290636390447617, 0.189859807491302, 0.164009973406792, 0.0829427018761635, 0.0939697995781898, -0.0459029376506805, 0.359121918678284, 0.228573128581047, 0.102487079799175, 0.184317916631699, -0.0392653718590736, 0.234983041882515, 0.282147973775864, 0.219778895378113 },
      { -0.120567925274372, -0.0643019750714302, -0.62560361623764, -0.000337924167979509, -0.0673106759786606, 0.197438657283783, -0.0490976050496101, -0.18262554705143, -0.0550765804946423, 0.0455072931945324, -0.745774686336517, 0.140461981296539, 0.0056594917550683, -0.1300318390131, -0.742987036705017, -0.0209854617714882, -0.0460812337696552, -0.0507471710443497, -0.0418611131608486, -0.00763441948220134, -0.046474065631628, -0.0284520648419857, -0.0471512787044048, 0.580873072147369, -0.0305814798921347, -0.133093148469925, -0.02530955709517, 0.0663419365882874, -0.011114870198071, -0.0446638613939285, -0.0319171994924545, -0.0146987130865455, -0.849323451519012, 0.0675572752952576, 0.0396661572158337, -0.0870210528373718, -0.562458395957947, -0.0380540825426579, -0.106729678809643, -0.00613287976011634, -0.023562528192997, 0.266539812088013, 0.175840094685555, -0.0950902700424194, -0.0274293404072523, 0.241799280047417, -0.00672818766906857, 0.0362655147910118, -0.00811214465647936, -0.0223894249647856, -0.110275655984879 },
      { 0.0985541269183159, 0.134678095579147, 0.304690629243851, 0.00642458908259869, -0.00365573726594448, -0.238358750939369, 0.0336222909390926, 0.261567503213882, 0.0402728952467442, -0.0868932232260704, 1.07861578464508, -0.0322705842554569, -0.0637054592370987, 0.0310043618083, 0.480158925056458, -0.00984786543995142, 0.0388491302728653, 0.0379329174757004, 0.0424329489469528, -0.502623021602631, -0.0102549074217677, 0.219719707965851, 0.00888066086918116, -0.484236478805542, 0.0341646410524845, 0.153739169239998, 0.0162835251539946, -0.0425245501101017, -0.00920772552490234, 0.0492565110325813, 0.0316881090402603, 0.0578148774802685, 0.252907574176788, -0.0761759653687477, -0.0524067990481853, 0.0773589387536049, 0.23573163151741, 0.0164057463407516, 0.0657134056091309, -0.189848303794861, -0.148845493793488, -0.461818069219589, 0.153887510299683, 0.0750422850251198, 0.0348294638097286, -0.255120903253555, 0.0173860881477594, -0.267672270536423, -0.0048666805960238, 0.0202874932438135, 0.245956465601921 },
      { 17.6066150665283, 22.1534156799316, -1.10363709926605, 2.71735596656799, 47.5707130432129, 40.0525512695312, 31.8953151702881, 39.9926452636719, 13.0087013244629, 49.2941627502441, 200.065994262695, -299.250091552734, 54.6281318664551, -10.1341257095337, -147.663757324219, -1.32547414302826, 10.2746315002441, 100.259399414062, 1.82973766326904, 403.673706054688, -5.92334747314453, 52.0417976379395, 264.560882568359, 449.346893310547, -11.0598754882812, 70.0621109008789, -35.2819900512695, 24.671760559082, 26.3118991851807, -42.6890411376953, 46.4782981872559, -5.24038219451904, -200.488235473633, 98.7243728637695, 22.8394718170166, -2.71198225021362, -13.6892576217651, 15.1328086853027, 21.687629699707, 109.575202941895, 316.693603515625, 212.609786987305, 523.891174316406, -7.45670080184937, -148.172897338867, 100.612365722656, 11.33167552948, 393.147399902344, 1.42006814479828, -14.6527509689331, 63.8606605529785 }
    },
    {
      { -0.065180778503418, -0.150914996862411, -0.0864668190479279, -0.125307247042656, -0.219344571232796, -1.23235309123993, 0.0622548796236515, -0.148465856909752, -0.294074505567551, -0.939068734645844, -0.840726673603058, -0.150931164622307, -0.244460180401802, -0.155071884393692, 0.162959069013596, -0.166789948940277, -0.397893130779266, -0.0393471904098988, 0.0467565953731537, -0.00865264609456062, -0.123111255466938, -0.241538241505623, -1.88347053527832, 2.29145908355713, -0.0873832106590271, 0.0627764165401459, 0.270995318889618, -0.36114114522934, -0.370655357837677, 0.106718651950359, -0.143510937690735, -0.0370270386338234, 0.174540624022484, 0.752952992916107, -0.313551694154739, 0.00828127097338438, 0.00775909144431353, -0.366840273141861, -0.543713450431824, -0.399846196174622, -0.33623594045639, -0.0494419038295746, -0.444024741649628, 0.0503564178943634, 0.0913928747177124, -0.765170514583588, 0.0759043768048286, 1.71270978450775, -0.0860534533858299, -0.059265423566103, 0.210590437054634 },
      { 0.484039723873138, 0.219069615006447, -0.236650839447975, -0.0294743832200766, 0.501120090484619, 2.28992366790771, -0.359122157096863, 0.334691345691681, 0.309552699327469, 1.79176115989685, 1.87431836128235, -1.26836538314819, 0.112988531589508, 0.202569931745529, -0.28563717007637, 0.311232775449753, 0.400201290845871, 0.770699739456177, -0.0647164881229401, 2.119713306427, 0.167931079864502, 0.00392426131293178, 2.59294390678406, 4.98267364501953, -0.0964829549193382, 0.832596123218536, 0.104350611567497, 1.05752122402191, -0.0455674342811108, 0.634580075740814, 0.816771805286407, 0.0596316121518612, -0.244600147008896, -0.591392338275909, 0.344017058610916, 0.0232798159122467, -0.31703707575798, 0.397257328033447, 0.887676000595093, 1.60922908782959, 1.89974129199982, 1.7610547542572, 3.71704125404358, -0.23331481218338, 0.0315834321081638, 1.58206760883331, -0.468059241771698, 3.76173257827759, -0.0792559757828712, -0.0937018617987633, -0.0760834962129593 },
      { -0.0150147285312414, -0.0539985150098801, 0.613840043544769, 0.447276413440704, -0.0688885599374771, -0.878174662590027, 0.586014688014984, 0.17329528927803, 0.184149965643883, -0.404451161623001, -0.843376934528351, 2.83121514320374, 0.360102564096451, 0.330502539873123, 0.388921111822128, 0.195970118045807, 0.129834845662117, -0.262745440006256, 0.314384698867798, -1.43444097042084, 0.515067100524902, 0.223538503050804, -0.680336892604828, -5.01322984695435, 0.356510698795319, -0.56078827381134, 0.0717655569314957, -0.217443019151688, 0.905248999595642, -0.291795372962952, -0.199531465768814, 0.170780941843987, 0.328212678432465, 0.544032454490662, 0.110983625054359, 0.189850628376007, 0.634488463401794, 0.115973629057407, -0.159929856657982, -0.385269999504089, -1.19892501831055, -1.03428292274475, -3.41878533363342, 0.335585296154022, 0.0806252062320709, -0.607511639595032, 0.714344382286072, -3.41126799583435, 0.624483942985535, 0.379214316606522, 0.255985587835312 },
      { 0.718806028366089, 0.851094305515289, 0.868852615356445, 0.585135817527771, 0.788512170314789, 0.661443054676056, 0.73153018951416, 0.690615832805634, 0.75353741645813, 0.72022408246994, 0.707112789154053, 0.0690499022603035, 0.708045482635498, 0.742008984088898, 0.850345134735107, 0.700690984725952, 0.780904471874237, 0.670776128768921, 0.692314028739929, 0.641218662261963, 0.694516003131866, 0.715827107429504, 0.621725142002106, 0.439232140779495, 0.805218517780304, 0.823870718479156, 0.730108201503754, 0.622507572174072, 0.444362580776215, 0.742105841636658, 0.71693879365921, 0.806774258613586, 0.991052687168121, 0.877410352230072, 0.801926970481873, 0.826027393341064, 0.847903847694397, 0.780853569507599, 0.749032139778137, 0.618077456951141, 0.682475686073303, 0.633741557598114, 0.497060865163803, 0.89488697052002, 0.821817338466644, 0.784452319145203, 0.460867017507553, 0.352934420108795, 0.593357980251312, 0.803326904773712, 0.824538350105286 },
      { -0.117931917309761, -0.126223102211952, -0.32174676656723, -0.0216592662036419, -0.220728129148483, 0.0458852648735046, -0.0421488918364048, 0.188896059989929, -0.136639297008514, -0.210663378238678, -0.0421129874885082, 0.596152067184448, -0.0920501351356506, -0.196735426783562, -0.29318705201149, -0.148024916648865, -0.139125242829323, 0.0685376152396202, 0.0175185911357403, 0.155535742640495, -0.117519304156303, 0.217651695013046, -0.0392900407314301, -0.00687263067811728, 0.0546357072889805, -0.172239005565643, -0.0680579841136932, -0.0251358468085527, -0.061526071280241, -0.0891641601920128, -0.0544249601662159, -0.0384657792747021, -0.386577695608139, -0.123199269175529, -0.19166325032711, -0.161289289593697, -0.314463317394257, -0.170270413160324, -0.0843372717499733, -0.0679590627551079, -0.0696734413504601, -0.0436464138329029, 0.056314866989851, -0.122225344181061, 0.0334340929985046, -0.255341082811356, -0.00578211434185505, 0.982059299945831, -0.00577048445120454, -0.0762730464339256, -0.184467181563377 },
      { 0.0510764308273792, 0.190040677785873, 0.136863738298416, 0.0491599701344967, 0.253393530845642, -0.0229010451585054, -0.00515885977074504, -0.322712063789368, 0.116129145026207, 0.164619505405426, -0.0516220405697823, -0.440408080816269, 0.0497493073344231, 0.0375031530857086, 0.179881379008293, 0.127766877412796, 0.112578168511391, -0.0217547751963139, 0.0380845181643963, -0.0910014063119888, -0.0683711245656013, -0.0603114068508148, 0.0330620370805264, -0.341716796159744, -0.0881112888455391, 0.0993854552507401, 0.0450888015329838, 0.00719340424984694, -0.0255016777664423, 0.100902110338211, -0.0138193434104323, 0.0467264875769615, 0.0905494764447212, 0.124208174645901, 0.171161413192749, 0.125890105962753, 0.109615556895733, 0.14502939581871, 0.11254994571209, 0.0689323991537094, 0.279914349317551, 0.199685379862785, 0.584188163280487, 0.0633445903658867, -0.0427483431994915, 0.296901106834412, 0.0578646622598171, -1.49531102180481, -0.0289860125631094, 0.0366412550210953, 0.106872498989105 },
      { 62.2494583129883, -96.4754028320312, 83.0819091796875, 8.88855266571045, -15.7965450286865, -85.8505477905273, 56.4693298339844, 61.4585342407227, 21.7861251831055, -22.4379043579102, -34.6582221984863, -433.601928710938, 73.0337295532227, 51.3913688659668, -67.1015167236328, 0.767322421073914, 32.4722023010254, -176.407440185547, -26.0634460449219, -253.258071899414, -0.111339934170246, 32.1725082397461, -51.4600296020508, -13.020378112793, 19.4322509765625, -53.8126792907715, -66.1938323974609, 46.8970489501953, 149.356826782227, -103.512504577637, -14.4138898849487, -8.13965511322021, -62.4649429321289, -742.360473632812, 29.3738307952881, -9.58422946929932, 66.1989440917969, 46.4147109985352, -89.656005859375, -36.2035255432129, -186.721786499023, -171.448226928711, 40.7951049804688, -3.71577501296997, -111.014801025391, -43.5981864929199, 46.8605613708496, -356.864410400391, 4.81441831588745, -16.7333164215088, -25.5409736633301 }
    },
    {
      { 0.0476044751703739, 0.362214773893356, 0.0889710113406181, -0.136719405651093, -0.143537551164627, 0.529134035110474, -0.0219790674746037, 0.144489675760269, -0.221079349517822, -0.639322638511658, -0.575065314769745, 0.157652780413628, -0.061602771282196, -0.163282722234726, 0.0721407309174538, -0.0639768242835999, -0.329034954309464, 0.687755703926086, -0.123478479683399, 1.47349333763123, -0.141005858778954, 0.105192415416241, 0.184625700116158, 0.214518681168556, -0.1579779535532, 0.691850543022156, -0.0142434127628803, -0.316175818443298, -0.369056910276413, -0.133336663246155, 0.0205020494759083, -0.193018361926079, 0.0758805721998215, 0.297011762857437, -0.116517446935177, 0.0200492236763239, 0.139823451638222, -0.275789827108383, 0.322585433721542, 0.124200962483883, 0.786776185035706, 1.05435156822205, 0.2351114153862, 0.0586561374366283, -0.0416156910359859, -0.0843512713909149, 0.119202993810177, 0.636501014232635, -0.0727210491895676, -0.0513264127075672, 0.462930619716644 },
      { 0.215971305966377, -0.262734532356262, -0.0639262124896049, 0.0227344296872616, 0.309032022953033, 0.342773288488388, -0.0681634768843651, -0.0199066121131182, 0.125670924782753, 1.17741167545319, 0.408436924219131, -0.55633157491684, 0.0380520112812519, 0.312011808156967, 0.0649692043662071, 0.154708921909332, 0.228813484311104, -0.353447169065475, 0.205319702625275, 0.0998400002717972, 0.165958151221275, -0.495619863271713, 0.0936328545212746, 0.508964121341705, 0.141621083021164, 0.291414737701416, 0.377053558826447, 0.911768317222595, 0.00980598013848066, 0.701418220996857, 0.665377616882324, 0.288228690624237, -0.00938629265874624, 0.400743186473846, -0.0181236620992422, -0.0384795963764191, -0.188811510801315, 0.226735815405846, -0.130134016275406, 0.530659437179565, 0.796178579330444, 0.509161055088043, 1.01012241840363, -0.239121973514557, 0.0400801338255405, 0.966299116611481, -0.413800358772278, 0.253588676452637, -0.107233233749866, -0.100330464541912, -0.544598639011383 },
      { -0.0484844744205475, -0.120596751570702, 0.0864302963018417, 0.228610292077065, -0.110696271061897, -0.881939649581909, 0.185628622770309, 0.202253296971321, 0.150622576475143, -0.464093148708344, -0.120288260281086, 0.310514241456985, -0.25553959608078, -0.0480877123773098, -0.0748109146952629, 0.0789297670125961, 0.102738603949547, 0.136563554406166, -0.0976809039711952, -1.47965574264526, 0.291307181119919, 0.346815437078476, -0.152485489845276, -0.991806924343109, -0.0655440613627434, -0.715200483798981, -0.170909821987152, -0.297469973564148, 0.609639167785645, -0.361840516328812, -0.324577331542969, -0.142333596944809, 0.0223675798624754, -0.282144784927368, 0.054512694478035, 0.100708119571209, 0.0862866640090942, 0.0844563022255898, -0.0527644641697407, -0.0808728188276291, -1.45565187931061, -1.16235852241516, -1.49507641792297, 0.202272400259972, 0.0246360972523689, -0.811441242694855, 0.37912392616272, -0.51354992389679, 0.348537117242813, 0.232829988002777, 0.515974760055542 },
      { 0.914784014225006, 1.02401697635651, 0.798155725002289, 0.781170010566711, 0.927139163017273, 0.986213207244873, 0.961092174053192, 0.851451635360718, 0.901927292346954, 0.984696686267853, 1.11345016956329, 1.02389907836914, 1.05633687973022, 1.00668168067932, 0.790299773216248, 0.88623172044754, 0.903931438922882, 0.74194997549057, 0.974170088768005, 0.936859965324402, 0.882720291614532, 0.93054735660553, 1.0133969783783, 1.18839859962463, 1.06304919719696, 1.06897640228271, 0.894468367099762, 0.84160453081131, 0.661004245281219, 0.920704066753387, 0.912156462669373, 1.04646229743958, 0.585081696510315, 0.808973014354706, 1.09558081626892, 0.955635488033295, 0.860743045806885, 0.902766346931458, 1.01986134052277, 0.887120962142944, 0.894780158996582, 1.00307953357697, 0.930860042572021, 0.98402339220047, 1.00568735599518, 1.08617532253265, 0.758416354656219, 0.849209189414978, 0.83365523815155, 0.944481194019318, 1.06357026100159 },
      { -0.049786277115345, -0.0281059611588717, 0.227954030036926, 0.00723370304331183, -0.136402100324631, 0.00604808935895562, -0.0688296779990196, 0.487285614013672, -0.0787055939435959, -0.219353452324867, 0.529345035552979, -0.414539694786072, -0.106190174818039, -0.170328199863434, 0.337070256471634, -0.120689779520035, -0.0690929889678955, 0.231003507971764, 0.0861532464623451, 0.0229271501302719, -0.0791129395365715, 0.384511500597, -0.0141348037868738, -0.653727889060974, 0.110989414155483, -0.247919723391533, -0.0419499538838863, -0.0277944225817919, -0.0314653739333153, -0.0683627799153328, -0.034956943243742, 0.0144320307299495, 0.428508311510086, -0.087172344326973, -0.351073741912842, -0.125082090497017, 0.162974715232849, -0.102549195289612, -0.116569988429546, 0.145820647478104, -0.00846929103136063, -0.345555275678635, 0.075721375644207, -0.0148399509489536, 0.0414121299982071, -0.449519038200378, -0.0178171880543232, 0.0217408966273069, -0.00157826521899551, -0.0546455420553684, -0.140391170978546 },
      { -0.0240191090852022, 0.0558525174856186, -0.160597413778305, 0.0199227407574654, 0.189082458615303, -0.0407521240413189, 0.00469158915802836, -0.760433077812195, 0.0532413981854916, 0.206786334514618, -1.745232462883, 0.602182865142822, 0.210662707686424, 0.0523876771330833, -0.228010520339012, 0.102464944124222, 0.0489848516881466, -0.118050910532475, -0.0375071130692959, 0.473004549741745, -0.0735193938016891, -0.391111671924591, 0.0368491113185883, 0.38367885351181, -0.174559727311134, 0.173694103956223, 0.0255151018500328, -0.0279936660081148, -0.0276801232248545, 0.0661658868193626, -0.0479503460228443, -0.0474057607352734, -0.0754169449210167, 0.074280247092247, 0.330731183290482, 0.101313792169094, -0.0731960460543633, 0.086429052054882, 0.101977750658989, -0.317003488540649, 0.592661321163177, 0.694504618644714, 0.0965401157736778, -0.00927139166742563, -0.0562116801738739, 0.53497326374054, 0.0572047308087349, 0.430385798215866, -0.0154154915362597, 0.0205405782908201, -0.130789071321487 },
      { 39.0729904174805, -89.6992950439453, 34.280345916748, -8.24222183227539, -2.29806351661682, -55.5586128234863, 17.1082096099854, 58.1679229736328, 21.769079208374, -34.0547294616699, -70.8722763061523, -169.875122070312, -63.4818954467773, 1.36056816577911, 56.064567565918, -9.87407112121582, 30.3074264526367, -374.512054443359, -22.4126243591309, -270.043029785156, 1.17534363269806, 12.0654582977295, -305.94091796875, -59.6177253723145, 26.598783493042, -68.7290267944336, -30.1116256713867, 39.3309555053711, 121.50244140625, -75.9782943725586, -40.4994659423828, 0.994220018386841, 103.428329467773, 97.3950729370117, -5.27660131454468, -6.30756139755249, 39.9119834899902, 34.5453567504883, -98.5861587524414, -62.6427536010742, -414.181793212891, -301.147216796875, -84.2122573852539, 5.96177434921265, 19.7086791992188, -69.7050247192383, 23.3369464874268, 31.3824043273926, -4.34305286407471, -8.12492752075195, -54.1859588623047 }
    },
    {
      { -0.0205236505717039, -0.0142400935292244, -0.0312465727329254, -0.0108241243287921, -0.0264325104653835, 0.0147584797814488, -0.00084386405069381, -0.0033724014647305, -0.00638880021870136, -0.0135864894837141, -0.0330607257783413, 0.014243527315557, -0.0216688942164183, -0.0124020157381892, 0.0164417643100023, -0.00852232985198498, -0.016784431412816, -0.0805755779147148, 0.00431242631748319, -0.0190615430474281, -0.0160250123590231, 0.00533151254057884, 0.0607954934239388, -0.105059817433357, -0.000256930332398042, -0.0124061219394207, 0.00353703461587429, -0.036310687661171, -0.0203941985964775, -0.00832179840654135, -0.023505637422204, -0.000666347099468112, 0.0096850236877799, -0.0488568730652332, -0.020523888990283, -0.0091479504480958, -0.0250339023768902, -0.0197881646454334, -0.00599215272814035, -0.0261441450566053, -0.028731195256114, -0.0379730090498924, -0.0585347637534142, 0.00286551425233483, -0.000955232826527208, -0.0328792333602905, -0.00654281070455909, -0.0550498180091381, -0.020972803235054, -0.00959345232695341, -0.00473914155736566 },
      { 0.00871511641889811, 0.00580994877964258, 0.00221111578866839, 0.00705581204965711, 0.019004812464118, -0.0116078844293952, 0.007375692948699, 0.0138546694070101, 0.00199664756655693, 0.0406814143061638, 0.0223644468933344, -0.0513266995549202, 0.0144616048783064, 0.0336625203490257, -0.0466021820902824, 0.00329822674393654, 0.00957022048532963, 0.102604478597641, 0.00297952420078218, 0.0402137786149979, -0.00132075138390064, -0.00777249643579125, -0.0451895818114281, -0.0743727162480354, 0.0044889566488564, 0.0120741920545697, 0.00246758037246764, 0.0396261513233185, 0.00427822163328528, 0.0238404422998428, 0.00711898365989327, 0.00573745090514421, -0.0218429584056139, 0.072840616106987, 0.0172196328639984, 0.00486725894734263, 0.0174381043761969, 0.0126588428393006, -0.0238269567489624, 0.00925562810152769, 0.0409497059881687, 0.0285399705171585, 0.0120800016447902, -0.0032100067473948, 0.00227449391968548, 0.041363887488842, 0.0105098485946655, -0.0734418854117393, 0.0165804792195559, 0.012258118018508, -0.00751887261867523 },
      { 0.0108894659206271, -0.00340487621724606, 0.0273803267627954, 0.00741444248706102, 0.00553701305761933, -0.00209195259958506, -0.00580587750300765, -0.011158567853272, 0.00910174567252398, -0.0176527723670006, -0.00200878898613155, -0.00339207611978054, 0.0198695436120033, -0.0176252759993076, 0.0305497720837593, 0.0038724064361304, 0.00585651164874434, -0.031776886433363, 0.00132529460825026, -0.0301262270659208, 0.0130948266014457, -4.29606297984719e-05, 0.00140373804606497, 0.131856843829155, -0.00127311621326953, -0.0045932293869555, -0.00253797066397965, -0.00627208640798926, 0.0229236874729395, -0.0127937328070402, 0.0107658123597503, -0.000813801772892475, 0.00406965659931302, -0.0376187115907669, 0.0064674187451601, -0.00472923927009106, 0.0242642257362604, 0.00312679121270776, 0.0183462910354137, 0.0213216133415699, -0.0137477796524763, -0.004939632024616, 0.00337152043357491, 0.00126481184270233, 0.000660408055409789, -0.0116907740011811, 0.000795399770140648, 0.119747675955296, 0.00268490891903639, -0.00308326026424766, 0.000892607087735087 },
      { 1.00092720985413, 1.0022691488266, 1.05500960350037, 0.991839468479156, 0.999805510044098, 1.00095987319946, 0.996289074420929, 1.00147151947021, 0.998571217060089, 0.996080040931702, 1.00280094146729, 1.01885485649109, 0.987727522850037, 0.99562793970108, 1.05088233947754, 1.00165283679962, 1.00045299530029, 0.999958634376526, 0.995029985904694, 1.01195859909058, 0.994901478290558, 1.00076067447662, 1.00736546516418, 1.01251184940338, 0.997410535812378, 0.999743640422821, 0.99816632270813, 0.995229661464691, 0.988905072212219, 1.00024271011353, 1.00030314922333, 0.997274219989777, 1.11207258701324, 0.993584334850311, 0.99400782585144, 1.00732982158661, 1.03490304946899, 1.0010814666748, 1.00882959365845, 1.00572669506073, 1.0021185874939, 1.0030380487442, 1.00228381156921, 1.00506865978241, 0.997758567333221, 0.993357241153717, 1.00046324729919, 1.01408421993256, 0.996714890003204, 0.997093737125397, 1.00149846076965 },
      { -0.01094751060009, -0.00873720925301313, -0.108776994049549, 0.000861097825691104, -0.0102993119508028, -0.0044019902125001, 0.00460514752194285, -0.00909044127911329, -0.00678138760849833, -0.00495504029095173, -0.0209958553314209, -0.0660677924752235, 0.000967134488746524, 0.00505382427945733, -0.116540096700191, 0.00178224942646921, -0.0068896459415555, -0.0118105513975024, -0.00451368931680918, -0.0317757613956928, -0.000158088281750679, -0.00531775411218405, -0.00245931348763406, 0.00486636068671942, -0.00222648656927049, -0.00623218202963471, -0.000819536449853331, -0.00756389042362571, -0.00178609031718224, -0.00563647365197539, -0.00664444034919143, -0.00351909385062754, -0.148761332035065, -0.00940726883709431, 0.000618454068899155, -0.00673180026933551, -0.0876847952604294, -0.00947166513651609, -0.0126397516578436, -0.0216396488249302, -0.0165320429950953, -0.0168600045144558, 0.0105894915759563, -0.00883858371526003, 0.00215587252750993, -0.000344056636095047, -0.0109855029731989, -0.0347729325294495, 1.10080927697709e-05, -0.0039990646764636, -0.00913120713084936 },
      { 0.00505410367622972, 0.0107659576460719, 0.0587048642337322, 0.000249194534262642, 0.00716048339381814, 0.00741483038291335, -0.0021858757827431, 0.00855289865285158, 0.00355835701338947, 0.00251434440724552, 0.0359527319669724, 0.0984358340501785, -0.00522287655621767, -0.00583360297605395, 0.0763701796531677, -0.00579678080976009, 0.00339959305711091, 0.00763491680845618, 0.000728341809008271, 0.0193836186081171, 0.00115624780301005, 0.00624013785272837, 0.00331905111670494, -0.0396853312849998, 0.00256912573240697, 0.00601867912337184, 0.000799389556050301, 0.00981887523084879, -0.00117157876957208, 0.00641122087836266, 0.0031120718922466, 0.00284187658689916, 0.0374018214643002, 0.00101430271752179, -0.0035567469894886, 0.00536711886525154, 0.0369647219777107, 0.00689979875460267, 0.00814168062061071, 0.0121790887787938, 0.0130578009411693, 0.019946962594986, -0.0184428412467241, 0.00358471903018653, -0.00279627554118633, 0.001350810052827, 0.00935026723891497, 0.0353825502097607, 0.000537700951099396, 0.00550740351900458, 0.0134973125532269 },
      { 4.40305137634277, -0.367791056632996, -4.52862167358398, -1.07837462425232, 3.14584922790527, -3.62262558937073, 0.132928475737572, -1.13510298728943, 2.30195808410645, 4.28453254699707, -2.63619613647461, -13.0943212509155, 4.46723127365112, -1.61029052734375, -29.4372158050537, 0.251501053571701, 3.12998962402344, 11.3860664367676, 1.37605154514313, -2.45167064666748, 1.6195011138916, 0.318560034036636, -16.90260887146, -4.97251224517822, -1.09470248222351, 2.3143470287323, -2.16489195823669, 3.98297166824341, 5.5012469291687, -3.37792229652405, 2.58859872817993, -1.20126521587372, -38.8114814758301, 42.2259254455566, 4.03897380828857, -0.538998484611511, -7.17605209350586, 3.99963617324829, -4.31873083114624, -1.8729932308197, 1.73970282077789, -1.72557067871094, -0.761396527290344, -0.913208544254303, -1.31354916095734, 5.27413654327393, -3.17125344276428, -6.07368040084839, -1.53467631340027, -0.180658042430878, 1.37784206867218 }
    },
    {
      { 0.055986076593399, -0.00936128757894039, 0.161373659968376, 0.08599803596735, 0.195281639695168, -0.327889680862427, 0.0113664753735065, -0.0224298611283302, 0.119356773793697, 0.344161450862885, 0.370166748762131, -0.113104641437531, 0.155161842703819, 0.126000002026558, -0.126539215445518, 0.0602365024387836, 0.197014898061752, 0.0761615410447121, -0.00782158505171537, -0.155455976724625, 0.103601306676865, -0.0239295791834593, -0.186526611447334, 0.342410504817963, 0.0346690826117992, -0.133811831474304, -0.0347856655716896, 0.207996189594269, 0.226179376244545, 0.0856339260935783, 0.0356179587543011, 0.0368405804038048, -0.109407864511013, 0.115909799933434, 0.181193470954895, -0.0270318929105997, 0.098513163626194, 0.205891758203506, -0.0537203699350357, 0.122333392500877, -0.0219190996140242, -0.0559133663773537, 0.38237527012825, -0.0509800910949707, 0.0169850215315819, 0.291393220424652, -0.0108784809708595, 0.0721830129623413, 0.0454461015760899, 0.0399330258369446, -0.143220499157906 },
      { -0.170964539051056, -0.0158479698002338, 0.0310723930597305, -0.0281728114932775, -0.242768928408623, -0.186279654502869, -0.00577803561463952, -0.0495618879795074, -0.0622535161674023, -0.703682363033295, -0.451066851615906, 0.485469847917557, -0.0989623069763184, -0.256728649139404, 0.25573006272316, -0.0963898003101349, -0.137688204646111, -0.323360532522202, -0.0224893931299448, -0.461651116609573, -0.041322935372591, 0.167284354567528, -0.0136457597836852, -0.148299172520638, -0.0101670566946268, -0.257531106472015, -0.144326761364937, -0.451312452554703, -0.015991572290659, -0.443335503339767, -0.254977077245712, -0.0733018592000008, 0.151478186249733, -0.203579902648926, -0.138478130102158, 0.0293358564376831, -0.0116580352187157, -0.156451746821404, 0.0812366530299187, -0.391695857048035, -0.592240929603577, -0.513283252716064, -0.662875115871429, 0.146194010972977, -0.032140426337719, -0.655556559562683, 0.140884339809418, -0.0957039073109627, 0.0709901452064514, 0.0447211712598801, 0.175264894962311 },
      { 0.00329107488505542, 0.0888636559247971, -0.232058703899384, -0.14339542388916, 0.0627213940024376, 0.476710468530655, -0.0614821016788483, -0.0404856838285923, -0.103320077061653, 0.232793927192688, 0.179549798369408, -0.277316749095917, -0.150372862815857, 0.0208020228892565, -0.168984442949295, -0.0630999654531479, -0.0650328993797302, 0.12947054207325, -0.0290764085948467, 0.610303819179535, -0.203596487641335, -0.0852708518505096, 0.100468941032887, 0.0170315280556679, -0.0312468744814396, 0.323441296815872, 0.049016397446394, 0.136068984866142, -0.350423693656921, 0.224968895316124, 0.0998645573854446, 0.0146204959601164, -0.0280615892261267, 0.109194688498974, -0.0748707354068756, -0.0291971992701292, -0.213741227984428, -0.0402264222502708, -0.0171728581190109, 0.0652247816324234, 0.579968690872192, 0.455603897571564, 0.634419500827789, -0.109400659799576, -0.0145497173070908, 0.346550524234772, -0.193496659398079, 0.00116220838390291, -0.207467436790466, -0.125250980257988, -0.102473244071007 },
      { 1.06230235099792, 0.991291105747223, 0.754699528217316, 1.13996744155884, 1.04297339916229, 1.01922929286957, 1.0548642873764, 1.05410635471344, 1.06669330596924, 1.05789375305176, 0.990976452827454, 0.978570342063904, 1.11596190929413, 1.03201460838318, 0.782467544078827, 1.07989466190338, 1.0526270866394, 1.07971858978271, 1.07177138328552, 0.984948694705963, 1.08669495582581, 1.03525817394257, 0.978690326213837, 0.930281281471252, 1.00912487506866, 1.00787842273712, 1.07116544246674, 1.09078574180603, 1.19831418991089, 1.05221307277679, 1.04525244235992, 1.01195597648621, 0.477267324924469, 1.06008231639862, 1.0566064119339, 1.00248503684998, 0.854973614215851, 1.04384541511536, 0.976744472980499, 1.02222204208374, 1.04274713993073, 1.02372372150421, 1.04075491428375, 0.977055549621582, 1.02996504306793, 1.0293824672699, 1.14553475379944, 1.0157778263092, 1.09978902339935, 1.04707503318787, 0.983850181102753 },
      { 0.0480358675122261, 0.0818257257342339, 0.562407255172729, -0.0103110997006297, 0.0954342111945152, -0.0102134896442294, -0.0147322108969092, -0.079889327287674, 0.054145235568285, 0.107169024646282, 0.0381458327174187, 0.423616468906403, 0.0253138933330774, 0.0914804562926292, 0.571455836296082, 0.0418195351958275, 0.0545467250049114, 0.00291942292824388, -0.0239804442971945, 0.153636962175369, 0.0429256521165371, -0.079815000295639, 0.0176051612943411, 0.140844658017159, -0.0178827587515116, 0.115061320364475, 0.0111093884333968, 0.0465261787176132, 0.0205624829977751, 0.0352501384913921, 0.047226183116436, 0.022692209109664, 0.736261248588562, 0.0319706983864307, 0.0652008801698685, 0.0773540958762169, 0.463759303092957, 0.0910039320588112, 0.0743984803557396, 0.118174754083157, 0.102575838565826, 0.175861030817032, -0.00802174489945173, 0.0715989246964455, -0.026259008795023, 0.148378849029541, 0.0329283848404884, 0.0435285121202469, 0.00782242696732283, 0.0341989360749722, 0.0995803698897362 },
      { -0.00177941808942705, -0.112830214202404, -0.286109656095505, -0.00426308903843164, -0.12423337996006, 0.0292960871011019, 0.0239597633481026, 0.163502141833305, -0.0389999896287918, -0.0854069963097572, 0.12118773907423, -0.660389065742493, -0.00309681030921638, 0.00485782511532307, -0.362379640340805, -0.0278349332511425, -0.035407580435276, -0.0144104650244117, -0.00199596211314201, -0.212095275521278, 0.0178898423910141, 0.039856169372797, -0.0223379377275705, 0.12040401250124, 0.0377541594207287, -0.122463569045067, -0.00588672701269388, -0.0422137565910816, 0.0120762567967176, -0.0367663726210594, -0.0175351072102785, -0.0124854128807783, -0.186177030205727, -0.0150525234639645, -0.0431549549102783, -0.0570103451609612, -0.189379513263702, -0.080609567463398, -0.0678478851914406, -0.0298499837517738, -0.262309581041336, -0.291907250881195, -0.0990673825144768, -0.0339412875473499, 0.0323339477181435, -0.177617117762566, -0.0573112517595291, -0.147098869085312, -0.000275201164186001, -0.0276149827986956, -0.0551064908504486 },
      { -41.1110877990723, 32.8596801757812, 0.593643069267273, 8.63203716278076, -13.9917182922363, 27.9574699401855, -14.0353441238403, -15.6742544174194, -23.2509670257568, -16.143180847168, 25.1236267089844, 151.929504394531, -39.438533782959, -1.32979226112366, 149.118530273438, -2.90847420692444, -30.2321453094482, 45.2645645141602, 8.1054801940918, 86.5462799072266, -7.15282821655273, -11.449821472168, 195.28010559082, 34.3855857849121, -5.41233539581299, 11.1205730438232, 37.1582679748535, -35.4110221862793, -76.4455795288086, 57.6136131286621, 4.45784378051758, 6.58525228500366, 179.62239074707, -250.021377563477, -27.911304473877, 1.85202777385712, 13.9265480041504, -36.756534576416, 59.9304885864258, 24.0757942199707, 102.11400604248, 97.4285049438477, 1.2966685295105, 1.9178808927536, 7.8709454536438, -8.80258464813232, -7.98355150222778, 83.576286315918, -4.97827816009521, -0.721746563911438, 8.0327672958374 }
    },
    {
      { 0.189606964588165, 0.246530786156654, -0.0384899936616421, -0.0167116280645132, 0.295423984527588, -0.0411487892270088, 0.0842482298612595, 0.0220725405961275, -0.101343154907227, 0.194082617759705, 0.17957216501236, 0.00711879599839449, 0.0520254634320736, -0.00769012421369553, -0.0397130362689495, 0.0712518990039825, -0.107517212629318, 0.517899394035339, 0.218642055988312, 0.621205389499664, 0.032461941242218, 0.0630797371268272, 0.0251691974699497, 0.794997990131378, 0.0376700609922409, 0.27983745932579, 0.0985638797283173, 0.384397089481354, -0.0693989098072052, 0.271134823560715, 0.29844868183136, 0.0647395402193069, -0.0233179684728384, 0.0631189793348312, 0.00132797448895872, 0.139702335000038, -0.0253642592579126, -0.0151150617748499, 0.149423405528069, 0.271809279918671, 0.499651998281479, 0.707660436630249, 0.523747324943542, 0.136644423007965, -0.0728965997695923, 0.350157529115677, 0.111183725297451, 0.711347937583923, 0.101798675954342, 0.0862168520689011, 0.177289307117462 },
      { -0.133852601051331, -0.312810868024826, -0.0129357287660241, 0.0577501021325588, -0.259823054075241, -0.0506824180483818, -0.481232881546021, -0.0553181059658527, 0.17510761320591, -0.527028024196625, -0.300028592348099, 0.039936937391758, 0.0304177384823561, -0.143416807055473, 0.108416333794594, -0.0337320268154144, 0.233421593904495, -0.567192256450653, -0.328093409538269, -0.445281594991684, 0.0235166810452938, -0.157213360071182, -0.0367419049143791, 0.117397546768188, -0.149353131651878, -0.205768883228302, -0.383747041225433, -0.27016893029213, 0.0399702563881874, -0.437352806329727, -0.0843938142061234, -0.112608268857002, 0.0473379902541637, -0.178584083914757, 0.0957277342677116, -0.152809306979179, -0.0800514370203018, -0.0730113312602043, 0.0257452875375748, -0.198231220245361, -0.456318467855453, -0.478752046823502, 0.0455038025975227, -0.20298108458519, 0.0784681662917137, -0.447047084569931, -0.184302479028702, -0.0597630627453327, -0.12088880687952, -0.151236981153488, -0.212800905108452 },
      { 0.00310885300859809, 0.109619230031967, 0.0829997807741165, -0.189627036452293, 0.0971178263425827, 0.0454317107796669, 0.405746608972549, 0.0920277014374733, -0.149246066808701, 0.201468303799629, -0.0446973107755184, -0.035557396709919, -0.573419630527496, 0.214861333370209, -0.0650038123130798, 0.0273023545742035, -0.204563841223717, 0.247823506593704, 0.10431233048439, -0.0399550050497055, -0.0219952184706926, 0.127987295389175, 0.0172300152480602, -0.61999374628067, 0.15683451294899, 0.0187006909400225, 0.325283050537109, -0.0146611351519823, -0.0604715459048748, 0.215843707323074, -0.105015806853771, 0.102452628314495, -0.00405591819435358, 0.0855506137013435, -0.242061287164688, 0.0667013749480247, 0.112501092255116, -0.21219190955162, -0.0849875137209892, 0.0111317941918969, 0.0334082916378975, -0.0637793019413948, -0.324358731508255, 0.0767953544855118, -0.0145019078627229, 0.148464843630791, 0.0389352478086948, -0.39421209692955, 0.0659227147698402, 0.078514039516449, 0.191592514514923 },
      { 0.00621548341587186, 0.0168726518750191, -0.211792349815369, 0.143781781196594, 0.0226599257439375, -0.00475799897685647, 0.000294698344077915, -0.0216534920036793, 0.0282444208860397, 0.0632082372903824, 0.026361832395196, -0.00400412548333406, 0.326488882303238, -0.12295076996088, -0.179581925272942, -0.0305758342146873, 0.0160100385546684, -0.0812884047627449, -0.00254511088132858, -0.0214852150529623, 0.0242755021899939, -0.0107718044891953, -0.00232327333651483, -0.0230404045432806, -0.0246391966938972, 0.0212435740977526, 0.0532924570143223, 0.0176476053893566, 0.0885847210884094, 0.00540856597945094, 0.015077511779964, 0.0118522131815553, -0.35463872551918, -0.00267232023179531, 0.132960945367813, -0.026770206168294, -0.147868722677231, 0.17122195661068, -0.020528644323349, -0.0303178932517767, 0.0358378477394581, 0.00755466660484672, -0.0443521924316883, 0.0206082947552204, 0.0127014219760895, 0.0700764879584312, -0.0196584835648537, -0.0264378432184458, -0.007444326300174, -0.0052818818949163, 0.00471329083666205 },
      { 1.0579491853714, 1.03396189212799, 1.33401072025299, 1.00867021083832, 1.04578125476837, 1.0672789812088, 1.05761277675629, 1.13886654376984, 1.11542701721191, 1.0185182094574, 1.21895337104797, 0.930464327335358, 0.946146547794342, 1.10592901706696, 1.38081502914429, 1.00457155704498, 1.13132333755493, 1.16007542610168, 1.26340866088867, 1.14803087711334, 1.03014540672302, 1.14811766147614, 0.996168732643127, 1.06826674938202, 1.17029464244843, 0.994265139102936, 1.01405906677246, 1.00340807437897, 1.00863337516785, 1.03447675704956, 1.03160738945007, 1.10651981830597, 1.47881126403809, 1.05255973339081, 0.992980241775513, 1.04665446281433, 1.25556802749634, 0.964452922344208, 1.02354311943054, 1.12334036827087, 1.05834662914276, 1.09458577632904, 1.16865396499634, 1.08214259147644, 1.07887017726898, 0.958662807941437, 1.05440580844879, 1.20589554309845, 1.06025457382202, 1.06019687652588, 1.07116663455963 },
      { -0.076291136443615, -0.0546290352940559, -0.172733545303345, -0.0257990341633558, -0.123998090624809, -0.0992328077554703, -0.0765438303351402, -0.19903726875782, -0.0970388352870941, -0.0166990607976913, -0.391731977462769, 0.0708598122000694, 0.131552830338478, -0.0574484020471573, -0.243312701582909, -0.00449525751173496, -0.10189700871706, -0.0875329524278641, -0.264816969633102, -0.0726052597165108, -0.0611156970262527, -0.189075618982315, 0.000106053157651331, -0.0834653824567795, -0.216385692358017, -0.00822984520345926, -0.0536270029842854, 0.0182430222630501, -0.0447611734271049, -0.0487088672816753, -0.0396123975515366, -0.180958330631256, -0.131252110004425, -0.0517872497439384, 0.0305387545377016, -0.0512678436934948, -0.135669931769371, 0.08831986784935, -0.0234095230698586, -0.0754300728440285, -0.0558659099042416, -0.0688602104783058, -0.168647140264511, -0.106666870415211, -0.0995750278234482, 0.00447731139138341, -0.0433297865092754, -0.337455242872238, -0.0720325782895088, -0.0638948380947113, -0.163956835865974 },
      { -9.05981540679932, 9.84351348876953, 122.90015411377, -13.1869134902954, 5.05948400497437, 15.5738019943237, 34.4825286865234, 21.4539527893066, -44.2754020690918, -15.1654691696167, -6.72973489761353, -18.6901149749756, -49.9808731079102, 85.2901916503906, 106.322212219238, -8.789963722229, -38.6718559265137, -280.153839111328, 2.26680994033813, -61.5265312194824, -14.4577169418335, 3.30185985565186, 12.6704387664795, -27.2989025115967, 23.3405914306641, -14.0795631408691, -48.5167808532715, -25.2925357818604, 19.5427169799805, -4.98333740234375, -16.2351360321045, 4.49859571456909, 150.359664916992, 258.494079589844, -57.9362602233887, -0.780064821243286, 122.668090820312, -34.5724411010742, 12.7295351028442, -7.02614450454712, -72.416374206543, -60.0824966430664, -39.3140029907227, 0.144285529851913, 0.646903216838837, -14.2796745300293, 14.9233884811401, -41.9849815368652, -3.32734847068787, -4.73374176025391, -11.0906839370728 }
    },
    {
      { 0.0417822897434235, -0.00923819560557604, 0.0851879343390465, 0.1827292740345, 0.0419919490814209, 0.0175083111971617, -0.000708402716554701, 0.029356537386775, 0.183202102780342, 0.017227329313755, 0.0147380959242582, -0.00540136219933629, 0.223449870944023, 0.254962146282196, -0.0324561335146427, 0.0172982085496187, 0.202160745859146, 0.097319133579731, -0.00867286697030067, 0.0200819969177246, 0.214804247021675, 0.0207460969686508, -0.027746208012104, 0.0357166863977909, 0.0279337987303734, 0.0152843799442053, 0.184403017163277, 0.0593325905501842, 0.210169225931168, -0.0668106153607368, 0.0654771327972412, -0.00587543100118637, -0.031297642737627, 0.0446993932127953, 0.200928047299385, 0.00332144787535071, 0.0251896139234304, 0.26798540353775, 0.0146589241921902, 0.0208661798387766, 0.00938697624951601, -0.0489173978567123, -0.00333066773600876, -0.00660804379731417, -0.0242236498743296, -5.59441032237373e-05, -0.0870030075311661, 0.0136332688853145, 0.0466689430177212, 0.0156226893886924, 0.00491815619170666 },
      { 0.0336260460317135, 0.0142420837655663, -0.0231932364404202, -0.17463992536068, -0.000649499532300979, 0.00704764621332288, 0.178180649876595, -0.0710310488939285, -0.238514676690102, 0.125727623701096, 0.0123180951923132, 0.0135300438851118, -0.260453283786774, -0.429329872131348, 0.0153188593685627, 0.0430148765444756, -0.253116577863693, -0.0810350850224495, 0.0398414209485054, 0.0624601766467094, 0.0527324452996254, 0.019606102257967, 0.00281767151318491, 0.01996248960495, 0.00420018751174212, -0.0295901782810688, 0.166990965604782, -0.0623037181794643, -0.0775145217776299, 0.15831147134304, 0.0268084593117237, -0.00639692414551973, -0.0234210379421711, -0.0219749975949526, -0.255115300416946, 0.04843445494771, 0.0244274474680424, -0.153717070817947, 0.00915176142007113, -0.00521692261099815, 0.054355014115572, 0.0607396923005581, 0.0292659141123295, 0.0204848479479551, -0.00248015718534589, 0.0606151446700096, 0.120348960161209, 0.0224419552832842, -0.0320504643023014, -0.0107017979025841, 0.101148888468742 },
      { -0.0851754173636436, -0.00212149182334542, -0.0874171778559685, 0.0183724146336317, -0.0766262337565422, -0.0177537780255079, -0.175281494855881, 0.0195654090493917, 0.0328558571636677, -0.108779169619083, -0.0204075817018747, -0.00117850152309984, -0.108385838568211, 0.11835703253746, 0.0187071431428194, -0.0732873231172562, 0.0382494814693928, 0.0112531678751111, -0.0114808799698949, -0.0552181713283062, -0.220853254199028, -0.0297386143356562, 0.0102874478325248, -0.0432981476187706, -0.0653161257505417, -0.0114644123241305, -0.309986174106598, 0.0173195078969002, -0.396725445985794, -0.0919777154922485, -0.0356248095631599, -0.0363935232162476, 0.065635047852993, 0.00579907046630979, 0.0387992970645428, -0.0447728224098682, -0.0738510489463806, 0.0747025534510612, -0.0114205358549953, -0.00663109915331006, -0.055731225758791, -0.0233028810471296, -0.068168893456459, -0.0271016303449869, 0.00351567193865776, -0.0628324821591377, -0.077968567609787, -0.0353482589125633, -0.0165434181690216, -0.0240634717047215, -0.0748064368963242 },
      { 0.0212140083312988, -0.00532779889181256, -0.00524527626112103, 0.0832302197813988, -0.0111550446599722, 0.00182492064777762, 0.014829209074378, 0.0129359737038612, 0.0228786561638117, -0.0224931053817272, 0.00480129150673747, 0.000286687514744699, 0.144449934363365, 0.022940244525671, -0.00747438240796328, 0.00653892988339067, 0.0123386764898896, -0.00697298720479012, -0.00642726430669427, -0.00639416417106986, 0.180864423513412, 0.0039211711846292, 0.00153700588271022, -0.00377175002358854, 0.0350230447947979, 0.00201931269839406, -0.0850819051265717, -0.00244846008718014, 0.325671821832657, -0.0099552609026432, -0.0142877930775285, 0.026269992813468, -0.0120635079219937, 0.014153684489429, 0.0691054686903954, 0.00754108373075724, -0.00323492451570928, -0.0715847164392471, -0.00546877784654498, 0.00161112239584327, -0.00631497427821159, -0.00714427139610052, 0.00340371462516487, 0.00652801524847746, 0.0114405546337366, -0.0178953725844622, -0.0457263700664043, -0.00056809390662238, 0.0207265503704548, 0.0300666280090809, -0.00393793452531099 },
      { 0.0183668434619904, 0.0424169786274433, 0.149516552686691, -0.0502612106502056, 0.0517053008079529, 0.0125854704529047, 0.0298724807798862, -0.00606285687536001, -0.0112165044993162, 0.0753715336322784, 0.0202591381967068, 0.0867112204432487, -0.018617270514369, 0.106151469051838, 0.110617473721504, 0.00174053967930377, -0.00488206837326288, 0.0545696914196014, -0.0319220907986164, 0.0548562854528427, -0.0806620195508003, 0.013606246560812, 0.0449765659868717, 0.0401497967541218, -0.0371050462126732, 0.0490962974727154, 0.0972272455692291, 0.114864595234394, -0.0576709434390068, 0.0450832396745682, 0.0755108520388603, 0.000941361475270241, 0.0860657840967178, -0.0205021481961012, -0.069029875099659, 0.0128808384761214, 0.155742064118385, 0.105524457991123, 0.0687267780303955, 0.0385828912258148, 0.045992661267519, 0.0668935403227806, 0.0227296203374863, 0.0281660687178373, -0.0022397383581847, 0.0856667906045914, 0.000797815038822591, 0.0299222636967897, -0.0217771027237177, 0.00223969062790275, 0.0250453297048807 },
      { 0.984356641769409, 0.970797598361969, 0.90904837846756, 0.962120234966278, 0.994451463222504, 1.01999998092651, 0.9776890873909, 1.02663362026215, 1.00964510440826, 0.941108524799347, 0.984894990921021, 0.897316038608551, 1.0044059753418, 0.963509857654572, 0.916380643844604, 1.01159870624542, 1.00773179531097, 0.963379263877869, 1.02142429351807, 0.937405526638031, 0.956540942192078, 0.98759537935257, 0.935828387737274, 0.982228517532349, 1.05657005310059, 0.966399490833282, 0.953268945217133, 0.837881922721863, 1.02197933197021, 0.969116389751434, 0.908178985118866, 1.01727020740509, 0.935596883296967, 1.01757907867432, 1.04142951965332, 0.975932538509369, 0.898088097572327, 0.856474101543427, 0.917796492576599, 0.959767639636993, 0.964368581771851, 0.92239910364151, 1.01018261909485, 0.972184479236603, 0.999032258987427, 0.925529658794403, 1.03581821918488, 0.991503953933716, 1.0186516046524, 0.991620063781738, 0.974460184574127 },
      { -7.98626518249512, -10.7955894470215, -68.7053298950195, 117.126342773438, -5.11392688751221, -4.74536323547363, -39.2622680664062, -0.836750984191895, 18.8323154449463, -8.74405765533447, -3.105553150177, 4.28675889968872, 15.4739665985107, -42.3609657287598, -1.87383210659027, -1.08836781978607, 19.195104598999, -46.7552833557129, -0.469722151756287, 3.15969967842102, 40.6997528076172, -3.37348747253418, 15.9057788848877, -1.92556893825531, -19.404914855957, -3.69598793983459, 33.6867370605469, 7.46743726730347, -35.890567779541, 5.02299499511719, -0.691323399543762, -1.22140467166901, 24.0953483581543, -54.3141822814941, 22.2848110198975, 0.756004810333252, -57.3915824890137, 8.04819393157959, -1.5413738489151, -2.23035001754761, 1.46164739131927, 11.8059778213501, -1.76955819129944, 1.3974974155426, 24.6450309753418, -3.31020498275757, 5.93461418151855, -1.1545478105545, -5.01517724990845, -10.4503917694092, -4.30063676834106 }
    }
  },
  {
    {
      { 0.939290702342987, 0.92669004201889, 0.939466059207916, 0.96964019536972, 0.902462065219879, 0.939913272857666, 0.861581206321716, 0.954078733921051, 0.926370680332184, 0.904682815074921, 0.887807369232178, 0.892564058303833, 1.00587677955627, 0.514238893985748, 0.984029769897461, 0.861586034297943, 0.947417974472046, 0.918216049671173, 0.849651575088501, 0.899894833564758, 0.988404929637909, 0.865638256072998, 0.890447080135345, 0.944159030914307, 0.8824263215065, 0.984453856945038, 0.97308611869812, 0.974662899971008, 0.914485216140747, 0.930266797542572, 0.889640271663666, 0.951455533504486, 0.908372819423676, 0.970015108585358, 0.906798362731934, 0.940833032131195, 0.914158940315247, 0.633471131324768, 1.04145336151123, 0.953913986682892, 0.928337812423706, 0.949030697345734, 0.923563778400421, 0.905734360218048, 0.880119800567627, 0.904042303562164, 0.899763405323029, 0.858990609645844, 0.947731256484985, 0.949886977672577, 0.921081006526947 },
      { 0.0878715664148331, 0.0733170062303543, -0.0442714057862759, 0.0644307509064674, 0.106100276112556, -0.054637610912323, 0.203633680939674, 0.0811557397246361, 0.0822447240352631, 0.0829328596591949, 0.107522629201412, 0.103078737854958, 0.0850826054811478, 0.995840311050415, 0.0674834921956062, 0.103843502700329, -0.0373660326004028, 0.0978799313306808, 0.199757248163223, 0.136752530932426, 0.0638290867209435, 0.19952604174614, 0.120189823210239, 0.0913735404610634, 0.0978341475129128, 0.0671482309699059, 0.0412883907556534, -0.104943670332432, 0.127727285027504, 0.113457269966602, 0.105379246175289, 0.060605350881815, -0.00110160873737186, 0.0715009570121765, 0.093402586877346, 0.119240306317806, 0.123600654304028, 0.718388020992279, -0.0324692726135254, 0.0546613521873951, 0.0918242856860161, 0.0683462247252464, 0.0867518112063408, 0.092814564704895, 0.0909784734249115, 0.112842164933681, 0.0864203721284866, 0.208637729287148, 0.107128910720348, 0.148412391543388, 0.0712840184569359 },
      { -0.0227114167064428, 0.00999292824417353, 0.0955987349152565, -0.00283683184534311, -0.00212280126288533, 0.109224796295166, -0.0638268291950226, -0.0177736785262823, 0.00312203844077885, 0.0169023424386978, 0.0017791228601709, -0.00219309167005122, -0.0412236489355564, -0.229804679751396, -0.0126214697957039, 0.0226547718048096, 0.0851304233074188, -0.00405307952314615, 0.00375810288824141, -0.0156184388324618, 7.93813305790536e-05, -0.0489614754915237, -0.00717228418216109, -0.0168229136615992, 0.0147021571174264, -0.0222050510346889, -0.0159054454416037, 0.11812250316143, -0.0269292593002319, -0.0153601616621017, 0.00293887965381145, 0.0298918075859547, 0.0975844487547874, -0.0147672221064568, 0.028033247217536, -0.0178103484213352, -0.0274439919739962, -0.1350427120924, 0.00867684651166201, 0.0365147776901722, 0.00966231059283018, -0.0122596146538854, -0.00354329100809991, -0.00706722447648644, 0.033262949436903, -0.00513905147090554, 0.0211655702441931, -0.0146300001069903, -0.0491086840629578, -0.0762073248624802, 0.0211659334599972 },
      { 0.0138564575463533, 0.00520094623789191, -0.0034013397525996, 0.00245627760887146, -0.000234955732594244, 0.0104227308183908, 0.0228952765464783, 0.00124568655155599, 0.0021166424266994, -0.00434907199814916, -0.000599379243794829, -0.00225363811478019, 0.00596135249361396, -0.000553334190044552, 0.00351673364639282, -1.69499489857117e-05, 0.00164220889564604, -0.000695021240971982, -0.0186182428151369, 0.0149132888764143, 0.000926289416383952, 0.00344564905390143, -0.00198434991762042, 0.00104769307654351, 0.000386436353437603, 0.00715163769200444, 0.00435853330418468, 0.00942245870828629, -0.00505618238821626, -0.00560007197782397, 0.00283912871964276, -0.021886209025979, 0.00748211657628417, -0.00335411261767149, 0.00246206973679364, -0.0173165164887905, 0.0044580684043467, -0.0222194939851761, 0.0384532809257507, -0.00235781050287187, -0.00327444588765502, 0.0186137687414885, -0.000729780120309442, -0.00248365360312164, 0.00167059921659529, -0.00370476278476417, -0.000341337115969509, -0.0194559805095196, 0.0227989833801985, 0.00817180518060923, 0.00703513901680708 },
      { 0.00586161762475967, -0.00194604764692485, 0.0233227964490652, 0.00269432947970927, -0.00160259660333395, -0.00125048728659749, -0.0149883972480893, -0.00215038470923901, -0.00362938735634089, 0.0112324971705675, -0.00554456561803818, 0.00325978570617735, -0.00295407301746309, -0.139684528112411, -0.00308246933855116, 0.00419813487678766, 0.0118192657828331, -0.0056341765448451, 0.011392435990274, -0.0107902931049466, 0.00756821967661381, -0.00755218975245953, 0.00713931676000357, 0.00301530887372792, -0.00349011877551675, -0.0123027255758643, 0.00207876460626721, 0.0151852490380406, 0.000109047759906389, 0.00613854406401515, -0.0082060219720006, 0.0100808469578624, -0.0185105465352535, 0.00327693508006632, -0.0145746516063809, -0.0160465873777866, 0.0036809139419347, 0.110052667558193, 0.0146429389715195, -0.000910411414224654, 0.00126428937073797, -0.0120127815753222, 0.00243337941356003, 0.01335225161165, 0.000495610933285207, -0.00140156911220402, 0.00796647649258375, -0.0217897277325392, -0.00103725423105061, 0.0187626481056213, -4.63765027234331e-05 },
      { -0.0103349480777979, -0.00189077388495207, -0.00753179471939802, -0.0074098901823163, -0.00413229735568166, -0.00539971562102437, -0.00788663513958454, 0.0043644281104207, -7.63019270380028e-05, -0.00993786193430424, 0.0125935208052397, 0.00132311426568776, 0.0012109731324017, 0.151921272277832, -0.00185890274588019, 0.000231632933719084, -0.00834161043167114, 0.00658634584397078, -0.0284535754472017, -0.0206346064805984, -0.0201013423502445, 0.00866381637752056, -0.00760069489479065, -0.00580684561282396, 0.00704042660072446, 0.00658769672736526, -0.00468762405216694, -0.0117246340960264, 0.00520383706316352, -0.00778230326250196, 0.0106551637873054, -0.0141268847510219, 0.00601712940260768, -0.00191507721319795, 0.0149103738367558, 0.0212738942354918, -0.00666288286447525, -0.0974557995796204, -0.0432352386415005, -0.0238135736435652, 0.00179909798316658, 0.0048824199475348, -0.00160922156646848, -0.008152618072927, -0.0066765658557415, 0.00133822544012219, -0.0233914945274591, 0.00189246179070324, -0.00588088482618332, -0.0268243178725243, -0.0126505792140961 },
      { -5.51440811157227, -7.57865238189697, -5.77112483978271, -10.1484022140503, 0.128809541463852, 3.66786789894104, -3.80385756492615, -6.9135799407959, -2.7126522064209, -1.07854676246643, -4.01524543762207, 0.980889737606049, -19.0391845703125, -35.8486442565918, -9.45209121704102, -10.3160390853882, -7.12206792831421, -1.54777228832245, -7.73997688293457, -4.3534574508667, -11.0040655136108, -28.2311973571777, -6.81161594390869, -4.65026044845581, -2.49066615104675, 3.86492276191711, 127.112403869629, -12.7406454086304, -0.758376717567444, -7.41025924682617, -6.01424837112427, -3.3549633026123, 1.64210796356201, -6.20106267929077, -0.681714236736298, -6.26016712188721, -3.36151623725891, -51.8091316223145, -2.01411557197571, 1.11804819107056, -1.41041851043701, 0.437416851520538, -2.91434097290039, -4.40381669998169, -1.73095560073853, -3.65264105796814, 1.01883018016815, -13.0558366775513, -7.26696395874023, -1.71422803401947, -4.70941162109375 }
    },
    {
      { 0.0154699757695198, -0.00629610614851117, 0.0108060296624899, 0.00435239961370826, -0.0175440907478333, -0.0404590927064419, -0.020375095307827, 0.00837989710271358, 0.0076624033972621, -0.0136331683024764, 0.000599606137257069, -0.0106923254206777, 8.31295910757035e-05, 0.000334066513460129, 0.00157872028648853, 0.00400612829253078, -0.0259948074817657, 0.00636891601607203, -0.00800539180636406, -0.00805000867694616, -0.00407447712495923, -0.000426155282184482, -0.0279824398458004, -0.016525074839592, -0.0142775112763047, 0.00455111591145396, -0.154071152210236, -0.0218410175293684, -0.0471890494227409, -0.0252353362739086, -0.00466226460412145, 0.00686591677367687, -0.0227412562817335, -0.012856719084084, -0.0544199422001839, -0.0126600358635187, 0.012440244667232, 0.0328639298677444, 0.0351333171129227, 0.00318273832090199, 0.0140465367585421, -0.0121946288272738, -0.0343830138444901, -0.0108572663739324, -0.0316315218806267, -0.021397652104497, -0.024129981175065, -0.0067303329706192, -0.0143720330670476, -0.0482072532176971, -0.00241615856066346 },
      { 1.03615081310272, 1.0332909822464, 1.06076323986053, 1.04447817802429, 1.09135699272156, 1.09984481334686, 1.06066262722015, 1.0365731716156, 1.02970087528229, 1.04621303081512, 1.06739020347595, 1.10789096355438, 1.0683286190033, 0.940054655075073, 1.05031645298004, 1.05987298488617, 1.09577012062073, 1.05462777614594, 1.06354928016663, 1.04040765762329, 1.0547548532486, 1.08104848861694, 1.07315540313721, 1.06225836277008, 1.10734236240387, 1.021399974823, 1.29402530193329, 1.08778202533722, 1.11574149131775, 1.08187818527222, 1.0666424036026, 1.00719559192657, 1.05948114395142, 1.05736947059631, 1.10115027427673, 1.07515704631805, 1.04265439510345, 0.922325849533081, 1.01232004165649, 1.01210379600525, 1.10563814640045, 1.05412721633911, 1.08645617961884, 1.06357252597809, 1.05092263221741, 1.06394612789154, 1.05483865737915, 1.02126395702362, 1.06616699695587, 1.07753491401672, 1.0236302614212 },
      { -0.0532884523272514, -0.0238530952483416, -0.122644394636154, -0.0413235723972321, -0.0484508909285069, -0.0622750334441662, -0.0491954497992992, -0.0419470109045506, -0.0403726994991302, -0.0439556837081909, -0.0753370746970177, -0.0884854570031166, -0.0657439380884171, 0.0527668334543705, -0.050830889493227, -0.0566365234553814, -0.082183800637722, -0.0458579324185848, -0.0491338036954403, -0.0281686391681433, -0.0447057113051414, -0.0797670856118202, -0.0458926782011986, -0.0497349239885807, -0.0715806633234024, -0.0321421697735786, -0.191059827804565, -0.0718823373317719, -0.066022552549839, -0.0577862411737442, -0.0694776922464371, -0.007479646243155, -0.0517779178917408, -0.0486778132617474, -0.073687382042408, -0.0445549376308918, -0.0549419224262238, 0.0294053424149752, -0.0502439737319946, -0.00053170247701928, -0.101263776421547, -0.0421811304986477, -0.0540668219327927, -0.0587125606834888, -0.021769231185317, -0.0413475856184959, -0.0276130139827728, -0.0144369201734662, -0.0633434802293777, -0.0302882920950651, -0.0163063146173954 },
      { 0.0166175030171871, 0.00358364544808865, -0.00581901567056775, 0.000375203118892387, 0.00072465231642127, 0.00028195945196785, 0.0174196343868971, 0.00908309500664473, 0.00855861697345972, -0.00363754760473967, 0.00434329593554139, -0.00174655439332128, 0.00350890378467739, -0.00195737998001277, 0.00578318443149328, 0.00370369432494044, -0.0140371397137642, 0.00365221872925758, -0.00396047485992312, 0.012009616009891, -0.00100998790003359, 0.00827534403651953, 0.00309115345589817, 0.000227530166739598, -0.00266162352636456, 0.00857057515531778, 0.0438039228320122, -0.0108877876773477, 0.00283402693457901, 0.00054888246813789, 0.00725212087854743, -0.00724357599392533, -0.00926177576184273, 0.000839419604744762, -0.000717562274076045, -0.00859679840505123, 0.0122489910572767, -0.0026938640512526, 0.0219108927994967, 0.00572579307481647, -0.00314411032013595, 0.0037137686740607, 0.000284359324723482, 0.00270493049174547, 0.00385543867014349, -0.000537358340807259, 0.00148070009890944, 0.00578681891784072, 0.0138331102207303, -0.00464318133890629, 0.00802363827824593 },
      { 0.00527844671159983, 0.00328849488869309, 0.0695882737636566, 0.00465156370773911, 0.02004131115973, -0.00105017470195889, -0.00294149178080261, -0.0062618050724268, -0.00219146301969886, 0.0204794518649578, -0.00475499173626304, 0.0201473906636238, 0.00371883786283433, -0.0647399500012398, -0.000346980203175917, 0.00395637797191739, 0.0458671562373638, -0.00222306465730071, 0.0252276696264744, -0.00748362438753247, 0.00868585240095854, -0.00329237384721637, -0.00555364787578583, 0.0110534345731139, 0.0267071053385735, -0.0044239666312933, -0.00866149831563234, 0.0311164371669292, -0.00489027472212911, 0.00307285529561341, -0.0116410572081804, 0.00835618656128645, 0.0425068289041519, 0.0071972175501287, 0.0114175667986274, -0.00862540770322084, 0.00155810383148491, -0.0457308702170849, 0.00614394759759307, -0.000318945094477385, 0.0248174499720335, -0.00110054004471749, 0.00106291961856186, 0.012928712181747, -0.00187273439951241, -0.00186118436977267, 0.0307694040238857, -0.00194926699623466, -0.000663100625388324, -0.00450903223827481, 0.00136432552244514 },
      { -0.00930519308894873, -0.00452669570222497, -0.00293065863661468, -0.00457705417647958, -0.0474383570253849, 0.00101836863905191, -0.00363871804438531, 0.00409826869145036, -0.0018701272783801, -0.00924787204712629, 0.0155418775975704, -0.0337679944932461, 1.11377466964768e-05, 0.0775834247469902, 0.00160965882241726, 0.000491109269205481, -0.0203929170966148, -0.00313657522201538, -0.0338331460952759, -0.00716025196015835, -0.00798253994435072, 0.00273782853037119, 0.00703321769833565, -0.00917438697069883, -0.0507761500775814, 0.0062337601557374, 0.0144925424829125, -0.0173857938498259, 0.00462019164115191, 0.00118101330008358, 0.0190925393253565, -0.00403760373592377, -0.0191175434738398, -0.0046862936578691, -0.0121384030207992, 0.0115974191576242, -0.00454418640583754, 0.0608821474015713, -0.014407642185688, -0.0153210759162903, -0.034503199160099, -0.000526435498613864, -0.000838687177747488, -0.0106194438412786, -0.00209594378247857, 0.00325696566142142, -0.0592262968420982, -0.00108043395448476, -7.37978698452935e-05, 0.0132012730464339, -0.0104731479659677 },
      { -4.87957668304443, -5.69003629684448, -32.2613372802734, -4.15610837936401, -6.49861907958984, 2.84936308860779, -5.10891628265381, -4.69995546340942, -2.94517064094543, 0.878837764263153, -5.81450223922729, -5.50967645645142, -6.67904281616211, -25.7455615997314, -5.53951930999756, -13.9582262039185, -3.48594737052917, -3.45270824432373, 4.15121555328369, -1.76824641227722, -5.14625215530396, -17.607364654541, -3.66261315345764, -3.5809063911438, -5.32638263702393, -4.65200710296631, -187.158737182617, 0.543666839599609, -1.84840703010559, -5.80686664581299, -4.39780950546265, -1.21205699443817, -0.422682851552963, -3.1781370639801, 3.12905716896057, -3.97846722602844, -2.58702230453491, -37.5979385375977, -3.33850812911987, 1.63078665733337, -9.58771991729736, 3.18589329719543, -1.07583391666412, -10.9557275772095, 2.90982007980347, -1.14576685428619, -2.62705230712891, -3.06796836853027, -2.33689284324646, -5.76096105575562, -2.93066644668579 }
    },
    {
      { 0.0756751075387001, 0.0599985867738724, 0.0867131501436234, 0.0840139091014862, 0.00554110435768962, 0.081903412938118, -0.00060680037131533, 0.0557890981435776, 0.0522514432668686, 0.0501731187105179, 0.0587012991309166, -0.0295263230800629, 0.0998058691620827, 0.0101950541138649, 0.112159386277199, 0.0830645933747292, 0.0641293972730637, 0.0110176960006356, 0.0476875752210617, 0.0297530349344015, 0.0988682508468628, 0.0249318983405828, 0.0328394547104836, 0.0502287745475769, -0.00915350951254368, 0.0822495669126511, 0.00420322315767407, 0.118242733180523, -0.0502346344292164, 0.0638368427753448, 0.0635402351617813, 0.0650227144360542, 0.0551345907151699, 0.0726039782166481, 0.0567186437547207, 0.0611405782401562, 0.0383938029408455, 0.090030625462532, 0.168282046914101, 0.0559281334280968, -0.00078430789289996, 0.0391342267394066, -0.00352992559783161, -0.0246603991836309, 0.0523004718124866, 0.0455918461084366, -0.0186879560351372, 0.0163502972573042, 0.0355321504175663, 0.0169686749577522, 0.076123408973217 },
      { -0.138369932770729, -0.200105860829353, -0.228624418377876, -0.134645655751228, -0.113341934978962, -0.274830639362335, -0.0609205923974514, -0.142788723111153, -0.208090350031853, -0.146094635128975, -0.152626946568489, -0.0692509934306145, -0.0798116773366928, -0.437820494174957, -0.174664288759232, -0.0880849659442902, -0.207115352153778, -0.17519523203373, -0.106365092098713, -0.115325562655926, -0.146871373057365, 0.0165438801050186, -0.163385793566704, -0.134575560688972, -0.0777083486318588, -0.221211284399033, -0.0356847755610943, -0.339690655469894, -0.104444652795792, -0.134810790419579, -0.151518553495407, -0.179757431149483, -0.219943180680275, -0.148242339491844, -0.0943563282489777, -0.141450762748718, -0.0967085957527161, -0.383219212293625, -0.288197040557861, -0.172078758478165, -0.0815088078379631, -0.117653213441372, -0.0560947395861149, -0.13239885866642, -0.217074900865555, -0.15126159787178, -0.111676797270775, -0.106733746826649, -0.0924928858876228, -0.062841922044754, -0.203450247645378 },
      { 1.05479681491852, 1.14516353607178, 1.06830787658691, 1.08498597145081, 1.09968709945679, 1.18351709842682, 1.04885852336884, 1.09431207180023, 1.12641906738281, 1.09749603271484, 1.08818197250366, 1.07379913330078, 1.01887488365173, 1.38412189483643, 1.09112012386322, 1.03635764122009, 1.12601375579834, 1.14480400085449, 1.09245824813843, 1.09409248828888, 1.11202728748322, 0.959037065505981, 1.10707628726959, 1.0855108499527, 1.06912648677826, 1.12319207191467, 1.017373919487, 1.20757865905762, 1.14033484458923, 1.1166318655014, 1.0948930978775, 1.12700057029724, 1.14539623260498, 1.08413434028625, 1.01497888565063, 1.10599648952484, 1.04784989356995, 1.24677002429962, 1.10766494274139, 1.13999283313751, 1.06836521625519, 1.08035171031952, 1.05904793739319, 1.13339471817017, 1.15861117839813, 1.11726486682892, 1.1004958152771, 1.1001980304718, 1.02784144878387, 1.06794512271881, 1.1474974155426 },
      { 0.028973501175642, 0.00295148836448789, 0.0188319776207209, -0.0108076734468341, 0.00082672230200842, 0.0124685959890485, 0.0282319504767656, 0.0150798577815294, 0.0199163444340229, -0.00116760493256152, 0.0064148586243391, 0.00323264859616756, -0.00519464164972305, -0.00709644658491015, 0.0107743153348565, 0.000648468383587897, -0.00563058629631996, 0.0045232349075377, -0.00738019635900855, 0.0161660499870777, -0.00845613330602646, 0.0259284842759371, 0.00636206427589059, 0.00294326548464596, -0.000300076586427167, 0.0166820846498013, 0.0107290269806981, 0.00493470253422856, 0.00982702989131212, 0.00282313395291567, 0.00324219139292836, -0.00909459218382835, 0.0151175549253821, -0.00112825690302998, -5.89627816225402e-05, -0.012856681831181, 0.018743596971035, -0.00767622515559196, 0.0552546195685863, 0.00570567743852735, 0.00539155723527074, 0.00883385818451643, 0.000479129754239693, 0.0129276216030121, 0.0114871077239513, 0.00150192307773978, 0.00860810559242964, 0.00419842358678579, 0.027316777035594, -0.00728863850235939, 0.00538403075188398 },
      { 0.00850050058215857, 0.00988008733838797, 0.0699728429317474, 0.0126999979838729, -0.0038589050527662, -0.0043485532514751, -0.00820638332515955, -0.0103142512962222, -0.00140781374648213, 0.0144841251894832, 0.00801411084830761, -0.0125797837972641, 0.0110097117722034, -0.16886043548584, -0.0143365086987615, -0.0155854038894176, 0.0471095629036427, 0.00970026291906834, 0.020316181704402, -0.00916690658777952, 0.0181952677667141, -0.00992396287620068, 0.0109759988263249, 0.014311627484858, -0.00566961150616407, -0.0163400489836931, 0.00217235996387899, 0.0276999939233065, -0.012791964225471, -0.0161130763590336, 0.00151734275277704, 0.00557855144143105, 0.0133698331192136, 0.0142920510843396, 0.00380874006077647, -0.0107975769788027, -7.19207746442407e-05, -0.120250053703785, 0.015019073151052, -0.00347039615735412, -0.0302047878503799, -0.00379112712107599, 0.00239662593230605, -0.0274168364703655, -0.00552340038120747, 0.000478390487842262, -0.0276910215616226, -0.00540839741006494, 0.00421891687437892, 0.0124277118593454, 0.00203468324616551 },
      { -0.0135298212990165, -0.0108025474473834, 0.0033860607072711, -0.00894083827733994, 0.016606668010354, 0.00135900150053203, -0.0072521585971117, 0.00391965219751, -0.00972749385982752, -0.0122027862817049, -0.0100661749020219, 0.0233915150165558, 0.00184302602428943, 0.16567125916481, 0.0140888206660748, 0.0166081711649895, -0.0190976466983557, -0.0027052101213485, -0.0399406589567661, -0.011604730039835, -0.0319206416606903, 0.00523660238832235, -0.00362765160389245, -0.0152653688564897, 0.0243748631328344, 0.025184178724885, -0.00240247393958271, -0.0180212166160345, 0.00903125014156103, 0.00503593822941184, 0.00293585518375039, -0.000340557278832421, -0.00589672103524208, -0.0107661774381995, 0.00678767450153828, 0.0196817256510258, -0.00220131827518344, 0.164892539381981, -0.0381679497659206, -0.0163046792149544, 0.0564438179135323, 0.00284997839480639, -0.0028394719120115, 0.027518792077899, 0.00126411276869476, -0.00507118366658688, 0.0494598969817162, -0.00137649837415665, 0.00249234773218632, -0.0157235842198133, -0.0129917720332742 },
      { -8.07236099243164, -7.53020429611206, -51.2090225219727, -10.1319236755371, 2.08204197883606, 6.75792169570923, -3.19653081893921, -10.710563659668, -4.72074317932129, 0.618454337120056, -3.84406924247742, 0.484294176101685, -20.5653343200684, -69.2355117797852, -13.3251686096191, 10.4876346588135, -14.4324979782104, -4.54859447479248, 1.79371023178101, -4.67462921142578, -13.194730758667, -63.0421905517578, -10.6397495269775, -5.41506958007812, 3.72197771072388, -14.1005344390869, 30.200439453125, -2.86790180206299, -8.12352180480957, -11.8642559051514, -4.83953905105591, -2.39561533927917, -6.69779968261719, -6.84417581558228, 1.29172706604004, -9.73726272583008, -2.96282386779785, -109.691673278809, -5.66589832305908, 0.605839550495148, -2.67602396011353, 0.942885220050812, 1.33243083953857, -1.04689848423004, 1.52455973625183, -2.97900247573853, -1.09455513954163, -9.44714736938477, -8.99505138397217, -6.02827978134155, -2.08379220962524 }
    },
    {
      { 0.14828933775425, -0.407556354999542, -0.107610583305359, -0.0416980870068073, -1.54811465740204, 0.0840534418821335, 0.0317929796874523, -0.0585002191364765, -0.890501618385315, -0.53799170255661, -0.960572063922882, -0.105323895812035, -0.161284476518631, 0.0448649376630783, -0.226359888911247, -1.50261747837067, -0.0346848256886005, -1.38759768009186, -0.32634648680687, 0.0320334285497665, -0.19293050467968, 0.0101544447243214, -0.170103833079338, -0.121908076107502, -1.35497140884399, -0.035065270960331, 0.473181545734406, -0.00905308499932289, 0.355516761541367, -0.337357699871063, -0.990931034088135, -0.0206653233617544, 0.108536757528782, -0.160865113139153, -2.09992814064026, -0.201994121074677, -0.0164962708950043, -0.064218707382679, -0.309794694185257, -0.0916733220219612, -1.97365784645081, -0.00321335718035698, -1.72357261180878, -1.64564263820648, -0.10867466032505, 0.0395487509667873, -0.876764476299286, -0.00902560073882341, 0.00441655050963163, 0.0177029352635145, -0.0796377435326576 },
      { -0.459590256214142, -0.321553319692612, -0.165726497769356, -0.129799500107765, 1.88014054298401, -0.533481419086456, -0.306849122047424, -0.155334398150444, -0.764747738838196, 0.46194651722908, 0.185087978839874, 2.74550127983093, -0.0457086190581322, -0.557302296161652, 0.0232702847570181, 2.23657917976379, -0.618429362773895, -0.573885321617126, 0.066753938794136, -0.236483544111252, 0.0153815494850278, 0.0744702592492104, 0.285579591989517, 0.0225861929357052, 3.15725231170654, 0.176760077476501, -0.692496538162231, -0.660035490989685, -0.13306075334549, -0.0348016582429409, 0.251199573278427, -0.268555134534836, -0.380900770425797, 0.0320925749838352, 1.61519157886505, 0.107239119708538, -0.192875906825066, 0.472472190856934, 0.22723314166069, -0.168320000171661, 1.85534596443176, -0.101921737194061, 1.93808138370514, 0.450171232223511, -0.0417807400226593, -0.290132164955139, 1.41100370883942, -0.18011049926281, -0.111171968281269, -0.314310103654861, -0.349212259054184 },
      { 1.07751095294952, 1.2359185218811, 1.87029147148132, 0.953579485416412, -1.03408432006836, 1.27382588386536, 1.06023478507996, 0.949672400951385, 1.42358577251434, 0.574566304683685, 1.48659932613373, -1.89695751667023, 0.900152266025543, 1.90622770786285, 0.802475035190582, -0.0632779374718666, 1.49054622650146, 1.28732657432556, 0.947374403476715, 0.973397374153137, 0.8485187292099, 0.721818208694458, 0.780536234378815, 0.784122228622437, -0.634128332138062, 0.606720566749573, 1.25937187671661, 1.53427803516388, 0.540732443332672, 0.95116126537323, 1.21892654895782, 0.934442460536957, 1.50715494155884, 0.782413244247437, 1.13211679458618, 0.860771775245667, 0.988148391246796, 1.23211252689362, 0.849353015422821, 0.94986766576767, -0.409130960702896, 0.929065585136414, 0.367754518985748, 0.841745555400848, 0.942933857440948, 1.03318881988525, -0.221455350518227, 0.981223464012146, 0.961263716220856, 1.13252639770508, 0.985012412071228 },
      { 0.166964933276176, 0.184074923396111, 0.37626650929451, 0.211117625236511, 0.0985723584890366, 0.266191601753235, 0.211764335632324, 0.264081120491028, 0.382070094347, 0.202612981200218, 0.156254321336746, -0.340616375207901, 0.260022699832916, -0.0117110926657915, 0.287366181612015, -0.0391391217708588, 0.584915637969971, 0.12628062069416, 0.235477656126022, 0.23469477891922, 0.258047431707382, 0.230954706668854, 0.175162106752396, 0.312281578779221, -0.24102458357811, 0.272997349500656, 0.17634429037571, 0.430673032999039, 0.258456259965897, 0.276510834693909, 0.142050251364708, 0.381856799125671, 0.432703852653503, 0.301054745912552, 0.0501189306378365, 0.251552581787109, 0.221627458930016, 0.0228645261377096, 0.122511520981789, 0.268544644117355, -0.00464295269921422, 0.18162414431572, 0.0554580390453339, 0.200733780860901, 0.146525084972382, 0.290597587823868, 0.162458628416061, 0.226266711950302, 0.240018621087074, 0.153209209442139, 0.22817662358284 },
      { -0.00326760718598962, -0.0634662210941315, -1.32503414154053, -0.00418394571170211, -0.43024817109108, -0.169132754206657, -0.00676356395706534, -0.0945148468017578, -0.021756537258625, -0.037135548889637, 0.0789150893688202, 0.40787610411644, -0.00298882741481066, 0.0249522663652897, -0.0208404846489429, 0.162420600652695, -0.884337484836578, 0.401295393705368, -0.19109696149826, -0.0131678646430373, -0.0117850294336677, -0.0475214794278145, 0.0719550549983978, -0.0619989037513733, 0.468909382820129, 0.0197428558021784, 0.0371264591813087, -0.61391270160675, 0.0192282237112522, -0.00615654978901148, 0.179026633501053, -0.149531841278076, -1.16225111484528, -0.0281688421964645, 0.0568542890250683, -0.044111005961895, -0.000774362531956285, -0.798508942127228, -0.0133344326168299, -0.0179582592099905, -0.356655061244965, 0.0236831493675709, -0.0684968233108521, -0.21741546690464, 0.0979797467589378, -0.0962378531694412, -0.919012844562531, -0.0158900562673807, -0.0569755248725414, -0.0131468093022704, -0.159537672996521 },
      { 0.017036359757185, 0.184869453310966, 0.159318178892136, -0.00230030226521194, 1.03278791904449, 0.0704460591077805, 0.0139340525493026, 0.0838440954685211, 0.113269604742527, 0.236549571156502, -0.249051198363304, 0.676563084125519, -0.017608318477869, 0.0701303109526634, 0.0168933682143688, -0.201211169362068, 0.473661690950394, 0.188711389899254, 0.256234049797058, 0.0193284545093775, 0.0028526175301522, 0.0440819934010506, -0.125162586569786, 0.0302879977971315, -0.551301121711731, -0.016632005572319, -0.0595428049564362, 0.335944354534149, -0.00632246909663081, -0.0188677962869406, -0.239542216062546, 0.108458623290062, 0.411414176225662, 0.008921611122787, -0.121147759258747, 0.0111631276085973, -0.00927915330976248, 0.950705230236053, 0.0654488429427147, 0.0319827646017075, 1.20906794071198, -0.0264238901436329, 0.0252351518720388, 0.585314333438873, -0.0361894443631172, 0.0120825245976448, 1.8155871629715, 0.00608486402779818, 0.00285657006315887, 0.00784115958958864, 0.286246925592422 },
      { 25.1535129547119, 60.5786361694336, 267.356079101562, -9.66460227966309, 479.208465576172, 35.1709899902344, 6.12946462631226, -0.35592994093895, 74.3677215576172, 23.5729465484619, 157.474884033203, 409.236450195312, 1.65480375289917, -571.003112792969, 27.6158103942871, 337.037658691406, -37.1921539306641, 148.226943969727, 39.8328590393066, -6.71125888824463, 19.2231597900391, -178.552032470703, 47.1318778991699, 16.7647743225098, 448.456420898438, -49.3198051452637, 32.9757766723633, -43.8171424865723, 22.7437705993652, 40.0647087097168, 118.901710510254, -1.05174124240875, 127.253242492676, 12.3972835540771, -9.03925895690918, -16.694938659668, 3.87681555747986, -700.209167480469, 42.2278137207031, -11.0269622802734, 519.335266113281, -0.353335946798325, 271.513336181641, 236.124160766602, 11.6794004440308, 60.8029823303223, 164.893524169922, -24.3629474639893, -11.6628513336182, 83.3883209228516, 71.7358932495117 }
    },
    {
      { 0.301870763301849, -0.348994970321655, 0.00657449895516038, -0.345914661884308, -0.378002673387527, 0.208042770624161, -0.118705846369267, -0.202120721340179, -0.213787853717804, -0.466002970933914, 0.00288659729994833, 2.10348701477051, -0.696021556854248, 0.0831292346119881, -0.555653870105743, 0.0817680284380913, 0.0961734279990196, 0.105929620563984, -0.179584860801697, -0.060399204492569, -0.475348830223083, 0.0976096987724304, -0.535606622695923, -0.398710042238235, 0.737593352794647, -0.0838938355445862, 0.787450194358826, 0.140712663531303, -0.437059998512268, -0.0907569229602814, -0.445703119039536, -0.15576608479023, 0.00257836771197617, -0.610520660877228, -0.702902853488922, -0.218362286686897, -0.0940189436078072, -0.367382884025574, -0.701472103595734, -0.199127063155174, 1.80711448192596, -0.149484664201736, -1.60226774215698, -0.149862080812454, -0.251467913389206, -0.159580990672112, -0.529300272464752, -0.0624779611825943, -0.0633224546909332, -0.0855600014328957, 0.067940317094326 },
      { -0.885970592498779, 0.164809182286263, -0.121893130242825, 0.109855569899082, 2.92936539649963, -0.510999262332916, 0.0136873982846737, 0.273298591375351, 0.331775397062302, 0.716775238513947, 1.49372851848602, 4.02839612960815, 0.101701758801937, -1.22360384464264, 0.469806492328644, 1.62480282783508, -0.464754074811935, 0.902287542819977, 0.347018331289291, 0.0299401916563511, 0.504496455192566, 0.0314236208796501, 1.42207050323486, 0.470299601554871, 6.08357381820679, 0.301197022199631, -0.698961555957794, -0.44877415895462, 1.81424415111542, 0.739668667316437, 1.79829835891724, -0.132398173213005, -0.104680418968201, 0.528024971485138, 2.19438982009888, 0.704196631908417, -0.203087851405144, 0.828179955482483, 1.35077559947968, -0.0559379011392593, 3.44326400756836, 0.275410950183868, 2.38152265548706, 1.31980538368225, 0.46268755197525, 0.370808213949203, 1.94460356235504, -0.109749361872673, 0.102883651852608, 0.111545540392399, -0.170896291732788 },
      { 0.970759212970734, 0.275733143091202, 1.12031292915344, 0.543973684310913, -2.36632680892944, 0.598577618598938, 0.395413160324097, 0.22933454811573, -0.367176204919815, -0.161532655358315, -1.22305154800415, -4.76309633255005, 0.730746567249298, 2.88127326965332, 0.122486248612404, -1.17121505737305, 0.669324159622192, -0.595022737979889, 0.118663817644119, 0.369523674249649, 0.207296907901764, 0.141637444496155, -0.260058522224426, 0.191897347569466, -3.5481321811676, 0.0414086654782295, 0.584908187389374, 0.627754390239716, -0.714446663856506, -0.161536201834679, -0.985355138778687, 0.452955037355423, 0.940128207206726, 0.186787039041519, -0.997057557106018, 0.16617988049984, 0.78932136297226, 1.78932893276215, -0.0634492859244347, 0.509256362915039, -3.58752799034119, 0.37345689535141, -0.680975079536438, -1.09554648399353, 0.0977175310254097, 0.161851823329926, -0.86238706111908, 0.644341886043549, 0.779251873493195, 0.147725015878677, 0.337505131959915 },
      { 0.332872003316879, 0.778731226921082, 0.807363092899323, 0.590571820735931, 0.572959721088409, 0.738529860973358, 0.693455815315247, 0.624551057815552, 0.853298664093018, 0.613063097000122, 0.72292172908783, 0.0116405300796032, 0.64582896232605, -0.00359579175710678, 0.753834843635559, 0.735758304595947, 1.01340746879578, 0.632421255111694, 0.666901767253876, 0.747156023979187, 0.695773899555206, 0.766249418258667, 0.607264578342438, 0.739388763904572, 0.0395653657615185, 0.766022622585297, 0.865547657012939, 0.871623337268829, 0.738431572914124, 0.584511399269104, 0.60247802734375, 0.923052251338959, 0.704340159893036, 0.747474491596222, 0.646687030792236, 0.539442241191864, 0.506905138492584, 0.0793223008513451, 0.377563506364822, 0.720489978790283, 0.348400115966797, 0.494665861129761, 0.612624883651733, 0.697145819664001, 0.637147307395935, 0.795366585254669, 0.60667496919632, 0.612069368362427, 0.520928502082825, 0.795753121376038, 0.880286395549774 },
      { 0.0270163342356682, -0.177974835038185, -1.09095740318298, -0.0545219555497169, 0.0773288160562515, -0.0809674859046936, 0.0337062776088715, -0.124939881265163, -0.101302854716778, 0.246724799275398, -0.0385561659932137, 1.54124212265015, -0.0525848194956779, 0.364881068468094, -0.187539920210838, -0.0121191162616014, -0.696090698242188, -0.104670539498329, 0.24299481511116, -0.104653462767601, -0.187587097287178, 0.0134217832237482, -0.18533219397068, -0.219081923365593, 1.52200412750244, -0.104603849351406, -0.165728375315666, -0.434562355279922, -0.22362032532692, 0.0460067577660084, -0.136543139815331, -0.286577522754669, -0.969906985759735, -0.111528068780899, 0.0694787576794624, -0.187593638896942, 0.00825821235775948, -1.2314966917038, -0.026107108220458, -0.0661988779902458, 0.693577408790588, 0.0626545995473862, 0.00245290342718363, 0.0840003490447998, 0.101183325052261, -0.225313976407051, 0.142954990267754, -0.0587398521602154, -0.190572679042816, -0.0368520133197308, -0.260750591754913 },
      { 0.0486633591353893, 0.217511102557182, 0.0779778584837914, 0.0262392237782478, -0.0450845956802368, 0.0119132837280631, -0.00921151135116816, 0.161733940243721, 0.215835824608803, -0.0602337643504143, 0.142156317830086, -1.0612815618515, -0.047780841588974, -0.227066487073898, 0.194572895765305, 0.0267985593527555, 0.374469608068466, 0.350725740194321, -0.312217593193054, 0.0347618982195854, 0.151199787855148, 0.00496244244277477, 0.124335817992687, 0.161035493016243, -3.05168437957764, 0.111791491508484, 0.126113384962082, 0.234886839985847, 0.182800427079201, -0.0332928188145161, 0.340534627437592, 0.166789814829826, 0.337672203779221, 0.0577984489500523, -0.162679269909859, 0.115372225642204, -0.0317422486841679, 1.37905812263489, 0.134025275707245, 0.0454656071960926, -0.634302079677582, -0.0746456831693649, -0.00399568304419518, 0.146114900708199, -0.0771344527602196, 0.0690802782773972, -0.344125270843506, -0.00683134561404586, -0.0138348918408155, 0.109159961342812, 0.162440299987793 },
      { 78.501823425293, -42.0774955749512, 284.538757324219, 12.1297693252563, -44.1862106323242, 47.4254951477051, -12.9638528823853, 29.8264923095703, -85.0844802856445, 33.9036254882812, -196.164642333984, -87.2196350097656, 90.6959533691406, -947.170471191406, 43.9780578613281, -368.317291259766, -54.6263160705566, -27.5369491577148, 67.9164352416992, -16.011833190918, 53.4658508300781, -139.290161132812, 48.1012954711914, 39.775146484375, 73.8028869628906, -114.353782653809, -217.557174682617, -55.900333404541, -200.776428222656, 19.7862758636475, -63.8443183898926, 2.27864146232605, 156.932662963867, 47.3132438659668, -109.914611816406, 1.3769599199295, 15.7391176223755, -1102.45422363281, -17.7568054199219, -12.190089225769, -354.714447021484, -27.3729934692383, -171.41015625, -266.6796875, 43.1927528381348, 54.198127746582, -49.3727836608887, -55.312557220459, 9.56275844573975, -93.6128311157227, -21.6781787872314 }
    },
    {
      { 0.332354992628098, 0.145584717392921, 0.138906925916672, -0.3503098487854, 0.0883655026555061, 0.157281741499901, -0.160240650177002, -0.187884613871574, 0.50248908996582, -0.0637032836675644, 1.10733234882355, -0.688989877700806, -0.656568169593811, 0.563827514648438, -0.314543724060059, 1.40728294849396, 0.121006660163403, 1.39532005786896, 0.266435235738754, -0.060209859162569, -0.281971484422684, 0.00121751963160932, -0.554257988929749, -0.315298616886139, -0.345645010471344, 0.0919201523065567, 0.413650333881378, 0.237718418240547, -0.679864525794983, 0.194703757762909, 0.640539586544037, -0.0844198316335678, -0.115896105766296, -0.543855547904968, 1.34944868087769, -0.13361069560051, -0.0998966991901398, 0.0801792964339256, -0.483284831047058, -0.149871557950974, 0.28382819890976, -0.160853430628777, 0.223552539944649, 0.551298975944519, -0.129113048315048, -0.0688736215233803, -0.333135485649109, -0.0599848441779613, -0.123230405151844, 0.1097317263484, 0.432659417390823 },
      { -0.820486545562744, -0.320156365633011, -0.046765822917223, 0.150991469621658, 1.07289552688599, -0.218628376722336, 0.289912432432175, 0.173878312110901, -0.157400980591774, 0.313445389270782, 0.984801888465881, 0.596819221973419, 0.125003680586815, -0.514390528202057, 0.228961065411568, 0.276651740074158, -0.0615440793335438, 0.299014449119568, -0.0745433568954468, 0.0751608163118362, 0.404083430767059, -0.00644182879477739, 1.2901930809021, 0.310387283563614, 0.109771654009819, -0.185355708003044, 0.412721693515778, -0.226886078715324, 2.14615225791931, 0.667700946331024, 1.34193074703217, -0.163666486740112, 0.231049001216888, 0.418225705623627, 0.468966275453568, 0.513555884361267, -0.1965601593256, -0.168182820081711, 0.707230448722839, -0.0949455574154854, 0.260104715824127, 0.244519829750061, 0.219915643334389, 0.425691485404968, 0.314593076705933, 0.211995646357536, 0.358389586210251, -0.0875042751431465, 0.118488624691963, 0.368713468313217, -0.665253877639771 },
      { 0.665226101875305, 0.331339865922928, 0.1491589397192, 0.34182807803154, -0.653678417205811, 0.163551077246666, -0.0514994561672211, 0.151902884244919, -0.533224105834961, -0.137574911117554, -1.9865859746933, -0.393734604120255, 0.577649056911469, -0.0543728098273277, 0.0613114610314369, -1.34447503089905, 0.0218283757567406, -0.36826229095459, 0.0951651036739349, 0.147371366620064, 0.164160713553429, 0.0787425115704536, -0.360580295324326, 0.132503122091293, -0.114881113171577, 0.133353367447853, -0.274801194667816, 0.0817669108510017, -0.865914404392242, -0.317161709070206, -1.51850771903992, 0.31337183713913, 0.176172807812691, 0.118315301835537, -1.50753474235535, 0.0994468405842781, 0.538996875286102, 0.77543830871582, -0.0465976446866989, 0.377857714891434, -1.02412605285645, 0.257040083408356, -0.279935598373413, -0.805148303508759, -0.0624263770878315, 0.13875338435173, 0.36751464009285, 0.43021959066391, 0.586318075656891, -0.369276612997055, 0.51091605424881 },
      { 0.574598729610443, 0.96608555316925, 0.912254393100739, 0.74710351228714, 0.884285986423492, 0.926128804683685, 0.93699449300766, 0.790620446205139, 1.10251557826996, 0.783027231693268, 1.0100291967392, 0.978703200817108, 0.749530076980591, 0.93796443939209, 0.890724897384644, 1.08616971969604, 0.978808462619781, 0.914555072784424, 0.802097320556641, 0.940381646156311, 0.801483452320099, 0.954885542392731, 0.810665607452393, 0.903487861156464, 1.08935523033142, 1.0009468793869, 0.675434589385986, 0.953036069869995, 0.928087651729584, 0.739293873310089, 0.890637457370758, 1.0277510881424, 0.889777362346649, 0.890260756015778, 0.927079260349274, 0.708184778690338, 0.710786879062653, 0.827274143695831, 0.644481122493744, 0.87111896276474, 0.937571346759796, 0.638094365596771, 1.01721823215485, 1.00031113624573, 0.868962287902832, 0.935971140861511, 0.923987448215485, 0.80166107416153, 0.696993768215179, 1.04531216621399, 1.07368767261505 },
      { 0.0323428399860859, -0.140966132283211, -0.208147540688515, -0.0341565534472466, 0.477779299020767, -0.0529656819999218, 0.0311381462961435, -0.109378859400749, -0.169531181454659, 0.379455238580704, -0.128865420818329, 0.125805497169495, -0.0134174739941955, -0.125555649399757, -0.226285681128502, -0.25595024228096, -0.148814037442207, -0.408872276544571, 0.576886832714081, -0.0944924205541611, -0.159855529665947, 0.0547292977571487, -0.21810220181942, -0.22539584338665, -0.522456109523773, -0.223071366548538, -0.074362687766552, -0.0773955807089806, -0.294196784496307, 0.019522013142705, -0.421646267175674, -0.237441062927246, -0.32706218957901, -0.0918751433491707, 0.147816672921181, -0.170492857694626, 0.0135126272216439, -0.74855774641037, 0.113858044147491, -0.0582191646099091, 0.0689251199364662, 0.0650689229369164, 0.0336121208965778, -0.0404258593916893, 0.0420317761600018, -0.199996590614319, 1.05170667171478, -0.0283300299197435, -0.165668517351151, -0.11470302939415, -0.158556014299393 },
      { 0.0477397292852402, 0.0808431804180145, 0.0187936015427113, 0.0160852503031492, -1.08876883983612, 0.0102612981572747, -0.0403199829161167, 0.139620810747147, 0.179080873727798, -0.325622230768204, 0.372736155986786, -0.45137631893158, -0.069733552634716, 0.530371844768524, 0.259723037481308, 0.285146504640579, 0.0905894786119461, 0.211428374052048, -0.786394774913788, 0.00342690269462764, 0.0912731811404228, -0.0446161217987537, 0.172587767243385, 0.158997356891632, 0.559979736804962, 0.230748414993286, 0.05162363499403, 0.0304914470762014, 0.24121356010437, -0.0437472686171532, 0.673373222351074, 0.129788488149643, 0.12275618314743, 0.0305516347289085, -0.20871764421463, 0.103886634111404, -0.0198352225124836, 1.32231152057648, -0.033822525292635, 0.0218716356903315, 0.0974152237176895, -0.0665449053049088, -0.00920839700847864, 0.107602812349796, -0.0476100295782089, 0.0270278230309486, -2.4177839756012, -0.0489652417600155, -0.0158012975007296, 0.185180127620697, -0.0817449390888214 },
      { 40.3806762695312, -27.0716438293457, -1.55672490596771, 25.3509979248047, -151.865081787109, 5.29329299926758, -23.297700881958, 28.2565364837646, -94.2181243896484, 32.6883811950684, -324.934143066406, -59.1773796081543, 98.0142593383789, -197.390533447266, 22.1382961273193, -334.545959472656, -57.5321502685547, -76.4410552978516, 76.8606414794922, -17.5921764373779, 43.7059669494629, 14.2580375671387, 43.3166580200195, 44.2763748168945, -95.9114532470703, -98.6729049682617, 358.812927246094, -93.0017318725586, -263.966156005859, 7.69864797592163, -74.6332015991211, 1.53197932243347, 42.2330551147461, 52.1527290344238, -63.4091682434082, 2.07615852355957, 9.89878177642822, -824.665283203125, 112.599960327148, -6.15652704238892, -71.948600769043, -13.4360675811768, -495.423095703125, -323.292419433594, 18.4450855255127, 56.5198173522949, -73.5109786987305, -35.1927947998047, 21.9911556243896, -315.257202148438, -28.6230354309082 }
    },
    {
      { 0.0838291123509407, -0.0828287526965141, -0.0601919256150723, -0.152656316757202, -0.16227325797081, 0.0564756207168102, -0.0250136367976665, -0.0917002186179161, -0.0401540286839008, -0.0832405015826225, 0.153177082538605, -0.582172751426697, -0.261612266302109, 0.233537301421165, -0.221054211258888, 0.230257958173752, -0.00724746566265821, 0.216813832521439, 0.00598519714549184, -0.0148697458207607, -0.201659888029099, -0.000558706989977509, -0.222696244716644, -0.132971748709679, -0.221414342522621, -0.0189503021538258, -0.113297462463379, -0.034580796957016, -0.013941259123385, -0.044800691306591, -0.124807715415955, -0.032276276499033, 0.121308147907257, -0.216762319207191, 0.488597333431244, -0.089057944715023, -0.0579785481095314, 0.198478192090988, -0.352467149496078, -0.0863739028573036, -0.00290617067366838, -0.0704626441001892, 0.162031799554825, -0.0123305395245552, -0.0303575657308102, 0.0257257707417011, -0.143297478556633, -0.0272016767412424, -0.0654704868793488, -0.00244998396374285, 0.102294497191906 },
      { -0.236135646700859, -0.0449830368161201, -0.0257612559944391, 0.0728975683450699, 0.542208909988403, -0.114583089947701, 0.0407259464263916, 0.0486805737018585, 0.0387106277048588, 0.170785248279572, 0.388737112283707, 0.165161371231079, 0.0461321473121643, -0.269833028316498, 0.127514913678169, 0.32423323392868, -0.392164945602417, 0.0796743035316467, 0.0787506476044655, 0.00629376713186502, 0.164341047406197, 0.0100526856258512, 0.496814846992493, 0.106893293559551, -0.157568067312241, 0.115302592515945, 0.168937563896179, -0.11314033716917, 0.515730321407318, 0.18910063803196, 0.59093964099884, -0.0547696352005005, -0.203118234872818, 0.169050484895706, 0.16194948554039, 0.178553357720375, -0.0613792836666107, -0.506181240081787, 0.508165895938873, -0.00809410121291876, 0.0911374539136887, 0.0904999971389771, 0.0165617056190968, 0.185800895094872, 0.0758463218808174, 0.0570066347718239, 0.451523184776306, -0.0325512550771236, 0.00256156967952847, 0.0707971304655075, -0.170335188508034 },
      { 0.227103695273399, 0.0293496958911419, 0.84741997718811, 0.150169774889946, -0.560610830783844, 0.10877450555563, 0.0403434298932552, 0.0989709571003914, -0.147679418325424, -0.0912571027874947, -0.4909827709198, 0.196312084794044, 0.243815258145332, 0.0158943645656109, 0.0531293042004108, -0.480288505554199, 0.479813426733017, -0.0555402860045433, -0.0155915133655071, 0.0660298839211464, 0.0874513983726501, 0.0285729318857193, -0.109779380261898, 0.0710333958268166, 0.13017150759697, -0.0394724979996681, -0.0882948115468025, 0.230553984642029, -0.313940078020096, -0.0583435222506523, -0.403167486190796, 0.0912071987986565, 0.546986043453217, 0.0634408518671989, -0.548279464244843, 0.0581583566963673, 0.207625404000282, 0.24439549446106, -0.011653327383101, 0.127006381750107, -0.256479889154434, 0.130154713988304, -0.106017656624317, -0.197164922952652, 0.00479372544214129, 0.00108741875737906, -0.175655201077461, 0.164177015423775, 0.259071171283722, -0.0503650605678558, 0.0667019262909889 },
      { 0.851817786693573, 1.00658226013184, 1.0591539144516, 0.892108380794525, 0.962416410446167, 0.978557527065277, 0.941121637821198, 0.914778590202332, 1.0196281671524, 0.935466945171356, 0.991150557994843, 1.02224540710449, 0.891942203044891, 0.97961562871933, 0.931027173995972, 1.00938403606415, 1.25379419326782, 0.96870481967926, 0.952056586742401, 0.967529773712158, 0.914283812046051, 0.961955368518829, 0.923279821872711, 0.962017714977264, 1.04350757598877, 0.965548813343048, 0.92578649520874, 1.12286698818207, 0.965315878391266, 0.914205610752106, 0.940423369407654, 1.022087931633, 1.0639955997467, 0.952212989330292, 0.969836711883545, 0.896435260772705, 0.880467355251312, 1.01163375377655, 0.801087200641632, 0.94726026058197, 0.977739453315735, 0.835519433021545, 1.02169263362885, 1.01587283611298, 0.949327051639557, 0.980083465576172, 0.976244390010834, 0.912190616130829, 0.870914161205292, 0.989743888378143, 1.0228271484375 },
      { -0.0254156906157732, -0.095105305314064, -1.09348046779633, -0.0142422635108232, 0.0229055341333151, -0.047779954969883, 0.0128371110185981, -0.0455444790422916, -0.0856403186917305, 0.064238153398037, -0.0748321786522865, 0.00198412500321865, -0.00265863048844039, -0.0669879764318466, -0.0437351912260056, -0.148171812295914, -0.684539794921875, -0.170621022582054, 0.0697926431894302, -0.0331740006804466, -0.0435432121157646, 0.0185824520885944, -0.0927341505885124, -0.0858000889420509, -0.144474565982819, -0.0234729405492544, -0.0332306437194347, -0.436579644680023, -0.0805822089314461, 0.0119589241221547, -0.150940239429474, -0.0988219901919365, -0.920728147029877, -0.0414294861257076, 0.0431918390095234, -0.0503789484500885, 0.0027387070003897, -0.479871600866318, -0.0251227244734764, -0.0194748174399137, -0.00931491144001484, 0.0368435606360435, 0.00704328482970595, -0.122964724898338, 0.0114983404055238, -0.0704855024814606, -0.0250439513474703, 0.00483282003551722, -0.0561088994145393, -0.0187160391360521, -0.0837112367153168 },
      { 0.0498711839318275, 0.118323959410191, 0.0918594151735306, 0.00206237309612334, -0.00973828881978989, 0.0110991857945919, -0.00724660791456699, 0.0479826405644417, 0.126539975404739, -0.0207025371491909, 0.107899375259876, -0.224790975451469, -0.0312359090894461, 0.225621849298477, 0.0550836995244026, 0.18851138651371, 0.355693250894547, 0.204217091202736, -0.117919974029064, 0.0110479230061173, 0.0264431275427341, -0.0152725810185075, 0.0584129206836224, 0.0564277172088623, 0.174272775650024, 0.0186180826276541, 0.00531726283952594, 0.239062413573265, 0.0766077637672424, -0.00928331259638071, 0.237492188811302, 0.0614296533167362, 0.323990285396576, 0.0104637201875448, -0.0508569106459618, 0.0254890583455563, -0.000918595120310783, 0.837993919849396, 0.072711206972599, 0.0163467023521662, 0.136184826493263, -0.0326241627335548, -0.00349630927667022, 0.16833969950676, -0.0180806890130043, 0.0238429494202137, -0.136313065886497, -0.0274698585271835, 0.00360886030830443, 0.039593331515789, 0.0597976632416248 },
      { 8.43302059173584, 1.83613932132721, 298.074279785156, 18.1294479370117, -26.847204208374, 5.1825532913208, -2.19239187240601, 18.8857402801514, -20.4108543395996, 6.46369647979736, -80.6891021728516, -17.0440311431885, 45.2215881347656, -75.1558380126953, 31.0299110412598, -91.5952072143555, -21.486156463623, -14.9130773544312, 11.2668991088867, -6.34212112426758, 33.5059127807617, -7.83033752441406, 31.4609489440918, 23.7120895385742, -49.2045516967773, -37.0843620300293, 387.267211914062, -39.2402839660645, -72.0948333740234, 18.5659217834473, -4.76370859146118, 2.79405617713928, 99.4197845458984, 27.2741355895996, -25.1457748413086, 3.85003781318665, 6.74954509735107, -290.577423095703, 1.84620308876038, 0.119689464569092, -53.3453750610352, -7.4045877456665, -199.557983398438, -99.4065704345703, 9.79404830932617, 15.3889465332031, -23.9719066619873, -6.40972328186035, 9.38085269927979, -51.8125267028809, -8.49858093261719 }
    },
    {
      { 0.000719039351679385, 0.00713090412318707, -0.00471930671483278, -0.000213386359973811, 0.0137438010424376, 0.00135234359186143, -0.0018567475490272, -0.000659259967505932, 0.00617020670324564, 0.00658245291560888, 0.00755935488268733, 0.0322379097342491, 5.94901393924374e-05, -0.0122325969859958, 0.00263680168427527, 0.0111250588670373, 0.00550108635798097, 0.00348271708935499, 0.00513527262955904, -0.000861899752635509, 0.00438941828906536, -0.000425791513407603, 0.00627391505986452, 0.000498885579872876, 0.0193623211234808, 0.0120074991136789, 0.00844047777354717, 0.00842792075127363, 0.0107204234227538, 0.0066266255453229, 0.0143801653757691, 0.00173061189707369, -0.00379702541977167, 0.000903083244338632, -0.00713778613135219, 0.00162955385167152, -0.00372385722585022, -0.00496697425842285, 0.00819207821041346, 0.00236906693316996, 0.00226387614384294, 0.000654106435831636, 0.00666289124637842, 0.00458278134465218, 0.00339376064948738, 0.00272241840139031, 0.00997177325189114, -0.000953545153606683, 0.000791803642641753, 0.000577789847739041, 0.00132955517619848 },
      { 0.000197469969862141, -0.000124256504932418, -0.00127501250244677, 0.000320180872222409, -0.0174857117235661, 0.00232641980983317, 0.00243842019699514, 0.00157323305029422, -0.00697058020159602, -0.0049101198092103, -0.00963267125189304, -0.0107630509883165, 0.00253677694126964, 0.0160632133483887, 0.00253564305603504, -0.0149183580651879, 0.0131402779370546, -0.002617058577016, -0.00435872189700603, 0.00104357802774757, 0.00152688310481608, -0.000222670394578017, -0.00813850667327642, 0.00516384094953537, 0.00132124021183699, -0.0125995706766844, 0.000561381049919873, 0.000582486973144114, -0.00897801853716373, -0.0020052595064044, -0.0159480813890696, 0.00246791425161064, 0.0132684037089348, 0.000700021046213806, -0.00343367946334183, -0.000746364006772637, 0.00328584620729089, 0.0155154196545482, -0.0127558503299952, 0.00163148168940097, -0.0158166978508234, 0.00491347489878535, -0.00406536413356662, -0.004017842002213, -0.00178299704566598, -0.00385644682683051, -0.01607845723629, 0.00142962008249015, 0.00567289814352989, 0.00126524129882455, 0.00107368908356875 },
      { -0.00214618048630655, 0.000337845965987071, -0.0409685559570789, -0.00409492244943976, 0.0170423742383718, -0.00499316398054361, -0.00352530926465988, -0.000556988117750734, 0.00321875070221722, 0.00293345865793526, 0.00399685930460691, -0.0144253987818956, -0.00512586114928126, -0.00491597503423691, -0.0045046336017549, 0.00740231573581696, -0.0227556396275759, -0.000222023532842286, -0.000243321090238169, -0.000343171181157231, -0.00438743596896529, 0.000574595062062144, 0.000536584586370736, -0.00264485366642475, -0.0113654769957066, 0.000884730601683259, -0.00117180030792952, -0.0127816190943122, -0.0015230622375384, 0.00198801304213703, 0.00316701177507639, -0.000761214236263186, -0.032343864440918, -0.00348609243519604, 0.00884483475238085, -0.00227366480976343, -0.00295078777708113, -0.00618929695338011, 0.000830172211863101, -0.00341190421022475, 0.00428851135075092, -0.00983158312737942, 0.000113228343252558, 0.0010880483314395, -0.00374321010895073, 0.00183364155236632, 0.00865055155009031, -0.00129922782070935, -0.00470663653686643, -0.00219264184124768, 0.00222234800457954 },
      { 1.0021960735321, 1.00104117393494, 0.999957919120789, 1.00223875045776, 1.00130331516266, 1.00152623653412, 1.00418961048126, 1.00196099281311, 1.00078570842743, 1.00106167793274, 1.00097370147705, 1.00209999084473, 1.00341272354126, 1.00334107875824, 1.00337302684784, 1.00205647945404, 0.988124966621399, 1.00170564651489, 1.0004004240036, 1.00158727169037, 1.00120890140533, 1.0010267496109, 1.0009548664093, 0.999697208404541, 1.00193989276886, 1.00474750995636, 0.998425483703613, 0.995425045490265, 1.00156533718109, 1.00231468677521, 1.00443744659424, 0.997124433517456, 1.00279223918915, 1.00117611885071, 1.00132501125336, 1.00373601913452, 1.00176894664764, 0.99952757358551, 1.00316762924194, 1.00180816650391, 1.00369036197662, 1.00533425807953, 0.999552190303802, 0.999772965908051, 1.00142669677734, 1.0012195110321, 1.00108456611633, 1.00190496444702, 1.0013040304184, 1.00119698047638, 0.999665975570679 },
      { 0.000513123406562954, 0.00197085342369974, 0.0583550781011581, -0.000731238687876612, 7.81917333370075e-06, -0.000983847072347999, -0.00197104271501303, -0.000628848385531455, 0.00182196812238544, -0.000239188317209482, 0.00143771979492158, -0.0135043011978269, -5.99704580963589e-05, 0.0100496858358383, -0.00133793987333775, 0.000863224151544273, 0.0317681469023228, 0.00192384875845164, 0.00161899405065924, -0.000899872393347323, -0.000374625466065481, -0.000133917230414227, 0.00205833464860916, 0.000478420028230175, -0.0210020020604134, -0.00736627727746964, -0.00147782161366194, 0.0186522547155619, -0.000324152671964839, -0.00362640130333602, 0.000515472900588065, 0.00275977072305977, 0.0346364378929138, 0.000515926105435938, -0.00163199752569199, -0.00171005947049707, -0.000258152693277225, 0.037369042634964, 0.00275523960590363, 0.000233673214097507, -0.00953406468033791, -0.000833386729937047, 0.000365129206329584, 0.00384173658676445, -0.000203429211978801, 0.000489025725983083, 0.00827497430145741, -0.00179924874100834, 2.06380154850194e-05, -1.95598804566544e-05, 0.00141192635055631 },
      { -0.000990285188890994, -0.00602459767833352, 0.00156226789113134, 0.00195587147027254, -0.00268749054521322, 0.00154438137542456, 0.000237285581533797, -0.00059592048637569, -0.00268011284060776, -0.00348641932941973, -0.00170522811822593, 0.0240340195596218, -0.000419024610891938, -0.0133942337706685, 0.000530919001903385, -0.00513851130381227, -0.0148904183879495, -0.00589505257084966, -0.00130373693536967, -0.000562800094485283, 0.000596990343183279, -0.000375979638192803, -0.000513813691213727, -0.000738849223125726, 0.0332411974668503, 0.00600055558606982, 0.00121392903383821, -0.0101990764960647, 0.000973691523540765, 0.000324646855005994, -0.00546161737293005, -0.00224076164886355, -0.0117494929581881, 0.000409309781389311, 0.000475727894809097, 0.000231629994232208, 0.000281356129562482, -0.0618638135492802, -0.00377398426644504, -0.00156063993927091, 0.0090753436088562, 0.000833452038932592, 0.000447935046395287, -0.00411531422287226, 0.00152723153587431, -0.00140745320823044, -0.0149211883544922, 0.000725653662811965, -0.00128799374215305, 0.000348468369338661, -0.00308309379033744 },
      { -0.42597696185112, -1.18248963356018, -18.8335971832275, -1.82488977909088, -1.6999089717865, -1.23847877979279, 0.237816646695137, -1.19762754440308, -0.705291986465454, -0.325328677892685, -0.440084099769592, -0.256473302841187, -1.22617733478546, 2.55625653266907, -2.51265931129456, 1.06255269050598, -1.9635773897171, -0.498623430728912, -0.342940181493759, -0.144835770130157, -1.75233459472656, 0.426897585391998, -2.28179502487183, -1.24072754383087, 2.39083528518677, 2.43629693984985, -2.30458807945251, -1.09402704238892, 0.74868243932724, -1.22975087165833, -1.37748348712921, -0.423212200403214, -3.82569479942322, -1.61718380451202, -0.499433577060699, -0.57524847984314, -0.541266024112701, 8.10883808135986, 2.05736064910889, -0.309856444597244, 3.38731384277344, -2.84307241439819, 1.43254041671753, 0.254186123609543, -0.80161589384079, -0.330915361642838, -0.946418881416321, 0.695298731327057, -0.0984209850430489, -1.35947906970978, -0.58078670501709 }
    },
    {
      { 0.000363671482773498, -0.0253048520535231, -0.00584701402112842, 0.030808063223958, -0.0126727437600493, 0.00916171167045832, -0.0109465876594186, -0.00853355973958969, -0.00823739171028137, 0.000454059103503823, -0.0243972893804312, -0.0235861986875534, 0.0235226061195135, -0.00126691616605967, 0.0131832119077444, -0.0199312828481197, -0.0148420119658113, -0.0119730243459344, 0.00223088427446783, -0.0164986476302147, 0.0170532874763012, 0.0107490364462137, -0.0321521759033203, 0.0168750360608101, -0.0107251023873687, 0.0458676144480705, 0.0148206762969494, -0.00797224510461092, -0.00207200925797224, 0.000522653688676655, -0.0285358522087336, -0.00884099025279284, 0.037130244076252, 0.019178370013833, 0.0185577720403671, -0.00161746924277395, 0.00629566889256239, -0.003462303429842, 0.0365303941071033, -0.00221775402314961, -0.00456048687919974, 0.0171723943203688, -0.00631952146068215, -0.0208094734698534, 0.0176323857158422, 0.000378212687792256, -0.00841437838971615, -0.00543354172259569, 0.00496034836396575, -0.00023807238903828, -0.0111427446827292 },
      { 0.00810387916862965, 0.0324564427137375, -0.00461507821455598, -0.0259610265493393, 0.0124030159786344, -0.0126157533377409, 0.0227100923657417, 0.0224534776061773, 0.0144454576075077, 0.0148570947349072, 0.0319937728345394, 0.00392264313995838, -0.00353273120708764, -0.0211758259683847, -0.00555232586339116, 0.0258200019598007, 0.000304851855617017, 0.00859183445572853, -0.0102103780955076, 0.021223958581686, -0.0136557137593627, -0.0188812091946602, 0.0437373779714108, -0.0122735938057303, -0.00490338541567326, -0.00661540031433105, -0.0040332218632102, 0.00318349059671164, 0.0386670678853989, 0.0109326839447021, 0.0308158174157143, 0.0120382448658347, -0.0385274179279804, -0.0173102412372828, -0.0031113603617996, 0.0115602789446712, 0.000296192651148885, 0.0142169529572129, -0.0209660511463881, 0.00692129041999578, 0.0165467858314514, -0.00442654965445399, 0.00726403295993805, 0.0193317234516144, -0.0183512885123491, 0.0177701823413372, 0.010846821591258, 0.00869769044220448, 0.00619413750246167, 0.0100438008084893, 0.027911338955164 },
      { -0.00468747410923243, -0.0211636945605278, 0.0140315582975745, 0.00715969316661358, -0.0178926773369312, 0.0021321305539459, -0.0228310283273458, -0.0170799717307091, -0.00232257461175323, -0.0161739103496075, -0.00548532092943788, 0.00552689330652356, -0.0113354437053204, 0.0173450279980898, -0.00203015096485615, -0.0121872602030635, 0.0216272287070751, -0.00110564054921269, -0.0017276470316574, -0.0174745917320251, 0.00176149839535356, 0.00663660420104861, -0.0178660172969103, -0.0081311734393239, 0.0111912814900279, -0.0264099929481745, -0.00221320381388068, 0.0066460995003581, -0.0384990535676479, 0.00675729056820273, -0.017329977825284, -0.0094361761584878, -0.0050787664949894, -0.00117510964628309, -0.00321136065758765, -0.0168488249182701, -0.00690784864127636, -0.00704609230160713, -0.0166856721043587, -0.00731179257854819, -0.00115750194527209, -0.00249639945104718, -0.00193394091911614, -0.00479246722534299, -0.000667328014969826, -0.019838435575366, -0.00781039800494909, -0.0142931649461389, -0.0270516462624073, -0.018354382365942, -0.0209044702351093 },
      { 0.00651022884994745, -0.00237140827812254, 0.010008804500103, -0.00160574831534177, 0.0021659373305738, 0.00951577257364988, 0.0102009894326329, 0.00241866870783269, -0.00206511770375073, 0.00365041987970471, -0.00179605325683951, 0.00327203865163028, -0.00233237817883492, 0.000353463372448459, 0.00281321932561696, 0.000659708282910287, 0.00708037707954645, 0.000947076827287674, 0.00334365502931178, 0.0111536728218198, 0.00437893066555262, 0.00190208933781832, -0.00446514366194606, 0.0036540636792779, 0.00419708248227835, 0.00176580587867647, 0.00176038686186075, 0.0107267731800675, -0.00175366166513413, -0.00293741375207901, 0.00622694147750735, -0.00269307196140289, 0.0209306608885527, -0.000242078749579377, 0.00225258059799671, 0.00586557108908892, 0.001838329131715, 0.000204940239200369, 0.016228836029768, 0.00547864474356174, -0.000482972507597879, -0.0040243798866868, 0.00360018154606223, -0.000436090835137293, 0.0108516328036785, 0.00124220119323581, 0.00158783618826419, 0.00641324417665601, 0.0109322480857372, 0.00561438966542482, 0.000514072657097131 },
      { 1.00302040576935, 1.02016770839691, 0.990649938583374, 1.00796937942505, 1.01126134395599, 0.999361932277679, 0.986905217170715, 1.0048416852951, 1.02029025554657, 1.00180065631866, 1.01338195800781, 1.00957918167114, 1.00968396663666, 1.02090048789978, 1.00477719306946, 1.01749873161316, 0.991417825222015, 1.01937079429626, 1.00487387180328, 0.983532428741455, 1.01034700870514, 0.995237469673157, 1.02846252918243, 1.00253987312317, 1.00787353515625, 1.00296556949615, 1.0025327205658, 0.994013965129852, 1.01845943927765, 1.00679242610931, 1.00925099849701, 1.01291334629059, 0.993677198886871, 1.00582933425903, 1.0029890537262, 1.0060658454895, 1.00804841518402, 1.01354956626892, 1.01809990406036, 1.00559341907501, 1.01038694381714, 1.0087388753891, 1.00821828842163, 1.01824927330017, 0.989350497722626, 1.01557528972626, 1.01087009906769, 0.996612906455994, 1.00910472869873, 0.995780766010284, 1.01121950149536 },
      { -0.00509880995377898, -0.0106207681819797, -0.00197035446763039, -0.0072894636541605, -0.00475885905325413, -0.00377620058134198, 0.0152133833616972, -0.00275154667906463, -0.0197844784706831, -0.0010877235326916, -0.0131652234122157, -0.00874292384833097, -0.00604290375486016, -0.0186368692666292, -0.00628632912412286, -0.0208881720900536, -0.00315477256663144, -0.0191580764949322, 0.00336674740538001, 0.0201769527047873, -0.0137182353064418, 0.00801673997193575, -0.0280154794454575, 0.00098810310009867, -0.0129208704456687, -0.00387567491270602, -8.011405589059e-05, -0.0038719882722944, -0.0156139712780714, -0.0114351101219654, -0.0112292319536209, -0.00560752162709832, -0.00682843942195177, -0.00212094932794571, 0.00426314817741513, -0.00451200967654586, -0.00641430262476206, -0.0083576338365674, -0.0320318154990673, -0.00774359796196222, -0.00342513551004231, -0.0101468488574028, -0.0123967947438359, -0.0185973197221756, 0.00463649583980441, -0.0128643317148089, -0.00684319064021111, 0.00775648513808846, -0.00511481612920761, 0.0133023550733924, -0.00716416630893946 },
      { -0.110618807375431, -4.19592475891113, 2.33073115348816, -0.406494081020355, 0.909158766269684, -7.1098484992981, 0.234799668192863, -0.233347743749619, -1.13725936412811, -1.2368950843811, 2.34095549583435, 1.31643652915955, 0.827186524868011, 2.40716648101807, 1.10748422145844, 1.69435119628906, 3.67702889442444, -1.86812400817871, -0.520876586437225, -2.49545121192932, -0.587324619293213, 7.0872220993042, -1.58458364009857, 1.11130118370056, 1.39903426170349, 8.56469345092773, -16.2867813110352, -1.44958186149597, 3.85558152198792, -0.778826057910919, -2.4932427406311, 0.0251144710928202, -10.2623662948608, 1.02043807506561, 1.53796899318695, 0.304670244455338, 0.293159663677216, 2.99633598327637, 11.6462688446045, 0.734061539173126, 0.646358251571655, 13.3684015274048, 1.52882528305054, 2.13776183128357, -4.10778093338013, -1.66183018684387, 0.387034118175507, 2.76884746551514, 0.931456744670868, -7.61808204650879, -1.24973142147064 }
    },
    {
      { -0.00766984513029456, 0.00127857481129467, -0.00908019673079252, 0.0670676678419113, -0.00108723167795688, 0.000321685540257022, 0.00360791315324605, 0.0133234839886427, 0.000199891641386785, 0.00256160367280245, -0.00416031712666154, 0.00297723314724863, 0.0593517944216728, -0.00506046693772078, 0.0550871714949608, -0.00263177463784814, -0.0121158678084612, 0.0020968469325453, 0.0035988874733448, 0.00209320592693985, 0.0734435170888901, 0.00320240622386336, -0.0029243326280266, 0.0430280975997448, 0.00421479297801852, 0.147307828068733, -0.000291348493192345, -0.0118547081947327, 0.0223750490695238, 0.0273982658982277, -0.00511144241318107, 0.0127438809722662, 0.0279954541474581, 0.056577455252409, -2.18762743315892e-05, 0.0176264643669128, 0.0183329693973064, -0.00123197806533426, 0.149479523301125, 0.0138038182631135, 0.00233043520711362, 0.0462642721831799, 0.00723750609904528, -0.000232269303523935, 0.0106604332104325, 0.0164563003927469, -0.000190045844647102, 0.000279670348390937, 0.0537497512996197, 0.00219515291973948, 0.00285555771552026 },
      { 0.004156778100878, 0.0114569859579206, -0.00846957508474588, -0.0435458533465862, 0.00415575364604592, -0.0102846100926399, -0.000170935119967908, 0.00481140101328492, -0.000390396598959342, 0.00636796746402979, 0.00546315778046846, 0.00468068616464734, -0.0196283180266619, 0.00578332040458918, -0.0262911058962345, 0.00832748413085938, -0.00640333304181695, -0.00167438492644578, -0.0115891061723232, -0.00477283447980881, -0.0540487878024578, -0.00765540543943644, 0.00494128279387951, -0.0343358665704727, -0.000795582251157612, -0.0805478766560555, 0.00154026539530605, 0.000853995501529425, 0.00895948614925146, -0.00460901949554682, 0.00188087217975408, -0.0219773426651955, -0.0435388125479221, -0.0455276519060135, 0.00105873018037528, -0.00254502589814365, -0.0157508626580238, -0.00255850260145962, -0.140486717224121, -0.0300094913691282, 0.00281261210329831, 0.0128768151625991, -0.00717332866042852, 0.00150483730249107, -0.020074374973774, 0.000522052869200706, 0.000497194880153984, 0.000818125321529806, 0.00529197510331869, 0.00272056716494262, 0.0168259982019663 },
      { -0.0160324387252331, -0.0123703377321362, 0.027706190943718, -0.0298683103173971, -0.00910458527505398, 0.0137937068939209, -0.00163291476201266, -0.0191417951136827, 0.00257641030475497, -0.00850966386497021, 0.00360266049392521, -0.00910031329840422, -0.0329357720911503, 0.000137865310534835, -0.0131680220365524, -0.00760239688679576, 0.0223745796829462, -0.00231652660295367, 0.00513573875650764, -0.000452625390607864, -0.00834590196609497, 0.000376117328414693, -0.00944903865456581, -0.0124906841665506, -0.00422545894980431, -0.0356871709227562, -0.00253072520717978, 0.0143646420910954, -0.0242035929113626, 0.0139093436300755, -0.000522295013070107, -0.00250622420571744, 0.0151264658197761, -0.00453113159164786, -0.00132664607372135, -0.0123249599710107, -0.000483467389130965, 0.00369418994523585, -0.0155257983133197, 0.0123532498255372, -0.00843133497983217, -0.0560573041439056, 0.00203305366449058, -0.00294846156612039, 0.0155202895402908, -0.016638420522213, -0.00386073789559305, 0.00386391626670957, -0.0677804574370384, -0.00581638887524605, -0.0154682695865631 },
      { -0.0147776184603572, -0.00246332609094679, 0.00458290008828044, 0.0343733206391335, 0.000141577591421083, 0.00425869133323431, -0.00122083653695881, 0.0069783185608685, -0.00161941326223314, 0.000439772935351357, -0.00186326482798904, -0.000429634732427076, 0.0246307086199522, 0.00178617902565747, 0.00795310456305742, -0.00126963073853403, 0.0025206923019141, -0.000479852751595899, 0.000675333722028881, 0.0022096554748714, 0.010537825524807, 0.00187964073847979, -0.00242112437263131, 0.00725709786638618, 0.000203229486942291, 0.00987284909933805, 0.0042717126198113, 0.00415300764143467, -0.00493151508271694, -0.00710559263825417, 0.000262741814367473, 0.00316723878495395, 0.00490684108808637, 0.0032995652873069, -0.000117091185529716, 0.00320200901478529, 0.00452719954773784, 0.000219702269532718, 0.0505606755614281, 0.00654615834355354, 7.77197419665754e-05, 0.0253462381660938, -0.000300619285553694, -0.001515285926871, 0.00875731743872166, 0.000116289986181073, 0.000643867126200348, -0.00102172303013504, 0.0410788357257843, 0.00372077594511211, -4.78363836009521e-05 },
      { 0.0119055453687906, 0.0134077183902264, -0.00695202359929681, -0.0117206564173102, 0.0060703856870532, 0.00464957160875201, 0.00415006000548601, -1.29767940961756e-05, 0.0122654354199767, 0.00455314991995692, 0.0104622608050704, 0.00647222436964512, -0.00243423599749804, 0.0131829679012299, 0.000242583424551412, 0.0127661461010575, 0.00454714661464095, 0.0113552389666438, 0.00605271756649017, 0.00640310300514102, 0.00729713682085276, 0.000715275935363024, 0.0279851537197828, 0.00147042982280254, 0.00727219879627228, 0.00240815640427172, -0.00268405815586448, 0.00227557960897684, 0.0166542883962393, 0.00423607509583235, 0.0124420179054141, 0.012349046766758, 0.00896485429257154, 0.00240735779516399, 0.00344567676074803, 0.00371681596152484, 0.0032105993013829, 0.0155838113278151, 0.0249803941696882, 0.0083888117223978, 0.00449457112699747, -0.00375614617951214, 0.0123728793114424, 0.0143414316698909, -0.0129298698157072, 0.0143172834068537, 0.00526773603633046, -0.000402213743655011, -0.00756048131734133, -0.00237559131346643, 0.00832234416157007 },
      { 1.00059843063354, 0.991738796234131, 0.993083000183105, 1.00048077106476, 0.998867809772491, 0.991068661212921, 0.995985686779022, 1.00034832954407, 0.989944458007812, 0.997113883495331, 0.989774942398071, 0.999117732048035, 0.990391552448273, 0.985018849372864, 0.994541227817535, 0.987573385238647, 0.993171036243439, 0.990960001945496, 0.997388422489166, 0.995016813278198, 0.988443851470947, 1.00221836566925, 0.973021268844604, 1.00015103816986, 0.995731592178345, 0.992221295833588, 1.00448441505432, 0.993835866451263, 0.988585889339447, 0.989306092262268, 0.987759292125702, 0.994985282421112, 0.989562928676605, 0.998701870441437, 1.00559282302856, 0.997265219688416, 0.995249569416046, 0.985220432281494, 0.94596540927887, 0.989016532897949, 1.00250267982483, 0.987573385238647, 0.986013948917389, 0.986846506595612, 1.00182366371155, 0.989518582820892, 0.999169647693634, 0.996939241886139, 0.991148471832275, 1.00467646121979, 0.992374420166016 },
      { 6.22884798049927, -3.29878354072571, 10.4156761169434, 3.31092858314514, 0.91766756772995, -3.58298563957214, -0.162179857492447, -3.51619935035706, -0.968527853488922, -0.549361407756805, 1.15855002403259, 0.5118128657341, 1.61255240440369, -0.542326807975769, 0.149143040180206, 1.3736537694931, 2.65462255477905, -0.0579372309148312, 0.463286072015762, -1.44107520580292, -3.30560779571533, 5.02175378799438, -3.89495325088501, 0.564043462276459, -0.161629617214203, 8.58342170715332, -9.55197525024414, 1.75710868835449, 2.86899304389954, -3.25768995285034, -0.738622546195984, -0.797677040100098, -3.61228537559509, 0.301571696996689, -0.922010958194733, -3.30048584938049, 0.173236027359962, 0.701237201690674, 2.84016966819763, -0.559445083141327, 0.414727747440338, 22.1161346435547, 0.923471331596375, 0.876826286315918, -2.88054132461548, -2.61504364013672, 0.00222086836583912, 0.444981396198273, 8.72086524963379, -5.83839750289917, -1.73493075370789 }
    }
  },
  {
    {
      { 0.939290702342987, 0.92669004201889, 0.939466059207916, 0.96964019536972, 0.902462065219879, 0.939913272857666, 0.861581206321716, 0.954078733921051, 0.926370680332184, 0.904682815074921, 0.887807369232178, 0.892564058303833, 1.00587677955627, 0.514238893985748, 0.984029769897461, 0.861586034297943, 0.947417974472046, 0.918216049671173, 0.849651575088501, 0.899894833564758, 0.988404929637909, 0.865638256072998, 0.890447080135345, 0.944159030914307, 0.8824263215065, 0.984453856945038, 0.97308611869812, 0.974662899971008, 0.914485216140747, 0.930266797542572, 0.889640271663666, 0.951455533504486, 0.908372819423676, 0.970015108585358, 0.906798362731934, 0.940833032131195, 0.914158940315247, 0.633471131324768, 1.04145336151123, 0.953913986682892, 0.928337812423706, 0.949030697345734, 0.923563778400421, 0.905734360218048, 0.880119800567627, 0.904042303562164, 0.899763405323029, 0.858990609645844, 0.947731256484985, 0.949886977672577, 0.921081006526947 },
      { 0.0878715664148331, 0.0733170062303543, -0.0442714057862759, 0.0644307509064674, 0.106100276112556, -0.054637610912323, 0.203633680939674, 0.0811557397246361, 0.0822447240352631, 0.0829328596591949, 0.107522629201412, 0.103078737854958, 0.0850826054811478, 0.995840311050415, 0.0674834921956062, 0.103843502700329, -0.0373660326004028, 0.0978799313306808, 0.199757248163223, 0.136752530932426, 0.0638290867209435, 0.19952604174614, 0.120189823210239, 0.0913735404610634, 0.0978341475129128, 0.0671482309699059, 0.0412883907556534, -0.104943670332432, 0.127727285027504, 0.113457269966602, 0.105379246175289, 0.060605350881815, -0.00110160873737186, 0.0715009570121765, 0.093402586877346, 0.119240306317806, 0.123600654304028, 0.718388020992279, -0.0324692726135254, 0.0546613521873951, 0.0918242856860161, 0.0683462247252464, 0.0867518112063408, 0.092814564704895, 0.0909784734249115, 0.112842164933681, 0.0864203721284866, 0.208637729287148, 0.107128910720348, 0.148412391543388, 0.0712840184569359 },
      { -0.0227114167064428, 0.00999292824417353, 0.0955987349152565, -0.00283683184534311, -0.00212280126288533, 0.109224796295166, -0.0638268291950226, -0.0177736785262823, 0.00312203844077885, 0.0169023424386978, 0.0017791228601709, -0.00219309167005122, -0.0412236489355564, -0.229804679751396, -0.0126214697957039, 0.0226547718048096, 0.0851304233074188, -0.00405307952314615, 0.00375810288824141, -0.0156184388324618, 7.93813305790536e-05, -0.0489614754915237, -0.00717228418216109, -0.0168229136615992, 0.0147021571174264, -0.0222050510346889, -0.0159054454416037, 0.11812250316143, -0.0269292593002319, -0.0153601616621017, 0.00293887965381145, 0.0298918075859547, 0.0975844487547874, -0.0147672221064568, 0.028033247217536, -0.0178103484213352, -0.0274439919739962, -0.1350427120924, 0.00867684651166201, 0.0365147776901722, 0.00966231059283018, -0.0122596146538854, -0.00354329100809991, -0.00706722447648644, 0.033262949436903, -0.00513905147090554, 0.0211655702441931, -0.0146300001069903, -0.0491086840629578, -0.0762073248624802, 0.0211659334599972 },
      { 0.0138564575463533, 0.00520094623789191, -0.0034013397525996, 0.00245627760887146, -0.000234955732594244, 0.0104227308183908, 0.0228952765464783, 0.00124568655155599, 0.0021166424266994, -0.00434907199814916, -0.000599379243794829, -0.00225363811478019, 0.00596135249361396, -0.000553334190044552, 0.00351673364639282, -1.69499489857117e-05, 0.00164220889564604, -0.000695021240971982, -0.0186182428151369, 0.0149132888764143, 0.000926289416383952, 0.00344564905390143, -0.00198434991762042, 0.00104769307654351, 0.000386436353437603, 0.00715163769200444, 0.00435853330418468, 0.00942245870828629, -0.00505618238821626, -0.00560007197782397, 0.00283912871964276, -0.021886209025979, 0.00748211657628417, -0.00335411261767149, 0.00246206973679364, -0.0173165164887905, 0.0044580684043467, -0.0222194939851761, 0.0384532809257507, -0.00235781050287187, -0.00327444588765502, 0.0186137687414885, -0.000729780120309442, -0.00248365360312164, 0.00167059921659529, -0.00370476278476417, -0.000341337115969509, -0.0194559805095196, 0.0227989833801985, 0.00817180518060923, 0.00703513901680708 },
      { 0.00586161762475967, -0.00194604764692485, 0.0233227964490652, 0.00269432947970927, -0.00160259660333395, -0.00125048728659749, -0.0149883972480893, -0.00215038470923901, -0.00362938735634089, 0.0112324971705675, -0.00554456561803818, 0.00325978570617735, -0.00295407301746309, -0.139684528112411, -0.00308246933855116, 0.00419813487678766, 0.0118192657828331, -0.0056341765448451, 0.011392435990274, -0.0107902931049466, 0.00756821967661381, -0.00755218975245953, 0.00713931676000357, 0.00301530887372792, -0.00349011877551675, -0.0123027255758643, 0.00207876460626721, 0.0151852490380406, 0.000109047759906389, 0.00613854406401515, -0.0082060219720006, 0.0100808469578624, -0.0185105465352535, 0.00327693508006632, -0.0145746516063809, -0.0160465873777866, 0.0036809139419347, 0.110052667558193, 0.0146429389715195, -0.000910411414224654, 0.00126428937073797, -0.0120127815753222, 0.00243337941356003, 0.01335225161165, 0.000495610933285207, -0.00140156911220402, 0.00796647649258375, -0.0217897277325392, -0.00103725423105061, 0.0187626481056213, -4.63765027234331e-05 },
      { -0.0103349480777979, -0.00189077388495207, -0.00753179471939802, -0.0074098901823163, -0.00413229735568166, -0.00539971562102437, -0.00788663513958454, 0.0043644281104207, -7.63019270380028e-05, -0.00993786193430424, 0.0125935208052397, 0.00132311426568776, 0.0012109731324017, 0.151921272277832, -0.00185890274588019, 0.000231632933719084, -0.00834161043167114, 0.00658634584397078, -0.0284535754472017, -0.0206346064805984, -0.0201013423502445, 0.00866381637752056, -0.00760069489479065, -0.00580684561282396, 0.00704042660072446, 0.00658769672736526, -0.00468762405216694, -0.0117246340960264, 0.00520383706316352, -0.00778230326250196, 0.0106551637873054, -0.0141268847510219, 0.00601712940260768, -0.00191507721319795, 0.0149103738367558, 0.0212738942354918, -0.00666288286447525, -0.0974557995796204, -0.0432352386415005, -0.0238135736435652, 0.00179909798316658, 0.0048824199475348, -0.00160922156646848, -0.008152618072927, -0.0066765658557415, 0.00133822544012219, -0.0233914945274591, 0.00189246179070324, -0.00588088482618332, -0.0268243178725243, -0.0126505792140961 },
      { -5.51440811157227, -7.57865238189697, -5.77112483978271, -10.1484022140503, 0.128809541463852, 3.66786789894104, -3.80385756492615, -6.9135799407959, -2.7126522064209, -1.07854676246643, -4.01524543762207, 0.980889737606049, -19.0391845703125, -35.8486442565918, -9.45209121704102, -10.3160390853882, -7.12206792831421, -1.54777228832245, -7.73997688293457, -4.3534574508667, -11.0040655136108, -28.2311973571777, -6.81161594390869, -4.65026044845581, -2.49066615104675, 3.86492276191711, 127.112403869629, -12.7406454086304, -0.758376717567444, -7.41025924682617, -6.01424837112427, -3.3549633026123, 1.64210796356201, -6.20106267929077, -0.681714236736298, -6.26016712188721, -3.36151623725891, -51.8091316223145, -2.01411557197571, 1.11804819107056, -1.41041851043701, 0.437416851520538, -2.91434097290039, -4.40381669998169, -1.73095560073853, -3.65264105796814, 1.01883018016815, -13.0558366775513, -7.26696395874023, -1.71422803401947, -4.70941162109375 }
    },
    {
      { 0.0154699757695198, -0.00629610614851117, 0.0108060296624899, 0.00435239961370826, -0.0175440907478333, -0.0404590927064419, -0.020375095307827, 0.00837989710271358, 0.0076624033972621, -0.0136331683024764, 0.000599606137257069, -0.0106923254206777, 8.31295910757035e-05, 0.000334066513460129, 0.00157872028648853, 0.00400612829253078, -0.0259948074817657, 0.00636891601607203, -0.00800539180636406, -0.00805000867694616, -0.00407447712495923, -0.000426155282184482, -0.0279824398458004, -0.016525074839592, -0.0142775112763047, 0.00455111591145396, -0.154071152210236, -0.0218410175293684, -0.0471890494227409, -0.0252353362739086, -0.00466226460412145, 0.00686591677367687, -0.0227412562817335, -0.012856719084084, -0.0544199422001839, -0.0126600358635187, 0.012440244667232, 0.0328639298677444, 0.0351333171129227, 0.00318273832090199, 0.0140465367585421, -0.0121946288272738, -0.0343830138444901, -0.0108572663739324, -0.0316315218806267, -0.021397652104497, -0.024129981175065, -0.0067303329706192, -0.0143720330670476, -0.0482072532176971, -0.00241615856066346 },
      { 1.03615081310272, 1.0332909822464, 1.06076323986053, 1.04447817802429, 1.09135699272156, 1.09984481334686, 1.06066262722015, 1.0365731716156, 1.02970087528229, 1.04621303081512, 1.06739020347595, 1.10789096355438, 1.0683286190033, 0.940054655075073, 1.05031645298004, 1.05987298488617, 1.09577012062073, 1.05462777614594, 1.06354928016663, 1.04040765762329, 1.0547548532486, 1.08104848861694, 1.07315540313721, 1.06225836277008, 1.10734236240387, 1.021399974823, 1.29402530193329, 1.08778202533722, 1.11574149131775, 1.08187818527222, 1.0666424036026, 1.00719559192657, 1.05948114395142, 1.05736947059631, 1.10115027427673, 1.07515704631805, 1.04265439510345, 0.922325849533081, 1.01232004165649, 1.01210379600525, 1.10563814640045, 1.05412721633911, 1.08645617961884, 1.06357252597809, 1.05092263221741, 1.06394612789154, 1.05483865737915, 1.02126395702362, 1.06616699695587, 1.07753491401672, 1.0236302614212 },
      { -0.0532884523272514, -0.0238530952483416, -0.122644394636154, -0.0413235723972321, -0.0484508909285069, -0.0622750334441662, -0.0491954497992992, -0.0419470109045506, -0.0403726994991302, -0.0439556837081909, -0.0753370746970177, -0.0884854570031166, -0.0657439380884171, 0.0527668334543705, -0.050830889493227, -0.0566365234553814, -0.082183800637722, -0.0458579324185848, -0.0491338036954403, -0.0281686391681433, -0.0447057113051414, -0.0797670856118202, -0.0458926782011986, -0.0497349239885807, -0.0715806633234024, -0.0321421697735786, -0.191059827804565, -0.0718823373317719, -0.066022552549839, -0.0577862411737442, -0.0694776922464371, -0.007479646243155, -0.0517779178917408, -0.0486778132617474, -0.073687382042408, -0.0445549376308918, -0.0549419224262238, 0.0294053424149752, -0.0502439737319946, -0.00053170247701928, -0.101263776421547, -0.0421811304986477, -0.0540668219327927, -0.0587125606834888, -0.021769231185317, -0.0413475856184959, -0.0276130139827728, -0.0144369201734662, -0.0633434802293777, -0.0302882920950651, -0.0163063146173954 },
      { 0.0166175030171871, 0.00358364544808865, -0.00581901567056775, 0.000375203118892387, 0.00072465231642127, 0.00028195945196785, 0.0174196343868971, 0.00908309500664473, 0.00855861697345972, -0.00363754760473967, 0.00434329593554139, -0.00174655439332128, 0.00350890378467739, -0.00195737998001277, 0.00578318443149328, 0.00370369432494044, -0.0140371397137642, 0.00365221872925758, -0.00396047485992312, 0.012009616009891, -0.00100998790003359, 0.00827534403651953, 0.00309115345589817, 0.000227530166739598, -0.00266162352636456, 0.00857057515531778, 0.0438039228320122, -0.0108877876773477, 0.00283402693457901, 0.00054888246813789, 0.00725212087854743, -0.00724357599392533, -0.00926177576184273, 0.000839419604744762, -0.000717562274076045, -0.00859679840505123, 0.0122489910572767, -0.0026938640512526, 0.0219108927994967, 0.00572579307481647, -0.00314411032013595, 0.0037137686740607, 0.000284359324723482, 0.00270493049174547, 0.00385543867014349, -0.000537358340807259, 0.00148070009890944, 0.00578681891784072, 0.0138331102207303, -0.00464318133890629, 0.00802363827824593 },
      { 0.00527844671159983, 0.00328849488869309, 0.0695882737636566, 0.00465156370773911, 0.02004131115973, -0.00105017470195889, -0.00294149178080261, -0.0062618050724268, -0.00219146301969886, 0.0204794518649578, -0.00475499173626304, 0.0201473906636238, 0.00371883786283433, -0.0647399500012398, -0.000346980203175917, 0.00395637797191739, 0.0458671562373638, -0.00222306465730071, 0.0252276696264744, -0.00748362438753247, 0.00868585240095854, -0.00329237384721637, -0.00555364787578583, 0.0110534345731139, 0.0267071053385735, -0.0044239666312933, -0.00866149831563234, 0.0311164371669292, -0.00489027472212911, 0.00307285529561341, -0.0116410572081804, 0.00835618656128645, 0.0425068289041519, 0.0071972175501287, 0.0114175667986274, -0.00862540770322084, 0.00155810383148491, -0.0457308702170849, 0.00614394759759307, -0.000318945094477385, 0.0248174499720335, -0.00110054004471749, 0.00106291961856186, 0.012928712181747, -0.00187273439951241, -0.00186118436977267, 0.0307694040238857, -0.00194926699623466, -0.000663100625388324, -0.00450903223827481, 0.00136432552244514 },
      { -0.00930519308894873, -0.00452669570222497, -0.00293065863661468, -0.00457705417647958, -0.0474383570253849, 0.00101836863905191, -0.00363871804438531, 0.00409826869145036, -0.0018701272783801, -0.00924787204712629, 0.0155418775975704, -0.0337679944932461, 1.11377466964768e-05, 0.0775834247469902, 0.00160965882241726, 0.000491109269205481, -0.0203929170966148, -0.00313657522201538, -0.0338331460952759, -0.00716025196015835, -0.00798253994435072, 0.00273782853037119, 0.00703321769833565, -0.00917438697069883, -0.0507761500775814, 0.0062337601557374, 0.0144925424829125, -0.0173857938498259, 0.00462019164115191, 0.00118101330008358, 0.0190925393253565, -0.00403760373592377, -0.0191175434738398, -0.0046862936578691, -0.0121384030207992, 0.0115974191576242, -0.00454418640583754, 0.0608821474015713, -0.014407642185688, -0.0153210759162903, -0.034503199160099, -0.000526435498613864, -0.000838687177747488, -0.0106194438412786, -0.00209594378247857, 0.00325696566142142, -0.0592262968420982, -0.00108043395448476, -7.37978698452935e-05, 0.0132012730464339, -0.0104731479659677 },
      { -4.87957668304443, -5.69003629684448, -32.2613372802734, -4.15610837936401, -6.49861907958984, 2.84936308860779, -5.10891628265381, -4.69995546340942, -2.94517064094543, 0.878837764263153, -5.81450223922729, -5.50967645645142, -6.67904281616211, -25.7455615997314, -5.53951930999756, -13.9582262039185, -3.48594737052917, -3.45270824432373, 4.15121555328369, -1.76824641227722, -5.14625215530396, -17.607364654541, -3.66261315345764, -3.5809063911438, -5.32638263702393, -4.65200710296631, -187.158737182617, 0.543666839599609, -1.84840703010559, -5.80686664581299, -4.39780950546265, -1.21205699443817, -0.422682851552963, -3.1781370639801, 3.12905716896057, -3.97846722602844, -2.58702230453491, -37.5979385375977, -3.33850812911987, 1.63078665733337, -9.58771991729736, 3.18589329719543, -1.07583391666412, -10.9557275772095, 2.90982007980347, -1.14576685428619, -2.62705230712891, -3.06796836853027, -2.33689284324646, -5.76096105575562, -2.93066644668579 }
    },
    {
      { 0.0756751075387001, 0.0599985867738724, 0.0867131501436234, 0.0840139091014862, 0.00554110435768962, 0.081903412938118, -0.00060680037131533, 0.0557890981435776, 0.0522514432668686, 0.0501731187105179, 0.0587012991309166, -0.0295263230800629, 0.0998058691620827, 0.0101950541138649, 0.112159386277199, 0.0830645933747292, 0.0641293972730637, 0.0110176960006356, 0.0476875752210617, 0.0297530349344015, 0.0988682508468628, 0.0249318983405828, 0.0328394547104836, 0.0502287745475769, -0.00915350951254368, 0.0822495669126511, 0.00420322315767407, 0.118242733180523, -0.0502346344292164, 0.0638368427753448, 0.0635402351617813, 0.0650227144360542, 0.0551345907151699, 0.0726039782166481, 0.0567186437547207, 0.0611405782401562, 0.0383938029408455, 0.090030625462532, 0.168282046914101, 0.0559281334280968, -0.00078430789289996, 0.0391342267394066, -0.00352992559783161, -0.0246603991836309, 0.0523004718124866, 0.0455918461084366, -0.0186879560351372, 0.0163502972573042, 0.0355321504175663, 0.0169686749577522, 0.076123408973217 },
      { -0.138369932770729, -0.200105860829353, -0.228624418377876, -0.134645655751228, -0.113341934978962, -0.274830639362335, -0.0609205923974514, -0.142788723111153, -0.208090350031853, -0.146094635128975, -0.152626946568489, -0.0692509934306145, -0.0798116773366928, -0.437820494174957, -0.174664288759232, -0.0880849659442902, -0.207115352153778, -0.17519523203373, -0.106365092098713, -0.115325562655926, -0.146871373057365, 0.0165438801050186, -0.163385793566704, -0.134575560688972, -0.0777083486318588, -0.221211284399033, -0.0356847755610943, -0.339690655469894, -0.104444652795792, -0.134810790419579, -0.151518553495407, -0.179757431149483, -0.219943180680275, -0.148242339491844, -0.0943563282489777, -0.141450762748718, -0.0967085957527161, -0.383219212293625, -0.288197040557861, -0.172078758478165, -0.0815088078379631, -0.117653213441372, -0.0560947395861149, -0.13239885866642, -0.217074900865555, -0.15126159787178, -0.111676797270775, -0.106733746826649, -0.0924928858876228, -0.062841922044754, -0.203450247645378 },
      { 1.05479681491852, 1.14516353607178, 1.06830787658691, 1.08498597145081, 1.09968709945679, 1.18351709842682, 1.04885852336884, 1.09431207180023, 1.12641906738281, 1.09749603271484, 1.08818197250366, 1.07379913330078, 1.01887488365173, 1.38412189483643, 1.09112012386322, 1.03635764122009, 1.12601375579834, 1.14480400085449, 1.09245824813843, 1.09409248828888, 1.11202728748322, 0.959037065505981, 1.10707628726959, 1.0855108499527, 1.06912648677826, 1.12319207191467, 1.017373919487, 1.20757865905762, 1.14033484458923, 1.1166318655014, 1.0948930978775, 1.12700057029724, 1.14539623260498, 1.08413434028625, 1.01497888565063, 1.10599648952484, 1.04784989356995, 1.24677002429962, 1.10766494274139, 1.13999283313751, 1.06836521625519, 1.08035171031952, 1.05904793739319, 1.13339471817017, 1.15861117839813, 1.11726486682892, 1.1004958152771, 1.1001980304718, 1.02784144878387, 1.06794512271881, 1.1474974155426 },
      { 0.028973501175642, 0.00295148836448789, 0.0188319776207209, -0.0108076734468341, 0.00082672230200842, 0.0124685959890485, 0.0282319504767656, 0.0150798577815294, 0.0199163444340229, -0.00116760493256152, 0.0064148586243391, 0.00323264859616756, -0.00519464164972305, -0.00709644658491015, 0.0107743153348565, 0.000648468383587897, -0.00563058629631996, 0.0045232349075377, -0.00738019635900855, 0.0161660499870777, -0.00845613330602646, 0.0259284842759371, 0.00636206427589059, 0.00294326548464596, -0.000300076586427167, 0.0166820846498013, 0.0107290269806981, 0.00493470253422856, 0.00982702989131212, 0.00282313395291567, 0.00324219139292836, -0.00909459218382835, 0.0151175549253821, -0.00112825690302998, -5.89627816225402e-05, -0.012856681831181, 0.018743596971035, -0.00767622515559196, 0.0552546195685863, 0.00570567743852735, 0.00539155723527074, 0.00883385818451643, 0.000479129754239693, 0.0129276216030121, 0.0114871077239513, 0.00150192307773978, 0.00860810559242964, 0.00419842358678579, 0.027316777035594, -0.00728863850235939, 0.00538403075188398 },
      { 0.00850050058215857, 0.00988008733838797, 0.0699728429317474, 0.0126999979838729, -0.0038589050527662, -0.0043485532514751, -0.00820638332515955, -0.0103142512962222, -0.00140781374648213, 0.0144841251894832, 0.00801411084830761, -0.0125797837972641, 0.0110097117722034, -0.16886043548584, -0.0143365086987615, -0.0155854038894176, 0.0471095629036427, 0.00970026291906834, 0.020316181704402, -0.00916690658777952, 0.0181952677667141, -0.00992396287620068, 0.0109759988263249, 0.014311627484858, -0.00566961150616407, -0.0163400489836931, 0.00217235996387899, 0.0276999939233065, -0.012791964225471, -0.0161130763590336, 0.00151734275277704, 0.00557855144143105, 0.0133698331192136, 0.0142920510843396, 0.00380874006077647, -0.0107975769788027, -7.19207746442407e-05, -0.120250053703785, 0.015019073151052, -0.00347039615735412, -0.0302047878503799, -0.00379112712107599, 0.00239662593230605, -0.0274168364703655, -0.00552340038120747, 0.000478390487842262, -0.0276910215616226, -0.00540839741006494, 0.00421891687437892, 0.0124277118593454, 0.00203468324616551 },
      { -0.0135298212990165, -0.0108025474473834, 0.0033860607072711, -0.00894083827733994, 0.016606668010354, 0.00135900150053203, -0.0072521585971117, 0.00391965219751, -0.00972749385982752, -0.0122027862817049, -0.0100661749020219, 0.0233915150165558, 0.00184302602428943, 0.16567125916481, 0.0140888206660748, 0.0166081711649895, -0.0190976466983557, -0.0027052101213485, -0.0399406589567661, -0.011604730039835, -0.0319206416606903, 0.00523660238832235, -0.00362765160389245, -0.0152653688564897, 0.0243748631328344, 0.025184178724885, -0.00240247393958271, -0.0180212166160345, 0.00903125014156103, 0.00503593822941184, 0.00293585518375039, -0.000340557278832421, -0.00589672103524208, -0.0107661774381995, 0.00678767450153828, 0.0196817256510258, -0.00220131827518344, 0.164892539381981, -0.0381679497659206, -0.0163046792149544, 0.0564438179135323, 0.00284997839480639, -0.0028394719120115, 0.027518792077899, 0.00126411276869476, -0.00507118366658688, 0.0494598969817162, -0.00137649837415665, 0.00249234773218632, -0.0157235842198133, -0.0129917720332742 },
      { -8.07236099243164, -7.53020429611206, -51.2090225219727, -10.1319236755371, 2.08204197883606, 6.75792169570923, -3.19653081893921, -10.710563659668, -4.72074317932129, 0.618454337120056, -3.84406924247742, 0.484294176101685, -20.5653343200684, -69.2355117797852, -13.3251686096191, 10.4876346588135, -14.4324979782104, -4.54859447479248, 1.79371023178101, -4.67462921142578, -13.194730758667, -63.0421905517578, -10.6397495269775, -5.41506958007812, 3.72197771072388, -14.1005344390869, 30.200439453125, -2.86790180206299, -8.12352180480957, -11.8642559051514, -4.83953905105591, -2.39561533927917, -6.69779968261719, -6.84417581558228, 1.29172706604004, -9.73726272583008, -2.96282386779785, -109.691673278809, -5.66589832305908, 0.605839550495148, -2.67602396011353, 0.942885220050812, 1.33243083953857, -1.04689848423004, 1.52455973625183, -2.97900247573853, -1.09455513954163, -9.44714736938477, -8.99505138397217, -6.02827978134155, -2.08379220962524 }
    },
    {
      { 0.14828933775425, -0.407556354999542, -0.107610583305359, -0.0416980870068073, -1.54811465740204, 0.0840534418821335, 0.0317929796874523, -0.0585002191364765, -0.890501618385315, -0.53799170255661, -0.960572063922882, -0.105323895812035, -0.161284476518631, 0.0448649376630783, -0.226359888911247, -1.50261747837067, -0.0346848256886005, -1.38759768009186, -0.32634648680687, 0.0320334285497665, -0.19293050467968, 0.0101544447243214, -0.170103833079338, -0.121908076107502, -1.35497140884399, -0.035065270960331, 0.473181545734406, -0.00905308499932289, 0.355516761541367, -0.337357699871063, -0.990931034088135, -0.0206653233617544, 0.108536757528782, -0.160865113139153, -2.09992814064026, -0.201994121074677, -0.0164962708950043, -0.064218707382679, -0.309794694185257, -0.0916733220219612, -1.97365784645081, -0.00321335718035698, -1.72357261180878, -1.64564263820648, -0.10867466032505, 0.0395487509667873, -0.876764476299286, -0.00902560073882341, 0.00441655050963163, 0.0177029352635145, -0.0796377435326576 },
      { -0.459590256214142, -0.321553319692612, -0.165726497769356, -0.129799500107765, 1.88014054298401, -0.533481419086456, -0.306849122047424, -0.155334398150444, -0.764747738838196, 0.46194651722908, 0.185087978839874, 2.74550127983093, -0.0457086190581322, -0.557302296161652, 0.0232702847570181, 2.23657917976379, -0.618429362773895, -0.573885321617126, 0.066753938794136, -0.236483544111252, 0.0153815494850278, 0.0744702592492104, 0.285579591989517, 0.0225861929357052, 3.15725231170654, 0.176760077476501, -0.692496538162231, -0.660035490989685, -0.13306075334549, -0.0348016582429409, 0.251199573278427, -0.268555134534836, -0.380900770425797, 0.0320925749838352, 1.61519157886505, 0.107239119708538, -0.192875906825066, 0.472472190856934, 0.22723314166069, -0.168320000171661, 1.85534596443176, -0.101921737194061, 1.93808138370514, 0.450171232223511, -0.0417807400226593, -0.290132164955139, 1.41100370883942, -0.18011049926281, -0.111171968281269, -0.314310103654861, -0.349212259054184 },
      { 1.07751095294952, 1.2359185218811, 1.87029147148132, 0.953579485416412, -1.03408432006836, 1.27382588386536, 1.06023478507996, 0.949672400951385, 1.42358577251434, 0.574566304683685, 1.48659932613373, -1.89695751667023, 0.900152266025543, 1.90622770786285, 0.802475035190582, -0.0632779374718666, 1.49054622650146, 1.28732657432556, 0.947374403476715, 0.973397374153137, 0.8485187292099, 0.721818208694458, 0.780536234378815, 0.784122228622437, -0.634128332138062, 0.606720566749573, 1.25937187671661, 1.53427803516388, 0.540732443332672, 0.95116126537323, 1.21892654895782, 0.934442460536957, 1.50715494155884, 0.782413244247437, 1.13211679458618, 0.860771775245667, 0.988148391246796, 1.23211252689362, 0.849353015422821, 0.94986766576767, -0.409130960702896, 0.929065585136414, 0.367754518985748, 0.841745555400848, 0.942933857440948, 1.03318881988525, -0.221455350518227, 0.981223464012146, 0.961263716220856, 1.13252639770508, 0.985012412071228 },
      { 0.166964933276176, 0.184074923396111, 0.37626650929451, 0.211117625236511, 0.0985723584890366, 0.266191601753235, 0.211764335632324, 0.264081120491028, 0.382070094347, 0.202612981200218, 0.156254321336746, -0.340616375207901, 0.260022699832916, -0.0117110926657915, 0.287366181612015, -0.0391391217708588, 0.584915637969971, 0.12628062069416, 0.235477656126022, 0.23469477891922, 0.258047431707382, 0.230954706668854, 0.175162106752396, 0.312281578779221, -0.24102458357811, 0.272997349500656, 0.17634429037571, 0.430673032999039, 0.258456259965897, 0.276510834693909, 0.142050251364708, 0.381856799125671, 0.432703852653503, 0.301054745912552, 0.0501189306378365, 0.251552581787109, 0.221627458930016, 0.0228645261377096, 0.122511520981789, 0.268544644117355, -0.00464295269921422, 0.18162414431572, 0.0554580390453339, 0.200733780860901, 0.146525084972382, 0.290597587823868, 0.162458628416061, 0.226266711950302, 0.240018621087074, 0.153209209442139, 0.22817662358284 },
      { -0.00326760718598962, -0.0634662210941315, -1.32503414154053, -0.00418394571170211, -0.43024817109108, -0.169132754206657, -0.00676356395706534, -0.0945148468017578, -0.021756537258625, -0.037135548889637, 0.0789150893688202, 0.40787610411644, -0.00298882741481066, 0.0249522663652897, -0.0208404846489429, 0.162420600652695, -0.884337484836578, 0.401295393705368, -0.19109696149826, -0.0131678646430373, -0.0117850294336677, -0.0475214794278145, 0.0719550549983978, -0.0619989037513733, 0.468909382820129, 0.0197428558021784, 0.0371264591813087, -0.61391270160675, 0.0192282237112522, -0.00615654978901148, 0.179026633501053, -0.149531841278076, -1.16225111484528, -0.0281688421964645, 0.0568542890250683, -0.044111005961895, -0.000774362531956285, -0.798508942127228, -0.0133344326168299, -0.0179582592099905, -0.356655061244965, 0.0236831493675709, -0.0684968233108521, -0.21741546690464, 0.0979797467589378, -0.0962378531694412, -0.919012844562531, -0.0158900562673807, -0.0569755248725414, -0.0131468093022704, -0.159537672996521 },
      { 0.017036359757185, 0.184869453310966, 0.159318178892136, -0.00230030226521194, 1.03278791904449, 0.0704460591077805, 0.0139340525493026, 0.0838440954685211, 0.113269604742527, 0.236549571156502, -0.249051198363304, 0.676563084125519, -0.017608318477869, 0.0701303109526634, 0.0168933682143688, -0.201211169362068, 0.473661690950394, 0.188711389899254, 0.256234049797058, 0.0193284545093775, 0.0028526175301522, 0.0440819934010506, -0.125162586569786, 0.0302879977971315, -0.551301121711731, -0.016632005572319, -0.0595428049564362, 0.335944354534149, -0.00632246909663081, -0.0188677962869406, -0.239542216062546, 0.108458623290062, 0.411414176225662, 0.008921611122787, -0.121147759258747, 0.0111631276085973, -0.00927915330976248, 0.950705230236053, 0.0654488429427147, 0.0319827646017075, 1.20906794071198, -0.0264238901436329, 0.0252351518720388, 0.585314333438873, -0.0361894443631172, 0.0120825245976448, 1.8155871629715, 0.00608486402779818, 0.00285657006315887, 0.00784115958958864, 0.286246925592422 },
      { 25.1535129547119, 60.5786361694336, 267.356079101562, -9.66460227966309, 479.208465576172, 35.1709899902344, 6.12946462631226, -0.35592994093895, 74.3677215576172, 23.5729465484619, 157.474884033203, 409.236450195312, 1.65480375289917, -571.003112792969, 27.6158103942871, 337.037658691406, -37.1921539306641, 148.226943969727, 39.8328590393066, -6.71125888824463, 19.2231597900391, -178.552032470703, 47.1318778991699, 16.7647743225098, 448.456420898438, -49.3198051452637, 32.9757766723633, -43.8171424865723, 22.7437705993652, 40.0647087097168, 118.901710510254, -1.05174124240875, 127.253242492676, 12.3972835540771, -9.03925895690918, -16.694938659668, 3.87681555747986, -700.209167480469, 42.2278137207031, -11.0269622802734, 519.335266113281, -0.353335946798325, 271.513336181641, 236.124160766602, 11.6794004440308, 60.8029823303223, 164.893524169922, -24.3629474639893, -11.6628513336182, 83.3883209228516, 71.7358932495117 }
    },
    {
      { 0.301870763301849, -0.348994970321655, 0.00657449895516038, -0.345914661884308, -0.378002673387527, 0.208042770624161, -0.118705846369267, -0.202120721340179, -0.213787853717804, -0.466002970933914, 0.00288659729994833, 2.10348701477051, -0.696021556854248, 0.0831292346119881, -0.555653870105743, 0.0817680284380913, 0.0961734279990196, 0.105929620563984, -0.179584860801697, -0.060399204492569, -0.475348830223083, 0.0976096987724304, -0.535606622695923, -0.398710042238235, 0.737593352794647, -0.0838938355445862, 0.787450194358826, 0.140712663531303, -0.437059998512268, -0.0907569229602814, -0.445703119039536, -0.15576608479023, 0.00257836771197617, -0.610520660877228, -0.702902853488922, -0.218362286686897, -0.0940189436078072, -0.367382884025574, -0.701472103595734, -0.199127063155174, 1.80711448192596, -0.149484664201736, -1.60226774215698, -0.149862080812454, -0.251467913389206, -0.159580990672112, -0.529300272464752, -0.0624779611825943, -0.0633224546909332, -0.0855600014328957, 0.067940317094326 },
      { -0.885970592498779, 0.164809182286263, -0.121893130242825, 0.109855569899082, 2.92936539649963, -0.510999262332916, 0.0136873982846737, 0.273298591375351, 0.331775397062302, 0.716775238513947, 1.49372851848602, 4.02839612960815, 0.101701758801937, -1.22360384464264, 0.469806492328644, 1.62480282783508, -0.464754074811935, 0.902287542819977, 0.347018331289291, 0.0299401916563511, 0.504496455192566, 0.0314236208796501, 1.42207050323486, 0.470299601554871, 6.08357381820679, 0.301197022199631, -0.698961555957794, -0.44877415895462, 1.81424415111542, 0.739668667316437, 1.79829835891724, -0.132398173213005, -0.104680418968201, 0.528024971485138, 2.19438982009888, 0.704196631908417, -0.203087851405144, 0.828179955482483, 1.35077559947968, -0.0559379011392593, 3.44326400756836, 0.275410950183868, 2.38152265548706, 1.31980538368225, 0.46268755197525, 0.370808213949203, 1.94460356235504, -0.109749361872673, 0.102883651852608, 0.111545540392399, -0.170896291732788 },
      { 0.970759212970734, 0.275733143091202, 1.12031292915344, 0.543973684310913, -2.36632680892944, 0.598577618598938, 0.395413160324097, 0.22933454811573, -0.367176204919815, -0.161532655358315, -1.22305154800415, -4.76309633255005, 0.730746567249298, 2.88127326965332, 0.122486248612404, -1.17121505737305, 0.669324159622192, -0.595022737979889, 0.118663817644119, 0.369523674249649, 0.207296907901764, 0.141637444496155, -0.260058522224426, 0.191897347569466, -3.5481321811676, 0.0414086654782295, 0.584908187389374, 0.627754390239716, -0.714446663856506, -0.161536201834679, -0.985355138778687, 0.452955037355423, 0.940128207206726, 0.186787039041519, -0.997057557106018, 0.16617988049984, 0.78932136297226, 1.78932893276215, -0.0634492859244347, 0.509256362915039, -3.58752799034119, 0.37345689535141, -0.680975079536438, -1.09554648399353, 0.0977175310254097, 0.161851823329926, -0.86238706111908, 0.644341886043549, 0.779251873493195, 0.147725015878677, 0.337505131959915 },
      { 0.332872003316879, 0.778731226921082, 0.807363092899323, 0.590571820735931, 0.572959721088409, 0.738529860973358, 0.693455815315247, 0.624551057815552, 0.853298664093018, 0.613063097000122, 0.72292172908783, 0.0116405300796032, 0.64582896232605, -0.00359579175710678, 0.753834843635559, 0.735758304595947, 1.01340746879578, 0.632421255111694, 0.666901767253876, 0.747156023979187, 0.695773899555206, 0.766249418258667, 0.607264578342438, 0.739388763904572, 0.0395653657615185, 0.766022622585297, 0.865547657012939, 0.871623337268829, 0.738431572914124, 0.584511399269104, 0.60247802734375, 0.923052251338959, 0.704340159893036, 0.747474491596222, 0.646687030792236, 0.539442241191864, 0.506905138492584, 0.0793223008513451, 0.377563506364822, 0.720489978790283, 0.348400115966797, 0.494665861129761, 0.612624883651733, 0.697145819664001, 0.637147307395935, 0.795366585254669, 0.60667496919632, 0.612069368362427, 0.520928502082825, 0.795753121376038, 0.880286395549774 },
      { 0.0270163342356682, -0.177974835038185, -1.09095740318298, -0.0545219555497169, 0.0773288160562515, -0.0809674859046936, 0.0337062776088715, -0.124939881265163, -0.101302854716778, 0.246724799275398, -0.0385561659932137, 1.54124212265015, -0.0525848194956779, 0.364881068468094, -0.187539920210838, -0.0121191162616014, -0.696090698242188, -0.104670539498329, 0.24299481511116, -0.104653462767601, -0.187587097287178, 0.0134217832237482, -0.18533219397068, -0.219081923365593, 1.52200412750244, -0.104603849351406, -0.165728375315666, -0.434562355279922, -0.22362032532692, 0.0460067577660084, -0.136543139815331, -0.286577522754669, -0.969906985759735, -0.111528068780899, 0.0694787576794624, -0.187593638896942, 0.00825821235775948, -1.2314966917038, -0.026107108220458, -0.0661988779902458, 0.693577408790588, 0.0626545995473862, 0.00245290342718363, 0.0840003490447998, 0.101183325052261, -0.225313976407051, 0.142954990267754, -0.0587398521602154, -0.190572679042816, -0.0368520133197308, -0.260750591754913 },
      { 0.0486633591353893, 0.217511102557182, 0.0779778584837914, 0.0262392237782478, -0.0450845956802368, 0.0119132837280631, -0.00921151135116816, 0.161733940243721, 0.215835824608803, -0.0602337643504143, 0.142156317830086, -1.0612815618515, -0.047780841588974, -0.227066487073898, 0.194572895765305, 0.0267985593527555, 0.374469608068466, 0.350725740194321, -0.312217593193054, 0.0347618982195854, 0.151199787855148, 0.00496244244277477, 0.124335817992687, 0.161035493016243, -3.05168437957764, 0.111791491508484, 0.126113384962082, 0.234886839985847, 0.182800427079201, -0.0332928188145161, 0.340534627437592, 0.166789814829826, 0.337672203779221, 0.0577984489500523, -0.162679269909859, 0.115372225642204, -0.0317422486841679, 1.37905812263489, 0.134025275707245, 0.0454656071960926, -0.634302079677582, -0.0746456831693649, -0.00399568304419518, 0.146114900708199, -0.0771344527602196, 0.0690802782773972, -0.344125270843506, -0.00683134561404586, -0.0138348918408155, 0.109159961342812, 0.162440299987793 },
      { 78.501823425293, -42.0774955749512, 284.538757324219, 12.1297693252563, -44.1862106323242, 47.4254951477051, -12.9638528823853, 29.8264923095703, -85.0844802856445, 33.9036254882812, -196.164642333984, -87.2196350097656, 90.6959533691406, -947.170471191406, 43.9780578613281, -368.317291259766, -54.6263160705566, -27.5369491577148, 67.9164352416992, -16.011833190918, 53.4658508300781, -139.290161132812, 48.1012954711914, 39.775146484375, 73.8028869628906, -114.353782653809, -217.557174682617, -55.900333404541, -200.776428222656, 19.7862758636475, -63.8443183898926, 2.27864146232605, 156.932662963867, 47.3132438659668, -109.914611816406, 1.3769599199295, 15.7391176223755, -1102.45422363281, -17.7568054199219, -12.190089225769, -354.714447021484, -27.3729934692383, -171.41015625, -266.6796875, 43.1927528381348, 54.198127746582, -49.3727836608887, -55.312557220459, 9.56275844573975, -93.6128311157227, -21.6781787872314 }
    },
    {
      { 0.332354992628098, 0.145584717392921, 0.138906925916672, -0.3503098487854, 0.0883655026555061, 0.157281741499901, -0.160240650177002, -0.187884613871574, 0.50248908996582, -0.0637032836675644, 1.10733234882355, -0.688989877700806, -0.656568169593811, 0.563827514648438, -0.314543724060059, 1.40728294849396, 0.121006660163403, 1.39532005786896, 0.266435235738754, -0.060209859162569, -0.281971484422684, 0.00121751963160932, -0.554257988929749, -0.315298616886139, -0.345645010471344, 0.0919201523065567, 0.413650333881378, 0.237718418240547, -0.679864525794983, 0.194703757762909, 0.640539586544037, -0.0844198316335678, -0.115896105766296, -0.543855547904968, 1.34944868087769, -0.13361069560051, -0.0998966991901398, 0.0801792964339256, -0.483284831047058, -0.149871557950974, 0.28382819890976, -0.160853430628777, 0.223552539944649, 0.551298975944519, -0.129113048315048, -0.0688736215233803, -0.333135485649109, -0.0599848441779613, -0.123230405151844, 0.1097317263484, 0.432659417390823 },
      { -0.820486545562744, -0.320156365633011, -0.046765822917223, 0.150991469621658, 1.07289552688599, -0.218628376722336, 0.289912432432175, 0.173878312110901, -0.157400980591774, 0.313445389270782, 0.984801888465881, 0.596819221973419, 0.125003680586815, -0.514390528202057, 0.228961065411568, 0.276651740074158, -0.0615440793335438, 0.299014449119568, -0.0745433568954468, 0.0751608163118362, 0.404083430767059, -0.00644182879477739, 1.2901930809021, 0.310387283563614, 0.109771654009819, -0.185355708003044, 0.412721693515778, -0.226886078715324, 2.14615225791931, 0.667700946331024, 1.34193074703217, -0.163666486740112, 0.231049001216888, 0.418225705623627, 0.468966275453568, 0.513555884361267, -0.1965601593256, -0.168182820081711, 0.707230448722839, -0.0949455574154854, 0.260104715824127, 0.244519829750061, 0.219915643334389, 0.425691485404968, 0.314593076705933, 0.211995646357536, 0.358389586210251, -0.0875042751431465, 0.118488624691963, 0.368713468313217, -0.665253877639771 },
      { 0.665226101875305, 0.331339865922928, 0.1491589397192, 0.34182807803154, -0.653678417205811, 0.163551077246666, -0.0514994561672211, 0.151902884244919, -0.533224105834961, -0.137574911117554, -1.9865859746933, -0.393734604120255, 0.577649056911469, -0.0543728098273277, 0.0613114610314369, -1.34447503089905, 0.0218283757567406, -0.36826229095459, 0.0951651036739349, 0.147371366620064, 0.164160713553429, 0.0787425115704536, -0.360580295324326, 0.132503122091293, -0.114881113171577, 0.133353367447853, -0.274801194667816, 0.0817669108510017, -0.865914404392242, -0.317161709070206, -1.51850771903992, 0.31337183713913, 0.176172807812691, 0.118315301835537, -1.50753474235535, 0.0994468405842781, 0.538996875286102, 0.77543830871582, -0.0465976446866989, 0.377857714891434, -1.02412605285645, 0.257040083408356, -0.279935598373413, -0.805148303508759, -0.0624263770878315, 0.13875338435173, 0.36751464009285, 0.43021959066391, 0.586318075656891, -0.369276612997055, 0.51091605424881 },
      { 0.574598729610443, 0.96608555316925, 0.912254393100739, 0.74710351228714, 0.884285986423492, 0.926128804683685, 0.93699449300766, 0.790620446205139, 1.10251557826996, 0.783027231693268, 1.0100291967392, 0.978703200817108, 0.749530076980591, 0.93796443939209, 0.890724897384644, 1.08616971969604, 0.978808462619781, 0.914555072784424, 0.802097320556641, 0.940381646156311, 0.801483452320099, 0.954885542392731, 0.810665607452393, 0.903487861156464, 1.08935523033142, 1.0009468793869, 0.675434589385986, 0.953036069869995, 0.928087651729584, 0.739293873310089, 0.890637457370758, 1.0277510881424, 0.889777362346649, 0.890260756015778, 0.927079260349274, 0.708184778690338, 0.710786879062653, 0.827274143695831, 0.644481122493744, 0.87111896276474, 0.937571346759796, 0.638094365596771, 1.01721823215485, 1.00031113624573, 0.868962287902832, 0.935971140861511, 0.923987448215485, 0.80166107416153, 0.696993768215179, 1.04531216621399, 1.07368767261505 },
      { 0.0323428399860859, -0.140966132283211, -0.208147540688515, -0.0341565534472466, 0.477779299020767, -0.0529656819999218, 0.0311381462961435, -0.109378859400749, -0.169531181454659, 0.379455238580704, -0.128865420818329, 0.125805497169495, -0.0134174739941955, -0.125555649399757, -0.226285681128502, -0.25595024228096, -0.148814037442207, -0.408872276544571, 0.576886832714081, -0.0944924205541611, -0.159855529665947, 0.0547292977571487, -0.21810220181942, -0.22539584338665, -0.522456109523773, -0.223071366548538, -0.074362687766552, -0.0773955807089806, -0.294196784496307, 0.019522013142705, -0.421646267175674, -0.237441062927246, -0.32706218957901, -0.0918751433491707, 0.147816672921181, -0.170492857694626, 0.0135126272216439, -0.74855774641037, 0.113858044147491, -0.0582191646099091, 0.0689251199364662, 0.0650689229369164, 0.0336121208965778, -0.0404258593916893, 0.0420317761600018, -0.199996590614319, 1.05170667171478, -0.0283300299197435, -0.165668517351151, -0.11470302939415, -0.158556014299393 },
      { 0.0477397292852402, 0.0808431804180145, 0.0187936015427113, 0.0160852503031492, -1.08876883983612, 0.0102612981572747, -0.0403199829161167, 0.139620810747147, 0.179080873727798, -0.325622230768204, 0.372736155986786, -0.45137631893158, -0.069733552634716, 0.530371844768524, 0.259723037481308, 0.285146504640579, 0.0905894786119461, 0.211428374052048, -0.786394774913788, 0.00342690269462764, 0.0912731811404228, -0.0446161217987537, 0.172587767243385, 0.158997356891632, 0.559979736804962, 0.230748414993286, 0.05162363499403, 0.0304914470762014, 0.24121356010437, -0.0437472686171532, 0.673373222351074, 0.129788488149643, 0.12275618314743, 0.0305516347289085, -0.20871764421463, 0.103886634111404, -0.0198352225124836, 1.32231152057648, -0.033822525292635, 0.0218716356903315, 0.0974152237176895, -0.0665449053049088, -0.00920839700847864, 0.107602812349796, -0.0476100295782089, 0.0270278230309486, -2.4177839756012, -0.0489652417600155, -0.0158012975007296, 0.185180127620697, -0.0817449390888214 },
      { 40.3806762695312, -27.0716438293457, -1.55672490596771, 25.3509979248047, -151.865081787109, 5.29329299926758, -23.297700881958, 28.2565364837646, -94.2181243896484, 32.6883811950684, -324.934143066406, -59.1773796081543, 98.0142593383789, -197.390533447266, 22.1382961273193, -334.545959472656, -57.5321502685547, -76.4410552978516, 76.8606414794922, -17.5921764373779, 43.7059669494629, 14.2580375671387, 43.3166580200195, 44.2763748168945, -95.9114532470703, -98.6729049682617, 358.812927246094, -93.0017318725586, -263.966156005859, 7.69864797592163, -74.6332015991211, 1.53197932243347, 42.2330551147461, 52.1527290344238, -63.4091682434082, 2.07615852355957, 9.89878177642822, -824.665283203125, 112.599960327148, -6.15652704238892, -71.948600769043, -13.4360675811768, -495.423095703125, -323.292419433594, 18.4450855255127, 56.5198173522949, -73.5109786987305, -35.1927947998047, 21.9911556243896, -315.257202148438, -28.6230354309082 }
    },
    {
      { 0.0838291123509407, -0.0828287526965141, -0.0601919256150723, -0.152656316757202, -0.16227325797081, 0.0564756207168102, -0.0250136367976665, -0.0917002186179161, -0.0401540286839008, -0.0832405015826225, 0.153177082538605, -0.582172751426697, -0.261612266302109, 0.233537301421165, -0.221054211258888, 0.230257958173752, -0.00724746566265821, 0.216813832521439, 0.00598519714549184, -0.0148697458207607, -0.201659888029099, -0.000558706989977509, -0.222696244716644, -0.132971748709679, -0.221414342522621, -0.0189503021538258, -0.113297462463379, -0.034580796957016, -0.013941259123385, -0.044800691306591, -0.124807715415955, -0.032276276499033, 0.121308147907257, -0.216762319207191, 0.488597333431244, -0.089057944715023, -0.0579785481095314, 0.198478192090988, -0.352467149496078, -0.0863739028573036, -0.00290617067366838, -0.0704626441001892, 0.162031799554825, -0.0123305395245552, -0.0303575657308102, 0.0257257707417011, -0.143297478556633, -0.0272016767412424, -0.0654704868793488, -0.00244998396374285, 0.102294497191906 },
      { -0.236135646700859, -0.0449830368161201, -0.0257612559944391, 0.0728975683450699, 0.542208909988403, -0.114583089947701, 0.0407259464263916, 0.0486805737018585, 0.0387106277048588, 0.170785248279572, 0.388737112283707, 0.165161371231079, 0.0461321473121643, -0.269833028316498, 0.127514913678169, 0.32423323392868, -0.392164945602417, 0.0796743035316467, 0.0787506476044655, 0.00629376713186502, 0.164341047406197, 0.0100526856258512, 0.496814846992493, 0.106893293559551, -0.157568067312241, 0.115302592515945, 0.168937563896179, -0.11314033716917, 0.515730321407318, 0.18910063803196, 0.59093964099884, -0.0547696352005005, -0.203118234872818, 0.169050484895706, 0.16194948554039, 0.178553357720375, -0.0613792836666107, -0.506181240081787, 0.508165895938873, -0.00809410121291876, 0.0911374539136887, 0.0904999971389771, 0.0165617056190968, 0.185800895094872, 0.0758463218808174, 0.0570066347718239, 0.451523184776306, -0.0325512550771236, 0.00256156967952847, 0.0707971304655075, -0.170335188508034 },
      { 0.227103695273399, 0.0293496958911419, 0.84741997718811, 0.150169774889946, -0.560610830783844, 0.10877450555563, 0.0403434298932552, 0.0989709571003914, -0.147679418325424, -0.0912571027874947, -0.4909827709198, 0.196312084794044, 0.243815258145332, 0.0158943645656109, 0.0531293042004108, -0.480288505554199, 0.479813426733017, -0.0555402860045433, -0.0155915133655071, 0.0660298839211464, 0.0874513983726501, 0.0285729318857193, -0.109779380261898, 0.0710333958268166, 0.13017150759697, -0.0394724979996681, -0.0882948115468025, 0.230553984642029, -0.313940078020096, -0.0583435222506523, -0.403167486190796, 0.0912071987986565, 0.546986043453217, 0.0634408518671989, -0.548279464244843, 0.0581583566963673, 0.207625404000282, 0.24439549446106, -0.011653327383101, 0.127006381750107, -0.256479889154434, 0.130154713988304, -0.106017656624317, -0.197164922952652, 0.00479372544214129, 0.00108741875737906, -0.175655201077461, 0.164177015423775, 0.259071171283722, -0.0503650605678558, 0.0667019262909889 },
      { 0.851817786693573, 1.00658226013184, 1.0591539144516, 0.892108380794525, 0.962416410446167, 0.978557527065277, 0.941121637821198, 0.914778590202332, 1.0196281671524, 0.935466945171356, 0.991150557994843, 1.02224540710449, 0.891942203044891, 0.97961562871933, 0.931027173995972, 1.00938403606415, 1.25379419326782, 0.96870481967926, 0.952056586742401, 0.967529773712158, 0.914283812046051, 0.961955368518829, 0.923279821872711, 0.962017714977264, 1.04350757598877, 0.965548813343048, 0.92578649520874, 1.12286698818207, 0.965315878391266, 0.914205610752106, 0.940423369407654, 1.022087931633, 1.0639955997467, 0.952212989330292, 0.969836711883545, 0.896435260772705, 0.880467355251312, 1.01163375377655, 0.801087200641632, 0.94726026058197, 0.977739453315735, 0.835519433021545, 1.02169263362885, 1.01587283611298, 0.949327051639557, 0.980083465576172, 0.976244390010834, 0.912190616130829, 0.870914161205292, 0.989743888378143, 1.0228271484375 },
      { -0.0254156906157732, -0.095105305314064, -1.09348046779633, -0.0142422635108232, 0.0229055341333151, -0.047779954969883, 0.0128371110185981, -0.0455444790422916, -0.0856403186917305, 0.064238153398037, -0.0748321786522865, 0.00198412500321865, -0.00265863048844039, -0.0669879764318466, -0.0437351912260056, -0.148171812295914, -0.684539794921875, -0.170621022582054, 0.0697926431894302, -0.0331740006804466, -0.0435432121157646, 0.0185824520885944, -0.0927341505885124, -0.0858000889420509, -0.144474565982819, -0.0234729405492544, -0.0332306437194347, -0.436579644680023, -0.0805822089314461, 0.0119589241221547, -0.150940239429474, -0.0988219901919365, -0.920728147029877, -0.0414294861257076, 0.0431918390095234, -0.0503789484500885, 0.0027387070003897, -0.479871600866318, -0.0251227244734764, -0.0194748174399137, -0.00931491144001484, 0.0368435606360435, 0.00704328482970595, -0.122964724898338, 0.0114983404055238, -0.0704855024814606, -0.0250439513474703, 0.00483282003551722, -0.0561088994145393, -0.0187160391360521, -0.0837112367153168 },
      { 0.0498711839318275, 0.118323959410191, 0.0918594151735306, 0.00206237309612334, -0.00973828881978989, 0.0110991857945919, -0.00724660791456699, 0.0479826405644417, 0.126539975404739, -0.0207025371491909, 0.107899375259876, -0.224790975451469, -0.0312359090894461, 0.225621849298477, 0.0550836995244026, 0.18851138651371, 0.355693250894547, 0.204217091202736, -0.117919974029064, 0.0110479230061173, 0.0264431275427341, -0.0152725810185075, 0.0584129206836224, 0.0564277172088623, 0.174272775650024, 0.0186180826276541, 0.00531726283952594, 0.239062413573265, 0.0766077637672424, -0.00928331259638071, 0.237492188811302, 0.0614296533167362, 0.323990285396576, 0.0104637201875448, -0.0508569106459618, 0.0254890583455563, -0.000918595120310783, 0.837993919849396, 0.072711206972599, 0.0163467023521662, 0.136184826493263, -0.0326241627335548, -0.00349630927667022, 0.16833969950676, -0.0180806890130043, 0.0238429494202137, -0.136313065886497, -0.0274698585271835, 0.00360886030830443, 0.039593331515789, 0.0597976632416248 },
      { 8.43302059173584, 1.83613932132721, 298.074279785156, 18.1294479370117, -26.847204208374, 5.1825532913208, -2.19239187240601, 18.8857402801514, -20.4108543395996, 6.46369647979736, -80.6891021728516, -17.0440311431885, 45.2215881347656, -75.1558380126953, 31.0299110412598, -91.5952072143555, -21.486156463623, -14.9130773544312, 11.2668991088867, -6.34212112426758, 33.5059127807617, -7.83033752441406, 31.4609489440918, 23.7120895385742, -49.2045516967773, -37.0843620300293, 387.267211914062, -39.2402839660645, -72.0948333740234, 18.5659217834473, -4.76370859146118, 2.79405617713928, 99.4197845458984, 27.2741355895996, -25.1457748413086, 3.85003781318665, 6.74954509735107, -290.577423095703, 1.84620308876038, 0.119689464569092, -53.3453750610352, -7.4045877456665, -199.557983398438, -99.4065704345703, 9.79404830932617, 15.3889465332031, -23.9719066619873, -6.40972328186035, 9.38085269927979, -51.8125267028809, -8.49858093261719 }
    },
    {
      { 0.000719039351679385, 0.00713090412318707, -0.00471930671483278, -0.000213386359973811, 0.0137438010424376, 0.00135234359186143, -0.0018567475490272, -0.000659259967505932, 0.00617020670324564, 0.00658245291560888, 0.00755935488268733, 0.0322379097342491, 5.94901393924374e-05, -0.0122325969859958, 0.00263680168427527, 0.0111250588670373, 0.00550108635798097, 0.00348271708935499, 0.00513527262955904, -0.000861899752635509, 0.00438941828906536, -0.000425791513407603, 0.00627391505986452, 0.000498885579872876, 0.0193623211234808, 0.0120074991136789, 0.00844047777354717, 0.00842792075127363, 0.0107204234227538, 0.0066266255453229, 0.0143801653757691, 0.00173061189707369, -0.00379702541977167, 0.000903083244338632, -0.00713778613135219, 0.00162955385167152, -0.00372385722585022, -0.00496697425842285, 0.00819207821041346, 0.00236906693316996, 0.00226387614384294, 0.000654106435831636, 0.00666289124637842, 0.00458278134465218, 0.00339376064948738, 0.00272241840139031, 0.00997177325189114, -0.000953545153606683, 0.000791803642641753, 0.000577789847739041, 0.00132955517619848 },
      { 0.000197469969862141, -0.000124256504932418, -0.00127501250244677, 0.000320180872222409, -0.0174857117235661, 0.00232641980983317, 0.00243842019699514, 0.00157323305029422, -0.00697058020159602, -0.0049101198092103, -0.00963267125189304, -0.0107630509883165, 0.00253677694126964, 0.0160632133483887, 0.00253564305603504, -0.0149183580651879, 0.0131402779370546, -0.002617058577016, -0.00435872189700603, 0.00104357802774757, 0.00152688310481608, -0.000222670394578017, -0.00813850667327642, 0.00516384094953537, 0.00132124021183699, -0.0125995706766844, 0.000561381049919873, 0.000582486973144114, -0.00897801853716373, -0.0020052595064044, -0.0159480813890696, 0.00246791425161064, 0.0132684037089348, 0.000700021046213806, -0.00343367946334183, -0.000746364006772637, 0.00328584620729089, 0.0155154196545482, -0.0127558503299952, 0.00163148168940097, -0.0158166978508234, 0.00491347489878535, -0.00406536413356662, -0.004017842002213, -0.00178299704566598, -0.00385644682683051, -0.01607845723629, 0.00142962008249015, 0.00567289814352989, 0.00126524129882455, 0.00107368908356875 },
      { -0.00214618048630655, 0.000337845965987071, -0.0409685559570789, -0.00409492244943976, 0.0170423742383718, -0.00499316398054361, -0.00352530926465988, -0.000556988117750734, 0.00321875070221722, 0.00293345865793526, 0.00399685930460691, -0.0144253987818956, -0.00512586114928126, -0.00491597503423691, -0.0045046336017549, 0.00740231573581696, -0.0227556396275759, -0.000222023532842286, -0.000243321090238169, -0.000343171181157231, -0.00438743596896529, 0.000574595062062144, 0.000536584586370736, -0.00264485366642475, -0.0113654769957066, 0.000884730601683259, -0.00117180030792952, -0.0127816190943122, -0.0015230622375384, 0.00198801304213703, 0.00316701177507639, -0.000761214236263186, -0.032343864440918, -0.00348609243519604, 0.00884483475238085, -0.00227366480976343, -0.00295078777708113, -0.00618929695338011, 0.000830172211863101, -0.00341190421022475, 0.00428851135075092, -0.00983158312737942, 0.000113228343252558, 0.0010880483314395, -0.00374321010895073, 0.00183364155236632, 0.00865055155009031, -0.00129922782070935, -0.00470663653686643, -0.00219264184124768, 0.00222234800457954 },
      { 1.0021960735321, 1.00104117393494, 0.999957919120789, 1.00223875045776, 1.00130331516266, 1.00152623653412, 1.00418961048126, 1.00196099281311, 1.00078570842743, 1.00106167793274, 1.00097370147705, 1.00209999084473, 1.00341272354126, 1.00334107875824, 1.00337302684784, 1.00205647945404, 0.988124966621399, 1.00170564651489, 1.0004004240036, 1.00158727169037, 1.00120890140533, 1.0010267496109, 1.0009548664093, 0.999697208404541, 1.00193989276886, 1.00474750995636, 0.998425483703613, 0.995425045490265, 1.00156533718109, 1.00231468677521, 1.00443744659424, 0.997124433517456, 1.00279223918915, 1.00117611885071, 1.00132501125336, 1.00373601913452, 1.00176894664764, 0.99952757358551, 1.00316762924194, 1.00180816650391, 1.00369036197662, 1.00533425807953, 0.999552190303802, 0.999772965908051, 1.00142669677734, 1.0012195110321, 1.00108456611633, 1.00190496444702, 1.0013040304184, 1.00119698047638, 0.999665975570679 },
      { 0.000513123406562954, 0.00197085342369974, 0.0583550781011581, -0.000731238687876612, 7.81917333370075e-06, -0.000983847072347999, -0.00197104271501303, -0.000628848385531455, 0.00182196812238544, -0.000239188317209482, 0.00143771979492158, -0.0135043011978269, -5.99704580963589e-05, 0.0100496858358383, -0.00133793987333775, 0.000863224151544273, 0.0317681469023228, 0.00192384875845164, 0.00161899405065924, -0.000899872393347323, -0.000374625466065481, -0.000133917230414227, 0.00205833464860916, 0.000478420028230175, -0.0210020020604134, -0.00736627727746964, -0.00147782161366194, 0.0186522547155619, -0.000324152671964839, -0.00362640130333602, 0.000515472900588065, 0.00275977072305977, 0.0346364378929138, 0.000515926105435938, -0.00163199752569199, -0.00171005947049707, -0.000258152693277225, 0.037369042634964, 0.00275523960590363, 0.000233673214097507, -0.00953406468033791, -0.000833386729937047, 0.000365129206329584, 0.00384173658676445, -0.000203429211978801, 0.000489025725983083, 0.00827497430145741, -0.00179924874100834, 2.06380154850194e-05, -1.95598804566544e-05, 0.00141192635055631 },
      { -0.000990285188890994, -0.00602459767833352, 0.00156226789113134, 0.00195587147027254, -0.00268749054521322, 0.00154438137542456, 0.000237285581533797, -0.00059592048637569, -0.00268011284060776, -0.00348641932941973, -0.00170522811822593, 0.0240340195596218, -0.000419024610891938, -0.0133942337706685, 0.000530919001903385, -0.00513851130381227, -0.0148904183879495, -0.00589505257084966, -0.00130373693536967, -0.000562800094485283, 0.000596990343183279, -0.000375979638192803, -0.000513813691213727, -0.000738849223125726, 0.0332411974668503, 0.00600055558606982, 0.00121392903383821, -0.0101990764960647, 0.000973691523540765, 0.000324646855005994, -0.00546161737293005, -0.00224076164886355, -0.0117494929581881, 0.000409309781389311, 0.000475727894809097, 0.000231629994232208, 0.000281356129562482, -0.0618638135492802, -0.00377398426644504, -0.00156063993927091, 0.0090753436088562, 0.000833452038932592, 0.000447935046395287, -0.00411531422287226, 0.00152723153587431, -0.00140745320823044, -0.0149211883544922, 0.000725653662811965, -0.00128799374215305, 0.000348468369338661, -0.00308309379033744 },
      { -0.42597696185112, -1.18248963356018, -18.8335971832275, -1.82488977909088, -1.6999089717865, -1.23847877979279, 0.237816646695137, -1.19762754440308, -0.705291986465454, -0.325328677892685, -0.440084099769592, -0.256473302841187, -1.22617733478546, 2.55625653266907, -2.51265931129456, 1.06255269050598, -1.9635773897171, -0.498623430728912, -0.342940181493759, -0.144835770130157, -1.75233459472656, 0.426897585391998, -2.28179502487183, -1.24072754383087, 2.39083528518677, 2.43629693984985, -2.30458807945251, -1.09402704238892, 0.74868243932724, -1.22975087165833, -1.37748348712921, -0.423212200403214, -3.82569479942322, -1.61718380451202, -0.499433577060699, -0.57524847984314, -0.541266024112701, 8.10883808135986, 2.05736064910889, -0.309856444597244, 3.38731384277344, -2.84307241439819, 1.43254041671753, 0.254186123609543, -0.80161589384079, -0.330915361642838, -0.946418881416321, 0.695298731327057, -0.0984209850430489, -1.35947906970978, -0.58078670501709 }
    },
    {
      { 0.000363671482773498, -0.0253048520535231, -0.00584701402112842, 0.030808063223958, -0.0126727437600493, 0.00916171167045832, -0.0109465876594186, -0.00853355973958969, -0.00823739171028137, 0.000454059103503823, -0.0243972893804312, -0.0235861986875534, 0.0235226061195135, -0.00126691616605967, 0.0131832119077444, -0.0199312828481197, -0.0148420119658113, -0.0119730243459344, 0.00223088427446783, -0.0164986476302147, 0.0170532874763012, 0.0107490364462137, -0.0321521759033203, 0.0168750360608101, -0.0107251023873687, 0.0458676144480705, 0.0148206762969494, -0.00797224510461092, -0.00207200925797224, 0.000522653688676655, -0.0285358522087336, -0.00884099025279284, 0.037130244076252, 0.019178370013833, 0.0185577720403671, -0.00161746924277395, 0.00629566889256239, -0.003462303429842, 0.0365303941071033, -0.00221775402314961, -0.00456048687919974, 0.0171723943203688, -0.00631952146068215, -0.0208094734698534, 0.0176323857158422, 0.000378212687792256, -0.00841437838971615, -0.00543354172259569, 0.00496034836396575, -0.00023807238903828, -0.0111427446827292 },
      { 0.00810387916862965, 0.0324564427137375, -0.00461507821455598, -0.0259610265493393, 0.0124030159786344, -0.0126157533377409, 0.0227100923657417, 0.0224534776061773, 0.0144454576075077, 0.0148570947349072, 0.0319937728345394, 0.00392264313995838, -0.00353273120708764, -0.0211758259683847, -0.00555232586339116, 0.0258200019598007, 0.000304851855617017, 0.00859183445572853, -0.0102103780955076, 0.021223958581686, -0.0136557137593627, -0.0188812091946602, 0.0437373779714108, -0.0122735938057303, -0.00490338541567326, -0.00661540031433105, -0.0040332218632102, 0.00318349059671164, 0.0386670678853989, 0.0109326839447021, 0.0308158174157143, 0.0120382448658347, -0.0385274179279804, -0.0173102412372828, -0.0031113603617996, 0.0115602789446712, 0.000296192651148885, 0.0142169529572129, -0.0209660511463881, 0.00692129041999578, 0.0165467858314514, -0.00442654965445399, 0.00726403295993805, 0.0193317234516144, -0.0183512885123491, 0.0177701823413372, 0.010846821591258, 0.00869769044220448, 0.00619413750246167, 0.0100438008084893, 0.027911338955164 },
      { -0.00468747410923243, -0.0211636945605278, 0.0140315582975745, 0.00715969316661358, -0.0178926773369312, 0.0021321305539459, -0.0228310283273458, -0.0170799717307091, -0.00232257461175323, -0.0161739103496075, -0.00548532092943788, 0.00552689330652356, -0.0113354437053204, 0.0173450279980898, -0.00203015096485615, -0.0121872602030635, 0.0216272287070751, -0.00110564054921269, -0.0017276470316574, -0.0174745917320251, 0.00176149839535356, 0.00663660420104861, -0.0178660172969103, -0.0081311734393239, 0.0111912814900279, -0.0264099929481745, -0.00221320381388068, 0.0066460995003581, -0.0384990535676479, 0.00675729056820273, -0.017329977825284, -0.0094361761584878, -0.0050787664949894, -0.00117510964628309, -0.00321136065758765, -0.0168488249182701, -0.00690784864127636, -0.00704609230160713, -0.0166856721043587, -0.00731179257854819, -0.00115750194527209, -0.00249639945104718, -0.00193394091911614, -0.00479246722534299, -0.000667328014969826, -0.019838435575366, -0.00781039800494909, -0.0142931649461389, -0.0270516462624073, -0.018354382365942, -0.0209044702351093 },
      { 0.00651022884994745, -0.00237140827812254, 0.010008804500103, -0.00160574831534177, 0.0021659373305738, 0.00951577257364988, 0.0102009894326329, 0.00241866870783269, -0.00206511770375073, 0.00365041987970471, -0.00179605325683951, 0.00327203865163028, -0.00233237817883492, 0.000353463372448459, 0.00281321932561696, 0.000659708282910287, 0.00708037707954645, 0.000947076827287674, 0.00334365502931178, 0.0111536728218198, 0.00437893066555262, 0.00190208933781832, -0.00446514366194606, 0.0036540636792779, 0.00419708248227835, 0.00176580587867647, 0.00176038686186075, 0.0107267731800675, -0.00175366166513413, -0.00293741375207901, 0.00622694147750735, -0.00269307196140289, 0.0209306608885527, -0.000242078749579377, 0.00225258059799671, 0.00586557108908892, 0.001838329131715, 0.000204940239200369, 0.016228836029768, 0.00547864474356174, -0.000482972507597879, -0.0040243798866868, 0.00360018154606223, -0.000436090835137293, 0.0108516328036785, 0.00124220119323581, 0.00158783618826419, 0.00641324417665601, 0.0109322480857372, 0.00561438966542482, 0.000514072657097131 },
      { 1.00302040576935, 1.02016770839691, 0.990649938583374, 1.00796937942505, 1.01126134395599, 0.999361932277679, 0.986905217170715, 1.0048416852951, 1.02029025554657, 1.00180065631866, 1.01338195800781, 1.00957918167114, 1.00968396663666, 1.02090048789978, 1.00477719306946, 1.01749873161316, 0.991417825222015, 1.01937079429626, 1.00487387180328, 0.983532428741455, 1.01034700870514, 0.995237469673157, 1.02846252918243, 1.00253987312317, 1.00787353515625, 1.00296556949615, 1.0025327205658, 0.994013965129852, 1.01845943927765, 1.00679242610931, 1.00925099849701, 1.01291334629059, 0.993677198886871, 1.00582933425903, 1.0029890537262, 1.0060658454895, 1.00804841518402, 1.01354956626892, 1.01809990406036, 1.00559341907501, 1.01038694381714, 1.0087388753891, 1.00821828842163, 1.01824927330017, 0.989350497722626, 1.01557528972626, 1.01087009906769, 0.996612906455994, 1.00910472869873, 0.995780766010284, 1.01121950149536 },
      { -0.00509880995377898, -0.0106207681819797, -0.00197035446763039, -0.0072894636541605, -0.00475885905325413, -0.00377620058134198, 0.0152133833616972, -0.00275154667906463, -0.0197844784706831, -0.0010877235326916, -0.0131652234122157, -0.00874292384833097, -0.00604290375486016, -0.0186368692666292, -0.00628632912412286, -0.0208881720900536, -0.00315477256663144, -0.0191580764949322, 0.00336674740538001, 0.0201769527047873, -0.0137182353064418, 0.00801673997193575, -0.0280154794454575, 0.00098810310009867, -0.0129208704456687, -0.00387567491270602, -8.011405589059e-05, -0.0038719882722944, -0.0156139712780714, -0.0114351101219654, -0.0112292319536209, -0.00560752162709832, -0.00682843942195177, -0.00212094932794571, 0.00426314817741513, -0.00451200967654586, -0.00641430262476206, -0.0083576338365674, -0.0320318154990673, -0.00774359796196222, -0.00342513551004231, -0.0101468488574028, -0.0123967947438359, -0.0185973197221756, 0.00463649583980441, -0.0128643317148089, -0.00684319064021111, 0.00775648513808846, -0.00511481612920761, 0.0133023550733924, -0.00716416630893946 },
      { -0.110618807375431, -4.19592475891113, 2.33073115348816, -0.406494081020355, 0.909158766269684, -7.1098484992981, 0.234799668192863, -0.233347743749619, -1.13725936412811, -1.2368950843811, 2.34095549583435, 1.31643652915955, 0.827186524868011, 2.40716648101807, 1.10748422145844, 1.69435119628906, 3.67702889442444, -1.86812400817871, -0.520876586437225, -2.49545121192932, -0.587324619293213, 7.0872220993042, -1.58458364009857, 1.11130118370056, 1.39903426170349, 8.56469345092773, -16.2867813110352, -1.44958186149597, 3.85558152198792, -0.778826057910919, -2.4932427406311, 0.0251144710928202, -10.2623662948608, 1.02043807506561, 1.53796899318695, 0.304670244455338, 0.293159663677216, 2.99633598327637, 11.6462688446045, 0.734061539173126, 0.646358251571655, 13.3684015274048, 1.52882528305054, 2.13776183128357, -4.10778093338013, -1.66183018684387, 0.387034118175507, 2.76884746551514, 0.931456744670868, -7.61808204650879, -1.24973142147064 }
    },
    {
      { -0.00766984513029456, 0.00127857481129467, -0.00908019673079252, 0.0670676678419113, -0.00108723167795688, 0.000321685540257022, 0.00360791315324605, 0.0133234839886427, 0.000199891641386785, 0.00256160367280245, -0.00416031712666154, 0.00297723314724863, 0.0593517944216728, -0.00506046693772078, 0.0550871714949608, -0.00263177463784814, -0.0121158678084612, 0.0020968469325453, 0.0035988874733448, 0.00209320592693985, 0.0734435170888901, 0.00320240622386336, -0.0029243326280266, 0.0430280975997448, 0.00421479297801852, 0.147307828068733, -0.000291348493192345, -0.0118547081947327, 0.0223750490695238, 0.0273982658982277, -0.00511144241318107, 0.0127438809722662, 0.0279954541474581, 0.056577455252409, -2.18762743315892e-05, 0.0176264643669128, 0.0183329693973064, -0.00123197806533426, 0.149479523301125, 0.0138038182631135, 0.00233043520711362, 0.0462642721831799, 0.00723750609904528, -0.000232269303523935, 0.0106604332104325, 0.0164563003927469, -0.000190045844647102, 0.000279670348390937, 0.0537497512996197, 0.00219515291973948, 0.00285555771552026 },
      { 0.004156778100878, 0.0114569859579206, -0.00846957508474588, -0.0435458533465862, 0.00415575364604592, -0.0102846100926399, -0.000170935119967908, 0.00481140101328492, -0.000390396598959342, 0.00636796746402979, 0.00546315778046846, 0.00468068616464734, -0.0196283180266619, 0.00578332040458918, -0.0262911058962345, 0.00832748413085938, -0.00640333304181695, -0.00167438492644578, -0.0115891061723232, -0.00477283447980881, -0.0540487878024578, -0.00765540543943644, 0.00494128279387951, -0.0343358665704727, -0.000795582251157612, -0.0805478766560555, 0.00154026539530605, 0.000853995501529425, 0.00895948614925146, -0.00460901949554682, 0.00188087217975408, -0.0219773426651955, -0.0435388125479221, -0.0455276519060135, 0.00105873018037528, -0.00254502589814365, -0.0157508626580238, -0.00255850260145962, -0.140486717224121, -0.0300094913691282, 0.00281261210329831, 0.0128768151625991, -0.00717332866042852, 0.00150483730249107, -0.020074374973774, 0.000522052869200706, 0.000497194880153984, 0.000818125321529806, 0.00529197510331869, 0.00272056716494262, 0.0168259982019663 },
      { -0.0160324387252331, -0.0123703377321362, 0.027706190943718, -0.0298683103173971, -0.00910458527505398, 0.0137937068939209, -0.00163291476201266, -0.0191417951136827, 0.00257641030475497, -0.00850966386497021, 0.00360266049392521, -0.00910031329840422, -0.0329357720911503, 0.000137865310534835, -0.0131680220365524, -0.00760239688679576, 0.0223745796829462, -0.00231652660295367, 0.00513573875650764, -0.000452625390607864, -0.00834590196609497, 0.000376117328414693, -0.00944903865456581, -0.0124906841665506, -0.00422545894980431, -0.0356871709227562, -0.00253072520717978, 0.0143646420910954, -0.0242035929113626, 0.0139093436300755, -0.000522295013070107, -0.00250622420571744, 0.0151264658197761, -0.00453113159164786, -0.00132664607372135, -0.0123249599710107, -0.000483467389130965, 0.00369418994523585, -0.0155257983133197, 0.0123532498255372, -0.00843133497983217, -0.0560573041439056, 0.00203305366449058, -0.00294846156612039, 0.0155202895402908, -0.016638420522213, -0.00386073789559305, 0.00386391626670957, -0.0677804574370384, -0.00581638887524605, -0.0154682695865631 },
      { -0.0147776184603572, -0.00246332609094679, 0.00458290008828044, 0.0343733206391335, 0.000141577591421083, 0.00425869133323431, -0.00122083653695881, 0.0069783185608685, -0.00161941326223314, 0.000439772935351357, -0.00186326482798904, -0.000429634732427076, 0.0246307086199522, 0.00178617902565747, 0.00795310456305742, -0.00126963073853403, 0.0025206923019141, -0.000479852751595899, 0.000675333722028881, 0.0022096554748714, 0.010537825524807, 0.00187964073847979, -0.00242112437263131, 0.00725709786638618, 0.000203229486942291, 0.00987284909933805, 0.0042717126198113, 0.00415300764143467, -0.00493151508271694, -0.00710559263825417, 0.000262741814367473, 0.00316723878495395, 0.00490684108808637, 0.0032995652873069, -0.000117091185529716, 0.00320200901478529, 0.00452719954773784, 0.000219702269532718, 0.0505606755614281, 0.00654615834355354, 7.77197419665754e-05, 0.0253462381660938, -0.000300619285553694, -0.001515285926871, 0.00875731743872166, 0.000116289986181073, 0.000643867126200348, -0.00102172303013504, 0.0410788357257843, 0.00372077594511211, -4.78363836009521e-05 },
      { 0.0119055453687906, 0.0134077183902264, -0.00695202359929681, -0.0117206564173102, 0.0060703856870532, 0.00464957160875201, 0.00415006000548601, -1.29767940961756e-05, 0.0122654354199767, 0.00455314991995692, 0.0104622608050704, 0.00647222436964512, -0.00243423599749804, 0.0131829679012299, 0.000242583424551412, 0.0127661461010575, 0.00454714661464095, 0.0113552389666438, 0.00605271756649017, 0.00640310300514102, 0.00729713682085276, 0.000715275935363024, 0.0279851537197828, 0.00147042982280254, 0.00727219879627228, 0.00240815640427172, -0.00268405815586448, 0.00227557960897684, 0.0166542883962393, 0.00423607509583235, 0.0124420179054141, 0.012349046766758, 0.00896485429257154, 0.00240735779516399, 0.00344567676074803, 0.00371681596152484, 0.0032105993013829, 0.0155838113278151, 0.0249803941696882, 0.0083888117223978, 0.00449457112699747, -0.00375614617951214, 0.0123728793114424, 0.0143414316698909, -0.0129298698157072, 0.0143172834068537, 0.00526773603633046, -0.000402213743655011, -0.00756048131734133, -0.00237559131346643, 0.00832234416157007 },
      { 1.00059843063354, 0.991738796234131, 0.993083000183105, 1.00048077106476, 0.998867809772491, 0.991068661212921, 0.995985686779022, 1.00034832954407, 0.989944458007812, 0.997113883495331, 0.989774942398071, 0.999117732048035, 0.990391552448273, 0.985018849372864, 0.994541227817535, 0.987573385238647, 0.993171036243439, 0.990960001945496, 0.997388422489166, 0.995016813278198, 0.988443851470947, 1.00221836566925, 0.973021268844604, 1.00015103816986, 0.995731592178345, 0.992221295833588, 1.00448441505432, 0.993835866451263, 0.988585889339447, 0.989306092262268, 0.987759292125702, 0.994985282421112, 0.989562928676605, 0.998701870441437, 1.00559282302856, 0.997265219688416, 0.995249569416046, 0.985220432281494, 0.94596540927887, 0.989016532897949, 1.00250267982483, 0.987573385238647, 0.986013948917389, 0.986846506595612, 1.00182366371155, 0.989518582820892, 0.999169647693634, 0.996939241886139, 0.991148471832275, 1.00467646121979, 0.992374420166016 },
      { 6.22884798049927, -3.29878354072571, 10.4156761169434, 3.31092858314514, 0.91766756772995, -3.58298563957214, -0.162179857492447, -3.51619935035706, -0.968527853488922, -0.549361407756805, 1.15855002403259, 0.5118128657341, 1.61255240440369, -0.542326807975769, 0.149143040180206, 1.3736537694931, 2.65462255477905, -0.0579372309148312, 0.463286072015762, -1.44107520580292, -3.30560779571533, 5.02175378799438, -3.89495325088501, 0.564043462276459, -0.161629617214203, 8.58342170715332, -9.55197525024414, 1.75710868835449, 2.86899304389954, -3.25768995285034, -0.738622546195984, -0.797677040100098, -3.61228537559509, 0.301571696996689, -0.922010958194733, -3.30048584938049, 0.173236027359962, 0.701237201690674, 2.84016966819763, -0.559445083141327, 0.414727747440338, 22.1161346435547, 0.923471331596375, 0.876826286315918, -2.88054132461548, -2.61504364013672, 0.00222086836583912, 0.444981396198273, 8.72086524963379, -5.83839750289917, -1.73493075370789 }
    }
  },
  {
    {
      { 0.693313539028168, 0.80438107252121, 0.92018735408783, 0.757844924926758, 0.747052848339081, 0.773248076438904, 0.644611656665802, 0.870806634426117, 0.870750427246094, 0.740375697612762, 0.770661771297455, 0.832322597503662, 0.222200021147728, 0.789603173732758, 0.62058824300766, 0.741308987140656, 0.8013995885849, 0.699948906898499, 0.779525756835938, 0.880277574062347, 0.806365489959717, 0.759503841400146, 0.757092654705048, 0.783462762832642, 0.761489510536194, 0.939522862434387, 0.882861196994781, 0.786723136901855, 0.693513095378876, 0.879397630691528, 0.754121840000153, 0.774092853069305, 0.937111556529999, 0.934636652469635, 0.767476558685303, 0.840050041675568, 0.817087471485138, 0.745438694953918, 0.904465556144714, 0.878835022449493, 0.835004925727844, 0.824941098690033, 0.695613443851471, 0.817794322967529, 0.885041117668152, 0.748997032642365, 0.796049237251282, 0.774727165699005, 0.64418226480484, 0.679900765419006, 0.793143808841705 },
      { 0.236669018864632, 0.161270007491112, -0.21421317756176, 0.240331843495369, 0.311697334051132, 0.0642815306782722, 0.437299489974976, 0.170098781585693, 0.155027702450752, 0.232704624533653, 0.268550187349319, 0.0719465240836143, 2.13537192344666, 0.257207661867142, 0.427344024181366, 0.163892194628716, 0.189804717898369, -0.145900651812553, 0.11736037582159, 0.175590202212334, 0.170743361115456, 0.142065837979317, 0.0684436708688736, 0.214990884065628, 0.28437265753746, 0.0788938626646996, 0.129566341638565, 0.0718825086951256, 0.214686304330826, 0.0789449959993362, 0.240378946065903, 0.183724105358124, -0.24119533598423, -0.258266746997833, 0.250142633914948, 0.175422310829163, 0.120022654533386, 0.272025972604752, 0.0622297301888466, 0.0632520169019699, 0.0817127823829651, 0.179567009210587, 0.283105105161667, 0.189893379807472, 0.229753732681274, 0.227491363883018, 0.179016381502151, 0.331316828727722, 0.155426949262619, 0.230612307786942, 0.167429149150848 },
      { 0.0482379347085953, 0.037472415715456, 0.243508517742157, -0.00867330282926559, -0.0906506404280663, 0.153763681650162, -0.0706696212291718, 0.0335613749921322, -0.0136926649138331, 0.0450677648186684, 0.0202137380838394, 0.0611480697989464, -0.867173194885254, -0.0598874948918819, 0.000280888110864908, 0.0567332394421101, 0.0164227113127708, 0.327555745840073, 0.0945991948246956, -0.0711710453033447, 0.0196030084043741, 0.0877307131886482, 0.176288098096848, 0.00775755289942026, -0.0276636742055416, -0.051010325551033, 0.00549245858564973, 0.130897209048271, 0.0620050132274628, 0.0177244488149881, -0.00067455240059644, 0.0149695333093405, 0.251566112041473, 0.272097885608673, -0.0428388379514217, -0.0453349687159061, 0.0538600198924541, -0.0132139837369323, -0.0197445936501026, 0.0584057494997978, 0.0952347666025162, -0.0766681209206581, 0.0324856713414192, 0.0385549701750278, -0.1364865899086, 0.00198171962983906, 0.00950684677809477, -0.103602737188339, 0.143235459923744, 0.0800435468554497, 0.0304081309586763 },
      { -0.0115646068006754, 0.0030950210057199, 0.00591920362785459, -0.00809772964566946, -0.0314149484038353, 0.00457627791911364, -0.0227186903357506, -0.0268973708152771, -0.0104937572032213, -0.00469963578507304, -0.0200589671730995, 0.0198441743850708, -0.011234506033361, 0.0183177143335342, -0.012894026003778, 0.0020406546536833, 0.00271260878071189, 0.114827573299408, -0.00538454856723547, 0.0541907697916031, -0.0488969013094902, -0.0022192623000592, 0.0058832080103457, -0.00819786824285984, -0.0212225038558245, 0.0102944308891892, -0.0168913789093494, 0.00843720324337482, -0.00842340942472219, 0.0140910884365439, -0.0386225357651711, -0.00305600161664188, 0.00701606040820479, 0.00457047717645764, 0.0270779207348824, 0.0531321093440056, -0.00270196655765176, 0.0371701568365097, 0.0203783996403217, 0.025904206559062, -0.00593496952205896, -0.0254764221608639, -0.0131783140823245, -0.00800241064280272, 0.0658211261034012, -0.0122279785573483, -0.00301954685710371, 0.0189127866178751, 0.0025235244538635, -0.0389195829629898, 0.00531022623181343 },
      { 0.00873971078544855, -0.0219389665871859, 0.0855513066053391, 0.0035913335159421, 0.00663334038108587, -0.000817543361335993, 0.0333704464137554, 0.044975221157074, 0.01655856333673, -0.0177655760198832, 0.00893149338662624, 0.0459986887872219, 0.0280717704445124, 0.0211404394358397, -0.0203558206558228, -0.00843743328005075, -0.00838212762027979, -0.0561319775879383, 0.0139257470145822, 0.0104217147454619, 0.0583786033093929, 0.00805555284023285, -0.0220214128494263, -0.0285163819789886, 0.0145301725715399, 0.00273615098558366, 0.0458393879234791, -0.0126338573172688, 0.0125335827469826, 0.0264672394841909, 0.0178507585078478, -0.010250368155539, 0.0877284556627274, 0.0901384577155113, 0.000524642935488373, 0.00108979747164994, 0.00547628849744797, -0.0418609753251076, -0.000989146530628204, 0.0394441038370132, 0.0083339037373662, 0.0806407481431961, 0.0178817510604858, -0.00827217567712069, 0.0281426906585693, 0.00198962166905403, 0.00252823787741363, -0.0198469124734402, -0.0103177418932319, 0.0409729517996311, 0.0062588257715106 },
      { 0.00126795622054487, 0.0302288178354502, -0.0361533835530281, 0.00471219047904015, 0.0497693046927452, -0.00906586088240147, -0.0208879169076681, -0.0470685176551342, 0.000896916026249528, 0.00797408260405064, -0.00916902907192707, -0.0278609599918127, -0.0460527464747429, -0.00815976783633232, 0.00363576062954962, 0.0144734727218747, 0.0127066550776362, 0.085814543068409, -0.00980754010379314, -0.0235924161970615, -0.0180457457900047, -0.00461190426722169, 0.0269781928509474, 0.0380054861307144, -0.00580924097448587, 0.00576500780880451, -0.0277836564928293, 0.010731871239841, -0.00517415767535567, -0.0168811473995447, 0.0245954189449549, 0.0288829300552607, -0.0428517982363701, -0.0348895229399204, -0.0046441862359643, -0.0106645748019218, -0.00372959487140179, 0.0162639319896698, -0.000502184266224504, -0.0520515069365501, -0.00588920572772622, -0.0437528789043427, -0.0149441044777632, 0.00912613421678543, -0.0227959118783474, 0.0152562707662582, 0.0133515195921063, 0.0202585496008396, 0.000492802704684436, -0.0157930310815573, -0.00922480411827564 },
      { -11.3918313980103, -6.31186962127686, -62.3916549682617, -0.258169174194336, -6.0090503692627, -6.42859172821045, -11.7954235076904, -31.7450504302979, -2.65929698944092, -4.54042291641235, 3.80715823173523, -69.2202987670898, 63.0678977966309, 10.1379709243774, -21.2421340942383, -4.96925640106201, -16.5684375762939, -1.9461555480957, -5.64665460586548, 9.96902656555176, -6.16215324401855, -22.072624206543, -0.583315491676331, 3.45643329620361, -4.59631586074829, 262.681396484375, -14.5764074325562, -14.8311862945557, -15.8599977493286, -27.7321529388428, -0.130849093198776, -3.41949248313904, -47.2728080749512, -74.0222625732422, -6.48098850250244, -10.899022102356, -4.96634435653687, -15.2639036178589, 249.657653808594, -36.1353416442871, 10.2663946151733, 17.825345993042, -10.3613500595093, -14.0368680953979, -66.4964294433594, -2.09956669807434, -11.7371788024902, -20.6973533630371, 5.37599086761475, -60.9133262634277, -6.57296323776245 }
    },
    {
      { -0.0135103305801749, -0.0774093121290207, 0.00155435036867857, -0.0220577996224165, -0.0165380649268627, -0.0148785002529621, 0.0200579222291708, -0.0710928291082382, -0.0708423182368279, -0.00659183831885457, -0.0176082290709019, 0.0406021177768707, -0.00225909892469645, 0.00054206425556913, -0.000472990534035489, 0.00204541510902345, -0.0640001744031906, 0.0149297760799527, -0.0718677639961243, -0.0702313482761383, -0.0929877832531929, -0.0338374339044094, -0.0105859516188502, -0.0515815503895283, -0.0688740760087967, -0.0142864817753434, -0.0719133391976357, -0.0019781393930316, -0.0195089131593704, -0.0793258547782898, -0.0725569650530815, -0.0637632757425308, -0.0103330090641975, 0.0142949232831597, -0.09074667096138, -0.0510043352842331, -0.0219891760498285, -0.0445781834423542, -0.0625450685620308, -0.0684736222028732, -0.0261483434587717, -0.0579633712768555, 0.0077263480052352, -0.0734743848443031, -0.0801046341657639, -0.0172881353646517, -0.0493666939437389, -0.068028636276722, -0.0139765338972211, -0.00727581093087792, -0.0493595451116562 },
      { 0.964631736278534, 1.07641279697418, 0.917156994342804, 0.953971683979034, 0.935666501522064, 0.98679780960083, 0.892845988273621, 0.999245226383209, 1.05366003513336, 0.930223643779755, 0.953649461269379, 0.839500427246094, 0.849242806434631, 0.938444495201111, 0.936307489871979, 0.911782324314117, 1.02606976032257, 0.936607718467712, 0.951771855354309, 0.980120241641998, 1.06156647205353, 0.986275732517242, 0.916431427001953, 1.05031979084015, 1.0151606798172, 0.905374526977539, 1.02654147148132, 0.94504326581955, 0.967778205871582, 1.02347040176392, 1.05116319656372, 1.01844799518585, 0.936245977878571, 0.912376642227173, 1.04450964927673, 0.976982116699219, 0.950687408447266, 0.999139130115509, 1.00108683109283, 1.0219167470932, 0.97729879617691, 1.02559924125671, 0.916960775852203, 1.0360289812088, 0.972487151622772, 0.946279466152191, 1.04350769519806, 0.985206723213196, 0.90371572971344, 0.945718705654144, 1.04493010044098 },
      { 0.0586194023489952, 0.0445432923734188, 0.0856573879718781, 0.0701944977045059, 0.100765690207481, 0.0448365435004234, 0.086379773914814, 0.0386849939823151, 0.0489287488162518, 0.0920188203454018, 0.0679809525609016, 0.116876929998398, 0.106690190732479, 0.0741144046187401, 0.0617627874016762, 0.0994615033268929, 0.0631132200360298, 0.00632676342502236, 0.094298854470253, 0.0965812057256699, 0.0539771318435669, 0.066552147269249, 0.103490024805069, 0.0160409137606621, 0.0694012343883514, 0.114243447780609, 0.0579415075480938, 0.0731999203562737, 0.0586710199713707, 0.0579883269965649, 0.0341037139296532, 0.0617630407214165, 0.0842212438583374, 0.0670284628868103, 0.0686376243829727, 0.0828195139765739, 0.0725106447935104, 0.0462770462036133, 0.0551680810749531, 0.0453271791338921, 0.0590679831802845, 0.0871410593390465, 0.075892485678196, 0.0564531311392784, 0.110498003661633, 0.0869160071015358, 0.0351887531578541, 0.0753638595342636, 0.120190188288689, 0.0698741301894188, 0.0368316881358624 },
      { -0.00151817651931196, -0.0115297995507717, 0.00586817227303982, -0.00202508922666311, 0.00650537805631757, -0.000184649237780832, 0.00657391734421253, 0.0137056140229106, -0.00712523935362697, 0.00644391495734453, -0.00295596360228956, 0.0310943126678467, 0.000889753515366465, -0.0187851600348949, -0.00349606154486537, 0.000974324473645538, -0.0114055974408984, -0.00630571926012635, 0.00706591969355941, -0.0153758684173226, 0.00557362847030163, 0.00310930795967579, -0.00116234831511974, -0.00896404962986708, -0.0112368892878294, -0.00513942539691925, -0.00733457552269101, 0.00127251469530165, -0.00202470505610108, -0.00433563906699419, 0.00269267382100224, 0.00474779075011611, 8.55078178574331e-05, 0.0240435153245926, -0.0278459042310715, -0.018996387720108, 7.69662219681777e-05, -0.0267780516296625, 0.0058073322288692, -0.00620636763051152, -0.00654153572395444, -0.00627362774685025, 0.00404699752107263, -0.00425732228904963, -0.0245458278805017, -0.00229075434617698, -0.00483921589329839, -0.0107742371037602, 0.00644746888428926, 0.00196284824050963, -0.0171650648117065 },
      { -0.00380045059137046, -0.0243843998759985, -0.0117974821478128, -0.0045028468593955, -0.00465337093919516, -0.0240560881793499, -0.00418688822537661, -0.0142781557515264, -0.020037928596139, -0.0181839987635612, -0.000671730376780033, -0.0427725054323673, -0.00579258892685175, 0.0275916680693626, 0.00412209145724773, -0.000879639061167836, -0.0115578342229128, 0.0481630638241768, -0.0184671804308891, -0.00817897543311119, -0.0206253584474325, -0.0287286322563887, -0.00446098623797297, 0.0123702073469758, -0.00912055838853121, -0.00314073916524649, -0.0125482250005007, -0.0147015312686563, -0.00389134883880615, -0.0115122199058533, -0.00437882728874683, -0.039701022207737, -0.010146951302886, -0.0178218521177769, -0.00194531842134893, -0.00209794077090919, -0.00277191260829568, 0.0212951991707087, -0.0046360744163394, -0.0181268472224474, 0.00455540418624878, -0.0163212101906538, -0.00310678966343403, -0.0263744741678238, -0.00676612369716167, 0.00379884988069534, -0.0185917448252439, 0.0123565355315804, -0.0128454929217696, -0.0186992343515158, -0.0198697168380022 },
      { 0.00292424089275301, 0.0160869359970093, 0.00560332927852869, 0.00569722475484014, -0.0177718587219715, 0.0335890203714371, 0.00134316924959421, 0.0156228579580784, 0.00824145507067442, 0.00297455117106438, -0.000137536873808131, 0.0158172249794006, 0.006979591678828, -0.0380236133933067, -0.00313184456899762, -0.00588225573301315, -0.00213415990583599, 0.0134559236466885, 0.00521318148821592, 0.00812076218426228, -0.00165402446873486, 0.00950725935399532, 0.00699910195544362, -0.0174342896789312, 0.0032022709492594, -0.00861464068293571, 0.00754679180681705, 0.00720381317660213, 0.00375974480994046, 0.00885187275707722, -0.0138492025434971, 0.0404519066214561, 0.0032782512716949, 0.00315665337257087, 0.00477020302787423, 0.00509601458907127, 0.00175841921009123, -0.00560165662318468, 0.00432352488860488, 0.0200861115008593, -0.0072505222633481, -0.00220840401016176, 0.00163271615747362, 0.0132255647331476, 0.0054053720086813, -0.0114868124946952, 0.0106723085045815, -0.0179786868393421, 0.00609555561095476, 0.00689137447625399, 0.0249949228018522 },
      { 1.95512652397156, 8.92400932312012, -2.72570967674255, 0.246015146374702, -0.274700373411179, -4.42808723449707, -5.23876094818115, 7.67569541931152, -2.0324125289917, -1.97654688358307, 0.334796756505966, 4.60253095626831, -9.92097568511963, 11.2833395004272, -3.65696811676025, -0.117707379162312, 6.80265188217163, -3.70300388336182, 0.0509192459285259, -3.29466199874878, -4.97698402404785, 14.0985307693481, -5.11204767227173, -3.12105321884155, -0.633242607116699, 13.6230382919312, 1.21454012393951, -1.3234988451004, 3.20322322845459, 8.14049530029297, -1.01147377490997, -5.29026222229004, 3.2419867515564, -17.2120571136475, 8.15035915374756, 7.91356039047241, 0.542776226997375, 8.08284473419189, -91.7490081787109, 13.6019449234009, 0.246427580714226, -2.02727603912354, -3.720383644104, 13.6971473693848, 25.2281818389893, -0.753160178661346, 3.35783886909485, 18.1014347076416, -4.90408086776733, 29.6256370544434, 1.65007174015045 }
    },
    {
      { 0.0405445098876953, 0.163054078817368, 0.508898735046387, 0.0189492143690586, 0.00974070560187101, 0.0644498914480209, -0.0305520724505186, 0.0918543189764023, 0.108320794999599, 0.0431696698069572, 0.0461016185581684, 0.278496026992798, -0.0015634692972526, -0.035625733435154, -0.0150577202439308, 0.0457852706313133, 0.131378591060638, 0.0926199033856392, 0.0541811585426331, 0.0545659773051739, 0.17940391600132, 0.124352566897869, 0.0683990642428398, 0.0865905955433846, 0.115868546068668, -0.027336360886693, 0.114908158779144, 0.0901217386126518, 0.0792325139045715, 0.118651799857616, 0.11883544921875, 0.0832059383392334, 0.503321290016174, 0.429801464080811, 0.171181827783585, 0.0347549058496952, 0.0983219295740128, 0.0807163864374161, -0.00397140253335238, 0.0968119874596596, 0.050274021923542, 0.0610673800110817, 0.0312542021274567, 0.113330855965614, 0.0891998633742332, 0.0348201207816601, 0.0897305607795715, 0.0407431125640869, 0.0450298823416233, 0.0395903289318085, 0.083502396941185 },
      { -0.278189271688461, -0.388494342565536, -1.00549578666687, -0.135486513376236, -0.143617987632751, -0.132499113678932, -0.028418792411685, -0.202657967805862, -0.389835089445114, -0.135443821549416, -0.22720105946064, -0.494883954524994, -0.334656864404678, -0.00971849169582129, -0.104817599058151, -0.167848616838455, -0.337145358324051, -0.883069038391113, -0.16384120285511, -0.115108028054237, -0.428992629051208, -0.226088359951973, -0.150497600436211, -0.367020487785339, -0.255502462387085, -0.0883535668253899, -0.336315512657166, -0.191759824752808, -0.319197833538055, -0.258015751838684, -0.307859003543854, -0.303051769733429, -1.14390981197357, -0.772663772106171, -0.299476742744446, -0.127777829766273, -0.144107386469841, -0.275884836912155, -0.0841559767723083, -0.245113417506218, -0.212397918105125, -0.303818315267563, -0.197105705738068, -0.363552510738373, -0.0917736515402794, -0.162901163101196, -0.324214965105057, 0.0104980962350965, -0.140902757644653, -0.299351364374161, -0.342198699712753 },
      { 1.19681358337402, 1.17314636707306, 1.36931145191193, 1.11196982860565, 1.11378526687622, 1.07152664661407, 1.05750417709351, 1.22189283370972, 1.22408151626587, 1.08107125759125, 1.16140007972717, 1.13242304325104, 1.37065935134888, 1.04137575626373, 1.13728272914886, 1.0735901594162, 1.17401170730591, 1.64046955108643, 1.10972833633423, 1.01955711841583, 1.2143040895462, 1.11649811267853, 1.01542055606842, 1.24316263198853, 1.11829829216003, 1.08449590206146, 1.20596706867218, 1.06912839412689, 1.20239007472992, 1.14004302024841, 1.17835700511932, 1.20272207260132, 1.52731251716614, 1.1515429019928, 1.06889665126801, 1.06317615509033, 1.09550416469574, 1.17109632492065, 1.02941179275513, 1.15498852729797, 1.1648964881897, 1.10250151157379, 1.15642845630646, 1.19974207878113, 0.990510404109955, 1.09013199806213, 1.21077024936676, 0.929754912853241, 1.05261254310608, 1.22675776481628, 1.21739900112152 },
      { 0.0126053420826793, 0.0175762251019478, -0.00569537794217467, -0.00473002204671502, -0.0179853122681379, 0.00526055879890919, -0.00403071846812963, -0.0670659840106964, 0.0132606169208884, -0.00744581501930952, 0.00666039763018489, 0.184743657708168, -0.00514725735411048, 0.0326378867030144, 0.00323953106999397, 0.00288430834189057, 0.0204224549233913, 0.196875497698784, 0.00460181431844831, 0.0369311235845089, -0.0291315373033285, 0.00706069124862552, -0.00241073942743242, 0.024505877867341, 0.022767310962081, 0.00607356708496809, 0.0148411737754941, 0.00427405443042517, 0.0092048579826951, -0.00854436866939068, -0.0342385247349739, 0.0014769493136555, 0.022703455761075, 0.11167573928833, 0.0951859205961227, 0.0443525761365891, 0.00109726726077497, 0.0769493877887726, 0.0396429225802422, 0.0173926148563623, 0.0100704543292522, 0.00231264042668045, 0.00859449990093708, 0.0156592186540365, 0.0488614588975906, 0.00923338904976845, 0.000721474702004343, 0.0531063079833984, -0.000434976333053783, -0.00420784344896674, 0.0219964031130075 },
      { 0.00369070959277451, 0.040242001414299, 0.188023701310158, 0.0285476818680763, 0.0067952130921185, -0.0256776958703995, 0.00624010758474469, 0.0662839859724045, 0.033224418759346, 0.0140519393607974, 0.000544258218724281, -0.20074787735939, -0.133280128240585, -0.0580502040684223, -0.00220327102579176, 0.00677434168756008, 0.028341680765152, -0.149505823850632, -0.0131865907460451, 0.0328737460076809, 0.0537651032209396, -0.00276350183412433, 0.028676824644208, -0.0277175381779671, 0.0239037182182074, -0.0054719066247344, 0.0237797908484936, -0.00205223658122122, 0.0128625221550465, 0.0493806004524231, 0.0131711969152093, 0.0369594171643257, 0.139933735132217, 0.136433184146881, -0.00322820921428502, 0.0152314230799675, 0.00467796297743917, -0.0459069237112999, -0.00324855884537101, 0.051638275384903, -0.0109042394906282, 0.0539908148348331, 0.000339337042532861, 0.0330189988017082, 0.0215416271239519, 0.00400088401511312, 0.0284811966121197, -0.0240526329725981, -0.00244487659074366, 0.0506052859127522, 0.0412490479648113 },
      { 0.0132659990340471, -0.0363772474229336, -0.037263885140419, -0.0212516933679581, 0.0322954617440701, 0.0492601245641708, 0.000852468423545361, -0.0598224364221096, -0.0164135824888945, 0.00534247979521751, 0.0101122232154012, 0.0968850627541542, 0.128137156367302, 0.0516299158334732, -0.00993370451033115, 0.0239878110587597, -0.0146867027506232, 0.125878319144249, 0.0302367303520441, -0.0238306242972612, 0.00113629340194166, 0.0120108872652054, -0.0288653522729874, 0.0398059040307999, -0.0210052970796824, 0.0114621436223388, -0.0207827370613813, 0.033589418977499, 0.00783486012369394, -0.0310151409357786, 0.040145218372345, -0.0467633344233036, -0.0473224930465221, -0.0283581800758839, -0.00950126722455025, -0.0114720808342099, -0.00400433503091335, 0.0183496735990047, 0.00814144778996706, -0.0536425746977329, 0.00961718615144491, 0.00243851658888161, -3.69840454368386e-05, -0.0213302131742239, -0.0105335125699639, 0.0160633735358715, -0.0192589871585369, 0.0229955539107323, 0.0292842779308558, -0.0123782865703106, -0.0491060763597488 },
      { -21.7376308441162, -7.93854570388794, -142.515441894531, -10.1252031326294, -9.73629856109619, -2.55911707878113, -3.18713235855103, -38.8436698913574, 7.20057106018066, -1.076531291008, -11.2255668640137, -12.077787399292, -52.4159736633301, -31.1143341064453, 10.0175848007202, -8.29776096343994, -10.5757694244385, 15.3428287506104, -1.87816047668457, 6.78302049636841, 13.9929552078247, -1.60256254673004, 8.81230068206787, 6.21223306655884, 1.82227826118469, 70.9604110717773, -5.20143795013428, -3.26402282714844, -22.04026222229, -31.4479846954346, 1.40808308124542, 8.07066535949707, -86.3448257446289, -258.995025634766, -21.22385597229, -41.5297584533691, -13.2084560394287, -23.751916885376, 67.1507110595703, -48.2368774414062, -2.26617503166199, -10.5543346405029, -2.18487906455994, -12.5612897872925, -55.0069732666016, -10.430869102478, -2.39557838439941, -83.7486801147461, 4.32529354095459, -87.8542709350586, -7.18691635131836 }
    },
    {
      { -0.0431879051029682, -0.923476874828339, 0.412545442581177, -0.0589045099914074, -0.148638889193535, -1.45620679855347, -0.227644458413124, 0.0229446552693844, -0.483914375305176, -0.598171412944794, 0.191370368003845, 1.11610746383667, -0.0652570202946663, -0.0639907419681549, -0.0691003054380417, -0.837555408477783, -0.160550579428673, 0.283848524093628, -2.19233584403992, -0.0327218137681484, -0.0513574779033661, -1.49754810333252, -1.86219429969788, 0.0323914475739002, 0.0795869305729866, -0.0305643398314714, -0.402473747730255, -1.32082509994507, -0.272742092609406, -0.0396717041730881, 0.0568116195499897, -0.773175835609436, 0.504354298114777, 0.376516968011856, 0.13192455470562, 0.0216064341366291, -1.17503237724304, 0.082824669778347, 0.432138204574585, -0.0712630972266197, 0.147906824946404, -0.0292944200336933, -0.138458758592606, -1.05023348331451, 0.0179975461214781, -0.0655534267425537, -0.279000282287598, 0.0150873083621264, -2.14287042617798, -0.0875005945563316, -0.183401599526405 },
      { 0.210462003946304, -0.853781461715698, -1.32309365272522, -0.0694694966077805, -0.0382151380181313, 1.43115854263306, 0.302815824747086, -0.128872394561768, -0.311956256628036, 0.375912874937057, -0.0931360945105553, -1.80483293533325, -0.204695537686348, -0.0901548489928246, -0.143043279647827, 1.64209365844727, -0.100738555192947, -1.81356167793274, 1.39187240600586, -0.159975066781044, -0.257295846939087, 0.856495082378387, 1.20913422107697, -0.408877402544022, -0.355574011802673, -0.259771198034286, -0.0697845667600632, 0.966855823993683, 0.350524663925171, -0.107351340353489, -0.44412961602211, -0.811442255973816, -1.50049436092377, -1.36374223232269, -0.397040367126465, -0.179744750261307, 1.72489476203918, -0.355636030435562, -0.645624339580536, -0.125221073627472, -0.633547186851501, -0.0822638496756554, 0.0660927221179008, -0.58119261264801, -0.228703990578651, 0.063570462167263, -0.764847457408905, 0.0670150592923164, 1.59516847133636, -0.0434181466698647, -0.464293420314789 },
      { 0.870758593082428, 2.05515766143799, 1.71320402622223, 0.797362923622131, 0.872028172016144, 0.848940968513489, 0.74864250421524, 0.937986433506012, 1.29427826404572, 0.797971546649933, 0.702544927597046, 1.22824347019196, 1.76252257823944, 0.97650283575058, 1.00789058208466, -0.119284197688103, 0.963494658470154, 2.03425526618958, 0.851087272167206, 0.932602405548096, 1.03853356838226, 1.05891668796539, 1.34766912460327, 1.10508406162262, 0.944546639919281, 1.06237864494324, 1.04381275177002, 0.888640940189362, 0.840255856513977, 0.840888381004333, 1.02265048027039, 1.97417604923248, 1.79814684391022, 1.63210332393646, 1.01406216621399, 0.94660758972168, 0.390216320753098, 1.06148302555084, 1.17898678779602, 0.935686469078064, 1.0428261756897, 0.841269969940186, 0.870430052280426, 1.69984066486359, 1.04200267791748, 0.713846981525421, 1.60504305362701, 0.668801069259644, -0.571479618549347, 0.899284303188324, 1.20363080501556 },
      { 0.170809715986252, 0.240276888012886, 0.510114371776581, 0.305457979440689, 0.239560514688492, 0.0303053092211485, 0.153733879327774, 0.209571644663811, 0.235524654388428, 0.215460956096649, 0.296587765216827, 0.794782996177673, 0.023177944123745, 0.209945470094681, 0.153689295053482, 0.0919607952237129, 0.140744477510452, 0.595066726207733, 0.088261216878891, 0.237148702144623, 0.257719874382019, 0.136098265647888, 0.0599764585494995, 0.25723472237587, 0.302482455968857, 0.192228779196739, 0.286473274230957, 0.0329859033226967, 0.183160156011581, 0.26619827747345, 0.352053731679916, 0.24337837100029, 0.405428141355515, 0.86619508266449, 0.287315964698792, 0.224058955907822, 0.0739437863230705, 0.277274578809738, 0.249061897397041, 0.245790645480156, 0.297222137451172, 0.241187691688538, 0.134723424911499, 0.241624236106873, 0.192873284220695, 0.290169596672058, 0.23576508462429, 0.267244815826416, -0.0363088920712471, 0.203440919518471, 0.182342946529388 },
      { 0.0631828755140305, 0.0810771286487579, -0.504505574703217, -0.0148016707971692, 0.0118866618722677, -0.161216348409653, 0.00446195155382156, -0.00135191681329161, 0.0184085220098495, 0.0255946703255177, -0.0154968425631523, -0.704290926456451, -0.0419743172824383, -0.144570350646973, 0.0581954307854176, -0.0196344014257193, 0.0827833935618401, -0.303542345762253, -0.0376806408166885, 0.00819532759487629, -0.0198311731219292, -0.0541971512138844, 0.0204758867621422, -0.0233821552246809, -0.0553238913416862, -0.0376750268042088, -0.0595297329127789, -0.0130982175469398, 0.0632060617208481, -0.0150071103125811, -0.0699212774634361, -0.000926313339732587, -0.32750791311264, -0.789152026176453, -0.016657704487443, 0.0124209066852927, -0.0966720581054688, -0.0500252656638622, 0.0431288927793503, 0.0155928144231439, 0.0225911438465118, 0.0280446391552687, 0.0621433481574059, 0.143286556005478, 0.0130358682945371, -0.0320215001702309, 0.0626087635755539, -0.0468639060854912, 0.782669723033905, -0.0104544963687658, 0.107020989060402 },
      { -0.148401126265526, -0.148681551218033, 0.214262932538986, 0.00351306353695691, 0.0225629452615976, 0.398321330547333, 0.022547772154212, -0.00767792342230678, 0.00458673667162657, 0.0691183358430862, 0.0202218201011419, 0.35517743229866, 0.143528535962105, 0.153236418962479, -0.0500170439481735, 0.222204178571701, 0.0313170813024044, 0.237963438034058, -0.0206865575164557, -0.00765357632189989, -0.0058362097479403, -0.0081128804013133, -0.125120267271996, 0.0293566398322582, 0.0773861333727837, 0.0452773347496986, 0.0158363599330187, 0.211319252848625, -0.136237755417824, 0.018319645896554, 0.0708429962396622, -0.0975727438926697, 0.131525859236717, 0.296128034591675, -0.0193203277885914, -0.0147459944710135, 0.0403673872351646, 0.0122937140986323, -0.0674446448683739, -0.0253630355000496, -0.0365319065749645, -0.027501767501235, 0.00860062893480062, -0.239233806729317, -0.0332533642649651, 0.0336060859262943, -0.10641685128212, 0.060845784842968, -0.56529575586319, 0.0182533357292414, -0.0259473863989115 },
      { 38.2538528442383, 71.6832046508789, -179.949981689453, 4.14967441558838, -10.8765602111816, 108.935035705566, 13.9890003204346, -11.4639186859131, 2.33479118347168, 34.2883110046387, -68.8818206787109, -30.2617530822754, -482.300109863281, -50.7641639709473, 30.9419651031494, 48.1990089416504, -9.56000423431396, 58.4192314147949, 176.808410644531, -10.5958156585693, 0.00948059931397438, 141.362167358398, -37.6131210327148, -11.1425371170044, -2.95368266105652, 81.1307678222656, 45.6069717407227, 179.393859863281, 21.038932800293, 7.35500478744507, -7.97978591918945, 108.85164642334, -119.871017456055, -287.721649169922, 15.2688722610474, -19.5762729644775, 100.332733154297, -30.9547691345215, 45.5807838439941, -18.1458549499512, 58.7357635498047, 9.28362655639648, 14.0939798355103, 152.703979492188, 22.3580303192139, -0.165988028049469, 70.719612121582, -180.27214050293, 270.927215576172, 54.1152114868164, 31.1988391876221 }
    },
    {
      { -0.055728442966938, 0.123756125569344, 0.201027542352676, -0.158084556460381, -0.22354331612587, -0.650153338909149, -0.162718743085861, -0.146828964352608, -0.0509916804730892, -0.551919460296631, 0.309545487165451, 1.09649014472961, -0.12471916526556, -0.0254400055855513, 0.110659956932068, -0.620542407035828, 0.115230605006218, 0.0782365426421165, -0.433259099721909, -0.168301984667778, 0.111305139958858, 0.00930285640060902, -0.725807011127472, -0.100224554538727, -0.0310919545590878, 0.491596519947052, -0.308815449476242, -0.414884954690933, -0.0149657344445586, -0.375421106815338, -0.0438128560781479, -0.0687725618481636, 0.274492859840393, 0.280664354562759, -0.0815671607851982, -0.0149658666923642, -0.787384390830994, -0.009372822009027, 0.813920021057129, -0.399800837039948, 0.101266972720623, -0.287875652313232, -0.259623885154724, 0.208108127117157, -0.147569179534912, -0.208580315113068, -0.362429767847061, 0.0923331826925278, -0.361198335886002, -0.379649072885513, -0.0608238875865936 },
      { 0.738249063491821, 0.185618817806244, -0.661525130271912, 0.17546558380127, 0.432043790817261, 0.29157343506813, 0.403654724359512, 0.233709096908569, 0.143102958798409, 0.933205783367157, 0.337933152914047, -1.63539958000183, -0.617108762264252, -0.0550926402211189, -0.0375712662935257, 1.79831957817078, 0.0624192133545876, -1.0441552400589, 1.15250813961029, -0.156386896967888, 0.0666530504822731, 0.886860072612762, 1.20178699493408, -0.11314956843853, -0.0815194696187973, -0.23514598608017, 0.566968858242035, 0.751398742198944, 0.754965901374817, 0.352857440710068, -0.290957421064377, 0.739328980445862, -0.553288280963898, -0.798789262771606, -0.262391209602356, -0.1566371768713, 2.13947033882141, -0.0735587850213051, -0.654502689838409, 0.274709016084671, 0.858768701553345, 0.380078703165054, 0.776910722255707, 0.459202080965042, -0.367721140384674, 0.540780246257782, 0.162016540765762, 0.0445213578641415, 1.4467978477478, 1.17925071716309, -0.0155292432755232 },
      { -0.110207937657833, -0.0728379040956497, 0.764368712902069, 0.241186320781708, 0.149962991476059, 0.180177941918373, 0.0808692872524261, 0.200831666588783, 0.0349440351128578, -0.209531918168068, -0.0969634801149368, 0.640946865081787, 2.63783097267151, 0.455350935459137, 0.256195217370987, -0.916725933551788, -0.041004478931427, 1.3131844997406, -1.22410261631012, 0.805343806743622, 0.00833656638860703, -0.560337007045746, -0.201807156205177, 0.548599421977997, 0.334688812494278, 0.521740853786469, -0.0179979111999273, -0.641793966293335, -0.168227031826973, 0.278749495744705, 0.507410168647766, -0.372923135757446, 0.608035087585449, 0.767229676246643, 0.871033072471619, 0.715027272701263, -0.791702806949615, 0.530493497848511, 0.510580718517303, 0.484355986118317, -0.456425845623016, 0.18878436088562, -0.206953182816505, -0.187294632196426, 1.11711597442627, 0.0687117204070091, 0.191073343157768, 0.0952982977032661, -1.10354661941528, -0.400756925344467, 0.117253243923187 },
      { 0.627527594566345, 0.664235413074493, 0.937787115573883, 0.739275097846985, 0.544497907161713, 0.65521514415741, 0.622211694717407, 0.729329347610474, 0.745107531547546, 0.627147376537323, 0.702893555164337, 0.969854593276978, 0.0604759380221367, 0.63637113571167, 0.703387379646301, 0.610615968704224, 0.746407866477966, 0.89330667257309, 0.45554256439209, 0.482203125953674, 0.663394808769226, 0.576700508594513, 0.607546031475067, 0.647887051105499, 0.743319988250732, 0.655554711818695, 0.685093760490417, 0.680523872375488, 0.651630163192749, 0.676286995410919, 0.894063472747803, 0.588239669799805, 0.805068910121918, 1.11780917644501, 0.473429381847382, 0.442118465900421, 0.62340247631073, 0.678086519241333, 0.843228280544281, 0.537532150745392, 0.691696166992188, 0.652425467967987, 0.593374252319336, 0.590455949306488, 0.368218988180161, 0.710681140422821, 0.787499129772186, 0.796124279499054, 0.528773546218872, 0.634414076805115, 0.756084024906158 },
      { 0.000216341490158811, 0.206569164991379, -0.402120232582092, -0.116963990032673, -0.0501270182430744, 0.336023777723312, 0.0627793148159981, -0.188776731491089, 0.0363490805029869, 0.168281242251396, -0.0373084507882595, -0.253669440746307, 0.231269702315331, 0.05148845911026, 0.0527635961771011, -0.0385961420834064, 0.0990270599722862, -0.343547016382217, 0.547577202320099, -0.0579841025173664, 0.00295076100155711, 0.278725355863571, 0.266159802675247, -0.0606502294540405, -0.103916734457016, 0.0206597764045, -0.0230224896222353, 0.331616312265396, -0.0346910133957863, -0.162112906575203, -0.183303788304329, 0.289628803730011, -0.232239961624146, -0.586976051330566, -0.0325466729700565, -0.0112956240773201, -0.10690151154995, -0.100110851228237, -0.110605552792549, -0.0839195027947426, -0.0462591275572777, -0.0572230443358421, 0.0969951078295708, 0.188818350434303, -0.0320335105061531, -0.118115067481995, -0.0253174677491188, 0.0606871768832207, 0.336258977651596, 0.0417921990156174, 0.0784677192568779 },
      { -0.101603753864765, -0.114571504294872, 0.144088864326477, 0.0726895853877068, 0.132877692580223, -0.230462267994881, -0.0348655506968498, 0.14785735309124, 0.0323958620429039, -0.0704556331038475, 0.0157501399517059, 0.159205049276352, -0.047650933265686, 0.00292378198355436, -0.0278848931193352, 0.170015200972557, 0.00328022032044828, 0.0643028095364571, -0.241488218307495, 0.0376037172973156, 0.0934306904673576, -0.0927414298057556, -0.333945721387863, 0.054467961192131, 0.123281374573708, -0.0249125082045794, 0.0639416873455048, 0.0831557810306549, -0.0595892556011677, 0.117970772087574, 0.0972235202789307, -0.287054598331451, 0.077456921339035, 0.190949007868767, -0.0307888872921467, -0.00100634468253702, 0.0834526792168617, 0.00922968797385693, 0.0654247254133224, 0.0700051784515381, 0.0507780537009239, 0.0374331995844841, -0.0264444947242737, -0.00355736771598458, -0.0450270660221577, 0.059338990598917, 0.0351959206163883, -0.0637680739164352, -0.151052057743073, -0.0121401883661747, -0.019077692180872 },
      { 52.3306694030762, -129.446502685547, -69.5518341064453, 22.3157024383545, 19.3160972595215, -57.7459030151367, 54.0924186706543, 38.4434547424316, -74.6158142089844, 27.4656810760498, -125.125, -104.489891052246, -819.3818359375, -68.6642303466797, 12.8850002288818, -5.92875671386719, -41.9710426330566, 70.1460494995117, 139.33171081543, -42.1279335021973, 44.7834548950195, -130.854690551758, -15.1667623519897, -11.6069612503052, -6.41031074523926, -485.985931396484, -87.8765335083008, -369.999786376953, 42.1388320922852, 98.053581237793, -13.4754447937012, -10.3252754211426, -52.3407859802246, -107.121559143066, 105.879516601562, 58.6956443786621, -205.19108581543, -45.0392875671387, 108.268119812012, 80.3501434326172, -113.031776428223, 43.0331230163574, 48.876148223877, -98.8819580078125, 187.265502929688, 11.868447303772, -51.0638465881348, -125.974983215332, 40.3522605895996, -70.7803955078125, -24.7035102844238 }
    },
    {
      { -0.0513769313693047, 1.4441020488739, 0.110576622188091, -0.123937666416168, -0.157015383243561, 0.368757516145706, -0.0789524763822556, -0.180140778422356, 0.281394064426422, -0.231626108288765, 0.0129799507558346, -0.0750023126602173, 0.226250350475311, 0.0115805510431528, 0.255917131900787, -0.431566625833511, 0.70882111787796, -0.0494999550282955, 2.49460339546204, -0.239902645349503, 0.538465619087219, 1.8795508146286, 1.17451691627502, -0.111026830971241, -0.0350060500204563, 1.25486826896667, -0.0526423566043377, -0.246844813227654, 0.0827520564198494, -0.386601120233536, -0.0212652143090963, 1.13009202480316, 0.161525711417198, 0.102985173463821, -0.125602856278419, -0.0655369535088539, 0.490763127803802, -0.0303628351539373, 0.40231454372406, -0.392736196517944, 0.347970455884933, -0.387707859277725, -0.136059403419495, 1.52216708660126, -0.22290113568306, -0.171943739056587, -0.0925052314996719, -0.00908934790641069, 0.337145447731018, -0.344363272190094, 0.173218980431557 },
      { 0.473863273859024, -0.641335546970367, 0.0058945300988853, 0.138333812355995, 0.349287837743759, 0.202550441026688, 0.288846731185913, 0.199136197566986, -0.384069979190826, 0.695582270622253, 0.437283545732498, 0.651728868484497, -0.85492330789566, 0.0136054269969463, -0.197164177894592, 1.26930439472198, -0.588095545768738, -0.193610802292824, -1.02875363826752, -0.069041796028614, -0.46673846244812, -0.399235755205154, 0.129319682717323, -0.0795769765973091, -0.0635749325156212, -0.43788942694664, 0.275420427322388, 0.258212774991989, 0.487257272005081, 0.344558447599411, -0.22739340364933, -0.191208705306053, -0.0477930270135403, 0.0670929998159409, -0.150492414832115, -0.088622935116291, 0.169858738780022, -0.0306257903575897, 0.42457589507103, 0.275773704051971, 0.66193962097168, 0.238904491066933, 0.56169319152832, -0.518648087978363, -0.265158176422119, 0.31696093082428, -0.044646080583334, 0.0203758087009192, 0.609915614128113, 0.771645724773407, -0.380355328321457 },
      { -0.141447752714157, -0.529674887657166, -0.0514045543968678, 0.0809243097901344, -0.00368224829435349, -0.811339616775513, -0.0817015692591667, 0.129591777920723, 0.00221235863864422, -0.364837765693665, -0.116153173148632, -0.449299186468124, 0.580315768718719, 0.0628232210874557, 0.154069393873215, -0.869282305240631, -0.140061557292938, 0.413076370954514, -0.814264059066772, 0.586217045783997, -0.150941655039787, -0.986683249473572, -1.0660343170166, 0.351872771978378, 0.148142457008362, 0.489136755466461, -0.167503833770752, -0.175163343548775, -0.217472195625305, 0.180726647377014, 0.251430004835129, -0.583725929260254, -0.0546181239187717, -0.0960134789347649, 0.589962482452393, 0.506769359111786, -0.414138674736023, 0.295866936445236, -0.261965572834015, 0.287459254264832, -0.628113567829132, -0.0316854789853096, -0.30441626906395, -0.493444055318832, 0.888334274291992, 0.0270079299807549, 0.00391230545938015, 0.0584476180374622, -0.447863727807999, -0.31062725186348, 0.075103260576725 },
      { 0.82302588224411, 0.858971774578094, 0.925020396709442, 0.915596067905426, 0.756116330623627, 0.777179777622223, 0.86914587020874, 0.852948606014252, 0.94958508014679, 0.796965181827545, 0.85687792301178, 0.867864906787872, 0.982708513736725, 0.905195653438568, 0.945018589496613, 0.883685290813446, 0.999786853790283, 0.97257661819458, 0.958338499069214, 0.65403950214386, 0.754563212394714, 0.790508329868317, 0.900276660919189, 0.816073834896088, 0.909586787223816, 0.758740186691284, 0.887606382369995, 0.665143311023712, 0.829519331455231, 0.795160233974457, 1.02147102355957, 0.733206808567047, 0.91886705160141, 0.896476149559021, 0.64906895160675, 0.612812280654907, 1.00247383117676, 0.862346053123474, 0.586630523204803, 0.72550231218338, 0.914366781711578, 0.837073385715485, 0.838174819946289, 0.776218295097351, 0.533117890357971, 0.906933605670929, 0.959062933921814, 0.964507758617401, 0.706598341464996, 0.733789503574371, 0.966383755207062 },
      { -0.0296034775674343, 0.42651641368866, 0.019955700263381, -0.10616672039032, -0.0688209608197212, 0.0893312469124794, -6.47807319182903e-05, -0.141736328601837, 0.115463264286518, 0.21860733628273, -0.0289499200880527, -0.00922534056007862, -0.447770148515701, 0.177766069769859, 0.049222432076931, -0.20224317908287, 0.245223924517632, -0.137988373637199, -0.182717740535736, -0.03232317045331, 0.128234624862671, 0.341620117425919, 0.250331193208694, -0.0646496936678886, -0.0798259526491165, -0.131786361336708, 0.0129753649234772, 0.195104390382767, -0.0477035492658615, -0.107998721301556, -0.129064947366714, 0.859048128128052, 0.0312104262411594, 0.0262611974030733, -0.0152408545836806, 0.0118854213505983, -0.0460725128650665, -0.0710692778229713, -0.0183940920978785, -0.0793222934007645, -0.0732799246907234, -0.00992235448211432, 0.0389004275202751, 0.388827472925186, -0.0139525113627315, -0.112354569137096, -0.00882372073829174, 0.0879096388816833, 0.580152153968811, 0.200997963547707, 0.193260505795479 },
      { -0.0408279746770859, -0.285076081752777, -0.0146341733634472, 0.0641901344060898, 0.125995233654976, 0.196217328310013, -0.00761431222781539, 0.104823879897594, -0.0233199913054705, -0.159192055463791, -0.000868676754180342, 0.011350424028933, 0.69937527179718, -0.137938812375069, -0.0597439929842949, 0.317822247743607, -0.212804779410362, -0.0381797254085541, 0.57159948348999, 0.0276485569775105, 0.132754981517792, -0.0562724396586418, -0.298132359981537, 0.0618717521429062, 0.09902273863554, 0.00689340569078922, 0.0448214821517467, -0.0228934418410063, -0.0115848295390606, 0.0703375861048698, 0.0854154750704765, -1.100182056427, -0.0137908086180687, 0.00365025969222188, -0.0177604425698519, -0.00798691157251596, 0.0653654560446739, -0.00273673608899117, 0.0364678800106049, 0.0672101378440857, 0.0877235606312752, 0.0865223929286003, -0.00821104925125837, -0.0266562923789024, -0.0446604341268539, 0.0431863442063332, 0.0309640951454639, -0.10249900072813, -0.778103828430176, -0.0375315807759762, -0.184318348765373 },
      { 56.9494323730469, -190.377136230469, -71.4398956298828, 18.1140804290771, 23.5387916564941, 615.438537597656, 21.4042091369629, 41.011589050293, -62.5326461791992, 14.6782236099243, -105.875625610352, -44.5517387390137, -276.755065917969, -21.1543121337891, -2.04952311515808, 17.7862434387207, -31.106481552124, 34.3352165222168, 49.2667503356934, -61.6632499694824, 93.9191284179688, 23.582160949707, 21.2283267974854, -3.60014653205872, -1.25991582870483, -744.399841308594, -87.9661102294922, 1007.30206298828, 37.8233413696289, 87.4657516479492, -11.0870885848999, -20.853832244873, -77.8645782470703, -55.6055641174316, 76.8302307128906, 28.3075408935547, -324.983551025391, -33.1192245483398, 535.33349609375, 78.963981628418, -177.240524291992, 35.9114265441895, 23.0696029663086, -196.699111938477, 179.384353637695, 20.0837821960449, -46.4833488464355, 5.1133337020874, 28.5349407196045, -290.256683349609, -27.8343162536621 }
    },
    {
      { 0.0301110744476318, 0.199598714709282, 0.0486482828855515, -0.0415577068924904, -0.0395406186580658, 0.148892894387245, -0.026266997680068, -0.0623245425522327, 0.0642528533935547, -0.104934751987457, -0.0228520892560482, 0.751485586166382, 0.0656356662511826, 0.00279472442343831, 0.0665837302803993, -0.137247830629349, 0.155246421694756, 0.101811066269875, 0.393073469400406, -0.085310623049736, 0.0626604110002518, 0.343043923377991, 0.347860842943192, -0.0410452447831631, -0.0148438308387995, 0.207409217953682, -0.0837975591421127, 0.173512667417526, 0.054238174110651, -0.135993346571922, -0.0192417241632938, 0.171599686145782, 0.0515491887927055, 0.0898872092366219, -0.0407730713486671, -0.0211085751652718, 0.212673142552376, -0.0170977003872395, -0.155204772949219, -0.138543218374252, 0.0627510473132133, -0.115189388394356, -0.0466858185827732, 0.256866961717606, -0.0775732696056366, -0.0659368932247162, -0.0281468089669943, -0.0042417123913765, -0.0508800558745861, -0.0849065184593201, 0.0632204860448837 },
      { 0.152777418494225, -0.044992096722126, -0.34482479095459, 0.039558220654726, 0.0833186060190201, 0.103006616234779, 0.0842302441596985, 0.10768011957407, -0.100021950900555, 0.225983813405037, 0.171480342745781, -1.08421587944031, -0.250560820102692, -0.00292656081728637, -0.0357968956232071, 0.479115068912506, -0.133998528122902, -0.494712769985199, -0.194815143942833, -0.0217569228261709, -0.0482089519500732, -0.0427301861345768, 0.0321560688316822, -0.040518656373024, -0.0478368811309338, -0.0403082221746445, 0.107361651957035, 0.123075343668461, 0.157989129424095, 0.11783142387867, -0.0961775928735733, 0.06855408847332, -0.162147700786591, -0.492333263158798, -0.0544062070548534, -0.0315233133733273, 0.0296279229223728, -0.0154587477445602, 0.187302678823471, 0.0927950367331505, 0.168746098875999, 0.116977542638779, 0.163052156567574, -0.0892320275306702, -0.0888924151659012, 0.101383671164513, -0.0474340803921223, 0.0106495581567287, 0.220294326543808, 0.205903828144073, -0.104075044393539 },
      { -0.0431749559938908, -0.0974707901477814, 0.370025515556335, 0.0334431193768978, 0.0278759207576513, -0.25344043970108, -0.0174404103308916, 0.00289785023778677, 0.0168071109801531, -0.102946624159813, -0.0360364019870758, 0.215048238635063, 0.142450839281082, 0.0334040969610214, 0.0239313095808029, -0.326707094907761, -0.0164608154445887, 0.4190534055233, -0.212864279747009, 0.199788808822632, -0.00117823481559753, -0.202274769544601, -0.321628332138062, 0.12815748155117, 0.0571583174169064, 0.060850016772747, -0.0215568002313375, -0.204265013337135, -0.0752079337835312, 0.0653005018830299, 0.0893284380435944, -0.144261687994003, 0.212421834468842, 0.429985791444778, 0.190682232379913, 0.173774167895317, -0.147780001163483, 0.11796635389328, -0.0931892320513725, 0.0990012586116791, -0.132276639342308, 0.0276109594851732, -0.0752512216567993, -0.0934374630451202, 0.305512577295303, 0.0153858829289675, 0.0237146131694317, 0.0218303650617599, -0.172526970505714, -0.0669671446084976, 0.0184410531073809 },
      { 0.931760609149933, 0.961363911628723, 1.18265283107758, 0.971300482749939, 0.920024156570435, 0.998758614063263, 0.941146075725555, 0.955587327480316, 0.994929432868958, 0.941395699977875, 0.948176264762878, 1.21994924545288, 1.0112429857254, 0.958177268505096, 0.96760368347168, 0.956264495849609, 0.991136908531189, 1.04437828063965, 1.00518488883972, 0.883004188537598, 0.943063795566559, 0.97122597694397, 0.968228936195374, 0.933904469013214, 0.973450720310211, 0.944415628910065, 0.967616021633148, 0.971424639225006, 0.943966329097748, 0.928116798400879, 1.02453899383545, 0.943062126636505, 1.05985689163208, 1.31969451904297, 0.88561600446701, 0.867898225784302, 1.00600552558899, 0.935150384902954, 0.915238320827484, 0.906769812107086, 0.969629526138306, 0.9321009516716, 0.939116895198822, 0.964538633823395, 0.838012278079987, 0.963461518287659, 0.992091357707977, 0.971293270587921, 0.912881255149841, 0.910946130752563, 0.983472406864166 },
      { -0.00604347605258226, 0.0427915006875992, -0.397792607545853, -0.0317093469202518, -0.0140720596536994, -0.127242848277092, 0.0290387682616711, -0.0557757504284382, 0.00320293521508574, 0.0482982471585274, -0.00993929523974657, -0.242430746555328, -0.164900884032249, 0.0339343883097172, 0.0176969468593597, -0.0729445740580559, 0.0332979448139668, -0.152772977948189, -0.101437129080296, -0.00953720696270466, -0.00509071629494429, 0.00958239007741213, 0.0639162510633469, -0.0156696382910013, -0.0337060280144215, -0.011690097860992, -0.0183594990521669, -0.0638497024774551, -0.0201578699052334, -0.0324414148926735, -0.0413543693721294, 0.100217834115028, -0.250077277421951, -0.53685736656189, -0.00126189505681396, 0.00701178656890988, -0.015992185100913, -0.0119092045351863, -0.010317724198103, -0.0258048623800278, -0.0230324883013964, -0.00607351493090391, 0.0204107388854027, 0.0236396715044975, -0.00274842977523804, -0.0334583446383476, -0.0176458954811096, 0.030994338914752, 0.182152047753334, 0.019487988203764, 0.0133103020489216 },
      { -0.0299436580389738, -0.00615760870277882, 0.140857115387917, 0.018733924254775, 0.0232841204851866, 0.242561489343643, -0.0176860075443983, 0.0504708327353001, 0.0149088203907013, -0.0282686818391085, 0.00379203376360238, 0.132528126239777, 0.271575063467026, -0.0161835141479969, -0.00729884766042233, 0.106109917163849, -0.0234232544898987, 0.0884221494197845, 0.269904106855392, 0.00851781386882067, 0.0334371849894524, 0.0290595162659883, -0.0756538286805153, 0.0196771007031202, 0.0510277040302753, 0.0126006808131933, 0.0301699675619602, 0.187768280506134, -0.0118936058133841, 0.0199259053915739, 0.0264845844358206, -0.156432181596756, 0.0887170732021332, 0.179455995559692, -0.00441428925842047, -0.0048386650159955, 0.0215235017240047, -0.0087973028421402, 0.0065645077265799, 0.0232049338519573, 0.0225855782628059, 0.00587032875046134, -0.00687639741227031, 0.049559623003006, -0.0167032666504383, 0.0160675700753927, 0.0317967273294926, -0.0319605991244316, -0.229749545454979, 0.00345294177532196, 0.00522549636662006 },
      { 27.3693866729736, -45.8137054443359, -43.104362487793, 6.56719017028809, 6.68450355529785, -8.04440879821777, 11.4481248855591, -0.0420891493558884, -41.0890159606934, 5.89526605606079, -40.8876037597656, -74.4263458251953, -86.7479934692383, -5.03716373443604, -9.01763153076172, 9.71935653686523, -13.1401329040527, 23.4947128295898, -23.8985748291016, -21.385570526123, 19.0242404937744, -7.76312875747681, 7.02602434158325, -0.445440500974655, 0.395374357700348, -185.923385620117, -34.6211013793945, -9.64646625518799, 21.5864486694336, 32.7262077331543, -1.7020800113678, -0.548571527004242, -26.1652355194092, -47.1197891235352, 14.2206830978394, 8.46063041687012, -113.605888366699, -5.36231184005737, 310.734283447266, 28.8297786712646, -47.5005226135254, 10.4878578186035, 12.8459444046021, -45.8639335632324, 64.3148422241211, 9.71772193908691, -22.8528156280518, -4.14370203018188, 19.9773712158203, 0.144677773118019, -15.5589275360107 }
    },
    {
      { -0.037905391305685, 0.0945251956582069, -0.00360233942046762, 0.0087510971352458, 0.00884975865483284, 0.0156197482720017, -0.0295649748295546, 0.0260043852031231, 0.0306409392505884, 0.0201126895844936, -0.0129190133884549, -0.438672751188278, -0.0166594795882702, 0.00407012552022934, 0.0461518540978432, 0.0164437163621187, 0.032513901591301, -0.0539214126765728, 0.171958476305008, 0.0331328399479389, 0.067889004945755, 0.0824854373931885, -0.0772853270173073, 0.0162835437804461, 0.0178257375955582, 0.0992886200547218, 0.0678635314106941, 0.0151591887697577, -0.0367104075849056, 0.0501548610627651, 0.0121837025508285, 0.0509686842560768, -0.013949254527688, -0.0452299155294895, 0.0330416411161423, 0.0166969541460276, -0.0678908005356789, 0.0124000292271376, 0.0246323514729738, 0.0492013990879059, 0.0311726443469524, 0.0362250842154026, -0.0203529261052608, 0.0879512876272202, 0.0298894625157118, 0.0136826746165752, 0.0403130501508713, -0.00145583716221154, 0.0231342315673828, 0.069199450314045, 0.00412113638594747 },
      { -0.0273617133498192, -0.0932160541415215, 0.199632912874222, -0.000176797795575112, -0.010826569981873, -0.0270018260926008, 0.0526137538254261, -0.0458093769848347, 0.00278004165738821, -0.0294353123754263, 0.0162915252149105, 0.85974383354187, -0.0594769828021526, 0.000794197956565768, -0.0852812752127647, -0.078598715364933, -0.0293850880116224, 0.1976128667593, -0.119201950728893, -0.00398802384734154, -0.0761391445994377, -0.0826264768838882, -0.00131909362971783, -0.0110781593248248, -0.00662712706252933, -0.0471462532877922, -0.0536701753735542, -0.0174216404557228, -0.0270365327596664, -0.0491773374378681, -0.00243580527603626, -0.0778320357203484, 0.121479585766792, 0.245696008205414, -0.015043007209897, -0.00307512213476002, -0.00871901120990515, -0.00517541123554111, 0.00864148046821356, -0.0388156026601791, -0.0413529351353645, -0.0612799637019634, 0.0415689200162888, -0.0448928996920586, 0.0181367788463831, -0.00609820289537311, -0.019417917355895, -0.00696447305381298, -0.0454111471772194, -0.210605472326279, 0.00662226928398013 },
      { 0.00936233997344971, 0.0179476123303175, -0.222304165363312, -0.0136462533846498, -0.0225618947297335, -0.0233616847544909, -0.0265038069337606, 0.00807257927954197, -0.0175853278487921, 0.00566504988819361, -0.0230283811688423, -0.344996780157089, 0.0713118463754654, -0.0150276282802224, 0.0548472031950951, 0.0538909062743187, -0.00152171717491001, -0.132102623581886, -0.0719336047768593, -0.0523986108601093, -0.0170634854584932, 0.00166666880249977, 0.0647960677742958, -0.0183240920305252, -0.00803334172815084, 0.0362142436206341, -0.0155185125768185, 0.0339761935174465, 0.0216671247035265, -0.00803802534937859, -0.0125120896846056, 0.0255426280200481, -0.160483688116074, -0.214182034134865, -0.0380482412874699, -0.0436084046959877, 0.0450083017349243, -0.0217790715396404, -0.00682974979281425, -0.0234099738299847, 0.00069227657513693, -0.00598774757236242, -0.023733688518405, -0.0244256313890219, -0.0900035575032234, -0.0150795774534345, -0.00550217879936099, -0.000395910668885335, 0.111589297652245, 0.09681785851717, -0.005643623881042 },
      { 1.01796615123749, 1.00237572193146, 0.910510122776031, 1.00334596633911, 1.01132810115814, 0.979126572608948, 1.01537311077118, 1.01065814495087, 0.993611752986908, 1.00192666053772, 1.0063236951828, 0.891778469085693, 0.986181616783142, 1.0103976726532, 1.00891220569611, 1.00739777088165, 0.998313784599304, 0.971904754638672, 1.01093506813049, 1.02972495555878, 1.00167727470398, 0.97946161031723, 1.0058354139328, 1.01570522785187, 1.00126695632935, 1.00665521621704, 1.00530064105988, 0.975798785686493, 1.00994384288788, 1.01657462120056, 0.999866902828217, 0.995352387428284, 0.973714649677277, 0.858775913715363, 1.03025925159454, 1.03418707847595, 0.994252264499664, 1.00995337963104, 0.978943169116974, 1.02714717388153, 1.00495982170105, 1.02162349224091, 1.00861418247223, 0.983847439289093, 1.04835844039917, 1.0023934841156, 1.00303626060486, 1.00759661197662, 0.995104372501373, 0.994199693202972, 1.00043475627899 },
      { -0.00186180509626865, 0.0362403765320778, 0.183980792760849, 0.0108890775591135, 0.00974032841622829, 0.0399897992610931, -0.0225930437445641, 0.016902968287468, 0.0171326138079166, 0.0116649065166712, 0.0123642245307565, 0.0822239741683006, 0.157880023121834, 0.0117605682462454, 0.00263070594519377, 0.0167357064783573, 0.0265116803348064, 0.0555220805108547, -0.0197422243654728, 0.00642137555405498, 0.0273064039647579, 0.0581484772264957, -0.0136424982920289, 0.00336404936388135, 0.00768534420058131, -0.00515647092834115, 0.0213140286505222, 0.0404377765953541, -0.000726127123925835, 0.0112931234762073, 0.0104013914242387, 0.0690092742443085, 0.120649456977844, 0.232821419835091, 0.0032708402723074, -0.00207143253646791, 0.00534332869574428, 0.00582823343575001, 0.0422026105225086, 0.00794364884495735, 0.0117984153330326, 0.00230592652224004, -0.00887822918593884, 0.0517009906470776, 0.00242898240685463, 0.0140652628615499, 0.0107210837304592, 0.000347027846146375, 0.0272130034863949, 0.0355485901236534, 0.0223822966217995 },
      { 0.015701811760664, -0.0328568480908871, -0.0691820681095123, -0.00460388464853168, 0.000394242379115894, -0.00793843436986208, 0.0126469396054745, -0.0151182133704424, -0.0119053749367595, -0.00680655520409346, -0.0163355935364962, -0.0531005226075649, -0.201324284076691, -0.0143003594130278, -0.00733870780095458, -0.0209584143012762, -0.0206104293465614, -0.0473051406443119, 0.0194059424102306, -0.00405981298536062, -0.00186074269004166, -0.0335822813212872, 0.0213030166924, -0.00345114129595459, -0.00548794725909829, -0.00779467029497027, -0.0139720262959599, -0.00269586406648159, 0.0129480008035898, -0.0073222448118031, -0.00492417439818382, -0.0794441476464272, -0.0408736951649189, -0.0724661275744438, -0.00328348111361265, 0.00271029840223491, -0.00647691590711474, 0.000287541566649452, -0.0250775534659624, -0.0078116562217474, -0.00692333607003093, 0.00638095801696181, 0.00487911561504006, -0.0232556760311127, 0.00464024161919951, -0.00458325585350394, -0.0110363224521279, 0.0018636571476236, -0.0470857471227646, -0.0041268477216363, -0.0229958761483431 },
      { -3.78997206687927, -1.40861487388611, -7.41892862319946, -3.13936018943787, 0.411217629909515, 63.2481536865234, -2.68469142913818, 3.58103656768799, 10.3896722793579, 1.09773504734039, 21.6920833587646, 50.0797996520996, 4.26781368255615, 1.58437049388885, -5.00416564941406, 0.324482887983322, 6.97885036468506, -6.9772572517395, 42.8984222412109, 5.83014249801636, 2.90662312507629, 34.95361328125, 0.580464541912079, 2.29768109321594, -0.633558928966522, -117.109367370605, 10.4633598327637, 41.6133117675781, 4.9993577003479, -7.31603860855103, 1.08115458488464, 1.62026214599609, -15.5511198043823, 9.29462337493896, -0.0776899382472038, -0.0504878796637058, 40.8125877380371, 1.17158305644989, -67.9477386474609, -6.63902521133423, 12.2866554260254, -4.05698776245117, -3.63965463638306, 4.62913513183594, -19.9347743988037, -4.25759744644165, 5.93426561355591, 8.84447002410889, 1.98291015625, -27.2907752990723, 0.562733709812164 }
    },
    {
      { 0.0461277030408382, 0.0234554875642061, 0.00387765699997544, -0.00479838903993368, -0.00300953653641045, -0.0419230274856091, -0.0642767399549484, -0.00107343681156635, 0.0120191024616361, -0.030023779720068, 0.0586219429969788, 0.0450994893908501, -0.00118819868657738, -0.0144834043458104, -0.0318279154598713, -0.0444203019142151, 0.0295817404985428, -0.0240404400974512, -0.0515961535274982, -0.0370872281491756, 0.0821679681539536, -0.0419740490615368, -0.0103006847202778, 0.0149444490671158, 0.0388110131025314, -0.0580331049859524, -0.0138194086030126, 0.00494831847026944, 0.00798554811626673, -0.0168166682124138, 0.0239946525543928, -0.0101662203669548, 0.00537318363785744, 0.029568325728178, -0.105645403265953, -0.0120258657261729, 0.00315358676016331, 0.0468700341880322, -0.0609715022146702, 0.00568703236058354, 0.05475889518857, -0.0306865368038416, -0.0488482266664505, -0.0149190872907639, 0.00996452104300261, -0.050143651664257, 0.0534543730318546, -0.00873606372624636, -0.0363215319812298, 0.0457840971648693, 0.0191196985542774 },
      { -0.117673739790916, -0.0487845651805401, -0.0117404097691178, -0.00648590736091137, -0.0211340077221394, -0.001801832113415, 0.11926182359457, -0.0887583717703819, -0.122722007334232, -0.0298182871192694, -0.251234829425812, -0.0565595403313637, 0.00820011645555496, 0.0088458564132452, 0.0560721419751644, 0.0423223003745079, -0.0357215888798237, -0.113055944442749, 0.0267473086714745, 0.0270577389746904, -0.134150356054306, -0.0131451524794102, 0.0148722529411316, -0.0680119246244431, -0.0782358795404434, 0.0238765217363834, -0.0718269497156143, 0.00238598487339914, -0.0241451859474182, 0.0331725291907787, -0.0704495683312416, 0.0123386662453413, -0.0178074296563864, -0.0133646130561829, 0.103347361087799, 0.0331671312451363, -0.00511844409629703, -0.077909380197525, 0.0455378741025925, 0.0319752842187881, -0.200993612408638, 0.0419344864785671, 0.0412717685103416, -0.0451690666377544, -0.0642252787947655, 0.0257338397204876, -0.127160459756851, 0.0286057759076357, 0.00984166283160448, -0.104928396642208, -0.127107873558998 },
      { 0.056741040199995, 0.0239651221781969, 0.00876673217862844, 0.0478989891707897, 0.0251423753798008, 0.023616224527359, -0.0335175842046738, 0.0232022330164909, 0.0635369941592216, 0.0569343604147434, 0.162240773439407, 0.027237955480814, -0.00502142310142517, 0.0311194788664579, -0.0110401790589094, -0.0188461598008871, 0.0099021103233099, 0.295387834310532, -0.00486573809757829, -0.00094345019897446, 0.0529214031994343, 0.036573126912117, -0.00436324812471867, 0.0586064048111439, 0.0495331026613712, 0.00129470380488783, 0.0563482232391834, -0.0173419415950775, 0.0185524709522724, -0.0271626822650433, 0.0517825037240982, 0.0135929910466075, 0.0229428093880415, -0.0290168318897486, 0.000840020715259016, -0.0419656820595264, 0.00339292269200087, 0.0578126385807991, -0.00708492891862988, -0.0447554402053356, 0.104298613965511, -0.0739249140024185, 0.0219233613461256, 0.0317700169980526, 0.135327413678169, 0.0229558870196342, 0.0462147928774357, -0.00699919415637851, -0.0149060310795903, 0.0551578365266323, 0.0709725171327591 },
      { 0.0160296894609928, 0.00540772499516606, -0.0143036432564259, 0.00375072727911174, 0.00356712401844561, -0.00128683482762426, -0.0490323044359684, 0.0480252578854561, 0.0110819572582841, -0.0119453119114041, 0.0106371752917767, -0.03751615062356, 9.13999756448902e-05, -0.00493136374279857, -0.0169709194451571, -0.00555396499112248, 0.00435347715392709, -0.0366783365607262, -0.00170808471739292, 0.0109475459903479, 0.0250159595161676, 0.00757963536307216, -0.0055444585159421, -0.00169894110877067, -0.00296456436626613, -0.000743797223549336, 0.00733654433861375, 0.00253961328417063, 0.00915304478257895, 0.0217807982116938, 0.0151634011417627, -0.00485667586326599, -0.0212875809520483, 0.000818673463072628, -0.00183924520388246, 0.0101826572790742, -0.00751975504681468, -0.00297071901150048, -0.0136976363137364, 0.0232422649860382, 0.00106512370985001, 0.0144198350608349, -0.034317746758461, 0.00595742790028453, -0.0620202235877514, -0.0152179272845387, 0.00913666747510433, -0.00281211803667247, -0.00866151042282581, 0.0124811958521605, 0.0158561393618584 },
      { 0.90239429473877, 0.920259118080139, 0.97216123342514, 0.968555152416229, 0.948054015636444, 0.937448501586914, 1.01590764522552, 0.919883072376251, 0.950585186481476, 0.973458766937256, 0.934502840042114, 0.977301418781281, 0.941056072711945, 0.958431541919708, 0.980622887611389, 0.926243960857391, 0.941475450992584, 0.90044754743576, 0.932416260242462, 0.952900767326355, 0.915154278278351, 0.919832348823547, 0.961236476898193, 0.958157181739807, 0.947206318378448, 0.984254896640778, 0.92948716878891, 0.922028601169586, 0.894081354141235, 0.940599322319031, 0.940510511398315, 0.93316239118576, 0.969916224479675, 0.969263553619385, 0.94996440410614, 0.959529519081116, 0.953340649604797, 1.00738716125488, 0.979641079902649, 0.925217032432556, 0.95024961233139, 0.963077366352081, 1.0038104057312, 0.928245365619659, 0.970599830150604, 0.949991464614868, 0.948135197162628, 0.992158114910126, 0.955953359603882, 0.914550364017487, 0.958413302898407 },
      { 0.0918208137154579, 0.0758096799254417, 0.0287768058478832, 0.0022899794857949, 0.0446797795593739, 0.0646644234657288, -0.00293092452920973, 0.0680581256747246, 0.0467628464102745, 0.0249515920877457, 0.0800969898700714, 0.035600695759058, 0.052103478461504, 0.0235949531197548, 0.00834097899496555, 0.0839215591549873, 0.0547362305223942, -0.0739042684435844, 0.0703920423984528, 0.0483701713383198, 0.0729939714074135, 0.0792625024914742, 0.0336530767381191, 0.0329835638403893, 0.0472720004618168, -0.0112250056117773, 0.0655582994222641, 0.0918351411819458, 0.108771778643131, 0.0532571487128735, 0.0401077270507812, 0.0616409704089165, 0.0274809189140797, 0.0303497593849897, 0.0322252884507179, 0.051146537065506, 0.0563435778021812, -0.0293136220425367, 0.0186298582702875, 0.0673162192106247, 0.0654600411653519, 0.0415656939148903, 0.00268446351401508, 0.0764209479093552, 0.0195935480296612, 0.0492205247282982, 0.0482880286872387, -0.00593781936913729, 0.0625875666737556, 0.0886300727725029, 0.0268028751015663 },
      { 9.3794116973877, 6.99833154678345, 7.05353355407715, -5.50295114517212, -5.36774873733521, -3.11339139938354, 21.0359439849854, 6.89768695831299, 11.4276180267334, 8.14081573486328, -6.83690071105957, -4.24995183944702, -3.64893531799316, -10.8201656341553, 4.47472286224365, 9.64506149291992, -9.29953575134277, 22.1660594940186, 1.84683632850647, -66.3639373779297, -4.09433317184448, -10.1492586135864, -11.3706359863281, 1.44899129867554, 2.46952342987061, 73.2162322998047, 9.10573101043701, -11.8251895904541, 4.14682292938232, -23.382043838501, 0.374557793140411, 1.28306674957275, 12.5128793716431, -7.63174676895142, 16.4087982177734, -50.646614074707, -9.52901649475098, -1.87579596042633, 11.1067085266113, -5.805739402771, 7.77080631256104, 16.5394554138184, 15.2962474822998, -1.24888622760773, -6.89144039154053, 6.45083856582642, 9.02828025817871, -24.2169704437256, -1.08877944946289, -16.541805267334, 6.69350814819336 }
    },
    {
      { -0.132921054959297, -0.144853636622429, 0.0640141293406487, 0.320276468992233, 0.0218300651758909, -0.0361033156514168, 0.465620785951614, 0.394672840833664, -0.0500308312475681, 0.000451465079095215, 0.147513210773468, 0.311659663915634, -0.0270443242043257, 0.0275439396500587, 0.158400774002075, -0.0791506543755531, 0.0515735037624836, 0.363533526659012, 0.0404715277254581, 0.513789653778076, -0.308560341596603, -0.0918288975954056, -0.142314806580544, 0.118409305810928, -0.0425170920789242, 0.313301116228104, -0.0677627176046371, -0.113306574523449, -0.0148152047768235, 0.626720726490021, -0.0335852429270744, 0.033820491284132, 0.056793250143528, 0.0425790287554264, 0.489664942026138, 0.389389187097549, -0.105125702917576, -0.163395196199417, 0.169931426644325, 0.663617730140686, 0.108178809285164, 0.209151193499565, 0.292173057794571, -0.0213946327567101, -0.0807588696479797, 0.203260183334351, -0.00931133236736059, -0.112549357116222, 0.0109138786792755, -0.163532301783562, 0.068952277302742 },
      { 0.282143443822861, -0.2997205555439, -0.192250207066536, -0.256474435329437, -0.309365272521973, -0.00213740370236337, -1.03962314128876, 0.370144605636597, 0.0710490345954895, 0.193377032876015, 0.173192247748375, -0.692614316940308, 0.392494827508926, -0.0607551708817482, -0.302605450153351, 0.385114639997482, -0.022595077753067, 0.23827999830246, -0.0356386378407478, -0.288645476102829, 0.336725473403931, 0.00291048921644688, 0.0134823201224208, 0.065613254904747, -0.0296708550304174, 0.425637722015381, 0.0386919267475605, 0.0117430537939072, 0.111749514937401, -0.612697243690491, 0.0117901507765055, -0.188274785876274, -0.250092715024948, -0.282769531011581, -0.00892826355993748, -0.322064131498337, 0.0295320358127356, 0.248128831386566, 0.0494518652558327, -0.810185730457306, -0.201947078108788, -0.19013486802578, -0.567074716091156, -0.217257723212242, 0.688409805297852, 0.24946491420269, 0.184263989329338, 0.0174407735466957, 0.146652668714523, 0.231497138738632, 0.182905822992325 },
      { -0.139572530984879, 0.207147166132927, 0.163517445325851, -0.124156624078751, 0.216268509626389, 0.00863119773566723, 0.611172258853912, -0.241534307599068, -0.00522874342277646, -0.235186949372292, -0.146546602249146, 0.256132990121841, -0.322185873985291, -1.91582530533196e-05, 0.207124143838882, -0.33989754319191, -0.0303302649408579, -1.18487298488617, -0.0386183969676495, -0.174726396799088, 0.00459009129554033, 0.0104116406291723, 0.0443200916051865, -0.0431628003716469, 0.0921675190329552, -0.451441138982773, 0.0328001007437706, -0.00847482681274414, -0.0873298272490501, -0.00205445429310203, -0.00928657036274672, 0.0675760805606842, 0.254524737596512, 0.234836101531982, -0.298571705818176, 0.0103423111140728, 0.015636719763279, -0.113912083208561, 0.0109394807368517, 0.267575562000275, 0.10585056245327, -0.194253116846085, 0.379227966070175, 0.109151490032673, -1.37786531448364, -0.471119999885559, -0.134527623653412, 0.0606358237564564, -0.27503490447998, -0.0761020705103874, -0.198536202311516 },
      { -0.0432418249547482, 0.0742650032043457, 0.130235642194748, 0.046088669449091, 0.0938893854618073, -0.000478899601148441, 0.310934633016586, -0.190048158168793, 0.016338225454092, 0.0244782660156488, -0.0163728054612875, 0.212007388472557, 0.0118612302467227, -0.0057984683662653, 0.0199561435729265, -0.0228433310985565, -0.00848639197647572, 0.262831687927246, 0.00804200861603022, 0.187333285808563, -0.0536690466105938, 0.00974649470299482, 0.0117705436423421, -0.0496705397963524, -0.0200123824179173, 0.0440593026578426, 0.025273572653532, 0.00327672017738223, -0.0387542396783829, 0.0703691616654396, 0.0960733592510223, -0.00336422468535602, 0.08702152967453, 0.157702699303627, -0.011614260263741, 0.18511700630188, 0.00795241631567478, -0.0436151549220085, 0.0144139984622598, 0.122424870729446, -0.0135653670877218, -0.180991366505623, 0.206700995564461, 0.0372811295092106, 0.72475004196167, 0.0921058654785156, -0.0343144200742245, -0.0170938596129417, -0.0326383858919144, -0.0229600612074137, -0.00192732166033238 },
      { 0.174995079636574, 0.071467712521553, -0.0455427207052708, -0.0125574357807636, -0.0302974469959736, 0.068191334605217, -0.4081971347332, 0.163146018981934, 0.0386605970561504, -0.0421330183744431, 0.132354855537415, 0.0292712170630693, 0.291258662939072, 0.00541244354099035, 0.00459242518991232, 0.0732426866889, 0.0444709993898869, 0.249336302280426, 0.0448056645691395, -0.0280916169285774, 0.150573387742043, 0.0901953354477882, 0.0354539342224598, 0.0519709885120392, -0.0580372735857964, -0.0577233396470547, 0.0893310457468033, 0.0559077225625515, 0.205832898616791, 0.026844235137105, -0.0600829571485519, 0.14075069129467, -0.0359958596527576, -0.0323551818728447, -0.11567460000515, -0.0871070697903633, 0.106247156858444, -0.0958863198757172, -0.0992188900709152, 0.123024456202984, 0.143028825521469, 0.393896490335464, -0.311364680528641, 0.111374199390411, 0.0305076073855162, -0.106206104159355, 0.111498795449734, 0.0430896990001202, 0.16240730881691, 0.065146878361702, 0.0475576743483543 },
      { 0.77638190984726, 0.880255937576294, 0.91465038061142, 0.985378265380859, 0.900543630123138, 0.90360689163208, 1.0915355682373, 0.796636760234833, 0.898903846740723, 1.04048466682434, 0.790254354476929, 0.920289099216461, 0.613628745079041, 0.957529246807098, 0.904316902160645, 0.961540102958679, 0.928727686405182, 1.19765114784241, 0.949999988079071, 0.798633575439453, 0.766729474067688, 0.853450894355774, 0.940598666667938, 0.864823400974274, 1.02237176895142, 1.05151522159576, 0.825377881526947, 0.916048347949982, 0.752172231674194, 0.9257892370224, 0.938807249069214, 0.815941333770752, 0.925707876682281, 0.911430656909943, 1.05849766731262, 0.796893239021301, 0.828489482402802, 1.06909620761871, 1.00012350082397, 0.69572901725769, 0.782736659049988, 0.669241487979889, 1.02665042877197, 0.813700199127197, 0.870004594326019, 1.08874380588531, 0.86992621421814, 0.890945076942444, 0.866825103759766, 0.871031582355499, 0.903217017650604 },
      { -3.49188423156738, -28.8264713287354, -92.7678833007812, 20.484712600708, 40.2478103637695, 42.8165740966797, -99.7446365356445, -198.21662902832, -27.2359828948975, 1.29982972145081, 30.5184745788574, -83.71533203125, 114.315940856934, 51.8320465087891, 6.87306594848633, -9.94032192230225, 33.1278610229492, -99.3484954833984, 26.2084617614746, 396.446899414062, -10.5080804824829, 14.3045768737793, 31.4384136199951, -15.097806930542, 1.09321451187134, -808.013549804688, -31.0236434936523, 70.5676116943359, -46.604061126709, -21.6220512390137, 15.6455450057983, 34.6695556640625, -58.336353302002, -58.2854843139648, 168.701751708984, 448.722961425781, 1.90622842311859, 65.022705078125, -323.465728759766, -2.27010583877563, 92.0282745361328, 130.161437988281, -52.7686996459961, 2.84730815887451, -123.714981079102, -3.99010109901428, -41.0550346374512, 70.3466720581055, 12.2069606781006, 104.306465148926, -4.41847038269043 }
    }
  },
  {
    {
      { 0.732901871204376, 0.673791229724884, 0.849120616912842, 0.786286234855652, 0.762789607048035, 0.877817273139954, 0.791789293289185, 0.899239838123322, 0.653760850429535, 0.813662052154541, 0.706915199756622, 0.220196515321732, 0.863410294055939, 0.906365036964417, 0.890374720096588, 0.960830211639404, 0.874392509460449, 0.731526017189026, 0.751415967941284, 0.684686660766602, 0.756359398365021, 0.748104453086853, 0.933958411216736, 0.870441019535065, 0.712883770465851, 0.835386157035828, 0.782159030437469, 0.797728955745697, 0.767299771308899, 0.759414255619049, 0.740223169326782, 0.674363374710083, 0.670708477497101, 0.744072914123535, 0.701710939407349, 0.763229131698608, 0.873173415660858, 0.725272238254547, 0.737559020519257, 0.742329895496368, 0.764297604560852, 0.800227463245392, 0.846812844276428, 0.890673637390137, 0.745569348335266, 0.751819550991058, 0.717626929283142, 0.593798518180847, 0.883229076862335, 0.812484562397003, 0.790722370147705 },
      { 0.267637848854065, 0.0731344148516655, 0.208879083395004, 0.194422915577888, 0.112290948629379, -0.252982020378113, 0.172662392258644, 0.0561512038111687, 0.364192336797714, 0.162810549139977, 0.314411967992783, 2.09177994728088, 0.206669554114342, 0.0629018321633339, 0.175797939300537, 0.0792398005723953, -0.318418890237808, 0.303535431623459, 0.18841265141964, 0.16847637295723, 0.208541959524155, 0.0738078057765961, 0.0957104712724686, 0.105412125587463, 0.220750972628593, -0.145188197493553, 0.329254508018494, 0.0972048789262772, 0.253472566604614, 0.0597410909831524, 0.278037399053574, 0.3249252140522, 0.262588232755661, 0.210274174809456, 0.173952713608742, 0.16290982067585, 0.156560227274895, 0.330338090658188, 0.612775027751923, 0.236301839351654, 0.364144653081894, 0.0955401957035065, 0.17111499607563, 0.0940982401371002, 0.0683902949094772, 0.168331146240234, 0.15737210214138, 0.460507750511169, 0.27588614821434, 0.115247659385204, 0.176396250724792 },
      { 0.0230823718011379, 0.152858540415764, -0.030283011496067, 0.0106909014284611, 0.0904139578342438, 0.348997533321381, 0.0485087186098099, 0.0600728206336498, 0.00401123240590096, 0.0537186414003372, -0.0124909225851297, -0.828484117984772, -0.0272386446595192, -0.0185711234807968, -0.0896004065871239, 0.0639259666204453, 0.419640630483627, -0.0133412722498178, 0.0416142269968987, 0.0912739410996437, 0.045267753303051, 0.175180032849312, -0.0661805793642998, 0.0855434387922287, 0.0618493296205997, 0.285617053508759, -0.108958162367344, 0.113545574247837, -0.0500102303922176, 0.166625171899796, -0.0184792876243591, 0.0177723653614521, -0.00618515908718109, 0.0183975081890821, 0.116274431347847, 0.0693105757236481, 0.0119183966889977, -0.0562353692948818, -0.263324379920959, 0.0317029841244221, -0.13127838075161, 0.0929893627762794, -0.0510754436254501, 0.00833727233111858, 0.165340319275856, 0.0675803050398827, 0.0762440264225006, -0.0128346225246787, -0.140722498297691, 0.0602374970912933, 0.0377154424786568 },
      { -0.0112269958481193, 0.0492942854762077, 0.0262944046407938, -0.000557430321350694, 0.00984367821365595, 0.0374634265899658, -0.00683072721585631, 0.0160924978554249, -0.0109630897641182, -0.00688197556883097, -0.0302974656224251, -0.00968982558697462, -0.0148914456367493, 0.0178528409451246, 0.089723639190197, 0.00563281821087003, 0.0325809605419636, 0.00413052225485444, 0.000365515443263575, 0.00567138846963644, -0.00668830471113324, 0.00133992242626846, 0.012378167361021, -0.00277202809229493, -0.00175986276008189, 0.0339486710727215, 0.0142610017210245, -0.000437343551311642, 0.0308056306093931, 0.0052076899446547, -0.00776539510115981, -0.0058854790404439, -0.00554907647892833, -0.0113116158172488, -0.00931859202682972, -0.0156725943088531, -0.00114653073251247, 0.00740314554423094, -0.00735427811741829, -0.0293489024043083, -0.0114089334383607, 0.0177642572671175, 0.0575172267854214, 0.0296753384172916, 0.00410267943516374, 0.0194540061056614, -0.00220763846300542, -0.0225240290164948, 0.078059196472168, -0.000222480448428541, 0.00510531757026911 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { -0.0169919673353434, 0.0897248536348343, -0.0126497242599726, 0.00518262851983309, -0.0186631176620722, -0.0152339283376932, 0.0107082054018974, -0.0123058883473277, -0.00809478852897882, 0.00404065009206533, 0.0157056637108326, -0.0112344967201352, 0.0132208960130811, -0.0021631654817611, -0.0254456214606762, -0.0230349320918322, -0.00971079710870981, -0.0146753117442131, -0.00366716622374952, -0.0400042980909348, -0.00395277282223105, -0.00364139163866639, 0.0100557953119278, -0.00695040402933955, -0.00119494122918695, -0.00801647268235683, -0.000674385693855584, -0.00321884686127305, -0.00256879534572363, -0.0141730681061745, 0.00432832213118672, -0.00956219248473644, 0.0355434976518154, 0.0231477599591017, -0.00720322178676724, 0.0307057946920395, -0.00723890447989106, -0.00126859790179878, -0.022504149004817, 0.0199084263294935, 0.0261177401989698, -6.57396667520516e-05, -0.00970006547868252, -0.00288332789205015, -0.00502874189987779, -0.0234342254698277, 0.0075692287646234, -0.00827719923108816, -0.0268462896347046, -0.000148139690281823, -0.00817624107003212 },
      { 4.66811275482178, -5.26705169677734, -21.3076858520508, -11.2151031494141, -13.0403528213501, -15.3777914047241, -17.4636325836182, -2.90620446205139, -11.2749357223511, 3.88649034500122, -2.00442790985107, 54.5592842102051, -8.8776683807373, 251.87646484375, -14.2993011474609, -29.9479846954346, -4.26519346237183, -11.5810222625732, -7.08601999282837, 1.57323551177979, -5.16423749923706, -1.84629821777344, 254.496124267578, 3.93210577964783, -23.9444828033447, -27.597095489502, -21.4005374908447, -4.09061717987061, -13.2187023162842, -4.44481945037842, -1.05054175853729, -11.6601810455322, -5.01846075057983, 0.631745994091034, 8.86234188079834, -6.00268459320068, -5.32001399993896, -7.98196315765381, 31.6005592346191, -1.6450412273407, 4.82454347610474, -55.7944221496582, -10.1468648910522, -22.035572052002, -6.09722423553467, -16.6830463409424, 0.984934628009796, -22.7493686676025, -39.6481246948242, -3.79120993614197, -9.44272899627686 }
    },
    {
      { -0.0459504500031471, 0.0519916564226151, -0.0595570728182793, -0.0417393110692501, -0.0396858602762222, 0.000448025151854381, -0.0784749314188957, -0.0795446038246155, 0.0236839428544044, 0.00391229148954153, -0.0509638786315918, -0.00224033952690661, -0.0900688767433167, -0.0587279088795185, -0.0768733993172646, -0.106527864933014, 0.00447138864547014, -0.047085203230381, -0.0121642267331481, -0.0358242280781269, 0.00568544957786798, -0.0110946670174599, -0.0109207844361663, -0.0476868376135826, -0.0144981853663921, 0.0290143322199583, -0.0671304613351822, -0.0865806639194489, -0.0784749835729599, -0.0363266803324223, -0.0188605058938265, 0.0165812261402607, 0.00114229996688664, -0.0112193990498781, -0.0639782100915909, -0.0337230488657951, -0.0605533309280872, -0.0649338886141777, -0.037537757307291, -0.0640289634466171, -0.028116999194026, 0.0383917130529881, -0.0610661581158638, -0.0870849117636681, -0.0130540430545807, -0.0427909605205059, 0.010020705871284, 0.0159148182719946, -0.0760742723941803, -0.0177635028958321, -0.0448058806359768 },
      { 0.954056739807129, 0.924094140529633, 0.971404373645782, 1.0264083147049, 1.00008368492126, 0.930078148841858, 1.0285370349884, 1.02008485794067, 0.885273516178131, 0.935454249382019, 0.994498431682587, 0.870366394519806, 1.01101362705231, 1.0110068321228, 0.990332663059235, 1.02611470222473, 0.952172875404358, 1.00142407417297, 0.952006816864014, 0.879895329475403, 0.917576014995575, 0.913053870201111, 0.924474477767944, 0.993485629558563, 0.966081082820892, 0.898434221744537, 0.982830584049225, 0.969350099563599, 1.02710711956024, 1.0049637556076, 0.943355679512024, 0.900365173816681, 0.997373104095459, 0.997571468353271, 1.01682960987091, 1.01358413696289, 1.01328039169312, 0.996626019477844, 0.927782952785492, 1.05467247962952, 0.915156841278076, 0.845450758934021, 0.980506002902985, 1.01978290081024, 0.952142000198364, 0.969028234481812, 0.926324367523193, 0.910498857498169, 0.962757706642151, 0.950248837471008, 1.03846871852875 },
      { 0.0749197006225586, 0.0419280864298344, 0.072004534304142, 0.0353585295379162, 0.0582795143127441, 0.0768945068120956, 0.0660304203629494, 0.053235836327076, 0.0866277515888214, 0.068132184445858, 0.0559648126363754, 0.088320791721344, 0.0777005404233932, 0.0579474456608295, 0.0937734842300415, 0.0558260977268219, 0.049440860748291, 0.0370457209646702, 0.0848045870661736, 0.0976749211549759, 0.0931013375520706, 0.104944095015526, 0.0889999642968178, 0.058757346123457, 0.054680310189724, 0.0600919276475906, 0.0761985331773758, 0.0938962325453758, 0.0757666751742363, 0.0494138821959496, 0.0797986835241318, 0.0828176662325859, 0.0438028872013092, 0.0469629690051079, 0.0529932528734207, 0.0512801110744476, 0.0583415105938911, 0.0725439190864563, 0.106916941702366, 0.0219463258981705, 0.11891620606184, 0.0963051989674568, 0.0802770256996155, 0.0635574981570244, 0.0787058100104332, 0.060910239815712, 0.0375935174524784, 0.0633435174822807, 0.109285660088062, 0.0707374513149261, 0.0276702679693699 },
      { -0.00115633173845708, -0.0179172791540623, -0.00559454597532749, -0.0157681349664927, -0.0187520477920771, -0.00361204613000154, -0.0180699154734612, -0.00949331186711788, 0.00487039377912879, -0.000442708231275901, -0.00201943446882069, 0.000942981045227498, -0.011018137447536, -0.00131826009601355, -0.0297304261475801, -0.00235070590861142, -0.00200155260972679, -0.008948914706707, -0.00219830824062228, 0.00334344361908734, -0.000408821098972112, -0.00204501743428409, -0.00260257232002914, -0.00481472024694085, -0.00240285624749959, 0.0193250384181738, -0.00828632619231939, -0.00109285546932369, -0.0288519039750099, -0.00909251254051924, -0.00200684438459575, 0.00481886183843017, -0.0101247234269977, -0.00617596786469221, -0.0113785546272993, -0.0129497842863202, -0.0184430629014969, -0.0149130970239639, -0.00428562751039863, -0.00228627398610115, -0.00180334004107863, 0.0192197058349848, -0.0155479460954666, -0.012816826812923, -0.00457220803946257, -0.00573954125866294, -0.00291681662201881, 0.000529678945895284, -0.0270686130970716, -0.000828074757009745, -0.0242873765528202 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { 0.00428924988955259, 0.00160533969756216, 0.00398974074050784, 0.0073116859421134, 0.0297752488404512, -0.000248076568823308, 0.00117216829676181, 0.00538820028305054, 0.00191415357403457, -0.00249006762169302, -0.00216877111233771, 0.000580474268645048, -0.00503412541002035, -0.00503960018977523, 0.00709900679066777, -0.00160366797354072, 0.000621880230028182, 0.00809730216860771, -0.0130123114213347, 0.0269168559461832, -0.00842948257923126, 0.00107926269993186, -0.00721931597217917, -0.0020924077834934, 6.78116994095035e-05, -0.00209413189440966, -0.00503140920773149, -0.00603016512468457, 0.00229433085769415, 0.0131408181041479, -0.00177131348755211, -0.00151548907160759, -0.0112220998853445, -0.000687621475663036, -0.00503729190677404, 0.00233120610937476, 0.00715618534013629, 0.00427976436913013, 0.000356729869963601, -0.0121755441650748, -0.0100491885095835, 0.00438194535672665, 0.00437231594696641, 0.00245664757676423, -0.00552255846560001, 0.00666380440816283, 0.0395799987018108, 0.00945600681006908, 0.00733082089573145, -0.000481119030155241, 0.013500889763236 },
      { 1.46471691131592, -3.95269441604614, 7.39901351928711, 7.28635740280151, 1.9844217300415, -0.395542740821838, 16.734411239624, 0.231749221682549, -3.22113275527954, -0.92182058095932, 0.530093967914581, -7.3266487121582, 9.33821487426758, -107.905319213867, 4.23176622390747, 13.0120792388916, -12.1504735946655, 7.43182611465454, 1.68878138065338, -2.36774492263794, -2.40548968315125, -4.03933191299438, -3.5534884929657, -0.417383700609207, 2.44366979598999, -28.1795845031738, 19.9978790283203, -0.460704058408737, 9.32700252532959, -2.05257654190063, 0.396540284156799, -5.94979667663574, -1.61296737194061, -4.92500638961792, 11.8743190765381, 3.28348231315613, 7.4947943687439, 1.79946506023407, -2.11171913146973, -0.455898851156235, -4.77866792678833, -0.242192178964615, 1.98249161243439, 6.90552806854248, 0.752136409282684, 6.65766668319702, -5.2786979675293, -4.63538312911987, 11.345775604248, 0.542663753032684, 5.85129642486572 }
    },
    {
      { 0.0870743691921234, 0.00186003686394542, 0.0737002044916153, 0.104680292308331, 0.0981228426098824, 0.387635737657547, 0.120479173958302, 0.13630074262619, -0.00808169785887003, 0.0045224130153656, 0.072374165058136, 0.0108564579859376, 0.140135779976845, -0.00383987464010715, 0.0893779546022415, 0.196574881672859, 0.372877478599548, 0.0825597867369652, 0.0679637789726257, 0.0446825139224529, 0.0330376513302326, 0.0747426152229309, -0.0278963036835194, 0.0362616255879402, 0.0726483091711998, 0.250072062015533, 0.0386681146919727, 0.0149956047534943, 0.152414843440056, 0.0501398853957653, 0.0179835055023432, 0.0330776050686836, -0.0200790651142597, 0.0046623251400888, 0.204900369048119, 0.0956417545676231, 0.0778846517205238, 0.0970444455742836, 0.134430229663849, 0.0982722640037537, 0.0409368872642517, 0.234620675444603, 0.0555047057569027, 0.147837445139885, 0.114049144089222, 0.121099956333637, 0.00397506495937705, -0.0314994379878044, 0.077553004026413, 0.0897815302014351, 0.0949658527970314 },
      { -0.126844704151154, -0.692128598690033, -0.110457316040993, -0.292141020298004, -0.206511124968529, -1.1303802728653, -0.313352853059769, -0.264049559831619, -0.106391049921513, -0.189647525548935, -0.217748388648033, -0.299453914165497, -0.315968602895737, -0.0654193460941315, -0.154174730181694, -0.279585659503937, -1.15763211250305, -0.275526463985443, -0.197493806481361, -0.139313325285912, -0.143274202942848, -0.147038325667381, -0.100546479225159, -0.21352182328701, -0.322213649749756, -0.564538538455963, 0.0147307300940156, -0.128938689827919, -0.279747158288956, -0.116096056997776, -0.121917359530926, -0.218623861670494, -0.258344888687134, -0.278149634599686, -0.325877815485001, -0.321152925491333, -0.281244277954102, -0.206561267375946, -0.25070521235466, -0.322541773319244, -0.128327041864395, -0.427434742450714, -0.138613447546959, -0.262563437223434, -0.185926899313927, -0.283496350049973, -0.182709842920303, -0.0625222325325012, -0.0714468434453011, -0.142650574445724, -0.343338757753372 },
      { 1.07487404346466, 1.47131454944611, 1.09667611122131, 1.18028008937836, 1.1300048828125, 1.66749668121338, 1.18697500228882, 1.1481214761734, 1.11153197288513, 1.16429734230042, 1.16060543060303, 1.3206672668457, 1.16356635093689, 1.02382600307465, 1.02358686923981, 1.16970992088318, 1.7281699180603, 1.1892329454422, 1.06642305850983, 1.11661088466644, 1.09320223331451, 1.01261985301971, 1.10123205184937, 1.14604067802429, 1.22797644138336, 1.16823506355286, 0.931143581867218, 1.08161687850952, 1.06628036499023, 1.06404781341553, 1.09814631938934, 1.1771547794342, 1.20554935932159, 1.20466530323029, 1.14000165462494, 1.17073941230774, 1.18868315219879, 1.08520615100861, 1.13416731357574, 1.21517074108124, 1.09477412700653, 1.03858876228333, 1.05996215343475, 1.12852704524994, 1.04246532917023, 1.19040381908417, 1.2027313709259, 1.135462641716, 0.995954632759094, 1.09618055820465, 1.2350138425827 },
      { 0.00173175835516304, 0.154109999537468, 0.00303939217701554, 0.0194307919591665, 0.00219364184886217, 0.0743304640054703, 0.0189997181296349, 0.0198545642197132, 0.00521860225126147, 0.0029417637269944, -0.00329539575614035, -0.00924340076744556, 0.0331640318036079, 0.0298804473131895, 0.0791739374399185, 0.00182440236676484, 0.0451612249016762, 0.0395587645471096, 0.0134966531768441, -0.00248926342464983, 0.00283337407745421, 0.00486992858350277, 0.00304415100254118, 0.00327972299419343, 0.0114963566884398, 0.157387003302574, 0.0453662015497684, 0.000999613432213664, 0.0889610424637794, -0.00568532152101398, 0.0050222366116941, 0.00687164394184947, 0.0188991464674473, 0.0157154109328985, 0.0156029481440783, 0.0276905000209808, 0.0320826321840286, 0.0432107523083687, 0.00251168524846435, -0.0135091161355376, -0.00367697747424245, 0.120692111551762, 0.0519488155841827, 0.026185629889369, 0.00341305718757212, 0.028017058968544, 0.00237279012799263, -0.0042480481788516, 0.060017928481102, 0.00179805315565318, 0.0375285111367702 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { 0.00746370106935501, 0.108510918915272, -0.00521124061197042, -0.0155043108388782, 0.0206321496516466, -0.00295231305062771, -0.00424538133665919, -0.00597122171893716, -0.00205409829504788, 0.00359590980224311, -2.91668347927043e-05, 0.00390927866101265, -0.00240155612118542, 0.00693121366202831, -0.00953732058405876, -0.00247903191484511, 0.0133432000875473, -0.0125905824825168, 0.0300947707146406, 0.00131448544561863, 0.0102648548781872, 0.00210434407927096, 0.000399075739551336, 0.0178652536123991, 0.0111453076824546, 0.00621974840760231, -0.000514571729581803, 0.0114610558375716, -0.00907539390027523, 0.0346782878041267, 0.00549191562458873, 0.00184541358612478, 0.0184793099761009, -0.00470052938908339, 0.0188653785735369, -0.00800103414803743, -0.0193989761173725, -0.00595172308385372, 0.00944056827574968, 0.030454633757472, 0.0150762340053916, 0.0438123792409897, -0.00439978484064341, 0.00157339789438993, 0.0355842486023903, -0.0139054981991649, -0.0254276525229216, -0.0300556682050228, -0.0143858399242163, -5.92426513321698e-05, -0.0240470971912146 },
      { -16.5451602935791, 11.0882291793823, -27.1612911224365, -5.28262853622437, -11.1594457626343, -29.9388618469238, -6.79686832427979, -13.3061971664429, -2.89906597137451, 0.846823751926422, -0.875512659549713, -40.4804878234863, -25.0676326751709, 48.4432678222656, -13.1691942214966, -40.6743812561035, -9.55444240570068, -23.0163898468018, -13.5119857788086, 3.54544234275818, -0.82737010717392, 6.03737831115723, 81.5941772460938, -8.64277839660645, -21.1221904754639, -185.732635498047, -87.6792449951172, 2.3937566280365, -24.0376472473145, 5.33100271224976, -7.54887056350708, -0.934348821640015, 0.805658221244812, 6.20752668380737, -21.2962474822998, -4.23045873641968, -13.3134088516235, -8.57929706573486, -11.8504009246826, -1.87477493286133, 0.989371418952942, -31.2184543609619, -20.9314765930176, -29.4960803985596, -7.67956876754761, -40.0643615722656, 6.91829872131348, 8.03067398071289, -14.7138595581055, -8.53815364837646, -18.2071075439453 }
    },
    {
      { -0.145724251866341, 0.102841332554817, 0.0390021875500679, -0.213276326656342, -1.43802869319916, 0.653183341026306, -1.54518067836761, -0.0675049871206284, -0.107809618115425, 0.0849460810422897, 0.0113022662699223, -0.0594085715711117, -0.575892925262451, 0.386710554361343, 0.014173517934978, -0.0718025639653206, 0.593783915042877, 0.0804029554128647, -0.512631058692932, -2.48055815696716, -0.485339790582657, -1.81005811691284, -0.00913405884057283, 0.528196334838867, -0.23888985812664, 1.0482634305954, 0.00987130030989647, -2.28722786903381, 0.15056648850441, -1.43124508857727, -0.0577476173639297, -0.117546930909157, -0.0971619263291359, -0.396629244089127, -0.752385258674622, -0.859860897064209, -0.261579513549805, 0.162896245718002, 0.163720950484276, 0.0667557120323181, -0.137407302856445, 0.993614137172699, 0.0173627845942974, -0.0596833899617195, -1.12632942199707, -0.049046166241169, -1.1437920331955, 0.00354718626476824, 0.0232973080128431, -1.0443788766861, -0.291249364614487 },
      { 0.086389847099781, -1.42926692962646, -0.159150093793869, -0.87860769033432, 1.05261790752411, -1.34482419490814, -0.170588493347168, -0.0884657949209213, -0.0434186197817326, -0.030025914311409, -0.229690939188004, -0.190184026956558, -0.0283352956175804, -0.679955303668976, -0.203140079975128, -0.156291872262955, -1.36942577362061, -0.354044556617737, 0.466962993144989, 0.64342188835144, 0.230214580893517, 1.21861219406128, -0.194570198655128, -0.581049501895905, 0.192900463938713, -1.8958797454834, 0.0792850181460381, 1.33004701137543, -0.430700868368149, 1.58088505268097, -0.040857408195734, -0.0102046169340611, -0.275805681943893, -1.65322852134705, -0.447878420352936, -0.947780013084412, -0.467168837785721, -0.389621675014496, -0.0905472785234451, -0.459809958934784, 0.0432809330523014, -1.55041754245758, -0.201109766960144, -0.102824948728085, 0.717594802379608, -0.0845629274845123, 0.416392773389816, -0.272169828414917, -0.24984535574913, 1.69199705123901, -0.598717391490936 },
      { 0.744180202484131, 1.7356950044632, 0.946078479290009, 1.84889829158783, 0.980153203010559, 1.42486119270325, 1.58273780345917, 0.83125764131546, 0.969960153102875, 0.698661386966705, 0.879912555217743, 1.74932909011841, 1.09994673728943, 1.1925950050354, 0.992820978164673, 0.990907907485962, 1.44974458217621, 1.07774209976196, 0.379342943429947, -0.728693783283234, 0.701862454414368, 1.34316349029541, 0.993448138237, 0.756009161472321, 0.848242282867432, 1.23459100723267, 0.673929631710052, 0.915109634399414, 1.03394460678101, 0.587335348129272, 0.796376645565033, 0.945037245750427, 1.09340083599091, 2.05979299545288, 1.62104034423828, 1.8727353811264, 1.17858827114105, 0.952061951160431, 0.719077050685883, 1.00617182254791, 0.793206036090851, 0.823684275150299, 0.94255268573761, 0.826486468315125, 0.941761791706085, 0.925784885883331, -0.860961139202118, 1.0477648973465, 1.04773247241974, 0.33591440320015, 1.14227938652039 },
      { 0.289123982191086, 0.458559483289719, 0.220933198928833, 0.189895510673523, 0.0222893599420786, 0.264636099338531, 0.273881614208221, 0.281310945749283, 0.15361051261425, 0.269086301326752, 0.290551632642746, 0.0202851705253124, 0.272712796926498, 0.274713158607483, 0.211238116025925, 0.226179495453835, 0.243708804249763, 0.237154170870781, 0.169793769717216, 0.172434270381927, 0.246662244200706, 0.0591263324022293, 0.18974968791008, 0.300196498632431, 0.246617689728737, 0.640025675296783, 0.249772697687149, 0.0868114531040192, 0.269251018762589, -0.0431909747421741, 0.307028323411942, 0.16405887901783, 0.233995899558067, 0.375130146741867, 0.213067203760147, 0.406601041555405, 0.245854631066322, 0.323130756616592, 0.283939301967621, 0.314355731010437, 0.274374306201935, 0.608825087547302, 0.24721996486187, 0.26967465877533, 0.0660541728138924, 0.265405982732773, 0.22998258471489, 0.18057869374752, 0.212719812989235, 0.043249886482954, 0.310007870197296 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { 0.0106490654870868, 0.218214675784111, -0.0159856360405684, -0.0370008610188961, 0.30138099193573, 0.0215441342443228, -0.146454438567162, -0.00289113214239478, 0.0365269482135773, -0.000889093906152993, 0.0242679081857204, 0.103609271347523, 0.050011932849884, -0.029912618920207, -0.0151595184579492, -0.0188252851366997, 0.130781769752502, -0.0170103739947081, 0.366613358259201, 0.879502713680267, 0.148287445306778, -0.0681551843881607, 0.0141223045065999, 0.0491655133664608, -0.0739532262086868, 0.0613776035606861, 0.0108878500759602, 0.0337084904313087, -0.0295330248773098, 0.345368474721909, -0.0265018623322248, 0.0207066275179386, -0.00393237499520183, -0.118217773735523, -0.055868074297905, -0.237624734640121, 0.0662406906485558, -0.0321813374757767, 0.00108266621828079, 0.0439742170274258, 0.00961124710738659, 0.146824315190315, -0.00596466194838285, 0.0123603921383619, 0.208136484026909, -0.0624425895512104, 0.61929988861084, 0.0347152650356293, -0.0308397095650434, -0.0363610275089741, 0.119175523519516 },
      { -26.5619144439697, 29.8970355987549, -3.93644905090332, 70.6431503295898, -44.9092864990234, -196.707290649414, 209.260009765625, 3.17994570732117, 12.7763223648071, -3.70835614204407, -0.858171284198761, -491.971221923828, -124.051467895508, 162.422622680664, 24.7311210632324, 17.6461429595947, -136.746994018555, -26.6350231170654, 0.86956512928009, 488.999359130859, 37.0578384399414, -44.1681060791016, -10.6341638565063, -27.2970428466797, 43.1982765197754, -767.073669433594, -191.404174804688, 108.423248291016, 17.917724609375, 162.168502807617, 2.18528461456299, 21.4226608276367, 7.12052726745605, 113.738265991211, 146.289031982422, 67.8593521118164, 61.1800498962402, -5.37334394454956, -51.0531120300293, 0.755031228065491, -18.2865734100342, -101.691184997559, -16.5094528198242, -3.83381462097168, 79.9591293334961, 29.9248123168945, 318.205810546875, 25.8040523529053, 29.4106845855713, 108.689117431641, 21.1811408996582 }
    },
    {
      { -0.232105985283852, -0.0894450172781944, -0.0721628367900848, -0.376704633235931, 0.196273788809776, 0.349521517753601, 0.181740313768387, -0.480681508779526, -0.235943496227264, -0.0284593105316162, -0.0724631175398827, -0.143469586968422, 0.0538324266672134, 1.01571261882782, -0.158025965094566, -0.375109910964966, 0.308476001024246, 0.0239811111241579, -0.393349766731262, -0.377805888652802, -0.644621968269348, -0.770535290241241, 0.444011569023132, 0.356552362442017, -0.034094326198101, 0.798749208450317, 0.0671789050102234, 0.000492343562655151, 0.0198856107890606, -0.0883383080363274, -0.16419044137001, -0.251410126686096, -0.21685254573822, -0.279946893453598, 0.406659185886383, -0.103998966515064, -0.171299129724503, 0.245193347334862, 0.154151096940041, -0.0269789192825556, -0.216822519898415, 0.989659667015076, -0.0729404613375664, -0.492099225521088, 0.0675630047917366, -0.275371313095093, -0.524925172328949, -0.0675685331225395, -0.0999826416373253, -0.774036705493927, -0.0993240922689438 },
      { 0.187018066644669, -0.782495379447937, -0.157519623637199, 0.234609425067902, 0.878704488277435, -0.406402915716171, 0.448970764875412, 0.417022913694382, 0.614950239658356, 1.05632543563843, 0.030671052634716, -0.739461958408356, 0.169005662202835, -0.626428246498108, -0.150331094861031, 0.31132572889328, -0.578593134880066, -0.127082973718643, 0.79679924249649, 1.41646254062653, 1.00903487205505, 1.19117224216461, -0.327522933483124, 0.305006295442581, 0.640949964523315, -1.23431372642517, 0.0565446093678474, 0.850396931171417, -0.394926488399506, 0.483381748199463, 0.185727193951607, 0.686308622360229, 0.187415242195129, 0.678061544895172, 0.389163076877594, 0.573961079120636, 0.307346761226654, -0.436459511518478, 0.227501943707466, -0.303895771503448, 0.655627548694611, -1.43528258800507, -0.189402475953102, 0.384625673294067, 0.80542653799057, 0.684947431087494, 1.51313292980194, 0.172748416662216, -0.477452367544174, 2.14918327331543, -0.0578984506428242 },
      { 0.345584064722061, 1.02916538715363, 0.64238303899765, 0.254975020885468, -0.596139907836914, 0.325958371162415, -0.498829305171967, 0.233911126852036, -0.0167964585125446, -0.46582043170929, 0.244192451238632, 2.78668928146362, -0.0694007948040962, 0.513856291770935, 0.941324174404144, 0.0926839634776115, 0.488844871520996, 0.576845288276672, -0.481509327888489, -1.50779807567596, -0.249564632773399, -0.0926887094974518, 0.627908647060394, -0.269125580787659, -0.105021730065346, 0.487463533878326, 0.107662998139858, -1.11560916900635, 0.853606641292572, -0.108047813177109, 0.266864687204361, -0.0905869603157043, 0.421344190835953, -0.0949663072824478, -0.381253242492676, -0.275794178247452, 0.043917216360569, 0.525215864181519, 0.0352349504828453, 0.425597697496414, -0.0729526281356812, 0.475759238004684, 0.739569246768951, 0.259828239679337, -0.967967391014099, -0.0272159781306982, -1.04217040538788, 0.26502588391304, 1.11996960639954, -0.823032975196838, 0.144831478595734 },
      { 0.675173699855804, 0.795345366001129, 0.542783200740814, 0.713565468788147, 0.73067432641983, 0.714411616325378, 0.779901146888733, 0.627106308937073, 0.620206654071808, 0.650776326656342, 0.741410374641418, 0.0652431771159172, 0.787745594978333, 0.691755056381226, 0.320343852043152, 0.595439016819, 0.709011197090149, 0.608393251895905, 0.641301214694977, 0.594194114208221, 0.700681626796722, 0.686654210090637, 0.656787514686584, 0.676970541477203, 0.604129254817963, 0.950279355049133, 0.807361841201782, 0.675123035907745, 0.491855621337891, 0.73835813999176, 0.679340481758118, 0.629984498023987, 0.635341167449951, 0.664128601551056, 0.711183249950409, 0.751934766769409, 0.694999635219574, 0.7734414935112, 0.685244858264923, 0.800927758216858, 0.5675408244133, 0.875702440738678, 0.466272056102753, 0.594405293464661, 0.804291605949402, 0.573550939559937, 0.63360732793808, 0.707827866077423, 0.383454531431198, 0.586636424064636, 0.802799165248871 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { -0.0415885411202908, 0.0709018781781197, -0.0296346358954906, 0.0499228946864605, -0.0105149624869227, 0.00352419749833643, 0.0265683773905039, 0.012791758403182, 0.0027297439519316, -0.0134072173386812, 0.0138138718903065, 0.174233764410019, 0.0999138131737709, -0.0143205197528005, -0.0170125439763069, 0.142404675483704, 0.0783758610486984, -0.0512179844081402, 0.364179015159607, 0.492226928472519, 0.0563514120876789, -0.0806184262037277, -0.00818888749927282, 0.0426320731639862, -0.0443610027432442, 0.0216708164662123, -0.00616229558363557, 0.216791853308678, -0.0535840094089508, -0.0585268549621105, -0.026170402765274, 0.00173124601133168, -0.0501457676291466, -0.0124309975653887, 0.110713519155979, -0.0530922636389732, 0.0702847242355347, -0.084077924489975, -0.00448528584092855, 0.0411566719412804, 0.0826103016734123, 0.087272472679615, -0.00532016903162003, 0.0237119384109974, 0.370017111301422, 0.00584230292588472, 0.0396288745105267, -0.0631208866834641, -0.0340388603508472, 0.00248055695556104, 0.0607485137879848 },
      { 3.07590079307556, 58.7049942016602, 60.8484535217285, -61.0769424438477, -166.505844116211, -108.713172912598, -89.5284652709961, 34.1797103881836, 43.0337142944336, -116.962501525879, -1.59336316585541, -874.17529296875, -125.779678344727, 113.314270019531, 89.170295715332, -1.42916488647461, -38.0687828063965, -36.3070983886719, -46.0087432861328, 99.4244766235352, 29.3967533111572, -73.2607879638672, -395.369049072266, -76.3688125610352, 101.872024536133, -458.315979003906, -120.603981018066, -14.4579448699951, 112.103149414062, -121.782814025879, 11.1135902404785, 52.2889938354492, -18.1775741577148, 10.32164478302, -137.711029052734, -72.2339172363281, -62.2888412475586, 2.07174873352051, -90.1171264648438, 6.06131029129028, 23.3068103790283, -122.691146850586, 21.761116027832, 64.0230712890625, -519.676147460938, 104.839553833008, 83.2127227783203, 48.2206153869629, 96.0827484130859, -203.627090454102, -40.877197265625 }
    },
    {
      { -0.154921606183052, -0.138801068067551, -0.0886319950222969, -0.0386528372764587, 1.85891342163086, 0.116988576948643, 1.98458707332611, -0.440944641828537, -0.119105741381645, 0.0068841683678329, -0.0657215267419815, 0.278500974178314, 0.759515583515167, 0.41253200173378, -0.224835976958275, -0.24415123462677, 0.0839061141014099, -0.00484479824081063, -0.263107746839523, 1.18525695800781, -0.388069838285446, 1.27094149589539, 1.293412566185, 0.132991790771484, 0.0452942214906216, 0.0620325095951557, -0.0344973877072334, 1.33711755275726, 0.00445499690249562, 0.34664112329483, -0.148317724466324, -0.170388266444206, -0.166903391480446, 0.014231794513762, 2.21547222137451, 0.754669010639191, -0.0156619437038898, 0.164136201143265, -0.0851936340332031, -0.00568974018096924, -0.168072506785393, -0.0828291997313499, -0.115772880613804, -0.470556497573853, -0.0553615875542164, -0.176636219024658, -0.331273198127747, -0.0438700765371323, -0.176195159554482, 0.455517679452896, 0.177476570010185 },
      { 0.102316692471504, -0.138459503650665, -0.132541432976723, 0.0106728710234165, -0.387112230062485, -0.0152484066784382, -0.682410478591919, 0.366993218660355, 0.487179934978485, 0.831620633602142, -0.0173665378242731, -0.666681945323944, -0.457958340644836, 0.446807712316513, -0.077086053788662, 0.274446398019791, -0.142297804355621, -0.0771408751606941, 0.598891258239746, 0.680704653263092, 0.842483162879944, 0.177192315459251, -0.607074797153473, 0.261869728565216, 0.487025707960129, 0.125181943178177, 0.0291985683143139, -0.182015106081963, -0.28930652141571, 0.125771760940552, 0.15938363969326, 0.617619395256042, 0.106174074113369, 0.552493751049042, -0.6347416639328, 0.322619795799255, 0.0618109442293644, -0.27246567606926, 0.3525270819664, -0.254909783601761, 0.526362121105194, 0.656657218933105, -0.122889935970306, 0.330350190401077, 0.472843259572983, 0.554377794265747, 1.41095244884491, 0.233123540878296, -0.365833133459091, 0.213640421628952, -0.401627123355865 },
      { 0.161956071853638, 0.357598870992661, 0.477759122848511, -0.0410063154995441, -0.852993428707123, -0.0317201502621174, -0.837579190731049, 0.133709952235222, -0.217718243598938, -0.493305772542953, 0.0988019928336143, 0.340467929840088, -0.179577603936195, -0.274745672941208, 0.710989892482758, -0.185350522398949, 0.126170486211777, 0.343497812747955, -0.610536336898804, -0.0226078871637583, -0.387142479419708, -1.19949090480804, 0.742739140987396, -0.204662084579468, -0.178870841860771, -0.104982309043407, 0.0722295269370079, -1.18433892726898, 0.527551889419556, -0.747798323631287, 0.100216701626778, -0.310163021087646, 0.295136630535126, -0.220423176884651, -0.979262053966522, -0.657077789306641, -0.00729972776025534, 0.26412832736969, -0.108812659978867, 0.232796713709831, -0.173743277788162, -0.44972038269043, 0.517098724842072, 0.204950824379921, -0.537118077278137, -0.140944257378578, 0.42178800702095, 0.139409318566322, 0.889811336994171, -0.405500859022141, 0.154977902770042 },
      { 0.86800092458725, 0.940522193908691, 0.698910355567932, 0.940228343009949, 0.965774595737457, 0.927536308765411, 1.07648873329163, 0.783596515655518, 0.844162583351135, 0.852409303188324, 0.910317778587341, 0.966441035270691, 1.0870201587677, 0.583229959011078, 0.516129672527313, 0.743554174900055, 0.925702512264252, 0.806211054325104, 0.847902536392212, 0.810103833675385, 0.891023218631744, 0.960025131702423, 0.691206693649292, 0.857487261295319, 0.764880001544952, 0.904140532016754, 0.983071148395538, 0.904358565807343, 0.704854369163513, 0.854417622089386, 0.849750995635986, 0.851262629032135, 0.826894044876099, 0.838777542114258, 1.04702508449554, 0.907272458076477, 0.917922377586365, 0.921722233295441, 0.859518468379974, 0.940645754337311, 0.75038069486618, 0.864598512649536, 0.667679309844971, 0.746449768543243, 0.727758884429932, 0.760854423046112, 0.867163360118866, 0.925872087478638, 0.565293669700623, 0.988796830177307, 1.04398965835571 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { -0.0343866534531116, -0.0235010515898466, -0.0213357489556074, 0.029836218804121, -0.0573214069008827, -0.0041979206725955, 0.0138583425432444, -0.00228074612095952, 0.0014088818570599, -0.0150632672011852, 0.0265778824687004, 0.283787608146667, -0.0188452061265707, 0.0329195857048035, -0.00821274518966675, 0.206914395093918, -0.00110629154369235, -0.0474741160869598, 0.306358814239502, -0.576214969158173, -0.0259019378572702, -0.00576836615800858, -0.178597837686539, 0.0121954930946231, -0.0272595323622227, 0.0129456771537662, -0.0208866987377405, 0.275880128145218, -0.0288459993898869, 0.227624580264091, -0.0115366000682116, 0.00439618853852153, -0.0721273496747017, -0.0270773265510798, 0.0961657464504242, -0.0315040983259678, 0.0223091412335634, -0.0618039779365063, -0.00705042341724038, 0.0400145165622234, 0.0758392140269279, 0.00929118320345879, -0.00462589086964726, -0.00127694942057133, 0.0885085016489029, 0.00659235520288348, -1.07502782344818, -0.187558174133301, -0.0341411866247654, 0.0325486920773983, -0.0538837090134621 },
      { 8.7891263961792, 24.5423488616943, 55.5760612487793, -61.6389617919922, 138.597885131836, -56.7167625427246, -187.784423828125, 42.1655120849609, 20.1744842529297, -133.718551635742, -3.67643427848816, -228.599884033203, -169.985107421875, 415.445495605469, 82.2858200073242, -9.76278591156006, -4.68936967849731, -25.5766658782959, -10.0327930450439, 70.7424926757812, 21.0501651763916, 15.519702911377, -775.783630371094, -66.7473831176758, 92.3905868530273, -41.5087280273438, 15.4587297439575, 135.844451904297, 80.0746612548828, 441.505462646484, 6.8740291595459, 16.3435459136963, -12.5693664550781, 18.5961112976074, -239.303192138672, -85.5198059082031, -89.2361907958984, 2.0148401260376, -55.3471527099609, 3.0804591178894, 36.3152122497559, -41.4646186828613, -1.19449174404144, 70.1657791137695, 976.143493652344, 65.9462356567383, 47.661563873291, 52.2880516052246, 104.974891662598, -372.422027587891, -44.8833389282227 }
    },
    {
      { -0.0745240226387978, 0.0485322587192059, -0.0325390994548798, -0.0181209314614534, 0.312438607215881, 0.185464024543762, 0.265635758638382, -0.155475988984108, -0.0393964350223541, -0.0160359106957912, -0.0264688264578581, 0.0840141102671623, 0.11603706330061, -0.130323335528374, -0.0780028402805328, -0.148854702711105, 0.176222652196884, -0.0100944144651294, -0.105904340744019, 0.0199765283614397, -0.117569155991077, 0.370572507381439, 0.209759622812271, 0.0869053304195404, 0.0228856261819601, 0.581281304359436, -0.0114018004387617, 0.0907049477100372, -0.00690028490498662, -0.0336150825023651, -0.0481339171528816, -0.0600114390254021, -0.0182123798877001, -0.0133398771286011, 0.331369459629059, 0.100911661982536, 0.00042542238952592, 0.0593015626072884, -0.0266100838780403, -0.0016334680840373, -0.0622070394456387, 0.675111591815948, -0.0376718305051327, -0.164555549621582, 0.131530612707138, -0.0544544532895088, -0.106774114072323, 0.00862216297537088, -0.0647001564502716, 0.199168264865875, 0.0358818508684635 },
      { 0.015855435281992, -0.375572770833969, -0.036322183907032, -0.0575745739042759, -0.00050092936726287, -0.0817106291651726, -0.0915985256433487, 0.120103724300861, 0.122525162994862, 0.28626075387001, -0.0166639648377895, -0.184014722704887, -0.139368340373039, 0.172746896743774, -0.027293398976326, 0.0845336094498634, -0.154801934957504, -0.0326156839728355, 0.220877453684807, 0.31382030248642, 0.209381192922592, 0.0504831224679947, -0.0521266013383865, 0.0165705010294914, 0.128049790859222, -0.91697108745575, 0.0129461083561182, 0.0924662202596664, -0.0864861831068993, 0.156505271792412, 0.0490497201681137, 0.190142273902893, 0.000470620900159702, 0.137408226728439, -0.0233834125101566, 0.146513000130653, -0.0223244111984968, -0.116656176745892, 0.115438379347324, -0.117148332297802, 0.142727553844452, -0.938739597797394, -0.043104849755764, 0.1111751049757, 0.0771802291274071, 0.208593904972076, 0.363298863172531, 0.0476154647767544, -0.120147041976452, 0.0416009016335011, -0.117719918489456 },
      { 0.081312470138073, 0.294714599847794, 0.149101659655571, 0.0277676973491907, -0.202058836817741, -0.0630153119564056, -0.149798572063446, 0.0531356036663055, -0.0311042349785566, -0.154069602489471, 0.0245347674936056, 0.0550108999013901, -0.00182306230999529, -0.0916862711310387, 0.241848230361938, 0.0156465284526348, -0.0301166400313377, 0.135874480009079, -0.230569571256638, -0.369517475366592, -0.103970766067505, -0.358362078666687, 0.0788939520716667, -0.0414791777729988, -0.0526372119784355, 0.188378036022186, 0.0246293116360903, -0.303911447525024, 0.170561105012894, -0.199678972363472, 0.0351490788161755, -0.0801023617386818, 0.106818541884422, -0.0642605647444725, -0.223598003387451, -0.164262980222702, 0.00693850591778755, 0.107535287737846, -0.019683139398694, 0.0914943218231201, -0.00546222319826484, 0.0655449107289314, 0.179555788636208, 0.0768562331795692, -0.162247851490974, -0.0489046014845371, -0.190229550004005, 0.0284698717296124, 0.306162565946579, -0.143168747425079, 0.0186692867428064 },
      { 0.950045585632324, 0.993768036365509, 0.905625402927399, 0.985568046569824, 0.964978873729706, 0.960328459739685, 0.995089590549469, 0.925150990486145, 0.946925640106201, 0.949243783950806, 0.9713494181633, 1.00610208511353, 1.01923823356628, 0.912792980670929, 0.837237954139709, 0.918305993080139, 0.967075228691101, 0.921331763267517, 0.942926347255707, 0.962912738323212, 0.966474831104279, 0.983156263828278, 0.944125473499298, 0.951889753341675, 0.921768844127655, 1.16034853458405, 0.979314982891083, 0.967605412006378, 0.902199029922485, 0.964358448982239, 0.952145397663116, 0.946488797664642, 0.938267350196838, 0.95344215631485, 0.982920825481415, 0.954801321029663, 0.971124351024628, 0.974461019039154, 0.949149608612061, 0.989025115966797, 0.902209639549255, 1.1392776966095, 0.887940168380737, 0.913001656532288, 0.956128835678101, 0.904066503047943, 0.955223262310028, 0.965840399265289, 0.850199162960052, 1.0028064250946, 1.0001757144928 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { -0.00919782277196646, 0.0634295344352722, -0.0070682168006897, 0.0170990135520697, 0.0715181678533554, 0.00670776795595884, 0.0310886688530445, -0.000301012099953368, -0.00318918353877962, -0.00797005835920572, 0.022174496203661, 0.120050638914108, 0.00861106719821692, 0.00358763663098216, -0.00272821285761893, 0.0408045202493668, 0.0661837980151176, -0.0151650346815586, 0.120843574404716, -0.0436028391122818, 0.0134417749941349, -0.000752612424548715, -0.00178261031396687, 0.00965246558189392, -0.00841226801276207, 0.0244011953473091, -0.00293024792335927, 0.109583958983421, -0.00550115434452891, 0.0979079529643059, -0.00261688651517034, -0.00167951884213835, -0.0245830528438091, -0.0144627597182989, 0.0351635813713074, 0.00574263744056225, 0.0247650817036629, -0.0216642376035452, -0.00446018809452653, 0.0147738698869944, 0.0213288683444262, 0.0664062350988388, -0.00273821619339287, -0.00223422888666391, 0.12683416903019, 0.00223853602074087, -0.105146944522858, -0.0323826670646667, -0.0128030106425285, 0.0101767927408218, 0.0190954357385635 },
      { 9.0477237701416, 16.8072032928467, 16.0557632446289, -25.47216796875, -19.6740036010742, -96.080207824707, -34.1956901550293, 18.3321571350098, 10.6028337478638, -43.7903442382812, 0.309434682130814, -66.7656784057617, -43.7648315429688, 293.324981689453, 30.2471752166748, 12.5222139358521, -78.65478515625, -2.79373741149902, 3.21277737617493, 24.8005771636963, 7.86204671859741, 4.92330408096313, -213.524108886719, -23.6816082000732, 43.592845916748, -366.825134277344, -0.88368558883667, 7.27822732925415, 16.6915798187256, 9.55507850646973, 1.76351463794708, 9.91563034057617, -1.88858151435852, 9.85189914703369, -45.9520835876465, -16.1440410614014, -28.6853675842285, 2.66871094703674, -22.8064289093018, 3.76869869232178, 11.9142398834229, -93.0669555664062, -2.10760307312012, 29.4745025634766, 3.45492815971375, 37.209587097168, 30.1764793395996, 3.77745509147644, 39.8882255554199, -134.60725402832, -11.4343318939209 }
    },
    {
      { 0.0203098449856043, -0.0271846167743206, 0.0191634893417358, 0.0342985540628433, 0.0974557399749756, -0.0794667899608612, 0.139079660177231, 0.0619138330221176, -0.017995847389102, 0.00419981544837356, 0.00907684490084648, -0.0324429869651794, 0.0572328791022301, -0.0601003803312778, 0.0320671126246452, 0.0690889656543732, -0.0670191869139671, 0.0118561536073685, 0.0317632406949997, 0.10970076918602, -0.00374495168216527, -0.083542600274086, 0.102662339806557, -0.0172012764960527, -0.0140834180638194, -0.256008505821228, -0.000917320663575083, 0.0450052544474602, 0.022614324465394, 0.0321383476257324, 0.010120578110218, -0.0341377183794975, -0.0130043420940638, 0.00791278295218945, 0.172323435544968, 0.046739399433136, 0.0460467822849751, -0.00307786837220192, -0.0492127984762192, 0.0106987711042166, 0.0293844416737556, -0.419971436262131, 0.0224916879087687, 0.0557017549872398, 0.0486131198704243, -0.0032057436183095, 0.00187872524838895, 0.0232000350952148, 0.0270509906113148, -0.0655022785067558, 0.0189878325909376 },
      { -0.00277372379787266, 0.157012432813644, 0.00112060410901904, -0.0109147094190121, -0.102018594741821, 0.0810530781745911, -0.0718715041875839, -0.0351219475269318, 0.0394503884017467, -0.0430803559720516, -0.00600409228354692, -0.0809712782502174, 0.00537818484008312, 0.00143339158967137, -0.00181653164327145, -0.0412505492568016, 0.0695913285017014, -0.00594444572925568, -0.0409289076924324, -0.164775937795639, 0.0101996101438999, -0.00121410156134516, -0.0759648308157921, 0.0185933467000723, -0.0102274604141712, 0.426853954792023, -0.0071632987819612, -0.0854906663298607, -0.00217351550236344, -0.0503894314169884, -0.0025951995048672, 0.0824226066470146, 0.00591317750513554, -0.0198203884065151, -0.113279722630978, -0.0352905318140984, -0.0278113428503275, 0.0177105739712715, 0.00710969138890505, 0.00524840364232659, -0.0211576167494059, 0.819410920143127, -0.00281555601395667, -0.0452920161187649, 0.00828691944479942, -0.0221994146704674, -0.0202367287129164, -0.0639722496271133, 0.0268944147974253, -0.0128262042999268, 0.0143389739096165 },
      { -0.00833675172179937, -0.0956375449895859, -0.041747160255909, -0.00463824672624469, 0.0218563489615917, -0.0201604291796684, -0.0431398712098598, -0.0181447137147188, -0.0292220488190651, 0.0187448263168335, 0.00247129518538713, 0.11826890707016, -0.0196074787527323, -0.0124054150655866, -0.0612544901669025, -0.0180723648518324, 0.0033102398738265, -0.0225784759968519, 0.0396080724895, 0.152900457382202, 0.0125482259318233, 0.0679212510585785, 0.0765096992254257, -0.00189497310202569, 0.0157360676676035, -0.10958543419838, -0.00291283498518169, -0.0655963942408562, -0.0350768342614174, 0.0119500001892447, -0.0142173366621137, -0.055929958820343, -0.0276372358202934, 0.0225481167435646, -0.0124084427952766, 0.0175615940243006, -0.00387778878211975, -0.0269920378923416, 0.00451833941042423, -0.0110674388706684, -0.0184507630765438, -0.294280767440796, -0.0499686636030674, -0.0157196167856455, -0.0197588503360748, 0.0035772433038801, 0.162988275289536, 0.0566657185554504, -0.0893487930297852, 0.0454140976071358, -0.000761705043260008 },
      { 1.0055456161499, 0.985038280487061, 1.02619671821594, 1.00620520114899, 1.00979900360107, 1.01760613918304, 1.01737821102142, 1.02368748188019, 1.00507426261902, 1.01131153106689, 1.0022611618042, 0.989973187446594, 0.996669828891754, 1.03315651416779, 1.04417538642883, 1.01595044136047, 1.00902688503265, 1.01583957672119, 1.01152336597443, 1.00126576423645, 1.00226247310638, 1.00261580944061, 0.995521366596222, 1.01450192928314, 1.01162087917328, 0.928276062011719, 1.00884139537811, 1.01195240020752, 1.02709627151489, 0.997952222824097, 1.00853073596954, 1.00586760044098, 1.01912426948547, 1.0084947347641, 1.02159094810486, 1.00654637813568, 1.01122152805328, 1.01020765304565, 1.01354467868805, 1.00510728359222, 1.01623141765594, 0.91830050945282, 1.0337290763855, 1.02656662464142, 0.990733444690704, 1.01433336734772, 1.00870251655579, 1.00855457782745, 1.04619944095612, 0.995169281959534, 1.00850570201874 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { -0.0010863218922168, -0.0327554307878017, 0.00315145705826581, -0.00419434066861868, -0.0434903055429459, -0.00254685874097049, -0.0087205907329917, -0.00294180680066347, 0.00426242081448436, 0.000761950796004385, -0.000796042499132454, -0.0590466670691967, -0.00580569123849273, 0.0121623845770955, -0.00116504984907806, 0.00180084153544158, -0.0270266998559237, 0.00249045249074697, -0.0222746543586254, -0.0011908010346815, -0.0113975666463375, 0.00599097413942218, -0.0200671255588531, -0.0117072332650423, 0.00284577463753521, -0.0043307775631547, 0.00280566792935133, 0.00697791157290339, -0.000655786192510277, 0.00559362536296248, 0.00341012352146208, 0.00309944245964289, 0.0079160388559103, -0.00516945868730545, -0.00643126294016838, -0.00106785458046943, -0.00550722796469927, 0.00285457214340568, 0.00050630240002647, -0.00411424180492759, 0.00143257610034198, -0.0319186113774776, 0.00241939537227154, -0.000247181917075068, 0.0394164547324181, 0.00560717098414898, -0.0921555683016777, -0.0115101523697376, 0.00336864963173866, -0.00230173394083977, -0.0172297246754169 },
      { -2.37012338638306, -3.39452743530273, -2.8198082447052, 7.53686046600342, 67.3106155395508, 21.8122901916504, 5.06326198577881, -8.83655071258545, -2.65075993537903, 17.3033466339111, -0.58323472738266, -8.60102558135986, 8.27830600738525, -57.3238258361816, -3.86640238761902, -1.22734761238098, 29.9052104949951, 0.360763370990753, -2.18687725067139, 16.9830150604248, 1.08401834964752, 0.263688206672668, -93.4855651855469, 13.616286277771, -10.6553430557251, 146.171112060547, 9.53755378723145, 26.0251369476318, -0.201033681631088, 39.5701904296875, -0.626832902431488, -3.28867888450623, 0.0521495938301086, 0.382052689790726, -2.53895401954651, 3.53054118156433, 5.26616048812866, -0.23735611140728, 15.87473487854, 0.191391304135323, 0.475935488939285, 57.211067199707, 0.636985421180725, -7.72182559967041, 19.2497024536133, -2.31907510757446, 1.88948118686676, 0.330613732337952, -11.8812456130981, 47.288703918457, -1.38919198513031 }
    },
    {
      { -0.267239332199097, 0.256949037313461, 0.615339457988739, -0.307853192090988, 0.578648209571838, -0.615050137042999, 0.699152290821075, 0.543999016284943, -0.144077941775322, 0.82164067029953, 0.0235209930688143, -0.0884782448410988, 0.837522029876709, -1.59564626216888, -0.101031810045242, 0.974016547203064, -0.560688436031342, -0.0252222716808319, 0.120344452559948, 0.53271609544754, -0.96241420507431, 0.225478485226631, -0.696810305118561, 0.962033867835999, -0.239043742418289, -0.817440927028656, 0.111725598573685, 1.17809104919434, -0.210492521524429, 1.17372369766235, 0.379477739334106, -0.00515645183622837, -0.301579684019089, -0.217642515897751, 0.567706346511841, 0.132329806685448, 0.642107427120209, 0.063270591199398, 1.12179601192474, 0.1552754342556, -0.106696851551533, -0.0969134271144867, 0.133160546422005, 0.532739341259003, 0.91722446680069, 0.756517350673676, -0.0926643833518028, -0.272829174995422, 0.259763330221176, -1.32774579524994, 0.134488433599472 },
      { 0.0796895623207092, -0.078570120036602, -1.04854834079742, 0.45468470454216, -0.226109027862549, -0.304975569248199, -0.201111406087875, -0.0431705936789513, -0.662891685962677, -0.833087861537933, 0.396438419818878, -0.312606543302536, 0.0304024759680033, -0.521006941795349, 0.447619587182999, -0.456691563129425, 0.119606643915176, 0.0775877013802528, 0.308871537446976, 1.28932201862335, 1.59771037101746, 0.0158168543130159, -0.534777104854584, -1.20337378978729, 0.0328958965837955, 0.677694380283356, -0.106652893126011, 0.0472625829279423, 1.40407633781433, -0.184638977050781, 0.285627901554108, -0.823834896087646, 0.572519719600677, 0.266039878129959, -0.427214026451111, 0.112895779311657, -0.266452044248581, 0.0847756117582321, -0.167852252721786, 0.225245177745819, 0.403981685638428, 0.113077588379383, -0.2852863073349, -0.529406547546387, 0.100277915596962, -0.426516205072403, 0.911110162734985, 0.0677049160003662, -0.416315823793411, -0.223007366061211, 0.0672609359025955 },
      { 0.288244843482971, 0.461118876934052, 0.773586332798004, 0.148536026477814, -0.171814754605293, 1.14499950408936, -0.472367763519287, -0.196194395422935, 1.18847382068634, 0.445751845836639, 0.154826775193214, 0.412420064210892, -0.0296772681176662, 0.108483493328094, -0.952047824859619, 0.625023722648621, 0.739038467407227, 0.0406855084002018, -0.137644201517105, -1.14237868785858, -0.26761656999588, -0.206960648298264, 0.478324502706528, 1.02578282356262, 0.341895997524261, 0.42658668756485, 0.0154460212215781, -0.551820933818817, -1.00689506530762, -0.367472857236862, -0.36622741818428, 1.04426991939545, 0.18467803299427, -0.0777314901351929, 0.346434354782104, 0.180496260523796, 0.205781608819962, 0.0370001420378685, -0.0418512746691704, 0.188796386122704, 0.10239066183567, 0.486380726099014, -0.196384936571121, 0.0699190273880959, -0.687583267688751, 0.527097940444946, -0.297220170497894, 0.41706719994545, 0.55514794588089, 0.819230496883392, 0.47041517496109 },
      { 0.328498393297195, 0.109524078667164, 0.603641927242279, 0.431696504354477, 0.535780370235443, 0.369290083646774, 0.577575623989105, 0.46611014008522, 0.269620269536972, 0.566968858242035, 0.267547488212585, 0.0282454956322908, 0.240244284272194, 1.19654703140259, 1.24827182292938, 0.355941385030746, 0.391518920660019, 0.435063272714615, 0.32614004611969, 0.201880127191544, 0.18111777305603, 0.237395793199539, 0.216292172670364, 0.603019535541534, 0.425763577222824, 0.298062264919281, 0.289233773946762, 0.258913338184357, 0.811091423034668, 0.268629640340805, 0.456539183855057, 0.373542934656143, 0.263971239328384, 0.296427577733994, 0.540714979171753, 0.487167626619339, 0.477432250976562, 0.449953734874725, 0.486292660236359, 0.215223029255867, 0.38319405913353, 0.119626596570015, 1.03984808921814, 0.524450302124023, 0.287598580121994, 0.607012629508972, 0.23631264269352, 0.281058371067047, 0.787814199924469, 0.324436515569687, 0.326475143432617 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { 0.621700763702393, 0.128681436181068, 0.416117042303085, 0.466917961835861, 0.06584782153368, 0.36367267370224, 0.425941467285156, 0.749069511890411, 0.4198018014431, 0.81061577796936, 0.447378039360046, 0.941932022571564, 0.536975264549255, 0.783711373806, 0.396720588207245, 0.451514452695847, 0.242983087897301, 0.53422874212265, 0.694388449192047, 0.748222231864929, 0.576748609542847, 1.15019452571869, 1.05031836032867, 0.638858735561371, 0.750483930110931, 0.318899631500244, 0.867905735969543, 0.944108963012695, 0.597842931747437, 0.84229850769043, 0.768141269683838, 0.424154341220856, 0.558156847953796, 0.895199418067932, 0.510747849941254, 0.44784539937973, 0.514968991279602, 0.501183271408081, 0.562181413173676, 0.468750178813934, 0.417570859193802, 0.320794224739075, 0.578858315944672, 0.740239202976227, 1.02749681472778, 0.416148871183395, 0.643657624721527, 0.732131600379944, 0.102780275046825, 0.780755341053009, 0.464333057403564 },
      { 154.153015136719, -28.832483291626, 488.143218994141, 16.1919040679932, 154.095291137695, 336.851806640625, -78.1681365966797, 52.4024887084961, -11.4591312408447, -125.359230041504, -10.6858263015747, -103.424491882324, -22.8364696502686, 673.930969238281, 645.1025390625, 47.2576789855957, 179.968002319336, -17.0757713317871, 19.9340076446533, -48.1320419311523, 32.4004783630371, -67.0014038085938, 1648.68237304688, -121.396141052246, 41.9632949829102, 566.936767578125, 28.123685836792, -141.563095092773, 216.451065063477, -123.020248413086, 14.9118185043335, 43.8957862854004, -81.1360626220703, -29.0161323547363, -9.39608764648438, -84.6565399169922, -0.988865077495575, -68.6919097900391, 185.983383178711, -44.8584251403809, 155.580001831055, 134.945007324219, -183.503219604492, 234.369567871094, -191.392944335938, -251.554458618164, -19.2123622894287, -2.46864223480225, 649.372680664062, 190.439849853516, 33.2383422851562 }
    },
    {
      { 0.224099278450012, 0.622755289077759, 0.434160083532333, -0.0247239731252193, -0.0231702122837305, 0.133455738425255, -0.0488359332084656, 0.758916437625885, 0.419962257146835, 0.192748174071312, -0.0819545984268188, -0.0534603483974934, -0.0148995956405997, 0.320162415504456, 0.435667186975479, 1.00044167041779, 0.142556428909302, -0.14374940097332, 0.0353072099387646, 0.0568697713315487, 0.0144724110141397, -0.116040468215942, 0.362617164850235, 0.390144884586334, -0.171132490038872, 0.0697649270296097, -0.128295376896858, 0.0449080988764763, 0.491191923618317, 0.0506317913532257, 0.24770775437355, 0.418735206127167, 0.0616499856114388, -0.0413039140403271, -0.138447299599648, -0.137688800692558, 0.121666394174099, 0.0650748759508133, 0.512790501117706, 0.0290330741554499, -0.0620231181383133, 0.302500575780869, 0.385962665081024, 0.520439684391022, -0.0178809240460396, 0.332321345806122, -0.0471085496246815, 0.168272152543068, -0.135348170995712, -0.199743196368217, -0.0442314632236958 },
      { -0.0162899494171143, 0.126416698098183, 0.326114982366562, 0.209264159202576, -0.00764567172154784, -0.35560730099678, -0.146861284971237, -0.764973163604736, -0.587597012519836, -0.000256022583926097, 0.124551460146904, 0.196729585528374, 0.044658437371254, 0.104682140052319, -0.0798740908503532, -0.42078047990799, -0.0514258779585361, 0.213561818003654, 0.21787978708744, 0.19673416018486, 0.110168613493443, 0.0395134910941124, 0.412327289581299, -0.212690562009811, 0.21999779343605, -0.338246554136276, 0.0278091859072447, -0.0060712331905961, -0.156547710299492, -0.0083717480301857, -0.113856181502342, -0.764512836933136, 0.358101010322571, -0.250972330570221, -0.00823777541518211, -0.0865470245480537, 0.347242295742035, -0.102416053414345, 0.710298955440521, -0.0142768444493413, -0.233748197555542, -0.717878758907318, -0.277394086122513, -0.48963525891304, 0.0248561389744282, -0.303087502717972, 0.146059408783913, -0.365996807813644, 0.926784992218018, -0.0159847997128963, 0.101600579917431 },
      { -0.283998876810074, -0.804323077201843, -0.419085294008255, -0.15931710600853, -0.0295751355588436, 0.278144180774689, 0.0299009867012501, 0.188134551048279, 0.112446703016758, -0.0377529710531235, -0.168825209140778, -0.114140123128891, 0.0402349941432476, -0.0126368813216686, -0.63472843170166, 0.13293768465519, -0.0557397156953812, -0.12821464240551, -0.367353498935699, -0.250450849533081, -0.161197409033775, 0.00958741549402475, -0.446282595396042, 0.108985669910908, -0.00280622346326709, 0.256461322307587, 0.0571833290159702, -0.0862194299697876, -0.142893776297569, -0.0226041935384274, -0.230596527457237, 0.324585378170013, -0.310265004634857, 0.168531164526939, 0.109624840319157, 0.135814011096954, -0.181188136339188, 0.0187776014208794, -0.845225512981415, -0.0124828610569239, 0.178076148033142, 0.294534474611282, 0.00135456130374223, -0.051702007651329, -0.0475226081907749, 0.108000330626965, -0.0450272336602211, 0.236233189702034, -1.40910375118256, 0.0927859991788864, -0.207806274294853 },
      { 0.100694879889488, 0.000229137192945927, -0.0176888071000576, 0.0591255873441696, 0.0588149949908257, 0.073469340801239, 0.0974852293729782, 0.0641721114516258, 0.138620868325233, 0.0507545657455921, 0.058982040733099, 0.0207062233239412, -0.00622595706954598, -0.103955142199993, 0.62569272518158, 0.0805450528860092, 0.115378633141518, -0.0695606097579002, 0.0191737394779921, 0.0106865204870701, 0.0179423596709967, 0.0207571759819984, 0.030772540718317, 0.0817612558603287, 0.017449414357543, 0.144942060112953, -0.00625379243865609, 0.0221731327474117, -0.148465394973755, 0.0226529650390148, 0.0441624969244003, 0.144926324486732, -0.0183777380734682, 0.0524782203137875, 0.0481774136424065, 0.0716730877757072, 0.055822629481554, -0.0393961295485497, 0.0939117446541786, 0.0431024618446827, 0.0864764079451561, 0.220483422279358, 0.0505928359925747, 0.0986887365579605, 0.016442459076643, 0.213396698236465, 0.0203134417533875, 0.0213161334395409, 0.629047453403473, 0.0440801121294498, 0.0786130130290985 },
      { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
      { 0.98148900270462, 1.1216698884964, 0.831165015697479, 0.892513573169708, 0.86572402715683, 0.909875452518463, 0.873786509037018, 0.883888900279999, 0.907537400722504, 0.908910691738129, 0.987796366214752, 0.88785195350647, 0.919976413249969, 0.939744412899017, 0.736922323703766, 0.715111255645752, 0.889418482780457, 1.02652418613434, 1.08945286273956, 0.959322810173035, 0.99712336063385, 0.974511027336121, 0.990904569625854, 0.863977015018463, 0.904896318912506, 0.901745438575745, 0.930163741111755, 0.992082834243774, 0.993868470191956, 0.946912705898285, 0.994423985481262, 0.897187829017639, 0.941024601459503, 0.957684397697449, 0.880294620990753, 0.880100905895233, 0.770584881305695, 0.989603161811829, 0.787922501564026, 0.911516785621643, 0.887528479099274, 0.938995897769928, 0.799316585063934, 0.944381177425385, 0.989226937294006, 0.716332614421844, 0.874548733234406, 0.935770869255066, 0.871802866458893, 0.911523163318634, 0.970747947692871 },
      { -32.0060653686523, -82.0628356933594, 245.264617919922, -56.0316276550293, 24.2496948242188, -84.2494506835938, -11.0268459320068, -5.18243646621704, -36.4850082397461, 53.9788093566895, 6.08013153076172, 79.0981292724609, 33.7436370849609, -401.403411865234, -72.028923034668, -93.9673614501953, -199.526794433594, 61.366527557373, -31.2955493927002, 20.7004833221436, 0.279486626386642, 25.6451034545898, -874.956359863281, 11.0563020706177, 35.1334762573242, -73.7290725708008, 78.0698776245117, 21.592041015625, 162.076263427734, 22.995943069458, 19.2710132598877, -108.410827636719, -1.68604779243469, 12.294059753418, 67.8447036743164, 1.67480003833771, -16.6808223724365, 37.6527061462402, 195.157363891602, 11.1053838729858, 55.5762596130371, -73.7420959472656, 392.657806396484, -29.2302513122559, 46.8241386413574, -90.5111312866211, 24.2647285461426, 16.5971908569336, -8.08323860168457, 26.5428848266602, -17.7993240356445 }
    }
  }
};


/** This function performs spectral adjustment to Sentinel-2 using the 
+++ Scheffler 2020 method
--- ard:       ARD image (single time step)
--- mask:      mask image
--- nc:        number of pixels
--- sid:       sensor ID
+++ Return:    SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int spectral_predict(ard_t ard, small *mask_, int nc, int sid){
int b, b_, p, s, c;
int b_src[_SPECHOMO_N_SRC_];
int b_dst[_SPECHOMO_N_DST_];
double xy, xx, yy, sam;
double weight[_SPECHOMO_N_CLS_], weight_select[_SPECHOMO_N_SIM_], max_weight;
int cluster_select[_SPECHOMO_N_SIM_], max_cluster, n_cluster;
double pred, wpred[_SPECHOMO_N_DST_], wsum;


  // get the correct source bands from brick
  for (b=0; b<_SPECHOMO_N_SRC_; b++){
    if ((b_src[b] = find_domain(ard.DAT, _SPECHOMO_SRC_DOMAIN_[b])) < 0){ 
      printf("Couldn't find source domain %s. ", _SPECHOMO_SRC_DOMAIN_[b]);
      return FAILURE;
    }
  }

  // get the correct destination bands from brick
  for (b=0; b<_SPECHOMO_N_DST_; b++){
    if ((b_dst[b] = find_domain(ard.DAT, _SPECHOMO_DST_DOMAIN_[b])) < 0){ 
      printf("Couldn't find target domain %s. ", _SPECHOMO_DST_DOMAIN_[b]);
      return FAILURE;
    }
  }

  // attention: removed default(none) for gcc back-compatibility (https://gcc.gnu.org/gcc-9/porting_to.html#ompdatasharing)
  // adding _SPECHOMO_CENTER_ and _SPECHOMO_COEFS_ to the firstprivate clause was no option as the build time and executables increased substantially
  // if something goes wrong, this might be a good place to start looking!
  // if you change something in the parallel block below, it is advised to temporarily add the default(none) and firstprivate clauses
  #pragma omp parallel private(s,c,b,b_,weight,weight_select,cluster_select,max_weight,max_cluster,n_cluster,xy,xx,yy,sam,wpred,wsum,pred) shared(ard,mask_,nc,b_src,b_dst,sid) // default(none)
  {

    #pragma omp for
    for (p=0; p<nc; p++){

      if (mask_ != NULL && !mask_[p]) continue;
      if (!ard.msk[p]) continue;

      memset(weight,         0, _SPECHOMO_N_CLS_*sizeof(double));
      memset(weight_select,  0, _SPECHOMO_N_SIM_*sizeof(double));
      memset(cluster_select, 0, _SPECHOMO_N_SIM_*sizeof(int));
      max_weight = -1.0;
      max_cluster = 0;

      // compute SAM and weight for each cluster
      for (s=0; s<_SPECHOMO_N_CLS_; s++){

        // SAM to each cluster center
        xy = xx = yy = 0.0;

        for (b=0; b<_SPECHOMO_N_SRC_; b++){
          xy += ard.dat[b_src[b]][p]         * _SPECHOMO_CENTER_[sid][b][s];
          xx += ard.dat[b_src[b]][p]         * ard.dat[b_src[b]][p];
          yy += _SPECHOMO_CENTER_[sid][b][s] * _SPECHOMO_CENTER_[sid][b][s];
        }

        if (xx > 0 && yy > 0){
          sam = acos(xy / sqrt(xx*yy));
        } else {
          sam = 1.0;
        }

        // weight for each cluster center
        // not exactly as in the paper, pragmatic suggestion by D. Scheffler
        if (sam > _SPECHOMO_POOR_SAM_){
          weight[s] = 0.00001;
        } else {
          weight[s] = 1.0 - sam / _SPECHOMO_POOR_SAM_;
        }

        #ifdef FORCE_DEBUG
        printf("\n");
        printf("on/off: %d\n", ard.msk[p]);
        printf("cluster number: %d\n", s);
        printf("ard: ");
        for (b=0; b<_SPECHOMO_N_SRC_; b++) printf("%05d ", ard.dat[b_src[b]][p]);
        printf("\n");
        printf("lib: ");
        for (b=0; b<_SPECHOMO_N_SRC_; b++) printf("%05d ", _SPECHOMO_CENTER_[sid][b][s]);
        printf("\n");
        printf("xx: %.2f, yy: %.2f, xy: %.2f, sam: %.8f, sam_deg: %.8f, weight: %.8f\n", xx, yy, xy, sam, sam*_R2D_CONV_, weight[s]);
        #endif

        // maximum weight of close clusters (-1 if no close cluster)
        if (weight[s] > max_weight && 
            s != _SPECHOMO_N_SIM_-1){

          max_weight  = weight[s];
          max_cluster = s;

        }

        #ifdef FORCE_DEBUG
        printf("max_weight: %.5f, max_cluster: %02d\n", max_weight, max_cluster);
        #endif

      }

      #ifdef FORCE_DEBUG
      print_dvector(weight,  "all weights", _SPECHOMO_N_CLS_, 2, 5);
      #endif


      n_cluster = 0;

      // find closest clusters, fill with global
      for (c=0; c<_SPECHOMO_N_SIM_; c++){

        // init with global cluster
        cluster_select[c] = _SPECHOMO_N_CLS_-1;
        weight_select[c]  = weight[_SPECHOMO_N_CLS_-1];

        if (max_weight > _SPECHOMO_MIN_WEIGHT_ ||
           (max_weight > _SPECHOMO_MED_WEIGHT_ &&
            max_weight > weight[_SPECHOMO_N_CLS_-1])){
        
          // copy closest cluster
          cluster_select[c] = max_cluster;
          weight_select[c]  = max_weight;

          // remove weight from closest cluster
          weight[max_cluster] = 0.0;

        } else break;

        // find the next closest cluster
        max_weight = -1.0;
        for (s=0; s<(_SPECHOMO_N_CLS_-1); s++){
          if (weight[s] >  max_weight){
            max_weight  = weight[s];
            max_cluster = s;
          }
        }

      }

      // number of close clusters
      n_cluster = c+1;
      if (n_cluster > _SPECHOMO_N_SIM_) n_cluster = _SPECHOMO_N_SIM_;

      #ifdef FORCE_DEBUG
      printf("found %d clusters\n", n_cluster);
      print_ivector(cluster_select, "cluster", _SPECHOMO_N_SIM_, 8);
      print_dvector(weight_select,  "weights", _SPECHOMO_N_SIM_, 2, 5);
      #endif
      
      // predict the target reflectance by using a weighted average of 
      // all close regressors

      memset(wpred, 0, _SPECHOMO_N_DST_*sizeof(double));
      wsum = 0.0;

      for (c=0; c<n_cluster; c++){

        s = cluster_select[c];

        for (b=0;  b <_SPECHOMO_N_DST_; b++){

          for (b_=0, pred=0; b_<_SPECHOMO_N_SRC_; b_++){
            pred += _SPECHOMO_COEFS_[sid][b][b_][s] * ard.dat[b_src[b_]][p];
          }
          pred += _SPECHOMO_COEFS_[sid][b][b_][s]; // offset

          wpred[b] += weight_select[c]*pred;

        }

        wsum += weight_select[c];

      }


      #ifdef FORCE_DEBUG
      printf("using %d regressors\n", n_cluster);
      printf("reflectance before:");
      for (b=0; b<_SPECHOMO_N_DST_; b++) printf(" %04d", ard.dat[b_dst[b]][p]); 
      printf("\n");
      printf("reflectance after: ");
      for (b=0; b<_SPECHOMO_N_DST_; b++) printf(" %04d", (short)(wpred[b] / wsum));
      printf("\n");
      //ard.dat[0][p] = n_cluster;
      #endif

      for (b=0; b <_SPECHOMO_N_DST_; b++) ard.dat[b_dst[b]][p] = (short)(wpred[b] / wsum);

    }

  }

  return SUCCESS;
};


/** This function performs spectral adjustment to Sentinel-2
--- ard:       ARD
--- mask:      mask image
--- nt:        number of ARD products over time
--- phl:       HL parameters
+++ Return:    SUCCESS/FAILURE/CANCEL
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int spectral_adjust(ard_t *ard, brick_t *mask, int nt, par_hl_t *phl){
int t, s, nc;
char sensor[NPOW_04];
bool adjust = false;
small *mask_ = NULL;


  if (!phl->sen.spec_adjust) return CANCEL;

  cite_me(_CITE_SPECADJUST_);


  // import mask (if available)
  if (mask != NULL){
    if ((mask_ = get_band_small(mask, 0)) == NULL){
      printf("Error getting processing mask."); return FAILURE;}
  }

  nc = get_brick_chunkncells(ard[0].DAT);

  // for each time step
  for (t=0; t<nt; t++){

    get_brick_sensor(ard[t].DAT, 0, sensor, NPOW_04);

    // is the sensor adjustable?
    for (s=0, adjust=false; s<_SPECHOMO_N_SEN_; s++){
      if (strcmp(sensor, _SPECHOMO_SENSOR_[s]) == 0){ 
        adjust = true; 
        break;
      }
    }

    #ifdef FORCE_DEBUG
    printf("adjusting %s: %d\n", sensor, adjust);
    #endif

    if (!adjust) continue;

    // perform the adjustment
    if (spectral_predict(ard[t], mask_, nc, s) == FAILURE){
      printf("failed to compute spectral prediction. "); return FAILURE;}

  }


  return SUCCESS;
}

